/*----------------------------------------------------------------------*\
 | Code to read and write binary versions of formal metadata.			|
 |																		|
 | The purpose is to encode the metadata so that they can be ingested	|
 | quickly without having to be parsed.  This should speed the process	|
 | of indexing collections and of carrying out searches on individual	|
 | metadata records.													|
 |																		|
 | The binary file consists of a header, an array of small structures	|
 | used to associate extension names with the numerical key values that	|
 | indicate their occurrence in the tree, an array of item structures	|
 | in which the pointers are all relative to the base of the array, and	|
 | a string pool.														|
 |																		|
 | The current version is intended to read files that were created on	|
 | the same machine.  This avoids problems with byte order, word size,	|
 | pointer size, and structure packing.									|
 |																		|
 | When a null pointer is given in the original parse tree, this is		|
 | encoded in the binary file using the pointer address -1 so that it	|
 | can be distinguished from a reference to the base of the array to	|
 | which the pointer's parent structure belongs.						|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "item.h"
#include "config.h"

/*----------------------------------------------------------------------*\
 | Definitions duplicated from local.c									|
\*----------------------------------------------------------------------*/

#define MAX_COMPOUND_COUNT	255
#define MAX_COMPONENT_COUNT	31

struct extension_info {
	int key;
	enum fgdc_keyword parent [1 + MAX_COMPOUND_COUNT];
	enum fgdc_keyword child [1 + MAX_COMPONENT_COUNT];
	};

extern int get_extension_count (void);
extern struct extension_info *get_extension_info (int i);

/*----------------------------------------------------------------------*/

struct binary_metadata_record {
	struct item *root;
	char *data;			/* so the memory can be freed */
	};

struct extension_id {
	int key;
	char name [64];
	};

struct header {
	char signature[8];
	struct {
		int offset;
		int length;
		int count;
		} extensions;
	struct {
		int offset;
		int length;
		int count;
		} structure;
	struct {
		int offset;
		int length;
		} pool;
	};

static struct item *null_item = (struct item *) -1;
static char * null_string = (char *) -1;

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

#ifndef LITTLE_ENDIAN
#define LITTLE_ENDIAN	0
#endif

#ifndef BIG_ENDIAN
#define BIG_ENDIAN		1
#endif

static int byte_order = BIG_ENDIAN;

static void get_byte_order (void) {
	short w = 255;
	unsigned char *s = (unsigned char *) &w;
	if (*s) byte_order = LITTLE_ENDIAN;
	else byte_order = BIG_ENDIAN;
	}

/*----------------------------------------------------------------------*\
 | read_file duplicates read_text in text.c								|
\*----------------------------------------------------------------------*/

static char *read_file (char *name) {
	long n;
	FILE *fp;
	char *dst,*end;
	long size;

	if (!name) return (NULL);

	dst = NULL;
	if (fp = fopen (name,"rb")) {
		if (fseek (fp,0L,SEEK_END) == 0) {
			size = 1L + ftell (fp);
			if (size > 1L) {
				if (dst = (char *) malloc (size)) {
					rewind (fp);
					n = fread (dst,1,size,fp);
					end = dst + n;
					*end = 0;
					}
				else
					fprintf (stderr,"Error: could not allocate space for file\n");
				}
			else
				fprintf (stderr,"Error: Input file is empty.\n");
			}
		fclose (fp);
		}
	else
		fprintf (stderr,"Error: could not open input file %s\n",name);
	return (dst);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

struct binary_metadata_record *read_binary (char *binary_file) {
	struct header *h;
	struct item *item_array = NULL;
	struct item *p;
	struct extension_id *extension_array;
	char *pool;
	char *data;
	int i;
	struct binary_metadata_record *result = NULL;

	get_byte_order();

	/*--------------------------------------------------------------*\
	 | Snarf the whole file in one shot.  That way the whole block	|
	 | can be freed at once when we're done with it.				|
	\*--------------------------------------------------------------*/

	if (data = read_file (binary_file)) {

		/*--------------------------------------------------------------*\
		 | Read signature.  Check to determine whether this file can be	|
		 | easily interpreted.  If not, fail (fix this later).			|
		\*--------------------------------------------------------------*/

		h = (struct header *) data;

		if (memcmp (h->signature,"FGDC",4) != 0) {
			fprintf (stderr,"Error: incorrect signature in binary input file %s\n",binary_file);
			free (data);
			return (NULL);
			}

		i = ((byte_order == LITTLE_ENDIAN) ? 0 : 1);
		if ((h->signature[4] - '0') != sizeof(void *) ||
			(h->signature[5] - '0') != sizeof(int) ||
			(h->signature[6] - '0') != i) {
			fprintf (stderr,"Error: incompatible binary file %s\n",binary_file);
			free (data);
			return (NULL);
			}

	/*\
		printf ("Header\n");
		printf ("  signature: %s\n",h->signature);
		printf ("  extensions\n");
		printf ("    offset: %lu\n",h->extensions.offset);
		printf ("    length: %lu\n",h->extensions.length);
		printf ("    count:  %lu\n",h->extensions.count);
		printf ("  structure\n");
		printf ("    offset: %lu\n",h->structure.offset);
		printf ("    length: %lu\n",h->structure.length);
		printf ("    count:  %lu\n",h->structure.count);
		printf ("  pool\n");
		printf ("    offset: %lu\n",h->pool.offset);
		printf ("    length: %lu\n",h->pool.length);
		fflush (stdout);
	\*/

		/* Read extension id array */

		if (h->extensions.count > 0)
			if (h->extensions.length/h->extensions.count != sizeof(struct extension_id)) {
				fprintf (stderr,"Error: incompatible binary file %s\n",binary_file);
				free (data);
				return (NULL);
				}

		if (h->extensions.count > 0)
			extension_array = (struct extension_id *)(data + h->extensions.offset);
		else
			extension_array = NULL;

		if (h->extensions.count > 0) {
			printf ("%d Extensions\n",h->extensions.count);
			for (i=0; i < h->extensions.count; i++)
				printf ("%4d\t%s\n",extension_array[i].key,extension_array[i].name);
			}

		/* Read the structure */

		if (h->structure.count > 0)
			item_array = (struct item *) (data + h->structure.offset);
		else
			item_array = NULL;

		/* Read the string pool */

		if (h->pool.length > 0)
			pool = data + h->pool.offset;
		else
			pool = NULL;

	/*\
		for (p=item_array, i=0; i < 10; i++,p++) {
			printf ("Item %d at 0x%lX\n",i,p-item_array);
			printf ("  next        = 0x%lX\n",p->next);
			printf ("  prev        = 0x%lX\n",p->prev);
			printf ("  parent      = 0x%lX\n",p->parent);
			printf ("  child       = 0x%lX\n",p->child);
			printf ("  indent      = %d\n",   p->indent);
			printf ("  line_number = %d\n",   p->line_number);
			printf ("  key         = %s\n",   text_of(p->key));
			printf ("  original    = %d\n",   p->original);
			printf ("  len         = %d\n",   p->len);
			printf ("  d           = 0x%lX\n",p->d);
			fflush (stdout);
			}
	\*/

		/* Adjust pointers */

		for (p=item_array, i=0; i < h->structure.count; i++,p++) {
			p->next   = ((p->next   == null_item) ? NULL : item_array + (int) p->next);
			p->prev   = ((p->prev   == null_item) ? NULL : item_array + (int) p->prev);
			p->parent = ((p->parent == null_item) ? NULL : item_array + (int) p->parent);
			p->child  = ((p->child  == null_item) ? NULL : item_array + (int) p->child);
			p->d      = ((p->d    == null_string) ? NULL : pool       + (int) p->d);
			}

	/*\
		for (p=item_array, i=0; i < 10; i++,p++) {
			printf ("Item %d at 0x%lX\n",i,p);
			printf ("  next        = 0x%lX\n",p->next);
			printf ("  prev        = 0x%lX\n",p->prev);
			printf ("  parent      = 0x%lX\n",p->parent);
			printf ("  child       = 0x%lX\n",p->child);
			printf ("  indent      = %d\n",   p->indent);
			printf ("  line_number = %d\n",   p->line_number);
			printf ("  key         = %s\n",   text_of(p->key));
			printf ("  original    = %d\n",   p->original);
			printf ("  len         = %d\n",   p->len);
			if (p->d)
				printf ("  d           = %s\n",  p->d);
			else
				printf ("  d           = (null)\n",  p->d);
			fflush (stdout);
			}
	\*/

		if (result = (struct binary_metadata_record *) malloc (sizeof (struct binary_metadata_record))) {
			result->root = item_array;
			result->data = data;
			}
		}
	return (result);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static int length_of (struct item *p) {
	int k = 0;
	struct item *q;

	if (p)
		if (p->key != Wblank){
			if (p->key == Wunknown) k += 1 + strlen (p->d);
			else
				for (q=p->child; q; q=q->next) k += length_of (q);
			}
	return (k);
	}

/*----------------------------------------------------------------------*/

static int number_of (struct item *p) {
	int k = 0;
	struct item *q;

	if (p) {
		k++;
		for (q=p->child; q; q=q->next) k += number_of (q);
		}
	return (k);
	}

/*----------------------------------------------------------------------*/

static void adjust_indent (struct item *p, int n) {
	struct item *q;

	if (p) {
		if (p->key != Wblank) p->indent = n;
		for (q=p->child; q; q=q->next) adjust_indent (q,n+1);
		}
	}

/*----------------------------------------------------------------------*/

static int current_item_ptr = 0;

static void build_ptr_list (struct item *p, struct item **item_ptr_array) {
	struct item *q;

	if (p) {
		item_ptr_array [current_item_ptr++] = p;
		for (q=p->child; q; q=q->next) build_ptr_list (q,item_ptr_array);
		}
	}

/*----------------------------------------------------------------------*/

int write_binary (char *binary_file, struct item *first) {
	int i,j;
	int ok = 0;
	FILE *fp;
	struct header header;
	int extension_count = 0;
	struct extension_id *extension_array = NULL;
	int item_count = 0;
	struct item **item_ptr_array = NULL;
	struct item *item_array = NULL;
	int pool_size = 0;
	char *pool = NULL;
	char *pool_ptr = NULL;
	struct item **qq;
	struct item *q;
	char *s;

	if (!binary_file) return (ok);

	get_byte_order();

	/* Adjust indent member of item structures */

	adjust_indent (first,0);

	/*------------------------------------------------------------------*\
	 | Create block describing extensions								|
	\*------------------------------------------------------------------*/

	if (extension_count = get_extension_count ()) {
		if (extension_array = (struct extension_id *) malloc (extension_count * sizeof (struct extension_id))) {
			memset (extension_array,0,extension_count * sizeof (struct extension_id));
			}
		else {
			fprintf (stderr,"Error: could not allocate array of %d extension identifiers\n",extension_count);
			return (ok);
			}
		for (i=0; i < extension_count; i++) {
			struct extension_info *e;
			if (e = get_extension_info (i)) {
				extension_array[i].key = e->key;
				strcpy (extension_array[i].name,text_of(e->key));
				}
			}
		}

	/*------------------------------------------------------------------*\
	 | Create block describing structure and string pool				|
	\*------------------------------------------------------------------*/

	/* Count items */

	item_count = number_of (first);

	/* Allocate an array of pointers to existing items */

	if (item_ptr_array = (struct item **) malloc (item_count * sizeof (struct item *))) {
		memset (item_ptr_array,0,item_count*sizeof(struct item *));
		current_item_ptr = 0;
		}
	else {
		fprintf (stderr,"Error: could not allocate array of pointers for %d items\n",item_count);
		return (ok);
		}

	/* Allocate an array of items into which pertinent data will be copied */

	if (item_array = (struct item *) malloc (item_count * sizeof (struct item))) {
		memset (item_array,0,item_count*sizeof(struct item));
		}
	else {
		fprintf (stderr,"Error: could not allocate array of %d items\n",item_count);
		return (ok);
		}

	/* Calculate the length of the string pool */

	pool_size = length_of (first);

	/* Allocate string pool */

	if (pool = (char *) malloc (pool_size)) {
		memset (pool,0,pool_size);
		pool_ptr = pool;
		}
	else {
		fprintf (stderr,"Error: could not allocate string pool of size %d\n",pool_size);
		return (ok);
		}

	/* Build list of pointers to real items */

	build_ptr_list (first,item_ptr_array);

	/* Build list of items in a contiguous block for output */

	qq = item_ptr_array;
	q = item_array;
	for (i=0; i < item_count; i++) {
		struct item *r = *qq;

		/*--------------------------------------------------------------*\
		 | Replace each element pointer with the array index of the		|
		 | corresponding element rather than its address.				|
		 | Use -1 (null_item) for a null pointer.						|
		\*--------------------------------------------------------------*/

		q->next = null_item;
		if (r->next)
			for (j=i; j < item_count; j++)
				if (item_ptr_array[j] == r->next) {
					q->next = (struct item *)j;
					}

		q->prev = null_item;
		if (r->prev)
			for (j=0; j < item_count; j++)
				if (item_ptr_array[j] == r->prev) {
					q->prev = (struct item *)j;
					}

		q->parent = null_item;
		if (r->parent)
			for (j=0; j < item_count; j++)
				if (item_ptr_array[j] == r->parent) {
					q->parent = (struct item *)j;
					}

		q->child = null_item;
		if (r->child)
			for (j=i; j < item_count; j++)
				if (item_ptr_array[j] == r->child) {
					q->child = (struct item *)j;
					}

		/* Copy data elements */

		q->key = r->key;
		q->indent = r->indent;
		q->line_number = r->line_number;	/* irrelevant */
		q->original = 0;

		if (q->key == Wunknown)
			if (r->d) {
				strcpy (pool_ptr,r->d);
				q->d = (char *)(pool_ptr - pool);	/* use offset in pool */
				q->len = 1 + strlen (r->d);
				pool_ptr += 1 + strlen (r->d);
				}
			else {
				q->d = null_string;
				q->len = 0;
				}
		else {
			q->d = null_string;	/* Don't encode containers or blanks */
			q->len = 0;
			}

		q++;
		qq++;
		}

	/*------------------------------------------------------------------*\
	 | Create file header.  Need size and offset of extensions block,	|
	 | size and offset of structure block, and size and offset of the	|
	 | string pool.														|
	\*------------------------------------------------------------------*/

	s = header.signature;
	*s++ = 'F';
	*s++ = 'G';
	*s++ = 'D';
	*s++ = 'C';
	*s++ = '0' + sizeof(void *);
	*s++ = '0' + sizeof(int);
	*s++ = ((byte_order == LITTLE_ENDIAN)? '0' : '1');
	*s   = 0;
	header.extensions.offset = sizeof (struct header);
	header.extensions.length = extension_count * sizeof (struct extension_id);
	header.extensions.count = extension_count;
	header.structure.offset = header.extensions.offset + header.extensions.length;
	header.structure.length = item_count * sizeof (struct item);
	header.structure.count = item_count;
	header.pool.offset = header.structure.offset + header.structure.length;
	header.pool.length = pool_size;

	/* Write output file */

	if (fp = fopen (binary_file,"wb")) {
		fwrite (&header,sizeof(struct header),1,fp);
		fwrite (extension_array,sizeof(struct extension_id),extension_count,fp);
		fwrite (item_array,sizeof(struct item),item_count,fp);
		fwrite (pool,1,pool_size,fp);
		fclose (fp);
		ok = 1;
		}
	else
		fprintf (stderr,"Error: could not create binary output file %s\n",binary_file);

	return (ok);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
