/*----------------------------------------------------------------------*\
 | Functions to read config file and return information from it.		|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "stricmp.h"

extern char *read_text (char *input_file);

/*----------------------------------------------------------------------*\

Configuration file syntax:

Summary:

The configuration file is encoded like the metadata; hierarchical
structure is indicated using indentation.  Most of these are simple,
and indicate how the program should treat input or output.  All can
be left out to elicit the default behavior.

input
	profile bio
	indent [strict | lax]
	blanks ignore
	extensions <file name>
	codeset [iso8859-1 | dos | mac | utf-8]
	encoding UTF-8
	tips <file name>
	upgrade [no]
	prefix <string>
	prune
	ext <file extension>
	language <code>
output
	errors <file name>
	tree <file name>
	text
		file <file name>
		cns <file name>
		indent <string>
		top_level [keep | skip]
		prefix <string>
		wrap <number>
	sgml
		file <file name>
		tags [astm | ps8]
		indent <string>
		blanks <string>
		skip_extensions
	xml
		file <file name>
		indent <string>
		blanks <string>
		doctype <string>
		skip_extensions
		stylesheet
			type <mime-type>
			href <url>
		order standard | as-is
		schema <url>
	html
		file <file name>
		faq <file name>
		err <file name>
		translate [off]
		encoding utf-8
		preformat >
		meta [off]
		base <url>
		style_file <filespec>
		stylesheet
			type <mime-type>
			href <url>
		script_file <filespec>
		body <args>
		header <html>
		header_file <filespec>
		footer <html>
		footer_file <filespec>
		omit_url
		link
			label
			link_faq
			link_html
			link_text
			link_sgml
			link_xml
			link_dif
		keyword
			prefix
			suffix
		blanks <string>
		element
			key <name>
			name
				prefix <html>
				suffix <html>
			value
				prefix <html>
				suffix <html>
				obeylines
			omit
	info
		file <file name>
	dif
		file <file name>

Compound types:

input = (profile) + (indent) + (blanks) + (extensions) + (encoding) +
        (tips) + (upgrade) + (prefix) + (prune) + (ext) + (language) +
        (codeset)
output = (tree) + (text) + (sgml) + (xml) + (html) + (dif) + (binary)
errors = (file)
text = (file) + (cns) + (indent) + (top_level) + (wrap)
sgml = (file) + (tags) + (indent) + (blanks) + (skip_extensions)
xml = (file) + (indent) + (blanks) + (skip_extensions) + (stylesheet) + (order) + (schema)
html = (file) + (faq) + (err) + (translate) + (encoding) + (preformat) + (meta) + (base) + (style_file) + (script_file) + (body) + (header | header_file) + (footer | footer_file) + (link) + (keywords) + (blanks) + 0{element}n + (stylesheet)
link = (label) + (link_faq) + (link_html) + (link_text) + (link_sgml) + (link_xml) + (link_dif)
keyword = (prefix) + (suffix)
element = key + (name) + (value)
name = (prefix) + (suffix)
value = (prefix) + (suffix) + (obeylines)
dif = (file)
binary = (file)
info = (file)
stylesheet = (type) + href

Scalar types:

profile
	Under input, the value is a name associated with a recognized
	profile of the FGDC standard.  Currently the only value that
	is recognized is "bio" to indicate the Biological Data Profile.

file
	Under output, the value given will be used as the name of the output
	file for the format under which the keyword appears.  If the
	characters "%s" appear in the value, they will be replaced with the
	name of the input file as given on the command line.  So, for example,

	text:
		file: %s.txt

	will result in the textual output being created with the name
	<input_file>.txt.  Note that this may produce an illegal file name
	under MS-DOS if <input_file> contains an extension already.  In that
	case, the file will not be overwritten.

	What appears in the configuration file for this option is overridden
	by the command-line option.  Thus it is possible to overwrite the
	input file by specifying its name as one of the output files using
	a command-line option -e, -t, -h, -s, or -d.

indent
	Under input, the value "strict" causes indented text to be processed
	as hierarchical rather than linear.  The default is to interpret
	indented unrecognized keywords as plain text.

	Under output, the value given will be used once for each level in the
	hierarchy, preceding the output data on each line.  This is a string
	that may be delimited by quotation marks (").  The backslash is used
	to include a character literally.

top_level
	Under output:text, the value "skip" causes the tree to be output at
	the node immediately below Metadata.  This is a convenience since,
	presumably, everything in the file is contained in Metadata, so if
	this is not done, everything is indented except the keyword Metadata.
	The default is to skip the Metadata keyword in text output, so the
	appearance of the keyword top_level WITHOUT the value "skip" causes
	the Metadata keyword (and its indentation throughout the file) to be
	retained.

blanks
	Under input, the value "ignore" causes blank lines to be ignored in
	the metadata.  Otherwise blank lines are included and processed as
	data.  The parser attempts to prevent the assignment of children to
	blank lines.

	Under output, the value associated with the keyword is output for
	every blank line in the input file, unless blanks are being ignored,
	in which case there won't be any in the parse tree when the output
	routine is called.  For SGML output, the default is ""; for HTML
	output, the default is "<P>\n".

extensions
	Under input, the argument is the full path to a local file containing
	extensions to the standard that are used in the current document.

encoding [or codeset (deprecated)]
	Under input, the argument is one of the values "ISO8859-1", "UTF-8",
	"DOS", or "MAC" (case is not significant).  This indicates the encoding
	of characters beyond ASCII 127.  Default is ISO8859-1.
	Under output:html the value UTF-8 causes the HTML files to be
	written using UTF-8 encoding (default is ISO-8859-1).

label
	Under output:html:link, the value given to "label" is the lead text
	in the line containing links to the other formats.  The default
	value is "Metadata also available as".

tips
	Under input, the argument is the full path to a local file containing
	tips to be used by xtme.

upgrade
	If given the value "no", mp will not attempt to upgrade the metadata
	to version 2 of the CSDGM.

language
	Under input, the argument is the 2-letter language code for the
	element names to be used in the file.  At this writing the
	following languages are supported:
		en	English
		es	Spanish
		id	Indonesian

tree
	If specified, this keyword causes the parse tree to be dumped to
	stdout.  If an argument is present, it is interpreted as the name
	of a file into which the parse tree should be dumped.

tags
	Under output:sgml, the value "astm" causes the ten-character tags
	described in the proposed ASTM D18.01.05 standard to be used.  This
	is the default behavior, however, so the eight-character tags are
	used only if output:sgml:tags is specified but "astm" is not given
	as the argument.

skip_extensions
	Under output:sgml, this keyword causes the sgml code generator to
	skip elements that are not part of the 19940608 version of the
	metadata standard.  By default extensions are included.

translate
	Under output:html, textual strings in metadata values are translated
	into html.  Specifically, the characters <, >, ", and & are converted
	into the corresponding entities (&lt;, &gt;, &quot;, and &amp;) and
	URLs of the form <URL:theURL> are rendered as <A HREF="theURL">&lt;URL:
	theURL&gt;</A> in the output.  If you don't want this behavior,
	specify translate off.

preformat
	Under output:html, preformat causes any groups of one or more lines
	that begin with > to be enclosed in <pre></pre> tags.  An optional
	single-character value associated with this tag allows the user to
	determine what character will be used for indicating the lines
	that should be rendered in this manner. Default is >.

faq
	Under output:html, the value indicates the name of the FAQ-style
	HTML output file.

err
	Under output:html, the value indicates the name of the output file
	for err2html.

meta
	Under output:html, the value off causes no meta tags to be generated.

body
	Under output:html, the argument is used to replace %s in <body %s>
	allowing the user to modify the background color of the html output.

header
	Under output:html, the argument is written at the beginning of the
	body in the html code, before the title and table of contents.

header_file
	Under output:html, the argument is the name of a file that can be read
	and the whose contents will become the header.  Ignored if header is
	specified.

footer
	Under output:html, the argument is written at the end of the
	body in the html code, after the "generated by mp" line.

footer_file
	Under output:html, the argument is the name of a file that can be read
	and the whose contents will become the footer.  Ignored if footer is
	specified.

key
	Under output:html:element, the argument is the name of the
	element for whom html prefix and suffix may be associated with
	the name, the value, or both.

prefix
	Under input, the argument is a special code for cns that indicates
	the following text is an element name.

	Under output:html:element:name, the argument is the html code to
	be output immediately before the name of the element indicated by
	the associated "key" keyword.

	Under output:html:element:value, the argument is the html code to
	be output immediately before the value associated with the element
	indicated by the associated "key" keyword.

	Under output:html:keywords, the argument is the html code to be
	output immediately before each element name.

suffix
	Under output:html:element:name, the argument is the html code to
	be output immediately following the name of the element indicated
	by the associated "key" keyword.

	Under output:html:element:value, the argument is the html code to
	be output immediately following the value associated with the element
	indicated by the associated "key" keyword.

	Under output:html:keywords, the argument is the html code to be
	output immediately following each element name.

obeylines
	Under output:html:element:value, the presence of this element causes
	<br> tags to be emitted at the end of every line of the element's
	value.  A child of obeylines is ignored.

prune
	Under input, this option causes mp to prune the whole tree of "dead"
	branches.

link
	Under output:html, specifies how the links to related files will
	be written.  If empty, no link line is written.

link_faq
	Under output:html:link, causes a link to be created for the FAQ-style
	HTML output.  If a value is given, this is the URL to be used in the
	link, with %s given where the name of the input file should be placed.

link_html
	Under output:html:link, causes a link to be created for the outline-
	style HTML output.  If a value is given, this is the URL to be used
	in the link, with %s given where the name of the input file should
	be placed.

link_text
	Under output:html:link, causes a link to be created for the indented
	text output.  If a value is given, this is the URL to be used
	in the link, with %s given where the name of the input file should
	be placed.

link_sgml
	Under output:html:link, causes a link to be created for the SGML
	output.  If a value is given, this is the URL to be used in the
	link, with %s given where the name of the input file should be
	placed.

link_xml
	Under output:html:link, causes a link to be created for the XML
	output.  If a value is given, this is the URL to be used in the
	link, with %s given where the name of the input file should be
	placed.

link_dif
	Under output:html:link, causes a link to be created for the DIF
	output.  If a value is given, this is the URL to be used in the
	link, with %s given where the name of the input file should be
	placed.

doctype
	Under output:xml, specifies that a DOCTYPE declaration should be
	included in the XML file.  If no value is specified, the default
	declaration is provided.  If a value is specified, it is used
	as the declaration, regardless of what it is.  It doesn't even
	have to BE a DOCTYPE declaration.

info
	Under output, may contain a child element "file" that is used
	to specify the file name of the info file generated by cns.

stylesheet
	Under output:xml, specifies a reference to an XSL stylesheet.
	Under output:html specifies the components of a link tag that
	will refer to an external CSS.

type
	Under output:xml:stylesheet, gives the type of the XSL stylesheet.
	Default is text/xsl

href
	Under output:xml:stylesheet, gives the network link to the XSL
	stylesheet.

order
	Under output:xml, anything other than the value "standard"
	causes the elements to be output in the order in which they
	appear in the parse tree.  If omitted, standard is assumed.

omit_url
	Under output:html, causes the page URL to be omitted.

omit
	Under output:element, causes this element not to be shown in
	outline-style HTML output

schema
	Under output:xml, causes the value of the schema element to be
	output as an attribute of the root (Metadata) element.  This
	actually creates two attributes, one of which defines the xsi
	namespace, the other specifies the value of the attribute
	xsi:noNamespaceSchemaLocation.

\*----------------------------------------------------------------------*/

enum config_word {
	Cinput,
	Cprofile,
	Cindent,
	Cblanks,
	Ccodeset,
	Cextensions,
	Coutput,
	Cerrors,
	Ctree,
	Ctext,
	Cfile,
	Ctop_level,
	Csgml,
	Cxml,
	Ctags,
	Cskip_extensions,
	Chtml,
	Cbody,
	Cheader,
	Cheader_file,
	Cfooter,
	Cfooter_file,
	Ctranslate,
	Cpreformat,
	Cmeta,
	Ckeyword,
	Celement,
	Ckey,
	Cname,
	Cvalue,
	Cprefix,
	Csuffix,
	Cunknown,
	Cempty,
	Cobeylines,
	Cbinary,
	Cupgrade,
	Ctips,
	Cprune,
	Cext,
	Cfaq,
	Cdif,
	Cbase,
	Cstyle_file,
	Cscript_file,
	CROOT,
	Clink,
	Clabel,
	Clink_faq,
	Clink_html,
	Clink_text,
	Clink_sgml,
	Clink_xml,
	Clink_dif,
	Cwrap,
	Cdoctype,
	Cinfo,
	Ccns,
	Cstylesheet,
	Ctype,
	Chref,
	Clanguage,
	Cerr,
	Corder,
	Comit_url,
	Comit,
	Cschema,
	Cencoding,
	CBLANK,
	Cnull
	};

struct keyword_list {
	enum config_word key;
	char *text;
	};

static struct keyword_list keyword_list[] = {
	{Cinput,			"input"},
	{Cprofile,			"profile"},
	{Cindent,			"indent"},
	{Cblanks,			"blanks"},
	{Ccodeset,			"codeset"},
	{Cextensions,		"extensions"},
	{Coutput,			"output"},
	{Cerrors,			"errors"},
	{Ctree,				"tree"},
	{Ctext,				"text"},
	{Cfile,				"file"},
	{Ctop_level,		"top_level"},
	{Csgml,				"sgml"},
	{Cxml,				"xml"},
	{Ctags,				"tags"},
	{Cskip_extensions,	"skip_extensions"},
	{Chtml,				"html"},
	{Cbody,				"body"},
	{Cheader,			"header"},
	{Cheader_file,		"header_file"},
	{Cfooter,			"footer"},
	{Cfooter_file,		"footer_file"},
	{Ctranslate,		"translate"},
	{Cpreformat,		"preformat"},
	{Cmeta,				"meta"},
	{Ckeyword,			"keyword"},
	{Celement,			"element"},
	{Ckey,				"key"},
	{Cname,				"name"},
	{Cvalue,			"value"},
	{Cprefix,			"prefix"},
	{Csuffix,			"suffix"},
	{Cobeylines,		"obeylines"},
	{Cbinary,			"binary"},
	{Ctips,				"tips"},
	{Cupgrade,			"upgrade"},
	{Cprune,			"prune"},
	{Cext,				"ext"},
	{Cfaq,				"faq"},
	{Cdif,				"dif"},
	{Cbase,				"base"},
	{Cstyle_file,		"style_file"},
	{Cscript_file,		"script_file"},
	{Clink,				"link"},
	{Clabel,			"label"},
	{Clink_faq,			"link_faq"},
	{Clink_html,		"link_html"},
	{Clink_text,		"link_text"},
	{Clink_sgml,		"link_sgml"},
	{Clink_xml,			"link_xml"},
	{Clink_dif,			"link_dif"},
	{Cwrap,				"wrap"},
	{Cdoctype,			"doctype"},
	{Cinfo,				"info"},
	{Ccns,				"cns"},
	{Cstylesheet,		"stylesheet"},
	{Ctype,				"type"},
	{Chref,				"href"},
	{Clanguage,			"language"},
	{Cerr,				"err"},
	{Corder,			"order"},
	{Comit_url,			"omit_url"},
	{Comit,				"omit"},
	{Cschema,			"schema"},
	{Cencoding,			"encoding"},
	{CROOT,				"ROOT"},
	{CBLANK,			"BLANK"},
	{Cunknown,			NULL}
	};

static int keyword_list_length = 0;

/*----------------------------------------------------------------------*/

static int compare (const void *e1, const void *e2) {
	struct keyword_list *p = (struct keyword_list *) e1;
	struct keyword_list *q = (struct keyword_list *) e2;
	return (stricmp (p->text,q->text));
	}

static enum config_word config_key_of (char *text) {
	int i,n;
	struct keyword_list k,*key;
	char string[128];

	if (keyword_list_length == 0) {
		for (i=0; keyword_list[i].text; i++);
		keyword_list_length = i;
		qsort (keyword_list,keyword_list_length,sizeof(struct keyword_list),compare);
		}
	if (!text) return (Cnull);
	if (*text) {
		n = strcspn (text," \t\n\v\f\r:=");
		if (n > 127) n = 127;
		memcpy (string,text,n);
		string[n] = 0;
		k.text = string;
		key = (struct keyword_list *) bsearch (&k,keyword_list,keyword_list_length,sizeof(struct keyword_list),compare);
		if (key) return (key->key);
		else return (Cunknown);
		}
	else return (Cempty);
	}

static char *config_text_of (enum config_word key) {
	int i;

	if (keyword_list_length == 0) {
		for (i=0; keyword_list[i].text; i++);
		keyword_list_length = i;
		qsort (keyword_list,keyword_list_length,sizeof(struct keyword_list),compare);
		}
	if (key == Cnull) return ("(null)");
	if (key == Cempty) return ("(empty)");
	for (i=0; i < keyword_list_length; i++)
		if (keyword_list[i].key == key) return (keyword_list[i].text);
	return ("(unknown)");
	}

/*----------------------------------------------------------------------*\
 | Error reporting mechanism											|
 |																		|
 | Since the config file may be processed before any other information	|
 | has been presented to the user, we should store warnings or error	|
 | messages in a text buffer that can be retrieved by the calling		|
 | program (typically mp or tkme) and presented to the user in some		|
 | appropriate way.														|
 |																		|
 | The function error_ptr is given a buffer size needed for the next	|
 | error message.  It returns the address of a buffer big enough for	|
 | the message, or NULL if space cannot be allocated.					|
 |																		|
 | The function config_errors simply returns a pointer to the buffer	|
 | that contains error messages.  This is NULL if no errors occurred.	|
\*----------------------------------------------------------------------*/

static char *error_buffer = NULL;
static int error_buffer_size = 0;
static int error_buffer_idx = 0;
#define GRANULARITY 4096

static char *error_ptr (int size) {
	int n;
	char *s;
	if (error_buffer_size - error_buffer_idx < 1 + size) {
		if (1 + size < GRANULARITY) n = GRANULARITY;
		else n = (1 + (1 + size)/GRANULARITY) * GRANULARITY;

		if (error_buffer)
			if (s = (char *) realloc (error_buffer,error_buffer_size + n)) {
				error_buffer = s;
				error_buffer_size += n;
				}
			else {
				free (error_buffer);
				error_buffer = NULL;
				error_buffer_size = 0;
				error_buffer_idx = 0;
				}
		else
			if (error_buffer = (char *) malloc (n))
				error_buffer_size = n;
		}
	return (error_buffer + error_buffer_idx);
	}

char *config_errors (void) {
	return (error_buffer);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

struct item {
	struct item *next;
	struct item *prev;
	struct item *parent;
	struct item *child;
	int indent;			/* number of leading whitespaces before text */
	int line_number;	/* line in source file */
	enum config_word key;
	int original;		/* zero if member d can be released thru free() */
	int len;			/* length of bloc pointed to by d */
	char *d;			/* text of item */
	};

static struct item *allocate_item (int fail) {
	struct item *p = NULL;

	if (p = (struct item *) malloc (sizeof(struct item)))
		memset (p,0,sizeof(struct item));
	else {
		fprintf (stderr,"Warning: could not allocate item\n");
		if (fail) exit (1);
		}
	return (p);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

#define UNKNOWN_TEXT	Cunknown
#define BLANK_LINE		CBLANK
#define TEXT_OF(x)		config_text_of(x)
#define KEY_OF(x)		config_key_of(x)

static int keep_blank_lines = 0;
static int allow_indented_text = 1;

static struct item *decode_text (char *dst) {
	char *end;
	char *b,*e,*s;
	struct item *p;
	struct item *first = NULL;
	struct item *current;
	int line_count;

	if (!dst) return (NULL);

	b = e = dst;
	end = dst + strlen (dst);

	/*--------------------------------------------------*\
	 | Start with an empty item							|
	\*--------------------------------------------------*/

	if (first == NULL) {
		first = allocate_item (1);
		current = first;
		}
	else {
		for (p=first; p->next; p=p->next);
		p->next = allocate_item (1);
		p->next->prev = p;
		current = p->next;
		}

	line_count = 0;
	while (b < end) {

		/*----------------------------------------------*\
		 | Find the first CR, LF, or EOF.  Zero the		|
		 | end-of-line character(s) if present.			|
		\*----------------------------------------------*/

		while (e < end) {
			if (*e == '\r') {
				*e = 0;
				if (*(e+1) == '\n') {
					e++;
					*e = 0;
					}
				break;
				}
			else
				if (*e == '\n') {
					*e = 0;
					break;
					}
			e++;
			}

		/*----------------------------------------------*\
		 | Clear trailing whitespace from the line.		|
		\*----------------------------------------------*/

		s = e;
		while (s >= b)
			if (*s == 0) s--;
			else
				if (isspace (*s)) *s-- = 0;
				else break;

		/*----------------------------------------------*\
		 | b points to the beginning of the line.		|
		 |												|
		 | If the first non-blank character in the line	|
		 | is '#', ignore the whole line.				|
		\*----------------------------------------------*/

		if (*b || keep_blank_lines) {
			for (s=b; *s && isspace(*s); s++);
			if (*s != '#') {
				current->original = 1;
				current->len = e - b + 1;
				current->d = b;
				current->line_number = 1 + line_count;
				current->next = allocate_item (1);
				current->next->prev = current;
				current = current->next;
				}
			}

		b = e + 1;
		line_count++;
		}
	if (first->next) {
		for (p=first; p->next->d; p=p->next);
		free (p->next);
		p->next = NULL;
		}
	else
		if (first->d == NULL) first = NULL;

	return (first);
	}

/*----------------------------------------------------------------------*\
 | Rather crude function that attempts to fix indented textual values.	|
 | Specifically, this routine looks for textual values that are nested	|
 | beneath other textual values and makes these siblings rather than	|
 | descendants.  It does not attempt to reclassify standard elements	|
 | as textual values; instead, it issues a warning in this case.		|
\*----------------------------------------------------------------------*/

static void equalize_indented_scalars (struct item *p) {
	struct item *q;

	while (p) {
		if (p->child)
			if (p->key == UNKNOWN_TEXT || p->key == BLANK_LINE) {
				if (p->child->key == UNKNOWN_TEXT || p->child->key == BLANK_LINE) {

					/*--------------------------------------------------*\
					 | Both p and p->child are either UNKNOWN_TEXT or	|
					 | BLANK_LINE.										|
					 |													|
					 | Convert p->child and all of its siblings to be	|
					 | siblings, not children, of p.					|
					\*--------------------------------------------------*/

					p->child->prev = p;			/* was NULL */

					p->child->parent = p->parent;	/* was p */

					for (q=p->child; q->next; q=q->next) {
						q->parent = p->parent;	/* was p */
						if (q->child) equalize_indented_scalars (q->child);
						}

					q->next = p->next;			/* was NULL */
					if (p->next) p->next->prev = q;
					p->next = p->child;
					p->child = NULL;
					}
				else {

					/*--------------------------------------------------*\
					 | p is of type UNKNOWN_TEXT but p->child is not.	|
					 | Don't second-guess at this point, but do issue	|
					 | a warning that some information might be lost in	|
					 | subsequent handling of this line.				|
					\*--------------------------------------------------*/

					char *fmt = "Warning (line %d of config file): element %s found in textual value, some information may be lost.\n";
					int m = 4 + strlen (fmt) + strlen (TEXT_OF (p->child->key));
					char *dst;
					if (dst = error_ptr (m))
						sprintf (dst,fmt,p->child->line_number,TEXT_OF (p->child->key));
					}
				}
			else
				equalize_indented_scalars (p->child);

		p = p->next;
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static struct item *decode_tree (struct item *first) {
	int i,j,k;
	char *s,*t;
	struct item *p,*q,*r;
	int item_count;
	struct item **list;

	if (!first) return (NULL);

	item_count = 0;
	for (p=first; p; p=p->next) {
		for (s=p->d,i=0; *s && isspace (*s); s++,i++);
		p->indent = i;
		p->d = s;
		p->key = KEY_OF (s);
		t = TEXT_OF (p->key);

		/*--------------------------------------------------------------*\
		 | If the current item can have a child on this line, determine	|
		 | whether it does, and store the child's information.  The		|
		 | rules prohibit such a child from having children of its own.	|
		\*--------------------------------------------------------------*/

		if (p->key != UNKNOWN_TEXT && p->key != BLANK_LINE) {
			s = p->d + strlen (t);
			if (isspace (*s) || *s == ':' || *s == '=') {
				*s++ = 0;
				p->len = strlen (p->d);
				}
			while (*s && isspace (*s)) s++;
			if (*s == ':' || *s == '=') s++;
			while (*s && isspace (*s)) s++;
			if (*s) {
				q = allocate_item (1);
				q->d = s;
				q->len = strlen (s);
				q->original = 1;
				q->indent = p->indent + 1;
				q->line_number = p->line_number;
				q->key = UNKNOWN_TEXT;
				q->parent = p;
				p->child = q;
				}
			}
		item_count++;
		}

	/*------------------------------------------------------------------*\
	 | Estimate the proper indentation to be assigned to blank lines.	|
	 | At this point the metadata are stored as one doubly-linked list,	|
	 | so you can go forward and backward using p->next and p->prev.	|
	\*------------------------------------------------------------------*/

	for (p=first; p; p=p->next)
		if (p->key == BLANK_LINE) {

			/* Find the most recent element that is not blank */

			for (r=p->prev; r; r=r->prev)
				if (r->key != BLANK_LINE) break;

			/* Find the next element that is not blank */

			for (q=p->next; q; q=q->next)
				if (q->key != BLANK_LINE) break;

			/*----------------------------------------------------------*\
			 | Blanks should be indented like the more indented of		|
			 | (previous nonblank) and (next nonblank).  If either of	|
			 | these nodes does not exist, the proper indentation is	|
			 | whichever does exist.  If neither exist, the file		|
			 | contains only blank lines, so zero will do.				|
			\*----------------------------------------------------------*/

			if (r && q)
				p->indent = ((r->indent > q->indent) ? r->indent : q->indent);
			else
				if (r) p->indent = r->indent;
				else
					if (q) p->indent = q->indent;
					else
						p->indent = 0;
			}

	/*------------------------------------------------------------------*\
	 | Figure out the topology of the file.  Create a list of items		|
	 | as an array of pointers.  For each item, determine whether it	|
	 | has a child and what its parent is.								|
	\*------------------------------------------------------------------*/

	if (list = (struct item **) malloc (item_count * sizeof(struct item *))) {
		memset (list,0,item_count*sizeof(struct item *));
		i = 0;
		for (p=first; p; p=p->next) list[i++] = p;

		for (i=0; i < item_count-1; i++) {
			p = list[i];
			q = list[i+1];

			/*----------------------------------------------------------*\
			 | The relationship between successive items in the file is	|
			 | determined by comparing their levels of indentation.		|
			\*----------------------------------------------------------*/

			j = q->indent - p->indent;

			/*----------------------------------------------------------*\
			 | If q is indented more than p, q is a child of p, and p	|
			 | is parent of q.											|
			 |															|
			 | Modified 30-Jan-1997 to not discard scalar child already	|
			 | in p (can only happen as text following element name.	|
			\*----------------------------------------------------------*/

			if (j > 0) {
				if (allow_indented_text &&
					p->key == UNKNOWN_TEXT && q->key == UNKNOWN_TEXT)
					q->parent = p->parent;
				else {
					if (p->child && (q->key == UNKNOWN_TEXT || q->key == BLANK_LINE)) {

						/*----------------------------------------------*\
						 | Modified 20-Feb-1998 to handle cases where a	|
						 | string immediately follows the name and a	|
						 | blank line follows them.  If the next non-	|
						 | blank is UNKNOWN_TEXT, keep the string that	|
						 | was placed on the same line as the element.	|
						 | If the next non-blank is a recognized		|
						 | element, the string after the element name	|
						 | was not permitted; discard it with a warning.|
						\*----------------------------------------------*/

						if (q->key == BLANK_LINE) {
							/* Point r at the next non-blank element */
							r = NULL;
							for (k=i+1; k < item_count; k++) {
								r = list[k];
								if (r->key != BLANK_LINE) break;
								}
							/* if r is an element, discard the text */
							if (r)
								if (r->key != UNKNOWN_TEXT && r->key != BLANK_LINE) {
									char *fmt = "Warning (line %d of config file): Extraneous text following element %s ignored\n";
									int m = 4 + strlen (fmt) + strlen (TEXT_OF (p->key));
									char *dst;
									if (dst = error_ptr (m))
										sprintf (dst,fmt,p->line_number,TEXT_OF (p->key));
									p->child = q;
									q->prev = NULL;
									}
								else {
									/* otherwise include the text */
									for (r=p->child; r->next; r=r->next);
									r->next = q;
									q->prev = r;
									}
							}
						else {
							/* q->key == UNKNOWN_TEXT */
							for (r=p->child; r->next; r=r->next);
							r->next = q;
							q->prev = r;
							}
						}
					else {
						if (p->child) {
							/* Should this refer to p->child->key? */
							char *fmt = "Warning (line %d of config file): Extraneous text following element %s ignored\n";
							int m = 4 + strlen (fmt) + strlen (TEXT_OF (p->key));
							char *dst;
							if (dst = error_ptr (m))
								sprintf (dst,fmt,p->line_number,TEXT_OF (p->key));
							}
						p->child = q;
						q->prev = NULL;
						}
					q->parent = p;
					p->next = NULL;
					}
				}

			/*----------------------------------------------------------*\
			 | If q is indented the same as p, q is just the next part	|
			 | of the compound type (or text field) of which p is a		|
			 | part.  The parent of q is that of p.						|
			\*----------------------------------------------------------*/

			if (j == 0) {
				q->parent = p->parent;
				}

			/*----------------------------------------------------------*\
			 | If q is indented less than p, q is the next element in a	|
			 | compound type begun earlier.  Find the previous element	|
			 | by searching backwards through the list for the nearest	|
			 | element q that was indented the same as or less than q;	|
			 | store q into its next pointer, set the parent of q to	|
			 | that of r, and sever the link from p to q.				|
			\*----------------------------------------------------------*/

			if (j < 0) {
				for (k=i; k >= 0; k--)
					if (list[k]->indent <= q->indent) {
						r = list[k];
						if (r->indent < q->indent) {
							char *fmt = "Warning: ambiguous indentation at line %d of config file\n";
							int m = 4 + strlen (fmt);
							char *dst;
							if (dst = error_ptr (m))
								sprintf (dst,fmt,q->line_number);
							}
						if (r->next == NULL) {
							r->next = q;
							q->prev = r;
							q->parent = r->parent;
							p->next = NULL;
							}
						else {
							struct item *rr;
							for (rr=r; rr->next && rr->next != q; rr=rr->next);
							rr->next = q;
							q->prev = rr;
							q->parent = rr->parent;
							if (rr != p) p->next = NULL;
							}
						break;
						}

				/*------------------------------------------------------*\
				 | If you fell through that loop, there is an error in	|
				 | the file's indentation that cannot easily be over-	|
				 | looked.  This indicates that a line somewhere in the	|
				 | middle of the file is indented less than all other	|
				 | lines in the file; the structure is unclear.			|
				\*------------------------------------------------------*/

				if (k < 0) {
					char *fmt = "Error: too little indentation at line %d of config file\n";
					int m = 4 + strlen (fmt);
					char *dst;
					if (dst = error_ptr (m))
						sprintf (dst,fmt,q->line_number);
					}
				}
			}

		/*--------------------------------------------------------------*\
		 | Traverse the tree, looking for situations in which scalar	|
		 | values have children that are themselves scalar values; in	|
		 | such cases, make them siblings, not children.				|
		\*--------------------------------------------------------------*/

		if (allow_indented_text) equalize_indented_scalars (first);

		free (list);
		}
	else {
		char *fmt = "Error: could not allocate array of pointers to items\n";
		int m = strlen (fmt);
		char *dst;
		if (dst = error_ptr (m))
			sprintf (dst,fmt);
		exit (1);
		}
	return (first);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

struct item *parse_config (char *input_file) {
	char *txt = read_text (input_file);
	if (txt == NULL) {
		char *fmt = "Error: could not open config file %s\n";
		int m = strlen (fmt) + strlen (input_file);
		char *dst;
		if (dst = error_ptr (m))
			sprintf (dst,fmt,input_file);
		return (NULL);
		}
	else
		return (decode_tree (decode_text (txt)));
	}

/*----------------------------------------------------------------------*\
 | Locate a node in the tree by matching the key.  The first argument	|
 | is the root node of the tree to be searched, the second is the key	|
 | to be matched.														|
\*----------------------------------------------------------------------*/

static struct item *find_option_key (struct item *p, enum config_word key) {
	struct item *q,*r;
	if (p->key == key) return (p);
	for (q=p->child; q; q=q->next)
		if (r = find_option_key (q,key)) return (r);
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Locate the next node in a list that matches a key.  First argument	|
 | is the current node; it will not be considered in the search.  This	|
 | routine searches only nodes at the same level in the tree as the		|
 | node given by the first argument.  It does not search children.		|
\*----------------------------------------------------------------------*/

static struct item *find_next_option_key (struct item *p, enum config_word key) {
	p = p->next;
	while (p) {
		if (p->key == key) return (p);
		p = p->next;
		}
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Locate the first child node in a list that matches a key.  The first	|
 | argument is the current node; its child list will searched.  This	|
 | routine searches only nodes that are children of the first argument.	|
\*----------------------------------------------------------------------*/

static struct item *find_child_option_key (struct item *p, enum config_word key) {
	p = p->child;
	while (p) {
		if (p->key == key) return (p);
		p = p->next;
		}
	return (NULL);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static struct item *first = NULL;

static void list_children (struct item *p) {
	while (p) {
		fprintf (stderr,"%s ",config_text_of(p->key));
		if (p->parent) fprintf (stderr,"(%s) ",config_text_of(p->parent->key));
		else fprintf (stderr,"(root) ");
		if (p->key == Cunknown)
			if (p->d) fprintf (stderr,"= \"%s\"\n",p->d);
			else fprintf (stderr,"= NULL\n");
		else fprintf (stderr,"\n");
		if (p->child) list_children (p->child);
		p = p->next;
		}
	}

static void unify_scalars (struct item *p) {
	struct item *q,*r;
	char *dst;

	while (p) {
		if (q = p->child)
			if (q->key == Cunknown) {
				if (r = q->next) {

					/*--------------------------------------------------*\
					 | This is a multi-line text value.  Find out how	|
					 | long each line is first, then allocate a block	|
					 | big enough to hold them all, and copy them to it	|
					\*--------------------------------------------------*/

					int n = 0;
					for (r=q; r; r=r->next)
						if (r->d) n += 1 + strlen (r->d);
						else n++;

					if (dst = malloc (1+n)) {
						*dst = 0;
						for (r=q; r; r=r->next) {
							if (r->d) strcat (dst,r->d);
							strcat (dst,"\n");	/* should this be " "? */
							}
						q->d = dst;
						q->len = 1 + n;
						q->original = 0;
						q->next = NULL;	/* should deallocate siblings, but this'll be okay for now */
						}
					else {
						/*----------------------------------------------*\
						 | Out of memory?  You're hosed, but there's no	|
						 | reason why you need to bomb out here.		|
						\*----------------------------------------------*/
						}
					}
				else {
					/* only one Cunknown; keep as is */
					}
				}
			else
				unify_scalars (q);	/* q is compound; check its children */
		p = p->next;
		}
	}

void read_configuration (char *config_file) {
	struct item *p,*q;

	if (first = parse_config (config_file)) {

		/* Combine text values into single elements */

		unify_scalars (first);

		/* Create a dummy element to be the parent of the whole tree */

		if (first->key != CROOT) {
			p = allocate_item (1);
			p->d = NULL;
			p->len = 0;
			p->original = 0;
			p->indent = 0;
			p->line_number = -1;
			p->key = CROOT;
			p->parent = NULL;
			p->child = first;
			for (q=p->child; q; q=q->next)
				q->parent = p;
			first = p;
			}

		/* Dump the parse tree  (debugging only) */

		/* list_children (first); */
		}
	}

/*----------------------------------------------------------------------*\
 | Three external functions are defined here:							|
 |																		|
 | p = find_option (q,"config_word");									|
 |																		|
 |	which uses q = first if q == NULL on entry,							|
 |	and returns q if q->key == key_of_config_word ("config_word")		|
 |																		|
 | p = find_next_option (q,"config_word");								|
 |																		|
 |	which skips q and looks only along the list q=q->next...			|
 |																		|
 | s = text_of_option (p);												|
 |																		|
 |	which returns p->child->d, not p->d.								|
\*----------------------------------------------------------------------*/

struct item *find_option (struct item *q, char *s) {
	enum config_word key;
	struct item *p;

	/* If you don't have a config data base, return NULL */

	if (!first) return (NULL);

	/* Find the key corresponding to the second argument */

	key = config_key_of (s);
	if (key == Cnull || key == Cunknown) return (NULL);

	/* Find the node corresponding to the first argument */

	if (!q) q = first;

	/* Find the first node in the subtree headed by q whose key matches s */

	p = find_option_key (q,key);

	return (p);
	}

struct item *find_child_option (struct item *q, char *s) {
	enum config_word key;
	struct item *p;

	/* If you don't have a config data base, return NULL */

	if (!first) return (NULL);

	/* Find the key corresponding to the second argument */

	key = config_key_of (s);
	if (key == Cnull || key == Cunknown) return (NULL);

	/* Find the node corresponding to the first argument */

	if (!q) q = first;

	/* Find the first node in the subtree headed by q whose key matches s */

	p = find_child_option_key (q,key);

	return (p);
	}

struct item *find_next_option (struct item *q, char *s) {
	enum config_word key;
	struct item *p;

	/* If you don't have a config data base, return NULL */

	if (!first) return (NULL);

	/* Find the key corresponding to the second argument */

	key = config_key_of (s);
	if (key == Cnull || key == Cunknown) return (NULL);

	/* Find the node corresponding to the first argument */

	if (!q) q = first;

	/* Find the next node in the subtree headed by q whose key matches s */

	p = find_next_option_key (q,key);

	return (p);
	}

char *text_of_option (struct item *p) {
	if (p && p->child) return (p->child->d);
	else return (NULL);
	}

char *name_of_option (struct item *p) {
	if (p)
		return (config_text_of (p->key));
	else
		return (NULL);
	}

static int in_config_tree (struct item *p, struct item *target) {
	struct item *q;
	int found = 0;

	if (p)
		if (p == target) found = 1;
		else
			for (q=p->child; q; q=q->next)
				if (in_config_tree (q,target)) {
					found = 1;
					break;
					}
	return (found);
	}

int in_config (struct item *target) {
	return (in_config_tree (first,target));
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
