# Explorer for DBF files
#
# Reads a DBF file and optionally a metadata record
# Shows fields and distinct values of the DBF
# Shows detailed metadata, if available, for each field and
# field value.
#
# Uses tcldbf and mq extensions (by me) to read DBF file and
# metadata.  tcldbf is a Tcl interface to Frank Warmerdam's
# dbf access routines in his shapefile library shapelib.
# mq is a Tcl interface to code used by mp and Tkme.
#
# Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)

package require Tk
package require dbf
package require mq

set version 1.0

set dbf_file {}
set met_file {}

set changed 0

# DBF handle
set d {}

# Metadata handle
set mf {}

set bgcolor "#bbddff"

#-------------------------------------------------------------------------
# Font setup
#-------------------------------------------------------------------------

if {[string compare $tcl_platform(platform) windows] == 0} {
	set font_face(list)		terminal
	set font_size(list)		9
	set font_bold(list)		0
	set font_italic(list)	0

	set font_face(value)	courier
	set font_size(value)	9
	set font_bold(value)	0
	set font_italic(value)	0

	set font_face(menu)		arial
	set font_size(menu)		9
	set font_bold(menu)		0
	set font_italic(menu)	0

	set font_face(output)	courier
	set font_size(output)	9
	set font_bold(output)	0
	set font_italic(output)	0

	set font_face(help)		arial
	set font_size(help)		9
	set font_bold(help)		0
	set font_italic(help)	0
	} \
else {
	set font_face(list)		fixed
	set font_size(list)		12
	set font_bold(list)		0
	set font_italic(list)	0

	set font_face(value)	fixed
	set font_size(value)	12
	set font_bold(value)	0
	set font_italic(value)	0

	set font_face(menu)		helvetica
	set font_size(menu)		12
	set font_bold(menu)		1
	set font_italic(menu)	0

	set font_face(output)	fixed
	set font_size(output)	12
	set font_bold(output)	0
	set font_italic(output)	0

	set font_face(help)		helvetica
	set font_size(help)		12
	set font_bold(help)		0
	set font_italic(help)	0
	}

proc font_apply {task} {
	global font_face font_size font_bold font_italic
	set weight(0) normal
	set weight(1) bold
	set slant(0) roman
	set slant(1) italic
	foreach part {list value menu output help} {
		set font_name $part
		append font_name _font
		set cmd [list font $task $font_name -family $font_face($part) -size $font_size($part) -weight $weight($font_bold($part)) -slant $slant($font_italic($part))]
		eval $cmd
		}
	}

font_apply create

#-------------------------------------------------------------------------
# A Notebook widget for Tcl/Tk
# $Revision: 1.3 $
#
# Copyright (C) 1996,1997,1998 D. Richard Hipp
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA  02111-1307, USA.
#
# Author contact information:
#   drh@acm.org
#   http://www.hwaci.com/drh/

#
# Create a new notebook widget
#
proc Notebook:create {w args} {
  global Notebook
  set Notebook($w,width) 400
  set Notebook($w,height) 300
  set Notebook($w,pages) {}
  set Notebook($w,top) 0
  set Notebook($w,pad) 5
  set Notebook($w,fg,on) black
  set Notebook($w,fg,off) grey50
  canvas $w -bd 0 -highlightthickness 0 -takefocus 0
  set Notebook($w,bg) [$w cget -bg]
  bind $w <1> "Notebook:click $w %x %y"
  bind $w <Configure> "Notebook:scheduleExpand $w"
  eval Notebook:config $w $args
}

#
# Change configuration options for the notebook widget
#
proc Notebook:config {w args} {
  global Notebook
  foreach {tag value} $args {
    switch -- $tag {
      -width {
        set Notebook($w,width) $value
      }
      -height {
        set Notebook($w,height) $value
      }
      -pages {
        set Notebook($w,pages) $value
      }
      -pad {
        set Notebook($w,pad) $value
      }
      -bg {
        set Notebook($w,bg) $value
      }
      -fg {
        set Notebook($w,fg,on) $value
      }
      -disabledforeground {
        set Notebook($w,fg,off) $value
      }
    }
  }

  #
  # After getting new configuration values, reconstruct the widget
  #
  $w delete all
  set Notebook($w,x1) $Notebook($w,pad)
  set Notebook($w,x2) [expr $Notebook($w,x1)+2]
  set Notebook($w,x3) [expr $Notebook($w,x2)+$Notebook($w,width)]
  set Notebook($w,x4) [expr $Notebook($w,x3)+2]
  set Notebook($w,y1) [expr $Notebook($w,pad)+2]
  set Notebook($w,y2) [expr $Notebook($w,y1)+2]
  set Notebook($w,y5) [expr $Notebook($w,y1)+30]
  set Notebook($w,y6) [expr $Notebook($w,y5)+2]
  set Notebook($w,y3) [expr $Notebook($w,y6)+$Notebook($w,height)]
  set Notebook($w,y4) [expr $Notebook($w,y3)+2]
  set x $Notebook($w,x1)
  set cnt 0
  set y7 [expr $Notebook($w,y1)+10]
  foreach p $Notebook($w,pages) {
    set Notebook($w,p$cnt,x5) $x
    set id [$w create text 0 0 -text $p -anchor nw -tags "p$cnt t$cnt"]
    set bbox [$w bbox $id]
    set width [lindex $bbox 2]
    $w move $id [expr $x+10] $y7
    $w create line \
       $x $Notebook($w,y5)\
       $x $Notebook($w,y2) \
       [expr $x+2] $Notebook($w,y1) \
       [expr $x+$width+16] $Notebook($w,y1) \
       -width 2 -fill white -tags p$cnt
    $w create line \
       [expr $x+$width+16] $Notebook($w,y1) \
       [expr $x+$width+18] $Notebook($w,y2) \
       [expr $x+$width+18] $Notebook($w,y5) \
       -width 2 -fill black -tags p$cnt
    set x [expr $x+$width+20]
    set Notebook($w,p$cnt,x6) [expr $x-2]
    if {![winfo exists $w.f$cnt]} {
      frame $w.f$cnt -bd 0
    }
    $w.f$cnt config -bg $Notebook($w,bg)
    place $w.f$cnt -x $Notebook($w,x2) -y $Notebook($w,y6) \
      -width $Notebook($w,width) -height $Notebook($w,height)
    incr cnt
  }
  $w create line \
     $Notebook($w,x1) [expr $Notebook($w,y5)-2] \
     $Notebook($w,x1) $Notebook($w,y3) \
     -width 2 -fill white
  $w create line \
     $Notebook($w,x1) $Notebook($w,y3) \
     $Notebook($w,x2) $Notebook($w,y4) \
     $Notebook($w,x3) $Notebook($w,y4) \
     $Notebook($w,x4) $Notebook($w,y3) \
     $Notebook($w,x4) $Notebook($w,y6) \
     $Notebook($w,x3) $Notebook($w,y5) \
     -width 2 -fill black
  $w config -width [expr $Notebook($w,x4)+$Notebook($w,pad)] \
            -height [expr $Notebook($w,y4)+$Notebook($w,pad)] \
            -bg $Notebook($w,bg)
  set top $Notebook($w,top)
  set Notebook($w,top) -1
  Notebook:raise.page $w $top
}

proc Notebook:which {w} {
	global Notebook
	return "[lindex $Notebook($w,pages) $Notebook($w,top)]"
	}

#
# This routine is called whenever the mouse-button is pressed over
# the notebook.  It determines if any page should be raised and raises
# that page.
#
proc Notebook:click {w x y} {
  global Notebook
  if {$y<$Notebook($w,y1) || $y>$Notebook($w,y6)} return
  set N [llength $Notebook($w,pages)]
  for {set i 0} {$i<$N} {incr i} {
    if {$x>=$Notebook($w,p$i,x5) && $x<=$Notebook($w,p$i,x6)} {
      Notebook:raise.page $w $i
      break
    }
  }
}

#
# For internal use only.  This procedure raised the n-th page of
# the notebook
#
proc Notebook:raise.page {w n} {
  global Notebook
  if {$n<0 || $n>=[llength $Notebook($w,pages)]} return
  set top $Notebook($w,top)
  if {$top>=0 && $top<[llength $Notebook($w,pages)]} {
    $w move p$top 0 2
  }
  $w move p$n 0 -2
  $w delete topline
  if {$n>0} {
    $w create line \
       $Notebook($w,x1) $Notebook($w,y6) \
       $Notebook($w,x2) $Notebook($w,y5) \
       $Notebook($w,p$n,x5) $Notebook($w,y5) \
       $Notebook($w,p$n,x5) [expr $Notebook($w,y5)-2] \
       -width 2 -fill white -tags topline
  }
  $w create line \
    $Notebook($w,p$n,x6) [expr $Notebook($w,y5)-2] \
    $Notebook($w,p$n,x6) $Notebook($w,y5) \
    -width 2 -fill white -tags topline
  $w create line \
    $Notebook($w,p$n,x6) $Notebook($w,y5) \
    $Notebook($w,x3) $Notebook($w,y5) \
    -width 2 -fill white -tags topline
  set Notebook($w,top) $n
  raise $w.f$n
}

#
# Change the page-specific configuration options for the notebook
#
proc Notebook:pageconfig {w name args} {
  global Notebook
  set i [lsearch $Notebook($w,pages) $name]
  if {$i<0} return
  foreach {tag value} $args {
    switch -- $tag {
      -state {
        if {"$value"=="disabled"} {
          $w itemconfig t$i -fg $Notebook($w,fg,off)
        } else {
          $w itemconfig t$i -fg $Notebook($w,fg,on)
        }
      }
      -onexit {
        set Notebook($w,p$i,onexit) $value
      }
    }
  }
}

#
# This procedure raises a notebook page given its name.  But first
# we check the "onexit" procedure for the current page (if any) and
# if it returns false, we don't allow the raise to proceed.
#
proc Notebook:raise {w name} {
  global Notebook
  set i [lsearch $Notebook($w,pages) $name]
  if {$i<0} return
  if {[info exists Notebook($w,p$i,onexit)]} {
    set onexit $Notebook($w,p$i,onexit)
    if {"$onexit"!="" && [eval uplevel #0 $onexit]!=0} {
      Notebook:raise.page $w $i
    }
  } else {
    Notebook:raise.page $w $i
  }
}

#
# Return the frame associated with a given page of the notebook.
#
proc Notebook:frame {w name} {
  global Notebook
  set i [lsearch $Notebook($w,pages) $name]
  if {$i>=0} {
    return $w.f$i
  } else {
    return {}
  }
}

#
# Try to resize the notebook to the next time we become idle.
#
proc Notebook:scheduleExpand w {
  global Notebook
  if {[info exists Notebook($w,expand)]} return
  set Notebook($w,expand) 1
  after idle "Notebook:expand $w"
}

#
# Resize the notebook to fit inside its containing widget.
#
proc Notebook:expand w {
  global Notebook
  set wi [expr [winfo width $w]-($Notebook($w,pad)*2+4)]
  set hi [expr [winfo height $w]-($Notebook($w,pad)*2+36)]
  Notebook:config $w -width $wi -height $hi
  catch {unset Notebook($w,expand)}
}
# End of code from D. Richard Hipp
#-------------------------------------------------------------------------

#-------------------------------------------------------------------------
# Older simple widgets from Brent Welch's book (as used in Tkme)
#-------------------------------------------------------------------------
# Create a scrolled listbox widget
# Modified from Brent Welch's book

proc ScrolledListbox { f args } {
	# Create listbox attached to scrollbars, pass thru $args
	frame $f
	eval {listbox $f.list \
		-selectmode "single" \
		-yscrollcommand [list $f.sy set] \
		-relief flat \
		} $args
	# Create scrollbars attached to the listbox
	scrollbar $f.sy -orient vertical -width 10 -command [list $f.list yview]
	# Arrange them in the parent frame
	pack $f.sy -side right -fill y
	pack $f.list -side left -fill both -expand true
	return $f
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Create a scrolled text widget
# Modified from Brent Welch's book

proc ScrolledText { f width height } {
	frame $f
	# The setgrid setting allows the window to be resized.
	text $f.text -width $width -height $height \
		-setgrid true -wrap none -padx 2 \
		-yscrollcommand [list $f.yscroll set]
	scrollbar $f.yscroll -orient vertical -width 10 -command [list $f.text yview]
	pack $f.yscroll -side right -fill y
	# The fill and expand are needed when resizing.
	pack $f.text -side left -fill both -expand true
	pack $f -side top -fill both -expand true
	return $f
	}

#-------------------------------------------------------------------------
# Procedures
#-------------------------------------------------------------------------

set diag {}
set undocumented_field {}

proc show_diagnostics {} {

	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	# If we have already read the DBF and metadata files, look through
	# the field list and determine whether we have any fields defined
	# in the metadata that are not present in the DBF
	# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	global d mf
	global attribute_label attribute_definition
	global undocumented_field

	if {[llength $d] > 0 && [llength $mf] > 0} {

		set field_name         {}
		set nonexistent_field  {}
		set unknown_field      {}
		set undocumented_field {}

		set field_info [$d fields]
		foreach item $field_info {
			lappend field_name "[lindex $item 0]"
			}
		foreach item [array names attribute_label] {
			if {[lsearch -exact $field_name $item] == -1} {
				# The attribute label does not match the name of a field in this dbf file
				lappend nonexistent_field "$item"
				}
			}
		foreach item $field_name {
			if {[llength [array names attribute_label $item]] == 0} {
				lappend unknown_field "$item"
				} \
			else {
				if {[llength [array names attribute_definition $item]] == 0} {
					lappend undocumented_field "$item"
					}
				}
			}
		if {[llength $nonexistent_field] > 0 ||
			[llength $unknown_field] > 0 ||
			[llength $undocumented_field] > 0} {

			global diag
			global bgcolor

			if {[llength $diag] == 0} {
				set diag [toplevel .diag]

				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				# Menu bar needs only allow user to close the window
				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -

				frame $diag.mbar -relief flat -borderwidth 2
				$diag.mbar configure -background $bgcolor
				button $diag.mbar.close -text "Done" -font menu_font -command {diag_done}
				$diag.mbar.close configure -background $bgcolor
				pack $diag.mbar.close -side left
				pack $diag.mbar -side top -expand false -fill x

				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				# Frame to hold the diagnostic messages
				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -

				frame $diag.f -background $bgcolor
				pack $diag.f -side top -fill both -expand true

				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				# Extra: fields in the metadata but not in the DBF
				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -

				labelframe $diag.f.1 -text "Extra" -background $bgcolor
				label $diag.f.1.explanation -justify left -text "Fields in the metadata\nbut not in the DBF" -background $bgcolor
				set w [ScrolledListbox $diag.f.1.listbox]
				$w configure -width 128
				set extra_list $w.list
				$extra_list configure -activestyle none
				$extra_list configure -selectmode single
				# button $diag.f.1.action -text "Remove" -command "drop_attributes $nonexistent_field"
				button $diag.mbar.remove -text "Remove extras" -command [list drop_attributes $nonexistent_field $extra_list]
				$diag.mbar.remove configure -background $bgcolor -font menu_font

				pack $diag.f.1.explanation -side top -padx 4 -pady 4
				foreach item $nonexistent_field {
					# pack $diag.f.1.action -side top
					pack $diag.mbar.remove -side left
					$extra_list insert end "$item"
					}
				pack $w -side top -fill y -expand true -padx 4 -pady 4

				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				# Missing: fields in the DBF but not in the metadata
				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -

				labelframe $diag.f.2 -text "Missing" -background $bgcolor
				label $diag.f.2.explanation -justify left -text "Fields in the DBF\nbut not in the metadata" -background $bgcolor
				set w [ScrolledListbox $diag.f.2.listbox]
				$w configure -width 128
				set missing_list $w.list
				$missing_list configure -activestyle none
				$missing_list configure -selectmode single
				# button $diag.f.2.action -text "Add" -command "add_attributes $unknown_field"
				button $diag.mbar.add -text "Add missing" -command [list add_attributes $unknown_field $missing_list]
				$diag.mbar.add configure -background $bgcolor -font menu_font

				pack $diag.f.2.explanation -side top  -padx 4 -pady 4
				foreach item $unknown_field {
					# pack $diag.f.2.action -side top
					pack $diag.mbar.add -side left
					$missing_list insert end "$item"
					}
				pack $w -side top -fill y -expand true -padx 4 -pady 4

				menu $missing_list.menu -tearoff false -font menu_font -title "Missing" -disabledforeground Navy
				$missing_list.menu add command -label "Missing field" -command {} -state disabled
				$missing_list.menu add command -label "Add to metadata" -command [list add_one_attribute $missing_list]
				$missing_list.menu add command -label "Ignore" -command [list ignore_attribute $missing_list]

				bind $missing_list <ButtonPress-3> [list tk_popup $missing_list.menu %X %Y 0]

				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				# Undefined: Attributes in the metadata that aren't defined
				# - - - - - - - - - - - - - - - - - - - - - - - - - - - -

				labelframe $diag.f.3 -text "Undefined" -background $bgcolor
				label $diag.f.3.explanation -justify left -text "Attributes you have\nnot defined yet" -background $bgcolor

				pack $diag.f.3.explanation -side top -anchor w -expand false -padx 4 -pady 4

				set sw [canvas $diag.f.3.sw -width 10 -height 10 -yscrollcommand [list $diag.f.3.yscroll set]]
				$sw configure -background $bgcolor
				scrollbar $diag.f.3.yscroll -orient vertical -command [list $diag.f.3.sw yview]
				pack $diag.f.3.yscroll -side right -fill y
				pack $sw -side top -fill both -expand true -padx 4 -pady 4

				set w [frame $sw.f -bd 0]
				$sw create window 0 0 -anchor nw -window $w

				$w configure -background $bgcolor -relief flat
				set k 0
				foreach item $undocumented_field {
					set the_label [label $w.label_$k -text "$item"]
					$the_label configure -font value_font -justify left -background $bgcolor
					set the_entry [entry $w.entry_$k -textvariable attribute_definition($item)]
					$the_entry configure -font value_font -width 60
					grid $the_label $the_entry
					grid $the_label -sticky w -padx 4
					incr k
					}

				pack $diag.f.1 -side left -fill y -expand 1 -padx 4 -pady 4
				pack $diag.f.2 -side left -fill y -expand 1 -padx 4 -pady 4
				pack $diag.f.3 -side left -fill both -expand 1 -padx 4 -pady 4

				if {$k > 0} {
					tkwait visibility $w.entry_0
					set bbox [grid bbox $w 0 0]
					set incr [lindex $bbox 3]
					set width [winfo reqwidth $w]
					set height [winfo reqheight $w]
					$sw configure -scrollregion "0 0 $width $height"
					$sw configure -yscrollincrement $incr
					set height [expr $incr * 10]
					$sw configure -width $width -height $height
					}

				wm title .diag "Warnings"

				# Position and size the window nicely
				# set the window's height to no more than that of the main window
				# (it is resizable, so the user can enlarge it)
				# set the windows upper left corner to a position to the right and
				# down from the main window's upper left corner.

				update
				set diag_geometry [wm geometry $diag]
				scan $diag_geometry   "%dx%d+%d+%d" dw dh dx dy

				set root_geometry [wm geometry .]
				scan $root_geometry "%dx%d+%d+%d" rw rh rx ry

				set dx [expr $rx + 32]
				set dy [expr $ry + 32]
				if {$dh > $rh} {set dh $rh}
				if {$dh < 256} {set dh 256}
				wm geometry $diag "${dw}x${dh}+${dx}+${dy}"
				}
			}
		}
	}

proc ignore_attribute {w} {
	if {[llength [$w curselection]] > 0} {
		$w delete [$w index active]
		}
	}

proc diag_done {} {
	global mf changed
	global attribute_definition attribute_definition_source
	global diag undocumented_field

	if {[llength $mf] == 0} {
		create_metadata
		}

	if {[set p [$mf find_first Metadata]]} {
		if {[set p [$mf find_in $p Entity_and_Attribute_Information]]} {
			if {[set q [$mf find_in $p Detailed_Description]]} {
				foreach label $undocumented_field {
					if {[string length $attribute_definition($label)] > 0} {
						if {[set r [$mf find_in $q Attribute]]} {
							while {$r} {
								if {[set u [$mf find_in $r Attribute_Label]]} {
									set s [$mf value_of -nonewline $u]
									if {[string compare -nocase $s $label] == 0} {
										if {[set v [$mf find_in $r Attribute_Definition]]} {
											$mf value_set $v "$attribute_definition($label)"
puts "Updated Attribute_Definition for $label"
											set changed 1
											} \
										else {
											if {[set v [$mf insert -child $r Attribute_Definition]]} {
puts "Inserted Attribute_Definition for $label"
												$mf value_set $v "$attribute_definition($label)"
puts "Assigned Attribute_Definition for $label"
												set changed 1
												}
											}
										}
									}
								set r [$mf find_next $r]
								}
							}
						}
					}
				}
			}
		}
	wm withdraw .diag
	catch {unset undocumented_field}
	}

#-------------------------------------------------------------------------
# Metadata file handling
#-------------------------------------------------------------------------

proc open_metadata {} {
	global mf met_file
	global attribute_label attribute_definition attribute_definition_source
	global enumerated_domain
	global codeset_name codeset_source
	global rdommin rdommax attrunit attrmres
	global unrepresentable_domain

	if {[metadata m -parse $met_file]} {
		if {[set p [$m find_first Entity_and_Attribute_Information]]} {
			if {[set p [$m find_in $p Detailed_Description]]} {
				# We currently assume that information should be taken
				# only from the first Detailed_Description in the file.
				#
				# If you want to get attribute information from every
				# Detailed_Description, uncomment the next line
				# as well as the commented "set p" and closing brace
				# at the end of this loop.
				# while {$p} {
					if {[set q [$m find_in $p Attribute]]} {
						set k 0
						while {$q} {
							if {[set r [$m find_in $q Attribute_Label]]} {
								set s [$m value_of -nonewline $r]
								if {[string length $s] > 0} {
									set label [string toupper $s]
									set label [string map {- _ % _ . _} $label]
									set attribute_label($label) "$s"
									if {[set u [$m find_in $q Attribute_Definition]]} {
										set s [$m value_of $u]
										if {[string length $s] > 0} {
											set attribute_definition($label) "$s"
											}
										}
									if {[set u [$m find_in $q Attribute_Definition_Source]]} {
										set s [$m value_of $u]
										if {[string length $s] > 0} {
											set attribute_definition_source($label) "$s"
											}
										}
									}
								}
							if {[set r [$m find_in $q Attribute_Domain_Values]]} {
								while {$r} {
									if {[set u [$m find_in $r Enumerated_Domain]]} {
										while {$u} {
											if {[set v [$m find_in $u Enumerated_Domain_Value]]} {
												set value [$m value_of -nonewline $v]
												if {[string length $value] > 0} {
													if {[set w [$m find_in $u Enumerated_Domain_Value_Definition]]} {
														set meaning [$m value_of $w]
														set enumerated_domain($label,$value) "$meaning"
														}
													}
												}
											set u [$m find_next $u]
											}
										}
									if {[set u [$m find_in $r Range_Domain]]} {
										set rdommin($label)  ""
										set rdommax($label)  ""
										set attrunit($label) ""
										set attrmres($label) ""

										if {[set v [$m find_in $u Range_Domain_Minimum]]} {
											set rdommin($label) [$m value_of -nonewline $v]
											}
										if {[set v [$m find_in $u Range_Domain_Maximum]]} {
											set rdommax($label) [$m value_of -nonewline $v]
											}
										if {[set v [$m find_in $u Attribute_Units_of_Measure]]} {
											set attrunit($label) [$m value_of -nonewline $v]
											}
										if {[set v [$m find_in $u Attribute_Measurement_Resolution]]} {
											set attrmres($label) [$m value_of -nonewline $v]
											}
										}
									if {[set u [$m find_in $r Codeset_Domain]]} {
										set codeset_name($label)  ""
										set codeset_source($label)  ""

										if {[set v [$m find_in $u Codeset_Name]]} {
											set codeset_name($label) [$m value_of -nonewline $v]
											}
										if {[set v [$m find_in $u Codeset_Source]]} {
											set codeset_source($label) [$m value_of -nonewline $v]
											}
										}
									if {[set u [$m find_in $r Unrepresentable_Domain]]} {
										set unrepresentable_domain($label) [$m value_of $u]
										}
									set r [$m find_next $r]
									}
								}
							set q [$m find_next $q]
							}
						}
					# If you want to get attribute information from every
					# Detailed_Description, uncomment the next two lines
					# as well as the commented "while" statement above
					# set p [$m find_next $p]
					# }
				}
			set mf $m

			# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
			# Make a backup of the metadata file
			# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

			set ext  [file extension $met_file]
			set backup $met_file.bak

			set fmt "text"
			if {[string compare -nocase $ext ".xml" ] == 0} {set fmt "xml" }
			if {[string compare -nocase $ext ".sgm" ] == 0} {set fmt "sgml"}
			if {[string compare -nocase $ext ".sgml"] == 0} {set fmt "sgml"}

			$mf write -format $fmt $backup
			puts "A copy of the metadata has been saved as $backup"
			}

		show_diagnostics
		}
	}

proc create_metadata {} {
	global mf d dbf_file changed

	if {$dbf_file == ""} return

	set layer_name [file tail [file rootname $dbf_file]]

	if {[metadata mf -create]} {
		if {[set p [$mf find_first Metadata]]} {
			if {[set p [$mf insert -child $p Entity_and_Attribute_Information]]} {
				if {[set p [$mf insert -child $p Detailed_Description]]} {
					if {[set q [$mf insert -child $p Entity_Type]]} {
						if {[set r [$mf insert -child $q Entity_Type_Label]]} {
							$mf value_set $r "$layer_name"
							}
						if {[set r [$mf insert -child $q Entity_Type_Definition]]} {
							# It would be nice to know what to put here
							}
						}
					foreach attr [$d fields] {
						set name   [lindex $attr 0]
						set type   [lindex $attr 1]
						set native [lindex $attr 2]
						set width  [lindex $attr 3]
						set prec   [lindex $attr 4]
						if {[llength $attr] > 5} {
							set definition [lindex $attr 5]
							} \
						else {
							set definition "($type of width $width"
							if {$native == "N"} {
								append definition " and precision $prec"
								}
							append definition ")"
							}
						if {[set q [$mf insert -child $p Attribute]]} {
							if {[set r [$mf insert -child $q Attribute_Label]]} {
								$mf value_set $r "$name"
								}
							if {[set r [$mf insert -child $q Attribute_Definition]]} {
								$mf value_set $r "$definition"
								}

							if {[set r [$mf insert -child $q Attribute_Domain_Values]]} {
								}
							}
						}
					}
				}
			}
puts "Metadata created"
		set changed 1
		}
	}

proc edit_edom {} {
	global d enumerated_domain
	global current_field
	global bgcolor

	set field $current_field

	if {[llength $d] > 0} {
		set distinct_values [lsort -dictionary -unique [$d values $field]]

		set max_val 100
		set distinct_count [llength $distinct_values]
		if {$distinct_count > $max_val} {
			tk_messageBox -type ok -message "You can only edit $max_val distinct values at once; this field has " -icon warning
			return
			}

		set edom_edit [toplevel .edom_edit]
		$edom_edit configure -background $bgcolor

		# - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		# Menu bar needs only allow user to close the window
		# - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		frame $edom_edit.mbar -relief flat -borderwidth 2
		$edom_edit.mbar configure -background $bgcolor
		button $edom_edit.mbar.close -text "Done" -font menu_font -command "edit_edom_done $field; wm withdraw .edom_edit"
		$edom_edit.mbar.close configure -background $bgcolor
		pack $edom_edit.mbar.close -side left
		pack $edom_edit.mbar -side top -expand false -fill x

		labelframe $edom_edit.f -text "Enumerated_Domain" -background $bgcolor

		set sw [canvas $edom_edit.f.sw -width 10 -height 10 -yscrollcommand [list $edom_edit.f.yscroll set]]
		$sw configure -background $bgcolor
		scrollbar $edom_edit.f.yscroll -orient vertical -command [list $edom_edit.f.sw yview]
		pack $edom_edit.f.yscroll -side right -fill y
		pack $sw -side top -fill both -expand true -padx 4 -pady 4

		set w [frame $sw.f -bd 0]
		$sw create window 0 0 -anchor nw -window $w

		$w configure -background $bgcolor -relief flat

		set k 0
		foreach item $distinct_values {
			if {[string length $item] == 0} {
				set item "(empty)"
				}
			set the_label [label $w.label_$k -text "$item"]
			$the_label configure -font value_font -justify left -background $bgcolor
			set the_entry [entry $w.entry_$k -textvariable enumerated_domain($field,$item)]
			$the_entry configure -font value_font -width 64
			grid $the_label $the_entry
			grid $the_label -sticky w  -padx 4
			grid $the_entry -sticky ew -padx 4
			incr k
			}

		pack $edom_edit.f -side top -fill both -expand 1 -padx 4 -pady 4

		if {$k > 0} {
			tkwait visibility $w.entry_0
			set bbox [grid bbox $w 0 0]
			set incr [lindex $bbox 3]
			set width [winfo reqwidth $w]
			set height [winfo reqheight $w]
			$sw configure -scrollregion "0 0 $width $height"
			$sw configure -yscrollincrement $incr
			set height [expr $incr * 10]
			$sw configure -width $width -height $height
			}

		wm title .edom_edit "Edit Enumerated_Domain for $field"

		# Position and size the window nicely
		# set the window's height to no more than that of the main window
		# (it is resizable, so the user can enlarge it)
		# set the windows upper left corner to a position to the right and
		# down from the main window's upper left corner.

		update
		set ee_geometry [wm geometry $edom_edit]
		scan $ee_geometry   "%dx%d+%d+%d" dw dh dx dy

		set root_geometry [wm geometry .]
		scan $root_geometry "%dx%d+%d+%d" rw rh rx ry

		set dx [expr $rx + 32]
		set dy [expr $ry + 32]
		if {$dh > $rh} {set dh $rh}
		if {$dh < 256} {set dh 256}
		wm geometry $edom_edit "${dw}x${dh}+${dx}+${dy}"
		}
	}

proc edit_edom_done {field} {
	global d mf enumerated_domain

	if {[llength $d] > 0 && [llength $mf] > 0} {
		set distinct_values [lsort -dictionary -unique [$d values $field]]
		foreach value $distinct_values {
			if {[llength [array names enumerated_domain "$field,$value"]] > 0} {
				update_edom $field $value $enumerated_domain($field,$value)
				}
			}
		}
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Given a field, value, and meaning, update the Enumerated_Domain
# element that matches the field and value.  If there is no such
# Enumerated_Domain element, create one and put this entry into it.
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc update_edom {field value meaning} {
	global mf changed

	if {[llength $mf] == 0} {
		create_metadata
		}

	if {[set p [$mf find_first Metadata]]} {
		if {[set p [$mf find_in $p Entity_and_Attribute_Information]]} {
			if {[set q [$mf find_in $p Detailed_Description]]} {
				if {[set r [$mf find_in $q Attribute]]} {
					while {$r} {
						if {[set u [$mf find_in $r Attribute_Label]]} {
							set s [$mf value_of -nonewline $u]
							if {[string compare -nocase $s $field] == 0} {
								if {[set v [$mf find_in $r Attribute_Domain_Values]]} {
									while {$v} {
										if {[set w [$mf find_in $v Enumerated_Domain]]} {
											if {[set x [$mf find_in $w Enumerated_Domain_Value]]} {
												set s [$mf value_of -nonewline $x]
												if {[string compare $s $value] == 0} {
													if {[set y [$mf find_in $w Enumerated_Domain_Value_Definition]]} {
														$mf value_set $y "$meaning"
puts "Updated Enumerated_Domain_Value_Definition for value $value of $field"
														set changed 1
														} \
													else {
														if {[set y [$mf insert -child $w Enumerated_Domain_Value_Definition]]} {
															$mf value_set $y "$meaning"
puts "Inserted Enumerated_Domain_Value_Definition for value $value of $field"
															set changed 1
															}
														}
													return
													}
												}
											}
										set v [$mf find_next $v]
										}

									# There is no Enumerated_Domain for this value.  Create one.
									# This is problematic because it's possible to have
									# more than one Attribute_Domain_Values.
									# I would prefer to add the new Enumerated_Domain
									# as a child of the first Attribute_Domain_Values.

									if {[set v [$mf find_in $r Attribute_Domain_Values]]} {
										if {[set w [$mf insert -child $v Enumerated_Domain]]} {
puts "Inserted Enumerated_Domain"
											if {[set x [$mf insert -child $w Enumerated_Domain_Value]]} {
												$mf value_set $x "$value"
puts "Inserted Enumerated_Domain_Value"
												set changed 1
												}
											if {[set x [$mf insert -child $w Enumerated_Domain_Value_Definition]]} {
												$mf value_set $x "$meaning"
puts "Inserted Enumerated_Domain_Value_Definition"
												set changed 1
												}
											}
										}
									return
									} \
								else {
									if {[set v [$mf insert -child $r Attribute_Domain_Values]]} {
										if {[set v [$mf insert -child $v Enumerated_Domain]]} {
											if {[set w [$mf insert -child $v Enumerated_Domain_Value]]} {
												$mf value_set $w "$value"
puts "Inserted Enumerated_Domain_Value for value $value of $field"
												}
											if {[set w [$mf insert -child $v Enumerated_Domain_Value_Definition]]} {
												$mf value_set $w "$meaning"
puts "Inserted Enumerated_Domain_Value_Definition for value $value of $field"
												}
											}
										set changed 1
										}
									}
								return
								}
							}
						set r [$mf find_next $r]
						}
					}
				}
			}
		}
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Given a field and definition, update the Attribute_Definition
# element that matches the field.  If there is no such
# Attribute element, create one and put this entry into it.
# Works for Attribute_Definition_Source as well; specify which element
# as the second argument.
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc update_attr {field element value} {
	global mf changed

	if {[llength $mf] == 0} {
		create_metadata
		}

	if {[string equal $element Attribute_Definition] == 0 ||
		[string equal $element Attribute_Definition_Source] == 0} {
		if {[set p [$mf find_first Metadata]]} {
			if {[set p [$mf find_in $p Entity_and_Attribute_Information]]} {
				if {[set q [$mf find_in $p Detailed_Description]]} {
					if {[set r [$mf find_in $q Attribute]]} {
						while {$r} {
							# Look at each Attribute element in turn
							if {[set u [$mf find_in $r Attribute_Label]]} {
								set s [$mf value_of -nonewline $u]
								# When you find the Attribute whose Attribute_Label
								# matches the database field you're on, work with it.
								if {[string compare -nocase $s $field] == 0} {
									if {[set v [$mf find_in $r $element]]} {
										# The element is there.  Get its value and
										# compare with the new value.  If the new
										# value is different, change to the new value.
										set s [$mf value_of $v]
										if {[string compare $s $value] != 0} {
											$mf value_set $v "$value"
puts "Updated $element for $field"
											set changed 1
											}
										} \
									else {
										# The element doesn't exist already.
										# If the new value is not empty, inser the element
										# and put the new value into it.
										if {[string length $value] > 0} {
											if {[set v [$mf insert -child $r $element]]} {
												$mf value_set $v "$value"
puts "Inserted $element for $field"
												set changed 1
												}
											}
										}
									return
									}
								}
							set r [$mf find_next $r]
							}
						}

					if {[string length $value] > 0} {
						# We reach this point only when there is no Attribute
						# whose Attribute_Label matches the given field name.
						# Create one and assign its label and definition.

						if {[set r [insert_attr $field]]} {
							if {[set u [$mf insert -child $r $element]]} {
								$mf value_set $u "$value"
puts "Inserted $element for $field"
								}
							set changed 1
							}
						}
					}
				}
			}
		}
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Given a field and codeset name, update the Codeset_Name of the first
# Codeset_Domain element that matches the field.  If there is no such
# Codeeset_Domain element, create one and put this entry into it.
# Works for Codeset_Source as well; specify which element as the second
# argument to the function.
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc update_codeset {field element value} {
	global mf changed

	if {[llength $mf] == 0} {
		create_metadata
		}

	if {[string equal $element Codeset_Name] == 0 ||
		[string equal $element Codeset_Source] == 0} {
		if {[set p [$mf find_first Metadata]]} {
			if {[set p [$mf find_in $p Entity_and_Attribute_Information]]} {
				if {[set q [$mf find_in $p Detailed_Description]]} {
					if {[set r [$mf find_in $q Attribute]]} {
						while {$r} {
							# Look at each Attribute element in turn
							if {[set u [$mf find_in $r Attribute_Label]]} {
								set s [$mf value_of -nonewline $u]
								# When you find the Attribute whose Attribute_Label
								# matches the database field you're on, work with it.
								if {[string compare -nocase $s $field] == 0} {

									# If the new value is empty, don't create a codeset to hold nothing

									if {[string length $value] == 0} {
										if {[set u [$mf find_in $r Codeset_Domain]]} {
											if {[set u [$mf find_in $u $element]]} {
												$mf delete $u
puts "Deleted $element"
												set changed 1
												}
											}
										return
										}

									if {[set u [$mf find_in $r Codeset_Domain]]} {
										if {[set u [$mf find_in $u $element]]} {
											$mf value_set $v "$value"
puts "Updated $element"
											} \
										else {
											if {[set v [$mf insert -child $u $element]]} {
												$mf value_set $v "$value"
puts "Inserted $element"
												}
											}
										} \
									else {
										if {[set u [$mf insert -child $r Attribute_Domain_Values]]} {
											if {[set u [$mf insert -child $u Codeset_Domain]]} {
												if {[set v [$mf insert -child $u $element]]} {
													$mf value_set $v "$value"
puts "Inserted $element"
													}
puts "Inserted Codeset_Domain"
												}
puts "Inserted Attribute_Domain_Values"
											}
										}
									set changed 1

									return
									}
								}
							set r [$mf find_next $r]
							}
						}

					if {[string length $value] > 0} {
						# We reach this point only when there is no Attribute
						# whose Attribute_Label matches the given field name.
						# Create one and assign its label and definition.

						if {[set r [insert_attr $field]]} {
							if {[set u [$mf insert -child $r Attribute_Domain_Values]]} {
puts "Inserted Attribute_Domain_Values"
								if {[set u [$mf insert -child $u Codeset_Domain]]} {
puts "Inserted Codeset_Domain"
									if {[set u [$mf insert -child $u $element]]} {
										$mf value_set $u "$value"
puts "Inserted $element"
										}
									}
								}
							set changed 1
							}
						}
					}
				}
			}
		}
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Update the Unrepresentable_Domain value for the given field
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc update_udom {field value} {
	global mf changed

	if {[llength $mf] == 0} {
		create_metadata
		}

	if {[set p [$mf find_first Metadata]]} {
		if {[set p [$mf find_in $p Entity_and_Attribute_Information]]} {
			if {[set q [$mf find_in $p Detailed_Description]]} {
				if {[set r [$mf find_in $q Attribute]]} {
					while {$r} {
						# Look at each Attribute element in turn
						if {[set u [$mf find_in $r Attribute_Label]]} {
							set s [$mf value_of -nonewline $u]
							# When you find the Attribute whose Attribute_Label
							# matches the database field you're on, work with it.
							if {[string compare -nocase $s $field] == 0} {

								# If the new value is empty, don't create a codeset to hold nothing

								if {[string length $value] == 0} {
									if {[set u [$mf find_in $r Unrepresentable_Domain]]} {
										$mf delete $u
puts "Deleted Unrepresentable_Domain"
										set changed 1
										}
									return
									}

								if {[set u [$mf find_in $r Unrepresentable_Domain]]} {
									$mf value_set $u "$value"
puts "Updated Unrepresentable_Domain"
									} \
								else {
									if {[set u [$mf insert -child $r Attribute_Domain_Values]]} {
puts "Inserted Attribute_Domain_Values"
										if {[set u [$mf insert -child $u Unrepresentable_Domain]]} {
											$mf value_set $u "$value"
puts "Inserted Unrepresentable_Domain"
											}
										}
									}
								set changed 1

								return
								}
							}
						set r [$mf find_next $r]
						}
					}

				if {[string length $value] > 0} {
					# We reach this point only when there is no Attribute
					# whose Attribute_Label matches the given field name.
					# Create one and assign its label and definition.

					if {[set r [insert_attr $field]]} {
						if {[set u [$mf insert -child $r Attribute_Domain_Values]]} {
puts "Inserted Attribute_Domain_Values"
							if {[set u [$mf insert -child $u Unrepresentable_Domain]]} {
puts "Inserted Unrepresentable_Domain"
								$mf value_set $u "$value"
puts "Inserted $element"
								}
							}
						set changed 1
						}
					}
				}
			}
		}
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Given a field and any of
#    Range_Domain_Minimum
#    Range_Domain_Maximum
#    Attribute_Units_of_Measure
#    Attribute_Measurement_Resolution
# Write the non-empty values of these elements into the Attribute
# element that matches the field.  If there is no such
# Attribute element, create one and put this entry into it.
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc update_rdom {field min max unit mres} {
	global mf changed

	if {[llength $mf] == 0} {
		create_metadata
		}

	if {[set p [$mf find_first Metadata]]} {
		if {[set p [$mf find_in $p Entity_and_Attribute_Information]]} {
			if {[set q [$mf find_in $p Detailed_Description]]} {
				if {[set r [$mf find_in $q Attribute]]} {
					while {$r} {
						if {[set u [$mf find_in $r Attribute_Label]]} {
							set s [$mf value_of -nonewline $u]
							if {[string compare -nocase $s $field] == 0} {

								# This is the right Attribute section.
								# Now find or make an Attribute_Domain_Values,
								# and within it find or make a Range_Domain

								if {[set u [$mf find_in $r Attribute_Domain_Values]]} {
									if {[set v [$mf find_in $u Range_Domain]]} {
										} \
									else {
										if {[set v [$mf insert -child $u Range_Domain]]} {
puts "Inserted Range_Domain for $field"
											set changed 1
											}
										}
									} \
								else {
									if {[set u [$mf insert -child $r Attribute_Domain_Values]]} {
puts "Inserted Attribute_Domain_Values for $field"
										if {[set v [$mf insert -child $u Range_Domain]]} {
puts "Inserted Range_Domain for $field"
											}
										set changed 1
										}
									}

								# v is now the Range_Domain, the parent of all these elements.

								set rdom(Range_Domain_Minimum)             "$min"
								set rdom(Range_Domain_Maximum)             "$max"
								set rdom(Attribute_Units_of_Measure)       "$unit"
								set rdom(Attribute_Measurement_Resolution) "$mres"

								foreach element [array names rdom] {
									if {[set w [$mf find_in $v $element]]} {
										# If the existing value is the same as the new value,
										# don't change it again, because that tells the user
										# that the existing metadata have changed, which has
										# not happened.

										if {[string compare $rdom($element) [$mf value_of -nonewline $w]] != 0} {
											# Different value.  Make the change.
											if {[string length $rdom($element)] > 0} {
												$mf value_set $w "$rdom($element)"
puts "Updated $element for $field"
												} \
											else {
												$mf delete $w
puts "Deleted $element for $field"
												}
											set changed 1
											}
										} \
									else {
										# If the element doesn't exist AND the new value is not empty,
										# then insert the element and set its value
										if {[string length $rdom($element)] > 0} {
											if {[set w [$mf insert -child $v $element]]} {
												$mf value_set $w "$rdom($element)"
puts "Inserted $element for $field"
												set changed 1
												}
											}
										}
									}

								unset rdom
								return
								}
							}
						set r [$mf find_next $r]
						}
					}

				# We reach this point only when there is no attribute
				# whose label matches the given field name.
				# Create one and assign its label and range domain values

				if {[set r [insert_attr $field]]} {
					if {[set r [$mf insert -child $r Attribute_Domain_Values]]} {
puts "Inserted Attribute_Domain_Values for $field"
						if {[set r [$mf insert -child $r Range_Domain]]} {
puts "Inserted Range_Domain for $field"

							if {[string length $min] > 0} {
								if {[set u [$mf insert -child $r Range_Domain_Minimum]]} {
									$mf value_set $u "$min"
puts "Inserted Range_Domain_Minimum for $field"
									}
								}

							if {[string length $max] > 0} {
								if {[set u [$mf insert -child $r Range_Domain_Maximum]]} {
									$mf value_set $u "$max"
puts "Inserted Range_Domain_Maximum for $field"
									}
								}

							if {[string length $unit] > 0} {
								if {[set u [$mf insert -child $r Attribute_Units_of_Measure]]} {
									$mf value_set $u "$unit"
puts "Inserted Attribute_Units_of_Measure for $field"
									}
								}

							if {[string length $mres] > 0} {
								if {[set u [$mf insert -child $r Attribute_Measurement_Resolution]]} {
									$mf value_set $u "$mres"
puts "Inserted Attribute_Measurement_Resolution for $field"
									}
								}
							}
						}
					set changed 1
					}
				}
			}
		}

	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Given a list of attribute labels, find the Attribute element that
# corresponds to each.  Delete it.
# Restricted to the first Detailed_Description
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc drop_attributes {attr_list w} {
	global mf changed

	if {[set p [$mf find_first Metadata]]} {
		if {[set p [$mf find_in $p Entity_and_Attribute_Information]]} {
			if {[set p [$mf find_in $p Detailed_Description]]} {
				set delete_list {}
				foreach label $attr_list {
					set done 0
					if {[set q [$mf find_in $p Attribute]]} {
						while {$q && !$done} {
							if {[set r [$mf find_in $q Attribute_Label]]} {
								set s [$mf value_of -nonewline $r]
								if {[string compare -nocase $s $label] == 0} {
									lappend delete_list $q
									set attrlabl($q) "$label"
									set done 1
									}
								}
							set q [$mf find_next $q]
							}
						}
					}
				if {[llength $delete_list] > 0} {
					foreach u [lsort -integer -decreasing $delete_list] {
						set label $attrlabl($u)
						$mf delete $u
puts "Deleted Attribute for $label"
						set changed 1

						# Since we've dropped this attribute from the metadata,
						# take its label out of the list in the diagnostics
						# window (the list is passed as the argument w)

						for {set i 0} {$i < [$w index end]} {incr i} {
							if {[string equal $label [$w get $i]]} {
								$w delete $i
								break
								}
							}
						}
					}
				unset attrlabl
				}
			}
		}
	if {[$w index end] == 0} {
		global diag
		$diag.mbar.remove configure -state disabled
		}
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Given a listbox widget, add the selected item to the metadata
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc add_one_attribute {w} {
	if {[llength [$w curselection]] > 0} {
		set label [$w get [$w index active]]
		add_attributes [list $label] $w
		}
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Given a list of attribute labels, create an Attribute element for each
# within the first Detailed_Description in the metadata, and assign the
# value of its Attribute_Label
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc add_attributes {attr_list w} {
	global mf changed

	if {[set p [$mf find_first Metadata]]} {
		if {[set p [$mf find_in $p Entity_and_Attribute_Information]]} {
			if {[set p [$mf find_in $p Detailed_Description]]} {
				foreach label $attr_list {
					if {[set q [insert_attr $label]]} {
						# Since we've added this attribute to the metadata,
						# take its label out of the list in the diagnostics
						# window (the list is passed as the argument w)

						for {set i 0} {$i < [$w index end]} {incr i} {
							if {[string equal $label [$w get $i]]} {
								$w delete $i
								break
								}
							}

						# It cannot have a definition yet, so add the field
						# name to the list of undefined attributes in the
						# diagnostics window

						global diag undocumented_field attribute_definition bgcolor

						set sw $diag.f.3.sw
						set dw $sw.f
						set k [llength $undocumented_field]

						lappend undocumented_field $label
						set the_label [label $dw.label_$k -text "$label"]
						$the_label configure -font value_font -justify left -background $bgcolor
						set the_entry [entry $dw.entry_$k -textvariable attribute_definition($label)]
						$the_entry configure -font value_font -width 60
						grid $the_label $the_entry
						grid $the_label -sticky w -padx 4

						tkwait visibility $dw.entry_$k
						set bbox [grid bbox $dw 0 0]
						set incr [lindex $bbox 3]
						set width [winfo reqwidth $dw]
						set height [winfo reqheight $dw]
						$sw configure -scrollregion "0 0 $width $height"
						$sw configure -yscrollincrement $incr
						set height [expr $incr * 10]
						$sw configure -width $width -height $height
						}
					}
				}
			}
		}
	if {[$w index end] == 0} {
		global diag
		$diag.mbar.add configure -state disabled
		}
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Given its label, insert an attribute into the metadata
# in the proper order (in the order it appears in the DBF)
# Create the Attribute_Label and set its value
# return the address of the Attribute element
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc insert_attr {label} {
	global d mf changed

	if {[llength $d] == 0 || [llength $mf] == 0} {return 0}

	# Get the list of fields in the DBF (for the order)

	set dbf_fields {}
	set field_info [$d fields]
	foreach item $field_info {
		lappend dbf_fields [string toupper [lindex $item 0]]
		}

	set label [string toupper $label]

	set the_index [lsearch -exact $dbf_fields $label]
	if {$the_index == -1} {return 0}

	# Find the first Detailed_Description in the metadata

	if {[set p [$mf find_first Entity_and_Attribute_Information]]} {
		if {[set p [$mf find_in $p Detailed_Description]]} {
			if {[set q [$mf find_in $p Attribute]]} {
				while {$q} {
					if {[set r [$mf find_in $q Attribute_Label]]} {
						set s [$mf value_of -nonewline $r]
						if {[string length $s] > 0} {
							set s [string toupper $s]
							# Find the position of this field in the DBF file
							set pos [lsearch -exact $dbf_fields $s]
							if {$pos > $the_index} {
								if {[set u [$mf insert -before $q Attribute]]} {
									if {[set v [$mf insert -child $u Attribute_Label]]} {
										$mf value_set $v "$label"
										set changed 1
										return $u
										}
									}
								}
							}
						}
					set q [$mf find_next $q]
					}

				# Arrive here if we didn't find a field in the metadata that
				# occurs later in the DBF file than the one we're inserting
				# This means we should just append the new attribute to the
				# existing list.

				if {[set q [$mf insert -child $p Attribute]]} {
					if {[set r [$mf insert -child $q Attribute_Label]]} {
						$mf value_set $r "$label"
						set changed 1
						return $q
						}
					}
				}
			}
		}

	# Arrive here if there is no Entity_and_Attribute_Information
	# or there is no Detailed_Description in the metadata

	return 0
	}

proc resequence_attributes {} {
	global d mf changed

	if {[llength $d] == 0 || [llength $mf] == 0} {return}

	# Get the list of fields in the DBF (for the order)

	set dbf_fields {}
	set field_info [$d fields]
	foreach item $field_info {
		lappend dbf_fields [string toupper [lindex $item 0]]
		}

	# Get the list of fields in the metadata

	set fields_not_in_dbf {}
	if {[set p [$mf find_first Entity_and_Attribute_Information]]} {
		if {[set p [$mf find_in $p Detailed_Description]]} {
			if {[set q [$mf find_in $p Attribute]]} {
				while {$q} {
					if {[set r [$mf find_in $q Attribute_Label]]} {
						set s [$mf value_of -nonewline $r]
						if {[string length $s] > 0} {
							set s [string toupper $s]
							if {[lsearch -exact $fields_not_in_dbf $s] != -1} {
								tk_messageBox -type ok -text "The attribute labeled $s appears more than once in the metadata.  Fix this problem before resequencing them." -icon warning
								return
								}
							if {[lsearch -exact $dbf_fields $s] == -1} {
								lappend fields_not_in_dbf "$s"
								}
							set address($s) $q
							}
						}
					set q [$mf find_next $q]
					}
				}

			# Detach all of the attributes

			foreach name [array names address] {
				$mf detach $address($name)
				}

			# Now reattach the attributes in the right order
			# First those that are in the DBF

			foreach name $dbf_fields {
				if {[llength [array names address $name]] > 0} {
					$mf attach -child $p $address($name)
					}
				}

			# After those, the attributes that aren't in the DBF

			foreach name $fields_not_in_dbf {
				if {[llength [array names address $name]] > 0} {
					$mf attach -child $p $address($name)
					}
				}
			}
		}
	}

#-------------------------------------------------------------------------
# DBF file handling
#-------------------------------------------------------------------------

proc open_dbf {} {
	global dbf_file d field_list

	if {[dbf d -open $dbf_file -readonly] != 0} {
		set counts [$d info]
		.tf.tip configure -text "File contains [lindex $counts 0] records, [lindex $counts 1] fields"

		set field_info [$d fields]
		foreach item $field_info {
			set name   [lindex $item 0]
			set type   [lindex $item 1]
			set native [lindex $item 2]
			set width  [lindex $item 3]
			set prec   [lindex $item 4]
			$field_list insert end "$name"
			}
		wm title . "DBF Explorer: $dbf_file"

		show_diagnostics
		.mbar.file.menu entryconfigure 0 -state disabled
		} \
	else {
		tk_messageBox -type ok -message "Error: could not open the input file $dbf_file" -icon warning
		}
	}

proc getmet {} {
	global met_file fs

	set types {{{Metadata files} {.met .MET}} {{XML files} {.xml .XML}} {{Text files} {.txt .TXT}} {{All files} *}}

	if {[string length $met_file] == 0 || [file exists $met_file] == 0} {
		set met_file [tk_getOpenFile -parent $fs -title "Select metadata" -filetypes $types -initialfile "$met_file"]
		}
	}

proc getdbf {} {
	global dbf_file fs

	set types {{{dBase files} {.dbf .DBF}} {{All files} *}}

	if {[string length $dbf_file] == 0 || [file exists $dbf_file] == 0} {
		set dbf_file [tk_getOpenFile -parent $fs -title "Select a dbase file" -filetypes $types -initialfile "$dbf_file"]
		}
	}

proc open_files {} {
	global dbf_file met_file

	if {[string length $dbf_file] > 0} {
		open_dbf
		}
	if {[string length $met_file] > 0} {
		open_metadata
		}
	resequence_attributes
	}

set fs {}

proc file_open {} {
	global met_file dbf_file fs
	if {[llength $fs] == 0} {
		set fs [toplevel .fs]

		global bgcolor

		$fs configure -background $bgcolor

		label $fs.metlab -text "Metadata:"
		label $fs.dbflab -text "DBF file:"

		$fs.metlab configure -background $bgcolor
		$fs.dbflab configure -background $bgcolor

		entry $fs.metname -width 32 -textvariable met_file
		entry $fs.dbfname -width 32 -textvariable dbf_file
		button $fs.metfind -command {getmet} -text "Browse..."
		button $fs.dbffind -command {getdbf} -text "Browse..."
		button $fs.ok -command {open_files; wm withdraw .fs} -text "OK"
		button $fs.close -command {wm withdraw .fs} -text "Cancel"

		grid $fs.metlab $fs.metname $fs.metfind -padx 4 -pady 4
		grid $fs.dbflab $fs.dbfname $fs.dbffind -padx 4 -pady 4

		grid $fs.metlab -sticky w
		grid $fs.dbflab -sticky w

		grid $fs.ok    -row 2 -column 0 -columnspan 2 -pady 4
		grid $fs.close -row 2 -column 2               -pady 4

		wm title $fs "Select input DBF and metadata files"

		update
		set fs_geometry [wm geometry $fs]
		scan $fs_geometry   "%dx%d+%d+%d" fw fh fx fy

		set root_geometry [wm geometry .]
		scan $root_geometry "%dx%d+%d+%d" rw rh rx ry

		set fx [expr $rx + 32]
		set fy [expr $ry + 32]
		wm geometry $fs "${fw}x${fh}+${fx}+${fy}"
		wm resizable $fs 0 0
		} \
	else {
		wm deiconify $fs
		}
	}

proc file_save {} {
	global mf changed met_file

	save_values

	if {[llength $mf] > 0} {
		if {$changed} {
			if {[string length $met_file] == 0} {
				file_save_as
				}
			if {[string length $met_file] > 0} {
				$mf write $met_file
				puts "Metadata updated: $met_file"
				set changed 0
				}
			} \
		else {
			puts "The metadata have not been changed."
			}
		} \
	else {
		puts "No metadata have been read."
		}
	}

proc file_save_as {} {
	global mf changed met_file

	save_values

	set types {{{Metadata files} {.met .MET}} {{XML files} {.xml .XML}} {{Text files} {.txt .TXT}} {{All files} *}}

	set met_file [tk_getSaveFile -title "Specify a metadata file" -filetypes $types -initialfile $met_file]
	if {[string length $met_file] == 0} return
	if {[file exists $met_file]} {
		if {[file writable $met_file]} {
			set ok 1
			} \
		else {
			set ok 0
			tk_messageBox -type ok -message "$met_file cannot be written.\nPlease use a different file name." -icon warning
			}
		} \
	else {
		if {[catch "open $met_file w" out]} {
			set ok 0
			tk_messageBox -type ok -message "$met_file cannot be written.\nPlease use a different file name." -icon warning
			} \
		else {
			close $out
			file delete $met_file
			set ok 1
			}
		}
	while {$ok == 0} {
		set met_file [tk_getSaveFile -title "Specify a metadata file" -filetypes $types]
		if {[string length $met_file] == 0} return
		if {[file writable $met_file]} {
			set ok 1
			} \
		else {
			set ok 0
			tk_messageBox -type ok -message "$met_file cannot be written.\nPlease use a different file name." -icon warning
			}
		}
	$mf write $met_file
	puts "Metadata updated: $met_file"
	set changed 0
	}

proc file_close {} {
	global d mf field_list value_list value_meaning met_file dbf_file
	global attribute_label attribute_definition attribute_definition_source enumerated_domain
	global current_key current_meaning changed

	save_values

	if {$changed} {
		set choice [tk_messageBox -type yesnocancel -default yes -message "Save changes before closing?" -icon question]
		switch -exact $choice {
			yes {
				file_save
				}
			no {
				}
			cancel {
				return
				}
			}
		}

	$field_list delete 0 end
	$value_list delete 0 end
	$value_meaning delete 1.0 end
	set current_field ""
	set current_index -1
	set current_value {}
	set value_index -1
	if {[llength $d] > 0} {
		$d forget
		set d {}
		}
	if {[llength $mf] > 0} {
		$mf forget
		set mf {}
		}
	catch {unset attribute_label}
	catch {unset attribute_definition}
	catch {unset attribute_definition_source}
	catch {unset enumerated_domain}
	catch {unset rdommin}
	catch {unset rdommax}
	catch {unset attrunit}
	catch {unset attrmres}

	set met_file {}
	set dbf_file {}

	wm title . "DBF Explorer: (no file loaded)"
	}

proc file_quit {} {
	global d mf field_list value_list value_meaning met_file dbf_file
	global attribute_label attribute_definition attribute_definition_source enumerated_domain
	global current_key current_meaning changed

	save_values

	if {$changed} {
		set choice [tk_messageBox -type yesnocancel -default yes -message "Save changes before closing?" -icon question]
		switch -exact $choice {
			yes {
				file_save
				}
			no {
				}
			cancel {
				return
				}
			}
		}

	if {$d != {}} {$d forget}
	exit
	}

proc edit_cut {} {}
proc edit_copy {} {}
proc edit_paste {} {}
proc edit_clear {} {}

set about_window {}

proc help_about {} {
	global version
	global about_window
	global bgcolor

	if {[llength $about_window] == 0} {
		set about_window [toplevel .about]
		$about_window configure -padx 4 -pady 4
		$about_window configure -background $bgcolor
		set about_version [label $about_window.version -justify left -text "DBF Explorer $version\nPeter N. Schweitzer\nU.S. Geological Survey\nReston, VA 20192"]
		$about_version configure -background $bgcolor
		$about_version configure -padx 8 -pady 8
		$about_version configure -font menu_font
		pack $about_version -side top -fill both -expand 1
		set about_ok [button $about_window.ok -text "Ok" -command {wm withdraw .about}]
		$about_ok configure -background $bgcolor
		pack $about_ok -side top
		} \
	else {
		wm deiconify $about_window
		}
	}

#-------------------------------------------------------------------------
# Whenever we change fields or values in an Enumerated_Domain, save the
# current information about this attribute to the metadata.
# This procedure should also be called when the user asks to close the
# file or quit the application.
#-------------------------------------------------------------------------

set current_field ""
set current_index -1
set current_value {}
set value_index -1

set current_definition {}
set current_definition_source {}
set current_key {}
set current_meaning {}

set current_codeset_name {}
set current_codeset_source {}

set stats_calculated(0) [list 0 0 0 0 0]

set current_udom {}

proc save_values {} {
	global current_field
	global attribute_definition        attrdef_text  current_definition
	global attribute_definition_source attrdefs_text current_definition_source

	global         codeset_name         codeset_source
	global current_codeset_name current_codeset_source
	global         codesetn_text        codesets_text

	global current_key value_meaning enumerated_domain current_meaning
	global rdommin rdommax attrunit attrmres

	global current_udom unrepresentable_domain udom_text

	if {[string length $current_field] == 0} {return}

	# If you have put text into the definition field already
	# Has the definition text changed from what was previously stored?
	# if so, get the new text and store it
	set t $attrdef_text.text
	set value [$t get 1.0 end-1c]
	if {[string compare $value $current_definition] != 0} {
		set attribute_definition($current_field) "$value"
		# update the metadata
		update_attr $current_field Attribute_Definition "$value"
		}

	# If you have put text into the definition source field already
	# Has the definition source text changed from what was previously stored?
	# if so, get the new text and store it
	set t $attrdefs_text.text
	set value [$t get 1.0 end-1c]
	if {[string compare $value $current_definition_source] != 0} {
		set attribute_definition_source($current_field) "$value"
		# update the metadata
		update_attr $current_field Attribute_Definition_Source "$value"
		}

	# If you have put text into the meaning field already
	if {[string length $current_key] > 0} {
		# Has the meaning text changed from what was previously stored?
		# if so, get the new text and store it
		set meaning [$value_meaning get 1.0 end-1c]
		if {[string compare $meaning $current_meaning] != 0} {
			set enumerated_domain($current_key) "$meaning"
			# update the metadata
			set k [string first , $current_key]
			set km1 [expr $k - 1]
			set kp1 [expr $k + 1]
			set the_field [string range $current_key 0 $km1]
			set the_value [string range $current_key $kp1 end]
			update_edom $the_field "$the_value" "$meaning"

			# Un-gray the value list item if it now has a meaning
			# Gray it if the meaning is now blank

			global value_list value_index
			if {[string length $meaning] > 0} {
				$value_list itemconfigure $value_index -background "#ffffff"
				} \
			else {
				$value_list itemconfigure $value_index -background "#dddddd"
				}
			}
		}

	# If you have put text into the codeset name already
	# Has the definition text changed from what was previously stored?
	# if so, get the new text and store it
	set t $codesetn_text.text
	set value [$t get 1.0 end-1c]
	if {[string compare $value $current_codeset_name] != 0} {
		set codeset_name($current_field) "$value"
		# update the metadata
		update_codeset $current_field Codeset_Name "$value"
		}

	# If you have put text into the codeset source field already
	# Has the codeset source text changed from what was previously stored?
	# if so, get the new text and store it
	set t $codesets_text.text
	set value [$t get 1.0 end-1c]
	if {[string compare $value $current_codeset_source] != 0} {
		set codeset_source($current_field) "$value"
		# update the metadata
		update_codeset $current_field Codeset_Source "$value"
		}

	# If there are values for rdommin, rdommax, attrunit, or attrmres,
	# store them in the metadata

	if {[string length $rdommin($current_field)] > 0 ||
		[string length $rdommax($current_field)] > 0 ||
		[string length $attrunit($current_field)] > 0 ||
		[string length $attrmres($current_field)] > 0} {
		update_rdom $current_field $rdommin($current_field) $rdommax($current_field) $attrunit($current_field) $attrmres($current_field)
		}

	# If you have put text into the unrepresentable domain field already
	# Has the text changed from what was previously stored?
	# if so, get the new text and store it
	set t $udom_text.text
	set value [$t get 1.0 end-1c]
	if {[string compare $value $current_udom] != 0} {
		set unrepresentable_domain($current_field) "$value"
		# update the metadata
		update_udom $current_field "$value"
		}
	}

proc set_field {the_listbox} {
	global d current_field current_index value_list value_meaning value_font
	global attribute_definition attribute_definition_source
	global attrdef_text attrdefs_text
	global current_definition current_definition_source
	global enumerated_domain current_key

	# If the first list element is empty, just return
	if {[string length [$the_listbox get 0]] == 0} {return}

	save_values

	set this_index [$the_listbox index active]
	set this_field [$the_listbox get $this_index]

	#---------------------------------------------------------------------
	# Determine whether the current element is different from the one
	# that was selected the last time this procedure was called.
	#---------------------------------------------------------------------

	set is_different 1
	if {$current_field == $this_field} {
		if {$current_index == $this_index} {
			set is_different 0
			}
		}
	if {$is_different} {
		$value_list delete 0 end
		$value_meaning delete 1.0 end

		set current_field $this_field
		set current_index $this_index

		set current_key {}

		set info [$d fields $current_field]
		set name   [lindex $info 0]
		set type   [lindex $info 1]
		set native [lindex $info 2]
		set width  [lindex $info 3]
		set prec   [lindex $info 4]

		# If you have read metadata and it contains a definition for this
		# attribute, show that definition in the appropriate widget

		set t $attrdef_text.text
		$t delete 1.0 end
		if {[array exists attribute_definition]} {
			if {[llength [array names attribute_definition $name]]} {
				$t insert end $attribute_definition($name)
				set current_definition "$attribute_definition($name)"
				}
			}

		set t $attrdefs_text.text
		$t delete 1.0 end
		if {[array exists attribute_definition_source]} {
			if {[llength [array names attribute_definition_source $name]]} {
				$t insert end $attribute_definition_source($name)
				set current_definition_source "$attribute_definition_source($name)"
				}
			}

		set all_values [$d values $this_field]
		set record_count [lindex [$d info] 0]

		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		# Show Codeset information if present in the metadata.  Show
		# only the first Codeset_Domain (this is a limitation of this
		# program, not a restriction inherent in the standard).
		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		global         codeset_name         codeset_source
		global current_codeset_name current_codeset_source
		global         codesetn_text        codesets_text

		set t $codesetn_text.text
		$t delete 1.0 end
		if {[array exists codeset_name]} {
			if {[llength [array names codeset_name $name]]} {
				$t insert end $codeset_name($name)
				set current_codeset_name "$codeset_name($name)"
				}
			}

		set t $codesets_text.text
		$t delete 1.0 end
		if {[array exists codeset_source]} {
			if {[llength [array names codeset_source $name]]} {
				$t insert end $codeset_source($name)
				set current_codeset_source "$codeset_source($name)"
				}
			}

		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		# Show unrepresentable domain if present
		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		global udom_text current_udom unrepresentable_domain

		set t $udom_text.text
		$t delete 1.0 end
		if {[array exists unrepresentable_domain]} {
			if {[llength [array names unrepresentable_domain $name]]} {
				$t insert end $unrepresentable_domain($name)
				set current_udom "$unrepresentable_domain($name)"
				}
			}

		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		# Calculate statistics if the field is numeric
		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		global stats_all_value
		global stats_nul_value
		global stats_min_value
		global stats_q25_value
		global stats_med_value
		global stats_q75_value
		global stats_max_value

		global stats_calculated

		if {$type == "Double" || $type == "Integer"} {

			set need_stats 1
			if {[array exists stats_calculated]} {
				if {[llength [array names stats_calculated $current_field]]} {
					set all_count [lindex $stats_calculated($current_field) 0]
					set nul_count [lindex $stats_calculated($current_field) 1]
					set min_value [lindex $stats_calculated($current_field) 2]
					set q25_value [lindex $stats_calculated($current_field) 3]
					set med_value [lindex $stats_calculated($current_field) 4]
					set q75_value [lindex $stats_calculated($current_field) 5]
					set max_value [lindex $stats_calculated($current_field) 6]
					set need_stats 0
					}
				}

			if {$need_stats} {
				set ordered [lsort $all_values]

				set all_count [llength $ordered]

				for {set k 0} {$k < $record_count} {incr k} {
					if {[string length [lindex $ordered $k]] > 0} {
						break
						}
					}
				set nul_count $k
				set ordered [lrange $ordered $k end]
				set ordered [lsort -real $ordered]
				set n [llength $ordered]

				set min_value [lindex $ordered 0]
				set max_value [lindex $ordered end]

				set k [expr {$n / 2}]
				if {[expr {$n & 1}]} {
					set med_value [lindex $ordered [expr {$k + 1}]]
					} \
				else {
					set med_value [expr {([lindex $ordered $k] + [lindex $ordered [expr $k + 1]])/2}]
					}

				set q25_value [lindex $ordered [expr {1 * $n / 4}]]
				set q75_value [lindex $ordered [expr {3 * $n / 4}]]
				set stats_calculated($current_field) [list $all_count $nul_count $min_value $q25_value $med_value $q75_value $max_value]
				unset ordered
				}

			$stats_all_value configure -text "$all_count"
			$stats_nul_value configure -text "$nul_count"
			$stats_min_value configure -text "$min_value"
			$stats_q25_value configure -text "$q25_value"
			$stats_med_value configure -text "$med_value"
			$stats_q75_value configure -text "$q75_value"
			$stats_max_value configure -text "$max_value"
			} \
		else {
			$stats_all_value configure -text ""
			$stats_nul_value configure -text ""
			$stats_min_value configure -text ""
			$stats_q25_value configure -text ""
			$stats_med_value configure -text ""
			$stats_q75_value configure -text ""
			$stats_max_value configure -text ""
			}

		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		# Show range domain if one was specified
		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		global rdommin rdommax attrunit attrmres
		global stats_rdom_min stats_rdom_max stats_rdom_unit stats_rdom_mres

		if {[llength [array names rdommin $current_field]] > 0} {
			$stats_rdom_min configure -textvariable rdommin($current_field)
			} \
		else {
			set rdommin($current_field) ""
			$stats_rdom_min configure -textvariable rdommin($current_field)
			}

		if {[llength [array names rdommax $current_field]] > 0} {
			$stats_rdom_max configure -textvariable rdommax($current_field)
			} \
		else {
			set rdommax($current_field) ""
			$stats_rdom_max configure -textvariable rdommax($current_field)
			}

		if {[llength [array names attrunit $current_field]] > 0} {
			$stats_rdom_unit configure -textvariable attrunit($current_field)
			} \
		else {
			set attrunit($current_field) ""
			$stats_rdom_unit configure -textvariable attrunit($current_field)
			}

		if {[llength [array names attrmres $current_field]] > 0} {
			$stats_rdom_mres configure -textvariable attrmres($current_field)
			} \
		else {
			set attrmres($current_field) ""
			$stats_rdom_mres configure -textvariable attrmres($current_field)
			}

		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		# Calculate number of distinct values and show them
		# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

		set stuff [lsort -dictionary -unique $all_values]
		set number_of_distinct_values [llength $stuff]
		set is_unique 0
		if {$number_of_distinct_values == [lindex [$d info] 0]} {
			global field_list
			$field_list itemconfigure $this_index -background "#ffff99"
			set is_unique 1
			}

		set k 0
		foreach item $stuff {
			if {$item == ""} {
				set item "(empty)"
				} \
			else {
				if {[string is space $item]} {
					set k [string length $item]
					set item "($k spaces)"
					}
				}
			$value_list insert $k "$item"

			# If this is a text field
			# but is not a unique key
			# and the metadata has been read
			# and the field has fewer than ~500 distinct values
			# and this value's meaning is not explained in the metadata
			# then gray the background of the field value in the value list
			# to draw attention to it.

			if {[string compare $type "String"] == 0} {
				if {$is_unique == 0} {
					if {$number_of_distinct_values < 500} {
						if {[array exists enumerated_domain]} {
							if {[llength [array names enumerated_domain $current_field,$item]] == 0} {
								$value_list itemconfigure $k -background "#dddddd"
								}
							}
						}
					}
				}

			incr k
			}

		$value_list configure -background white

		set the_text [format "%-10s %s(%d" $name $type $width]
		if {$type == "Double"} {
			append the_text ".$prec"
			}
		append the_text "):  [llength $stuff] distinct values"
		if {$is_unique} {
			append the_text ", unique"
			}
		.tf.tip configure -text "$the_text"
		.tf.count configure -text ""
		}
	}


proc set_value {the_listbox} {
	global d current_field current_value value_index value_list value_meaning
	global enumerated_domain
	global current_key current_meaning changed

	# If the first list element is empty, just return
	if {[string length [$the_listbox get 0]] == 0} {return}

	save_values

	set this_index [$the_listbox index active]
	set this_value [$the_listbox get $this_index]

	#---------------------------------------------------------------------
	# Determine whether the current element is different from the one
	# that was selected the last time this procedure was called.
	#---------------------------------------------------------------------

	set is_different 1
	if {$current_value == $this_value} {
		if {$value_index == $this_index} {
			set is_different 0
			}
		}
	if {$is_different} {
		set value_index   $this_index
		set current_value $this_value
		set current_key   "$current_field,$current_value"

		$value_meaning delete 1.0 end
		if {[array exists enumerated_domain]} {
			if {[llength [array names enumerated_domain $current_key]]} {
				$value_meaning insert end "$enumerated_domain($current_key)"
				set current_meaning "$enumerated_domain($current_key)"
				}
			}

		set stuff [$d values $current_field]
		set n [llength $stuff]
		if {$current_value == "(empty)"} {
			set k 0
			foreach item $stuff {
				if {$item == ""} {
					incr k
					}
				}
			} \
		else {
			set k 0
			foreach item $stuff {
				if {$item == $current_value} {
					incr k
					}
				}
			}

		set the_text "$k / $n"
		.tf.count configure -text "$the_text"
		}
	}

#-------------------------------------------------------------------------
# Main program
#-------------------------------------------------------------------------

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Create menu bar
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

frame .mbar -relief flat -bd 3
.mbar configure -background $bgcolor
pack .mbar -side top -expand false -fill x
menubutton .mbar.file -text "File" -menu .mbar.file.menu
menubutton .mbar.edit -text "Edit" -menu .mbar.edit.menu
menubutton .mbar.help -text "Help" -menu .mbar.help.menu
.mbar.file configure -bg $bgcolor -font menu_font
.mbar.edit configure -bg $bgcolor -font menu_font
.mbar.help configure -bg $bgcolor -font menu_font
pack .mbar.file .mbar.edit .mbar.help -side left

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Constitute menus
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

menu .mbar.file.menu -tearoff false -title "File"
	.mbar.file.menu configure -bg $bgcolor -font menu_font
	.mbar.file.menu add command -label "Open"       -command {file_open}
	.mbar.file.menu add command -label "Save"       -command {file_save}
	.mbar.file.menu add command -label "Save As"    -command {file_save_as}
	.mbar.file.menu add command -label "Close"      -command {file_close}
	.mbar.file.menu add command -label "Quit"       -command {file_quit}

menu .mbar.edit.menu -tearoff false -title "Edit"
	.mbar.edit.menu configure -bg $bgcolor -font menu_font
	.mbar.edit.menu add command -label "Cut"        -command {edit_cut}
	.mbar.edit.menu add command -label "Copy"       -command {edit_copy}
	.mbar.edit.menu add command -label "Paste"      -command {edit_paste}
	.mbar.edit.menu add command -label "Clear"      -command {edit_clear}

menu .mbar.help.menu -tearoff false -title "Help"
	.mbar.help.menu configure -bg $bgcolor -font menu_font
	.mbar.help.menu add command -label "About"      -command {help_about}

#-------------------------------------------------------------------------
# Paned window containing Tix components
#-------------------------------------------------------------------------

set paned [panedwindow .pw -orient horizontal]
$paned configure -bg $bgcolor

set attrlist [ScrolledListbox $paned.attrlist]
$attrlist configure -width 128
$paned add $attrlist
$paned paneconfigure $attrlist -padx 4 -pady 4

set field_list $attrlist.list
$field_list configure -font list_font
$field_list configure -exportselection false
# $field_list configure -activestyle none
bind $field_list <ButtonPress-1> {%W activate @%x,%y; set_field %W}

set w [frame .pw.info]
$w configure -width 480

set nb $w.nb
Notebook:create $nb -pages {Description Values Codeset Range Statistics Other} -bg $bgcolor
$nb configure -width 480
pack $nb -expand yes -fill both -side top

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Descriptive information about the attribute
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

set f [Notebook:frame $nb Description]
$f configure -background $bgcolor

set attrdef [labelframe $f.f1 -text "Attribute Definition"]
$attrdef configure -background $bgcolor
pack $attrdef -fill both -expand 1 -padx 4 -pady 4

set attrdef_text [ScrolledText $attrdef.t 64 10]
set t $attrdef_text.text
$t configure -wrap word -font value_font
pack $attrdef_text -fill both -expand 1 -padx 4 -pady 4

set attrdefs [labelframe $f.f2 -text "Attribute Definition Source"]
$attrdefs configure -background $bgcolor
pack $attrdefs -fill both -expand 1 -padx 4 -pady 4

set attrdefs_text [ScrolledText $attrdefs.t 64 3]
set t $attrdefs_text.text
$t configure -wrap word -font value_font
pack $attrdefs_text -fill both -expand 1 -padx 4 -pady 4

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Enumeration of the attribute values and their meanings
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

set f [Notebook:frame $nb Values]
$f configure -background $bgcolor

set edom [labelframe $f.f -text "Enumerated Domain"]
$edom configure -background $bgcolor
pack $edom -fill both -expand 1 -padx 4 -pady 4

set edom_pw [panedwindow $edom.pw -orient horizontal -background $bgcolor]

set value_listbox [ScrolledListbox $edom_pw.value_list]
set value_textbox [ScrolledText    $edom_pw.meaning 40 10]

$edom_pw add $value_listbox -width 128
$edom_pw add $value_textbox -width 256

set value_list $value_listbox.list
bind $value_list <ButtonPress-1> {%W activate @%x,%y; set_value %W}
$value_list configure -font list_font
$value_list configure -exportselection true
$value_list configure -activestyle none

set value_meaning $value_textbox.text
$value_meaning configure -wrap word -font value_font

pack $edom_pw -fill both -expand 1 -padx 4 -pady 4

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Codeset domain: name and source
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

set f [Notebook:frame $nb Codeset]
$f configure -background $bgcolor

set codesetn [labelframe $f.f1 -text "Codeset Name"]
$codesetn configure -background $bgcolor
pack $codesetn -fill both -expand 1 -padx 4 -pady 4

set codesetn_text [ScrolledText $codesetn.t 64 10]
set t $codesetn_text.text
$t configure -wrap word -font value_font
pack $codesetn_text -fill both -expand 1 -padx 4 -pady 4

set codesets [labelframe $f.f2 -text "Codeset Source"]
$codesets configure -background $bgcolor
pack $codesets -fill both -expand 1 -padx 4 -pady 4

set codesets_text [ScrolledText $codesets.t 64 3]
set t $codesets_text.text
$t configure -wrap word -font value_font
pack $codesets_text -fill both -expand 1 -padx 4 -pady 4

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Range domain
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

set f [Notebook:frame $nb Range]
$f configure -background $bgcolor

proc copy_min_max {} {
	global current_field stats_table rdommin rdommax
	set rdommin($current_field) [$stats_table.min_value cget -text]
	set rdommax($current_field) [$stats_table.max_value cget -text]
	save_values
	}

set stats_copy [button $f.f2 -text "Use actual range" -command {copy_min_max}]
grid $stats_copy -padx 8 -pady 8

set stats_rdom [labelframe $f.f3 -text "Range Domain"]
$stats_rdom configure -background $bgcolor
$stats_rdom configure -padx 8 -pady 8

set rdommin(0) {}
set rdommax(0) {}
set attrunit(0) {}
set attrmres(0) {}

set stats_rdom_min_label  [label $stats_rdom.min_label  -background $bgcolor -text "Range_Domain_Minimum"]
set stats_rdom_max_label  [label $stats_rdom.max_label  -background $bgcolor -text "Range_Domain_Maximum"]
set stats_rdom_unit_label [label $stats_rdom.unit_label -background $bgcolor -text "Units_of_Measure"]
set stats_rdom_mres_label [label $stats_rdom.mres_label -background $bgcolor -text "Measurement_Resolution"]

set stats_rdom_min  [entry $stats_rdom.min  -textvariable rdommin(0)  -justify right]
set stats_rdom_max  [entry $stats_rdom.max  -textvariable rdommax(0)  -justify right]
set stats_rdom_unit [entry $stats_rdom.unit -textvariable attrunit(0) -justify right]
set stats_rdom_mres [entry $stats_rdom.mres -textvariable attrmres(0) -justify right]

grid $stats_rdom_min_label  $stats_rdom_min
grid $stats_rdom_max_label  $stats_rdom_max
grid $stats_rdom_unit_label $stats_rdom_unit
grid $stats_rdom_mres_label $stats_rdom_mres

grid $stats_rdom_min_label  -ipadx 4 -ipady 4 -sticky w
grid $stats_rdom_max_label  -ipadx 4 -ipady 4 -sticky w
grid $stats_rdom_unit_label -ipadx 4 -ipady 4 -sticky w
grid $stats_rdom_mres_label -ipadx 4 -ipady 4 -sticky w

grid $stats_rdom -padx 8 -pady 8

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Statistics describing numerical attributes
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

set f [Notebook:frame $nb Statistics]
$f configure -background $bgcolor

set stats_table [labelframe $f.f1 -text "Summary of actual values"]
$stats_table configure -background $bgcolor
$stats_table configure -padx 8 -pady 8

set stats_all_label [label $stats_table.all_label -background $bgcolor -text "total values"   ]
set stats_nul_label [label $stats_table.nul_label -background $bgcolor -text "empty"          ]
set stats_min_label [label $stats_table.min_label -background $bgcolor -text "minimum"        ]
set stats_q25_label [label $stats_table.q25_label -background $bgcolor -text "25th percentile"]
set stats_med_label [label $stats_table.med_label -background $bgcolor -text "median"         ]
set stats_q75_label [label $stats_table.q75_label -background $bgcolor -text "75th percentile"]
set stats_max_label [label $stats_table.max_label -background $bgcolor -text "maximum"        ]

foreach q {all nul min q25 med q75 max} {
	eval "\$stats_${q}_label configure -background \"#bbddff\" -font menu_font"
	eval "set stats_${q}_value \[label \$stats_table.${q}_value -text \"\"\]"
	eval "\$stats_${q}_value configure -font value_font -background \"#dddddd\" -width 12 -anchor e"

	eval "grid \$stats_${q}_label \$stats_${q}_value"
	eval "grid \$stats_${q}_label -ipadx 4 -ipady 4 -sticky w"
	eval "grid \$stats_${q}_value -ipadx 4 -ipady 4"
	}

grid $stats_table -padx 8 -pady 8

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

set f [Notebook:frame $nb Other]

set udom [labelframe $f.f1 -text "Unrepresentable Domain"]
$udom configure -background $bgcolor

set udom_text [ScrolledText $udom.text 40 4]
pack $udom_text -fill both -expand 1 -padx 4 -pady 4

pack $udom -fill both -expand 1 -padx 4 -pady 4

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

$paned add $w
$paned paneconfigure $w -padx 4 -pady 4

pack .pw -fill both -expand 1

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Create the tip line as a framed label
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

frame .tf -relief flat -bd 1
.tf configure -background $bgcolor
pack .tf -side bottom -expand false -fill x
label .tf.tip -text "DBF Explorer (Peter N. Schweitzer, U.S. Geological Survey, Reston VA 20192)"
.tf.tip configure -background $bgcolor
.tf.tip configure -font value_font
label .tf.count -text ""
.tf.count configure -background $bgcolor
.tf.count configure -font value_font
pack .tf.tip   -fill both -expand false -side left
pack .tf.count -fill both -expand false -side right

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

proc list_up {} {
	global field_list
	set w $field_list
	set the_item [$w index active]
	if {$the_item > 0} {
		$w selection clear $the_item
		set the_item [expr $the_item - 1]
		$w selection set $the_item
		$w activate $the_item
		$w see $the_item
		set_field $w
		}
	}

proc list_down {} {
	global field_list
	set w $field_list
	set the_item [$w index active]
	set max_item [$w index end]
	if {$the_item < $max_item} {
		$w selection clear $the_item
		set the_item [expr $the_item + 1]
		$w selection set $the_item
		$w activate $the_item
		$w see $the_item
		set_field $w
		}
	}

proc list_home {} {
	global field_list
	set w $field_list
	set the_item [$w index active]
	if {$the_item > 0} {
		$w selection clear $the_item
		set the_item 0
		$w selection set $the_item
		$w activate $the_item
		$w see $the_item
		set_field $w
		}
	}

proc list_end {} {
	global field_list
	set w $field_list
	set the_item [$w index active]
	set max_item [$w index end]
	if {$the_item < $max_item} {
		$w selection clear $the_item
		set the_item [expr $max_item - 1]
		$w selection set $the_item
		$w activate $the_item
		$w see $the_item
		set_field $w
		}
	}

proc list_pageup {} {
	}

proc list_pagedown {} {
	}

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Bindings
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

event add <<list-up>>       <Alt-Key-Up>    <Alt-Key-KP_8>
event add <<list-down>>     <Alt-Key-Down>  <Alt-Key-KP_2>
event add <<list-home>>     <Alt-Key-Home>  <Alt-Key-KP_7>
event add <<list-end>>      <Alt-Key-End>   <Alt-Key-KP_1>
event add <<list-pageup>>   <Alt-Key-Prior> <Alt-Key-KP_9>
event add <<list-pagedown>> <Alt-Key-Next>  <Alt-Key-KP_3>

event add <<open>>          <Control-Key-o>
event add <<save>>          <Control-Key-s>
event add <<quit>>          <Control-Key-q>
event add <<Cut>>           <Control-Key-x>
event add <<Copy>>          <Control-Key-c>
event add <<Paste>>         <Control-Key-v>

event add <<Edit>>          <Alt-e>

bind . <<list-up>>       {list_up}
bind . <<list-down>>     {list_down}
bind . <<list-home>>     {list_home}
bind . <<list-end>>      {list_end}
bind . <<list-pageup>>   {list_pageup}
bind . <<list-pagedown>> {list_pagedown}

bind . <<open>>          {file_open}
bind . <<save>>          {file_save}
bind . <<quit>>          {file_quit}

bind . <<Edit>>          {edit_edom}

#-------------------------------------------------------------------------
