/*----------------------------------------------------------------------*\
 | dbfmeta.c is a program that reads a DBF file and generates from it	|
 | a Detailed_Description for the Entity_and_Attribute_Information of	|
 | metadata compatible with the Content Standard for Digital Geospatial	|
 | Metadata of the FGDC.  It uses extensions to the FGDC standard		|
 | Attribute_Type, Attribute_Width, Attribute_Precision which were		|
 | suggested by ESRI and are used by Arc8.								|
 |																		|
 | Based on the dbf access routines developed by Frank Warmerdam as		|
 | part of his shapelib C library for reading and writing shapefiles.	|
 | <http://shapelib.maptools.org/>  Thanks, Frank!						|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <unistd.h>
#include <shapefil.h>

#include "stricmp.h"

static char version[] = "1.13";

/*----------------------------------------------------------------------*\
 | Revision history:													|
 |																		|
 | 1.13 (20040414) PNS Fixed bug in which <attrmres> XML tag was not	|
 |                     properly closed.  Thanks to John Graves for		|
 |                     letting me know about this problem.				|
 |																		|
 | 1.12 (20040213) PNS Fix code calculating the maximum number of		|
 |                     nonblank characters in a string field.			|
 |                     Added code to read integer fields as long long	|
 |                     (64-bit) signed integers.  This is more complex	|
 |                     but might be more reliable for long numerical	|
 |                     code attributes the DBF file thinks are integer.	|
 |                     Thanks to Hugh Phillips for pointing out the		|
 |                     issue with long integers.						|
 |																		|
 | 1.11 (20030312) PNS Modified to allow use to omit descriptions of	|
 |                     specific fields by name.							|
 |																		|
 | 1.10 (20020729) PNS For character fields, find the length of the		|
 |                     longest value, treating trailing blanks as if	|
 |                     they didn't exist.  Report this to the user.		|
 |																		|
 | 1.9  (20020614) PNS Fixed bug introduced by last fix; enlarge the	|
 |                     space allocated for each string value by one to	|
 |                     hold the terminating nul byte (D'oh!).			|
 |																		|
 | 1.8  (20020611) PNS Fix code so that the single real number can		|
 |                     occur more than once and still be considered a	|
 |                     special value.  Modified dbfopen.c to not strip	|
 |                     space from string values by default.				|
 |																		|
 | 1.7  (20020528) PNS Open the input file before worrying about naming	|
 |                     the output file.  Fixed bug that caused the		|
 |                     verbose statistical report to not show integer	|
 |                     values.  Describe in an Enumerated_Domain a		|
 |                     singleton real number.  Thanks to Hugh Phillips	|
 |                     for catching these problems.						|
 |																		|
 | 1.6  (20020411) PNS Modified to output SGML or XML if the output file|
 |                     name looks like it has that extension.			|
 |																		|
 | 1.5  (20020206) PNS Modified to open dbf file read-only and to report|
 |                     duplicate names to stdout rather than stderr.	|
 |																		|
 | 1.4  (20010926) PNS Added code to report the number of blank records	|
 |                     in a string field.								|
 |																		|
 | 1.3  (20010227) PNS Revised handling of attribute information by		|
 |                     reading all fields at once.  This enables me to	|
 |                     check for duplicate field names.					|
 |																		|
 | 1.2  (20010126) PNS Added code to skip attributes whose names match	|
 |                     those that are intrinsic to Arc/Info.			|
 |                     Fixed bug in counting unique string values.		|
 |																		|
 | 1.1  (20010125) PNS Modified handling of enumerated domains to give	|
 |                     some statistical information about the values.	|
 |																		|
 | 1.0  (20010124) PNS First release.									|
\*----------------------------------------------------------------------*/

struct field_info {
	char name [16];
	DBFFieldType type;
	int width;
	int precision;
	};

static char *type_of (DBFFieldType t) {
	if (t == FTString) return ("String");
	if (t == FTInteger) return ("Integer");
	if (t == FTDouble) return ("Double");
	if (t == FTInvalid) return ("Invalid");
	return ("Unknown");
	}

#define INTEGERS_PROBABLY_ARE_CATEGORIES	20
#define MAX_TEXT_LENGTH	4096
static char default_output[] = "dbfmeta.out";

/*----------------------------------------------------------------------*\
 | Special function to read integer values from the shapefile.  Looks	|
 | like DBFReadIntegerAttribute but accommodates values that might be	|
 | longer than can fit into a standard 32-bit integer.  Some data use	|
 | long integer values as codes.										|
\*----------------------------------------------------------------------*/

static long long llvalue = 0;

static long long *DBFReadLongIntegerAttribute (DBFHandle df, int record, int field) {
	long long *result;
	int ndec;
	int negative = 0;
	char *s;
	char *string = (char *) DBFReadStringAttribute (df,record,field);
	if (!(s = strchr (string,'E'))) s = strchr (string,'e');
	if (s) {
		/* Must decode scientific notation to get integer value */
		s++;
		if (*s == '+') s++;
		ndec = 0;
		while (*s) {
			ndec *= 10;
			ndec += *s - '0';
			s++;
			}
		s = string;
		while (isspace(*s)) s++;

		if (*s == '-' || *s == '+') {
			if (*s == '-') negative = 1;
			s++;
			}
		llvalue = 0;
		while (isdigit(*s)) {
			llvalue *= 10;
			llvalue += *s - '0';
			s++;
			}
		if (*s == '.') {
			s++;
			while (isdigit(*s) && ndec > 0) {
				llvalue *= 10;
				llvalue += *s - '0';
				s++;
				ndec--;
				}
			}
		if (negative) llvalue = -llvalue;
		result = &llvalue;
		}
	else {
		llvalue = strtoll (string,&s,0);
		if (s == string) result = NULL;
		else result = &llvalue;
		}
	return (result);
	}

/*----------------------------------------------------------------------*\
 | lltoa converts a long long integer value to a string in a static		|
 | buffer, returning the address of the buffer.							|
\*----------------------------------------------------------------------*/

#ifndef HAVE_LLTOA

static char *lltoa (long long value, char *buffer, int radix) {
	char *t = buffer;
	char *s = buffer;
	char c;
	int negative = 0;

	if (value < 0) {
		value = -value;
		negative = 1;
		}

	do {
		*s++ = '0' + (value % radix);
		value /= radix;
		} while (value > 0);
	if (negative) *s++ = '-';
	*s-- = 0;

	while (s > t) {
		c = *t;
		*t++ = *s;
		*s-- = c;
		}

	return (buffer);
	}

#endif

/*----------------------------------------------------------------------*/

static char *intrinsic_attribute [] = {
	"AREA",
	"PERIMETER",
	"LENGTH",
	"FNODE_",
	"TNODE_",
	"LPOLY_",
	"RPOLY_",
	NULL
	};

static int prev_name_intrinsic = 0;

static int intrinsic (char *name) {
	int i;

	for (i=0; intrinsic_attribute[i]; i++)
		if (stricmp (name,intrinsic_attribute[i]) == 0) break;

	if (intrinsic_attribute[i]) {
		prev_name_intrinsic = 1;
		return (1);
		}

	/*------------------------------------------------------------------*\
	 | Dicy part.  Look for cover# and cover-ID, which show up in DBFs	|
	 | as NAME_ and NAME_ID immediately following an intrinsic.			|
	\*------------------------------------------------------------------*/

	if (prev_name_intrinsic) {
		char *s = name;
		while (*s && isalnum(*s)) s++;
		if (strcmp (s,"_") == 0 || strcmp (s,"_ID") == 0) return (1);
		else {
			prev_name_intrinsic = 0;
			return (0);
			}
		}
	else
		return (0);
	}

struct counted_value {
	char *text;
	long long number;
	int count;
	};

/* Neat-o string compare function from Dr. Dobbs article */

static int cvs_compare (const void *e1, const void *e2) {
	const struct counted_value *v1 = (const struct counted_value *) e1;
	const struct counted_value *v2 = (const struct counted_value *) e2;
	const char *s1 = v1->text;
	const char *s2 = v2->text;
	int q;

	while (*s1 && *s2)
		if (isdigit(*s1) && isdigit(*s2)) {
			if (q = atoi (s1) - atoi (s2)) return (q);
			while (isdigit(*s1)) s1++;
			while (isdigit(*s2)) s2++;
			}
		else
			if (q = tolower(*s1) - tolower(*s2)) return (q);
			else {
				s1++;
				s2++;
				}
	return (*s1 - *s2);
	}

static int cvll_compare (const void *e1, const void *e2) {
	const struct counted_value *v1 = (const struct counted_value *) e1;
	const struct counted_value *v2 = (const struct counted_value *) e2;
	const long long n1 = v1->number;
	const long long n2 = v2->number;
	return (n1 - n2);
	}

int main (int argc, char *argv[]) {
	char *input_file = NULL;
	char output_file [FILENAME_MAX];
	FILE *out;
	DBFHandle df;
	int rc,fc,max_width;
	int i,j,k;
	DBFFieldType ft[1024];
	char *string,*s,*e;
	double d,dmin,dmax,dneg;
	int mcount,max_length;
	long long m,mmin,mmax;
	long long *pllValue;
	struct counted_value *pv;
	int count,ncount,pcount,zcount;
	int many_negative_values;
	char *pool;
	char *pool_ptr;
	struct field_info *field = NULL;
	int do_xml = 0;

	int dialog = 0;
	int strict = 0;
	int detail = 0;
	char response [MAX_TEXT_LENGTH];
	char definition [MAX_TEXT_LENGTH];
	char units [MAX_TEXT_LENGTH];
	double resolution;
	char number [64];

	char *not[64] = {NULL};

	*output_file = 0;
	if (argc > 1)
		for (i=1; i < argc; i++)
			if (memcmp (argv[i],"-o",2) == 0) {
				i++;
				if (i < argc) strcpy (output_file,argv[i]);
				}
			else
				if (memcmp (argv[i],"-i",2) == 0)
					dialog = 1;
				else
					if (memcmp (argv[i],"-strict",7) == 0)
						strict = 1;
					else
						if (memcmp (argv[i],"-detail",7) == 0)
							detail = 1;
						else
							if (memcmp (argv[i],"-not",4) == 0) {
								i++;
								if (i < argc) {
									for (j=0; not[j]; j++);
									if (j < sizeof(not) - 1) {
										not[j++] = argv[i];
										not[j] = NULL;
										}
									}
								}
							else
								input_file = argv[i];

	printf ("dbfmeta %s - Peter N. Schweitzer (U.S. Geological Survey)\n",version);

	if (input_file) {
		if (df = DBFOpen (input_file,"rb")) {

			/*----------------------------------------------------------*\
			 | If the user didn't give us an output file, try to write	|
			 | to a default name.  But don't write over a file with		|
			 | that name if one exists already.							|
			\*----------------------------------------------------------*/

			if (!*output_file) {
				char *reason;
				if (access (default_output,F_OK) != 0)
					if (out = fopen (default_output,"w")) {
						fclose (out);
						strcpy (output_file,default_output);
						}
					else
						reason = "cannot be created";
				else
					reason = "exists already";
				if (!*output_file) {
					printf ("You didn't specify an output file name, and my default name %s %s.\n",default_output,reason);
					printf ("Where do you want me to put the output?: ");
					fgets (response,4096,stdin);
					s = response + strlen (response) - 1;
					if (*s == '\n') *s-- = 0;
					if (*s == '\r') *s-- = 0;
					while (s >= response && isspace(*s)) *s-- = 0;
					s = response;
					while (*s && isspace(*s)) s++;
					if (*s) strcpy (output_file,s);
					else {
						printf ("I need to put it somewhere!\n");
						exit (1);
						}
					}
				else
					printf ("Writing output to file %s\n",default_output);
				}

			rc = DBFGetRecordCount (df);
			fc = DBFGetFieldCount (df);

			printf ("Opened file %s\n",input_file);
			printf ("  Number of fields: %d\n",fc);
			printf ("  Number of records: %d\n",rc);

			/* Allocate space for the field info */

			if (field = (struct field_info *) malloc (fc * sizeof (struct field_info))) {
				for (i=0; i < fc; i++)
					field[i].type = DBFGetFieldInfo (df,i,field[i].name,&field[i].width, &field[i].precision);

				/*------------------------------------------------------*\
				 | Look at the names to make sure there aren't any		|
				 | duplicate names.  Warn user if there are.			|
				\*------------------------------------------------------*/

				for (i=0; i < fc - 1; i++)
					for (j=i+1; j < fc; j++)
						if (strcmp (field[i].name,field[j].name) == 0)
							printf ("WARNING: Fields %d and %d have the same name \"%s\"\n",i+1,j+1,field[i].name);
				}

			/* Allocate space for the values */

			if (pv = (struct counted_value *) malloc (rc * sizeof (struct counted_value))) {
				memset (pv,0,rc*sizeof(struct counted_value));
				}
			else {
				fprintf (stderr,"Error: could not allocate space for %d struct counted_text\n",rc);
				exit (1);
				}

			/* Create a pool for string values */

			max_width = 0;
			for (j=0; j < fc; j++) {
				if (field[j].width > max_width) max_width = field[j].width;
				}
			max_width++;	/* need one more for terminating NUL */
			if (pool = (char *) malloc (rc*max_width)) {
				memset (pool,0,rc*max_width);
				pool_ptr = pool;
				}
			else {
				fprintf (stderr,"Error: could not allocate string pool of size %d\n",rc*max_width);
				exit (1);
				}

			if (s = strrchr (output_file,'.'))
				if ((stricmp (s,".sgm") == 0) ||
					(stricmp (s,".sgml") == 0)) do_xml = 1;
				else
					if (stricmp (s,".xml") == 0) do_xml = 1;

			if (out = fopen (output_file,"w")) {
				if (do_xml) {
					fprintf (out,"<detailed>\n");
					fprintf (out,"<enttyp>\n");
					fprintf (out,"<enttypl>%s</enttypl>\n",input_file);
					fprintf (out,"<enttypd></enttypd>\n");
					fprintf (out,"</enttyp>\n");
					}
				else {
					fprintf (out,"Detailed_Description:\n");
					fprintf (out,"  Entity_Type:\n");
					fprintf (out,"    Entity_Type_Label: %s\n",input_file);
					fprintf (out,"    Entity_Type_Definition:\n");
					}

				/* Read each field and describe what you can */

				for (j=0; j < fc; j++) {
					int skip;		/* skip this field?  0 means no, 1 means yes */
					*definition = 0;
					*units = 0;
					resolution = 0.0;

					printf ("%s",field[j].name);
					if (field[j].type == FTInteger)
						printf (" is an %s",type_of(field[j].type));
					else
						printf (" is a %s",type_of(field[j].type));
					printf (" of width %d",field[j].width);
					if (field[j].type == FTDouble)
						printf (" and precision %d",field[j].precision);
					printf (".\n");

					skip = 0;
					for (k=0; not[k]; k++)
						if (stricmp (field[j].name,not[k]) == 0) {
							printf ("  Description omitted at your request.\n");
							skip = 1;
							break;
							}

					if (intrinsic (field[j].name)) {
						printf ("  Intrinsic to ArcInfo, not documented here.\n");
						skip = 1;
						}

					if (!skip) {
						if (dialog) {
							printf ("  What is its definition?: ");
							fgets (response,4096,stdin);
							s = response + strlen (response) - 1;
							if (*s == '\n') *s-- = 0;
							if (*s == '\r') *s-- = 0;
							while (s >= response && isspace(*s)) *s-- = 0;
							if (strlen (response) == 0)
								printf ("  You'll need to fill in Attribute_Definition for this attribute.\n");
							s = response;
							while (*s && isspace(*s)) s++;
							strcpy (definition,s);
							}

						if (do_xml) {
							fprintf (out,"<attr>\n");
							fprintf (out,"<attrlabl>%s</attrlabl>\n",field[j].name);
							fprintf (out,"<attrdef>%s</attrdef>\n",definition);
							if (!strict) {
								fprintf (out,"<attrtype>%s</attrtype>\n",type_of(field[j].type));
								fprintf (out,"<attwidth>%d</attwidth>\n",field[j].width);
								if (ft[j] == FTDouble)
									fprintf (out,"<atprecis>%d</atprecis>\n",field[j].precision);
								}
							fprintf (out,"<attrdomv>\n");
							}
						else {
							fprintf (out,"  Attribute:\n");
							fprintf (out,"    Attribute_Label: %s\n",field[j].name);
							fprintf (out,"    Attribute_Definition: %s\n",definition);
							if (!strict) {
								fprintf (out,"    Attribute_Type: %s\n", type_of (field[j].type));
								fprintf (out,"    Attribute_Width: %d\n",field[j].width);
								if (ft[j] == FTDouble)
									fprintf (out,"    Attribute_Precision: %d\n",field[j].precision);
								}
							fprintf (out,"    Attribute_Domain_Values:\n");
							}
						switch (field[j].type) {
							case FTString:

								/*------------------------------------------*\
								 | Strings are always Enumerated_Domains.	|
								 | If they're really place names or other	|
								 | self-sufficient labels, the user can		|
								 | reasonably omit them.  But we can't make	|
								 | that decision for the user.				|
								\*------------------------------------------*/

								max_length = 0;
								count = 0;
								for (i=0; i < rc; i++)
									if (string = (char *) DBFReadStringAttribute (df,i,j)) {
										/* Have we seen this attribute value already? */
										for (k=0; pv[k].text; k++)
											if (strcmp (string,pv[k].text) == 0) {
												pv[k].count++;
												break;
												}
										/* If not, add it to our list */
										if (pv[k].text == NULL) {
											pv[k].text = pool_ptr;
											strcpy (pv[k].text,string);
											pool_ptr += 1 + strlen (string);
											pv[k].count = 1;
											count++;

											for (e = string + strlen (string) - 1; e >= string && isspace(*e); e--);
											if (e - string + 1 > max_length) max_length = e - string + 1;
											}
										}
								qsort (pv,count,sizeof (struct counted_value),cvs_compare);
								for (k=0; k < count; k++) {
									for (s=pv[k].text; *s; s++)
										if (*s == '\r' || *s == '\n') *s = ' ';
									if (do_xml) {
										fprintf (out,"<edom>\n");
										if (*pv[k].text)
											fprintf (out,"<edomv>%s</edomv>\n",pv[k].text);
										else
											fprintf (out,"<edomv>(no value)</edomv>\n");
										fprintf (out,"<edomvd></edomvd>\n");
										fprintf (out,"</edom>\n");
										}
									else {
										fprintf (out,"      Enumerated_Domain:\n");
										if (*pv[k].text)
											fprintf (out,"        Enumerated_Domain_Value: %s\n",pv[k].text);
										else
											fprintf (out,"        Enumerated_Domain_Value: (no value)\n");
										fprintf (out,"        Enumerated_Domain_Value_Definition: \n");
										}
									}
								printf ("  Documented as an Enumerated_Domain\n");
								if (k > 1) {
									printf ("  There are %d distinct values.\n",k);
									mcount = 0;
									for (k=0; k < count; k++)
										if (pv[k].count > 1) mcount++;
									if (mcount == 0)
										printf ("  Each value occurs only once in the data.\n");
									else {
										printf ("  %d values occur more than once in the data.\n",mcount);
										if (detail) {
											printf ("   count | value\n");
											for (k=0; k < count; k++)
												printf ("  %6d | \"%s\"\n",pv[k].count,pv[k].text);
											}
										}
									if (*pv[0].text == 0)
										printf ("  For %d records this field is empty.\n",pv[0].count);
									else
										printf ("  All records have a value in this field.\n");
									}
								else
									if (*pv[0].text) {
										printf ("  All records have the same value for this field.\n");
										printf ("  The value is \"%s\".\n",pv[0].text);
										}
									else
										printf ("  For all records in the file this field is empty.\n");
								printf ("  Excluding trailing blanks, the longest string is %d characters.\n",max_length);
								memset (pv,0,rc*sizeof(struct counted_value));
								pool_ptr = pool;
								break;

							case FTInteger:

								/*--------------------------------------*\
								 | Read the first value to use to find	|
								 | the minimum and maximum values.		|
								\*--------------------------------------*/

								if (pllValue = DBFReadLongIntegerAttribute (df,0,j))
									mmin = mmax = *pllValue;
								else {
									fprintf (stderr,"Error: could not interpret long integer value from \"%s\"\n",DBFReadStringAttribute(df,0,j));
									mmin = mmax = (long long) 0;
									}
								count = 0;
								for (i=0; i < rc; i++) {
									if (pllValue = DBFReadLongIntegerAttribute (df,i,j)) {
										m = *pllValue;
										if (m < mmin) mmin = m;
										if (m > mmax) mmax = m;
										for (k=0; k < count; k++)
											if (m == pv[k].number) {
												pv[k].count++;
												break;
												}
										if (k == count) {
											pv[k].number = m;
											pv[k].count = 1;
											count++;
											}
										}
									else
										fprintf (stderr,"Error: could not interpret long integer value from \"%s\"\n",DBFReadStringAttribute(df,i,j));
									}
								qsort (pv,count,sizeof(struct counted_value),cvll_compare);

								ncount = pcount = zcount = 0;
								for (k=0; k < count; k++)
									if (pv[k].number <  0) ncount++;
									else
										if (pv[k].number >  0) pcount++;
										else
											if (pv[k].number == 0) zcount++;
								printf ("  %d negative, %d positive, %d zero values\n",ncount,pcount,zcount);

								/*------------------------------------------*\
								 | Dicy decisions here.  If there aren't	|
								 | very many different numbers, perhaps		|
								 | they really are categories.  If there	|
								 | are lots of different numbers, chances	|
								 | are they are not categories but are		|
								 | instead counts.							|
								\*------------------------------------------*/

								if (count <= INTEGERS_PROBABLY_ARE_CATEGORIES) {
									if (do_xml)
										for (k=0; k < count; k++) {
											fprintf (out,"<edom>\n");
											fprintf (out,"<edomv>%s</edomv>\n",lltoa(pv[k].number,number,10));
											fprintf (out,"<edomvd></edomvd>\n");
											fprintf (out,"</edom>\n");
											}
									else
										for (k=0; k < count; k++) {
											fprintf (out,"      Enumerated_Domain:\n");
											fprintf (out,"        Enumerated_Domain_Value: %s\n",lltoa(pv[k].number,number,10));
											fprintf (out,"        Enumerated_Domain_Value_Definition: \n");
											}
									printf ("  Documented as Enumerated_Domain\n");
									mcount = 0;
									for (k=0; k < count; k++)
										if (pv[k].count > 1) mcount++;
									if (mcount == 0)
										printf ("  Each value occurs only once in the data.\n");
									else {
										printf ("  %d values occur more than once in the data.\n",mcount);
										if (detail) {
											printf ("   count | value\n");
											for (k=0; k < count; k++)
												printf ("  %6d | %s\n",pv[k].count,lltoa(pv[k].number,number,10));
											}
										}
									}
								else {

									/*--------------------------------------*\
									 | If only one or two negative values	|
									 | are used, write them as Enumerated,	|
									 | then write the positives as a Range.	|
									\*--------------------------------------*/

									m = 0;
									for (k=0; k < count; k++)
										if (pv[k].number < 0) m++;

									if (m == 1 || m == 2) {
										for (k=0; k < count; k++)
											if (pv[k].number < 0) {
												if (do_xml) {
													fprintf (out,"<edom>\n");
													fprintf (out,"<edomv>%s</edomv>\n",lltoa(pv[k].number,number,10));
													fprintf (out,"<edomvd></edomvd>\n");
													fprintf (out,"</edom>\n");
													}
												else {
													fprintf (out,"      Enumerated_Domain:\n");
													fprintf (out,"        Enumerated_Domain_Value: %s\n",lltoa(pv[k].number,number,10));
													fprintf (out,"        Enumerated_Domain_Value_Definition: \n");
													}
												printf ("  Value %s documented using Enumerated_Domain\n",lltoa(pv[k].number,number,10));
												printf ("  This value occurs %d times in the data.\n",pv[k].count);
												}

										/*----------------------------------*\
										 | Having documented the negatives	|
										 | as Enumerated, find the smallest	|
										 | positive; it will become the new	|
										 | Range_Domain_Minimum.			|
										\*----------------------------------*/

										mmin = mmax;
										for (k=0; k < count; k++)
											if (pv[k].number >= 0 && pv[k].number < mmin) mmin = pv[k].number;
										printf ("  %d distinct positive values documented using Range_Domain\n",count-m);
										}
									else
										printf ("  %d distinct values documented using Range_Domain\n",count);

									if (dialog) {
										printf ("  What are the units of its values?: ");
										fgets (response,4096,stdin);
										s = response + strlen (response) - 1;
										if (*s == '\n') *s-- = 0;
										if (*s == '\r') *s-- = 0;
										while (s >= response && isspace(*s)) *s-- = 0;
										if (strlen (response) == 0)
											printf ("  You'll need to fill in Attribute_Definition for this attribute.\n");
										s = response;
										while (*s && isspace(*s)) s++;
										strcpy (units,s);
										}

									if (do_xml) {
										fprintf (out,"<rdom>\n");
										fprintf (out,"<rdommin>%s</rdommin>\n",lltoa(mmin,number,10));
										fprintf (out,"<rdommax>%s</rdommax>\n",lltoa(mmax,number,10));
										fprintf (out,"<attrunit>%s</attrunit>\n",units);
										fprintf (out,"<attrmres>1</attrmres>\n");
										fprintf (out,"</rdom>\n");
										}
									else {
										fprintf (out,"      Range_Domain:\n");
										fprintf (out,"        Range_Domain_Minimum: %s\n",lltoa(mmin,number,10));
										fprintf (out,"        Range_Domain_Maximum: %s\n",lltoa(mmax,number,10));
										fprintf (out,"        Attribute_Units_of_Measure: %s\n",units);
										fprintf (out,"        Attribute_Measurement_Resolution: 1\n");
										}
									}
								break;

							case FTDouble:

								/*------------------------------------------*\
								 | Doubles are always a Range_Domain, so 	|
								 | figure the maximum and minimum.  We won't|
								 | be able to get a good estimate of the	|
								 | resolution, nor do we know the units.	|
								 | Give the user a place to put that info.	|
								\*------------------------------------------*/

								dneg = 0.0;
								many_negative_values = 0;
								ncount = pcount = zcount = 0;
								dmin = dmax = DBFReadDoubleAttribute (df,0,j);
								for (i=0; i < rc; i++) {
									d = DBFReadDoubleAttribute (df,i,j);
									if (d <  0.0) {
										ncount++;
										if (dneg == 0.0) dneg = d;
										else
											if (d != dneg)
												many_negative_values = 1;
										}
									if (d == 0.0) zcount++;
									if (d >  0.0) pcount++;
									if (d < dmin) dmin = d;
									if (d > dmax) dmax = d;
									}

								/*--------------------------------------*\
								 | Hugh: Why not do the same thing with	|
								 | single negative real numbers as you	|
								 | do with single negative integers?	|
								 |										|
								 | Having documented the negatives as	|
								 | as Enumerated, find the smallest		|
								 | positive; it will become the new		|
								 | Range_Domain_Minimum.				|
								\*--------------------------------------*/

								if (ncount > 0 && many_negative_values == 0) {
									dmin = dmax;
									for (i=0; i < rc; i++) {
										d = DBFReadDoubleAttribute (df,i,j);
										if (d >= 0.0 && d < dmin) dmin = d;
										}
									}

								printf ("  Documented as Range_Domain\n");
								printf ("  %d negative, %d positive, %d zero values\n",ncount,pcount,zcount);
								if (dialog) {
									printf ("  What are the units of its values?: ");
									fgets (response,4096,stdin);
									s = response + strlen (response) - 1;
									if (*s == '\n') *s-- = 0;
									if (*s == '\r') *s-- = 0;
									while (s >= response && isspace(*s)) *s-- = 0;
									if (strlen (response) == 0)
										printf ("  You'll need to fill in Attribute_Units_of_Measure for %s.\n",field[j].name);
									s = response;
									while (*s && isspace(*s)) s++;
									strcpy (units,s);

									printf ("  What is the resolution of its values?: ");
									fgets (response,4096,stdin);
									s = response + strlen (response) - 1;
									if (*s == '\n') *s-- = 0;
									if (*s == '\r') *s-- = 0;
									while (s >= response && isspace(*s)) *s-- = 0;
									s = response;
									while (*s && isspace(*s)) s++;
									resolution = strtod (s,&e);
									if (e == s)
										printf ("  You'll need to fill in Attribute_Measurement_Resolution for %s.\n",field[j].name);
									}

								/* If only one negative value, enumerate it */

								if (ncount == 1) {
									if (do_xml) {
										fprintf (out,"<edom>\n");
										fprintf (out,"<edomv>%.*lf</edomv>\n",field[j].precision,dneg);
										fprintf (out,"<edomvd></edomvd>\n");
										}
									else {
										fprintf (out,"      Enumerated_Domain:\n");
										fprintf (out,"        Enumerated_Domain_Value: %.*lf\n",field[j].precision,dneg);
										fprintf (out,"        Enumerated_Domain_Value_Definition:\n");
										}
									}

								/* Express the rest as a range domain */

								if (do_xml) {
									fprintf (out,"<rdom>\n");
									fprintf (out,"<rdommin>%.*lf</rdommin>\n",field[j].precision,dmin);
									fprintf (out,"<rdommax>%.*lf</rdommax>\n",field[j].precision,dmax);
									fprintf (out,"<attrunit>%s</attrunit>\n",units);
									if (resolution > 0.0)
										fprintf (out,"<attrmres>%.*lf</attrmres>\n",field[j].precision,resolution);
									else
										fprintf (out,"<attrmres></attrmres>\n");
									fprintf (out,"</rdom>\n");
									}
								else {
									fprintf (out,"      Range_Domain:\n");
									fprintf (out,"        Range_Domain_Minimum: %.*lf\n",field[j].precision,dmin);
									fprintf (out,"        Range_Domain_Maximum: %.*lf\n",field[j].precision,dmax);
									fprintf (out,"        Attribute_Units_of_Measure: %s\n",units);
									if (resolution > 0.0)
										fprintf (out,"        Attribute_Measurement_Resolution: %.*lf\n",field[j].precision,resolution);
									else
										fprintf (out,"        Attribute_Measurement_Resolution:\n");
									}
								break;

							case FTInvalid:
								fprintf (out,"<udom></udom>\n");
								printf ("  %s left undocumented as Unrepresentable_Domain\n",field[j].name);
								break;

							default:
								printf ("  %s has an unrecognized type %d\n",field[j].name,field[j].type);
								break;
							}
						if (do_xml) {
							fprintf (out,"</attrdomv>\n");
							fprintf (out,"</attr>\n");
							}
						}
					}
				if (do_xml) {
					fprintf (out,"</detailed>\n");
					}
				fclose (out);
				}
			else {
				fprintf (stderr,"Error: could not create output file %s\n",output_file);
				DBFClose (df);
				exit (1);
				}
			DBFClose (df);
			}
		else {
			fprintf (stderr,"Error: could not open input file %s\n",input_file);
			exit (1);
			}
		}
	else {
		fprintf (stderr,"Usage: %s input_file [-i] [-not field1] [-not field2] [-strict] [-o output_file]\n",argv[0]);
		fprintf (stderr,"       -i                interactive; asks you about the data\n");
		fprintf (stderr,"       -not label        omit description of field with this label\n");
		fprintf (stderr,"       -strict           omit ESRI extensions for type, width, and precision\n");
		fprintf (stderr,"       -detail           show distribution of enumerated and integer values\n");
		fprintf (stderr,"       -o output_file    write output to the specified file\n");
		fprintf (stderr,"\n");
		fprintf (stderr,"Output format: Indented text is the default.  If the name of\n");
		fprintf (stderr,"               your output file ends in .xml, .sgml, or .sgm,\n");
		fprintf (stderr,"               then the output will be written using XML tags\n");
		fprintf (stderr,"               (the SGML and XML representations are the same).\n");
		exit (0);
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
