/*----------------------------------------------------------------------*\
 | dbfuniq.c is a program that reads a DBF file and generates from it	|
 | a list of distinct values for each character variable, sorted.		|
 |																		|
 | Based on the dbf access routines developed by Frank Warmerdam as		|
 | part of his shapelib C library for reading and writing shapefiles.	|
 | <http://gdal.velocet.ca/projects/shapelib/>  Thanks, Frank!			|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <unistd.h>
#include <shapefil.h>

#include "stricmp.h"

static char version[] = "1.5";

/*----------------------------------------------------------------------*\
 | Revision history:													|
 |																		|
 | 1.5 (20020206) PNS Modified to open dbf file read-only and to report	|
 |                    duplicate names to stdout rather than stderr.		|
 |																		|
 | 1.4 (20010926) PNS Added code to report the number of blank records	|
 |                    in a string field.								|
 |																		|
 | 1.3 (20010227) PNS Revised handling of attribute information by		|
 |                    reading all fields at once.  This enables me to	|
 |                    check for duplicate field names.					|
 |																		|
 | 1.2 (20010126) PNS Added code to skip attributes whose names match	|
 |                    those that are intrinsic to Arc/Info.				|
 |                    Fixed bug in counting unique string values.		|
 |																		|
 | 1.1 (20010125) PNS Modified handling of enumerated domains to give	|
 |                    some statistical information about the values.	|
 |																		|
 | 1.0 (20010124) PNS First release.									|
\*----------------------------------------------------------------------*/

struct field_info {
	char name [16];
	DBFFieldType type;
	int width;
	int precision;
	};

static char *type_of (DBFFieldType t) {
	if (t == FTString) return ("String");
	if (t == FTInteger) return ("Integer");
	if (t == FTDouble) return ("Double");
	if (t == FTInvalid) return ("Invalid");
	return ("Unknown");
	}

#define INTEGERS_PROBABLY_ARE_CATEGORIES	20
#define MAX_TEXT_LENGTH	4096
static char default_output[] = "dbfmeta.out";

static char *intrinsic_attribute [] = {
	"AREA",
	"PERIMETER",
	"LENGTH",
	"FNODE_",
	"TNODE_",
	"LPOLY_",
	"RPOLY_",
	NULL
	};

static int prev_name_intrinsic = 0;

static int intrinsic (char *name) {
	int i;

	for (i=0; intrinsic_attribute[i]; i++)
		if (stricmp (name,intrinsic_attribute[i]) == 0) break;

	if (intrinsic_attribute[i]) {
		prev_name_intrinsic = 1;
		return (1);
		}

	/*------------------------------------------------------------------*\
	 | Dicy part.  Look for cover# and cover-ID, which show up in DBFs	|
	 | as NAME_ and NAME_ID immediately following an intrinsic.			|
	\*------------------------------------------------------------------*/

	if (prev_name_intrinsic) {
		char *s = name;
		while (*s && isalnum(*s)) s++;
		if (strcmp (s,"_") == 0 || strcmp (s,"_ID") == 0) return (1);
		else {
			prev_name_intrinsic = 0;
			return (0);
			}
		}
	else
		return (0);
	}

struct counted_value {
	char *text;
	int number;
	int count;
	};

/* Neat-o string compare function from Dr. Dobbs article */

static int cvs_compare (const void *e1, const void *e2) {
	const struct counted_value *v1 = (const struct counted_value *) e1;
	const struct counted_value *v2 = (const struct counted_value *) e2;
	const char *s1 = v1->text;
	const char *s2 = v2->text;
	int q;

	while (*s1 && *s2)
		if (isdigit(*s1) && isdigit(*s2)) {
			if (q = atoi (s1) - atoi (s2)) return (q);
			while (isdigit(*s1)) s1++;
			while (isdigit(*s2)) s2++;
			}
		else
			if (q = tolower(*s1) - tolower(*s2)) return (q);
			else {
				s1++;
				s2++;
				}
	return (*s1 - *s2);
	}

static int cvi_compare (const void *e1, const void *e2) {
	const struct counted_value *v1 = (const struct counted_value *) e1;
	const struct counted_value *v2 = (const struct counted_value *) e2;
	const int n1 = v1->number;
	const int n2 = v2->number;
	return (n1 - n2);
	}

int main (int argc, char *argv[]) {
	char *input_file = NULL;
	char output_file [FILENAME_MAX];
	FILE *out;
	DBFHandle df;
	int rc,fc,max_width;
	int i,j,k;
	DBFFieldType ft[1024];
	char *string,*s,*e;
	double d,dmin,dmax;
	int m,mmin,mmax,mcount;
	struct counted_value *pv;
	int count,ncount,pcount,zcount;
	char *pool;
	char *pool_ptr;
	struct field_info *field = NULL;

	int dialog = 0;
	int strict = 0;
	int detail = 0;
	char response [MAX_TEXT_LENGTH];
	char definition [MAX_TEXT_LENGTH];
	char units [MAX_TEXT_LENGTH];
	double resolution;

	*output_file = 0;
	if (argc > 1)
		for (i=1; i < argc; i++)
			if (memcmp (argv[i],"-o",2) == 0) {
				i++;
				if (i < argc) strcpy (output_file,argv[i]);
				}
			else
				if (memcmp (argv[i],"-i",2) == 0)
					dialog = 1;
				else
					if (memcmp (argv[i],"-strict",7) == 0)
						strict = 1;
					else
						if (memcmp (argv[i],"-detail",7) == 0)
							detail = 1;
						else
							input_file = argv[i];

	printf ("dbfuniq %s - Peter N. Schweitzer (U.S. Geological Survey)\n",version);

	if (input_file) {

		/*--------------------------------------------------------------*\
		 | If the user didn't give us an output file, try to write to a	|
		 | default name.  But don't write over a file with that name if	|
		 | one exists already.											|
		\*--------------------------------------------------------------*/

		if (!*output_file) {
			char *reason;
			if (access (default_output,F_OK) != 0)
				if (out = fopen (default_output,"w")) {
					fclose (out);
					strcpy (output_file,default_output);
					}
				else
					reason = "cannot be created";
			else
				reason = "exists already";
			if (!*output_file) {
				printf ("You didn't specify an output file name, and my default name %s %s.\n",default_output,reason);
				printf ("Where do you want me to put the output?: ");
				fgets (response,4096,stdin);
				s = response + strlen (response) - 1;
				if (*s == '\n') *s-- = 0;
				if (*s == '\r') *s-- = 0;
				while (s >= response && isspace(*s)) *s-- = 0;
				s = response;
				while (*s && isspace(*s)) s++;
				if (*s) strcpy (output_file,s);
				else {
					printf ("I need to put it somewhere!\n");
					exit (1);
					}
				}
			else
				printf ("Writing output to file %s\n",default_output);
			}

		/* Open and process the input file */

		if (df = DBFOpen (input_file,"rb")) {

			rc = DBFGetRecordCount (df);
			fc = DBFGetFieldCount (df);

			/* Allocate space for the field info */

			if (field = (struct field_info *) malloc (fc * sizeof (struct field_info))) {
				for (i=0; i < fc; i++)
					field[i].type = DBFGetFieldInfo (df,i,field[i].name,&field[i].width, &field[i].precision);

				/*------------------------------------------------------*\
				 | Look at the names to make sure there aren't any		|
				 | duplicate names.  Warn user if there are.			|
				\*------------------------------------------------------*/

				for (i=0; i < fc - 1; i++)
					for (j=i+1; j < fc; j++)
						if (strcmp (field[i].name,field[j].name) == 0)
							printf ("WARNING: Fields %d and %d have the same name \"%s\"\n",i+1,j+1,field[i].name);
				}

			/* Allocate space for the values */

			if (pv = (struct counted_value *) malloc (rc * sizeof (struct counted_value))) {
				memset (pv,0,rc*sizeof(struct counted_value));
				}
			else {
				fprintf (stderr,"Error: could not allocate space for %d struct counted_text\n",rc);
				exit (1);
				}

			/* Create a pool for string values */

			max_width = 0;
			for (j=0; j < fc; j++) {
				if (field[j].width > max_width) max_width = field[j].width;
				}
			if (pool = (char *) malloc (rc*max_width)) {
				memset (pool,0,rc*max_width);
				pool_ptr = pool;
				}
			else {
				fprintf (stderr,"Error: could not allocate string pool of size %d\n",rc*max_width);
				exit (1);
				}

			if (out = fopen (output_file,"w")) {
				/* Read each field and describe what you can */

				for (j=0; j < fc; j++) {
					*definition = 0;
					*units = 0;
					resolution = 0.0;

					if (intrinsic (field[j].name)) {}
					else {
						if (dialog) {
							printf ("  What is its definition?: ");
							fgets (response,4096,stdin);
							s = response + strlen (response) - 1;
							if (*s == '\n') *s-- = 0;
							if (*s == '\r') *s-- = 0;
							while (s >= response && isspace(*s)) *s-- = 0;
							if (strlen (response) == 0)
								printf ("  You'll need to fill in Attribute_Definition for this attribute.\n");
							s = response;
							while (*s && isspace(*s)) s++;
							strcpy (definition,s);
							}

						fprintf (out,"Attribute_Label: %s\n",field[j].name);
						switch (field[j].type) {
							case FTString:

								/*------------------------------------------*\
								 | Strings are always Enumerated_Domains.	|
								 | If they're really place names or other	|
								 | self-sufficient labels, the user can		|
								 | reasonably omit them.  But we can't make	|
								 | that decision for the user.				|
								\*------------------------------------------*/

								count = 0;
								for (i=0; i < rc; i++)
									if (string = (char *) DBFReadStringAttribute (df,i,j)) {
										/* Have we seen this attribute value already? */
										for (k=0; pv[k].text; k++)
											if (strcmp (string,pv[k].text) == 0) {
												pv[k].count++;
												break;
												}
										/* If not, add it to our list */
										if (pv[k].text == NULL) {
											pv[k].text = pool_ptr;
											strcpy (pv[k].text,string);
											pool_ptr += 1 + strlen (string);
											pv[k].count = 1;
											count++;
											}
										}
								qsort (pv,count,sizeof (struct counted_value),cvs_compare);
								for (k=0; k < count; k++) {
									for (s=pv[k].text; *s; s++)
										if (*s == '\r' || *s == '\n') *s = ' ';
									if (*pv[k].text)
										fprintf (out,"%s\n",pv[k].text);
									else
										fprintf (out,"(no value)\n");
									}
								memset (pv,0,rc*sizeof(struct counted_value));
								pool_ptr = pool;
								break;

							case FTInteger:
								break;

							case FTDouble:
								break;

							case FTInvalid:
								break;

							default:
								break;
							}
						}
					}
				fclose (out);
				}
			else {
				fprintf (stderr,"Error: could not create output file %s\n",output_file);
				DBFClose (df);
				exit (1);
				}
			DBFClose (df);
			}
		else {
			fprintf (stderr,"Error: could not open input file %s\n",input_file);
			exit (1);
			}
		}
	else {
		fprintf (stderr,"Usage: %s input_file [-i] [-strict] [-o output_file]\n",argv[0]);
		fprintf (stderr,"       -i                interactive; asks you about the data\n");
		fprintf (stderr,"       -strict           omit ESRI extensions for type, width, and precision\n");
		fprintf (stderr,"       -detail           show distribution of enumerated and integer values\n");
		fprintf (stderr,"       -o output_file    write output to the specified file\n");
		exit (0);
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
