/*----------------------------------------------------------------------*\
 | Code to write DIF version of formal metadata.						|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "item.h"
#include "stricmp.h"

/*----------------------------------------------------------------------*\
 | Functions to retrieve a keyword's text given its symbolic name and	|
 | vice versa.															|
\*----------------------------------------------------------------------*/

enum dif_keyword {
	DEntry_ID,
	DEntry_Title,
	DStart_date,
	DStop_date,
	DSensor_Name,
	DSource_Name,
	DInvestigator,
	DTechnical_Contact,
	DAuthor,
	DData_Center,
	DOriginating_Center,
	DCampaign_or_Project,
	DStorage_Medium,
	DParameter,
	DDiscipline,
	DCoverage,
	DLocation,
	DKeyword,
	DRevision_Date,
	DScience_Review_Date,
	DFuture_Review_Date,
	DReference,
	DQuality,
	DSummary,
	DGroup,
	DEnd_Group,
	DData_Center_Name,
	DDataset_ID,
	DData_Center_Contact,
	DLast_name,
	DFirst_name,
	DMiddle_name,
	DEmail_address,
	DPhone,
	DAddress,
	DCampaign,
	DProject,
	DLat_Long_Coverage,
	DMinimum_latitude,
	DMaximum_latitude,
	DMinimum_longitude,
	DMaximum_longitude,
	DCoord_System,
	DRA_Dec_Coverage,
	DMinimum_RA,
	DMaximum_RA,
	DMinimum_Dec,
	DMaximum_Dec,
	DEpoch,
	DEquinox,
	DGal_Coverage,
	DMinimum_Gal_Lat,
	DMaximum_Gal_Lat,
	DMinimum_Gal_Long,
	DMaximum_Gal_Long,
	DEarth,
	Dunknown,
	Dblank,
	Dnull
	};

struct dif_keyword_list {
	enum dif_keyword key;
	char *text;
	};

static struct dif_keyword_list dif_keyword_list[] = {
	{DEntry_ID,				"Entry_ID"},
	{DEntry_Title,			"Entry_Title"},
	{DStart_date,			"Start_date"},
	{DStop_date,			"Stop_date"},
	{DSensor_Name,			"Sensor_Name"},
	{DSource_Name,			"Source_Name"},
	{DInvestigator,			"Investigator"},
	{DTechnical_Contact,	"Technical_Contact"},
	{DAuthor,				"Author"},
	{DData_Center,			"Data_Center"},
	{DOriginating_Center,	"Originating_Center"},
	{DCampaign_or_Project,	"Campaign_or_Project"},
	{DStorage_Medium,		"Storage_Medium"},
	{DParameter,			"Parameter"},
	{DDiscipline,			"Discipline"},
	{DCoverage,				"Coverage"},
	{DLocation,				"Location"},
	{DKeyword,				"Keyword"},
	{DRevision_Date,		"Revision_Date"},
	{DScience_Review_Date,	"Science_Review_Date"},
	{DFuture_Review_Date,	"Future_Review_Date"},
	{DReference,			"Reference"},
	{DQuality,				"Quality"},
	{DSummary,				"Summary"},
	{DGroup,				"Group"},
	{DEnd_Group,			"End_Group"},
	{DData_Center_Name,		"Data_Center_Name"},
	{DDataset_ID,			"Dataset_ID"},
	{DData_Center_Contact,	"Data_Center_Contact"},
	{DLast_name,			"Last_name"},
	{DFirst_name,			"First_name"},
	{DMiddle_name,			"Middle_name"},
	{DEmail_address,		"Email_address"},
	{DPhone,				"Phone"},
	{DAddress,				"Address"},
	{DCampaign,				"Campaign"},
	{DProject,				"Project"},
	{DLat_Long_Coverage,	"Lat_Long_Coverage"},
	{DMinimum_latitude,		"Minimum_latitude"},
	{DMaximum_latitude,		"Maximum_latitude"},
	{DMinimum_longitude,	"Minimum_longitude"},
	{DMaximum_longitude,	"Maximum_longitude"},
	{DCoord_System,			"Coord_System"},
	{DRA_Dec_Coverage,		"RA_Dec_Coverage"},
	{DMinimum_RA,			"Minimum_RA"},
	{DMaximum_RA,			"Maximum_RA"},
	{DMinimum_Dec,			"Minimum_Dec"},
	{DMaximum_Dec,			"Maximum_Dec"},
	{DEpoch,				"Epoch"},
	{DEquinox,				"Equinox"},
	{DGal_Coverage,			"Gal_Coverage"},
	{DMinimum_Gal_Lat,		"Minimum_Gal_Lat"},
	{DMaximum_Gal_Lat,		"Maximum_Gal_Lat"},
	{DMinimum_Gal_Long,		"Minimum_Gal_Long"},
	{DMaximum_Gal_Long,		"Maximum_Gal_Long"},
	{DEarth,				"Earth"},
	{Dunknown,				"(unknown)"},
	{Dblank,				""},
	{Dnull,					NULL}
	};

static int dif_keyword_list_length = 0;

/*----------------------------------------------------------------------*/

static int compare (const void *e1, const void *e2) {
	struct dif_keyword_list *p = (struct dif_keyword_list *) e1;
	struct dif_keyword_list *q = (struct dif_keyword_list *) e2;
	return (strcmp (p->text,q->text));
	}

static enum dif_keyword dif_key_of (char *text) {
	int i;
	struct dif_keyword_list k,*key;
	char string[64];

	if (dif_keyword_list_length == 0) {
		for (i=0; dif_keyword_list[i].text; i++);
		dif_keyword_list_length = i;
		qsort (dif_keyword_list,dif_keyword_list_length,sizeof(struct dif_keyword_list),compare);
		}
	if (!text) return (Dnull);
	if (*text) {
		char *s = string;
		char *t = text;
		while (*t && !isspace(*t) && *t != ':' && *t != '=') *s++ = *t++;
		*s = 0;
		k.text = string;
		key = (struct dif_keyword_list *) bsearch (&k,dif_keyword_list,dif_keyword_list_length,sizeof(struct dif_keyword_list),compare);
		if (key) return (key->key);
		else return (Dunknown);
		}
	else return (Dblank);
	}

static char *dif_text_of (enum dif_keyword key) {
	int i;

	if (dif_keyword_list_length == 0) {
		for (i=0; dif_keyword_list[i].text; i++);
		dif_keyword_list_length = i;
		qsort (dif_keyword_list,dif_keyword_list_length,sizeof(struct dif_keyword_list),compare);
		}
	if (key == Dnull) return ("(null)");
	if (key == Dblank) return ("(blank)");
	for (i=0; i < dif_keyword_list_length; i++)
		if (dif_keyword_list[i].key == key) return (dif_keyword_list[i].text);
	return ("(unknown)");
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static struct item *parse_dif (char *name) {
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Code to determine whether a given string is a DIF parameter, and if	|
 | it is a DIF minor parameter, give the corresponding major parameter.	|
\*----------------------------------------------------------------------*/

struct dif_parameter {
	char *major;
	char *minor[64];
	};

static struct dif_parameter dif_parameter_list[] = {
	{"ANATOMICAL PARAMETERS",{
		"Body Regions",
		"Cardiovascular System",
		"Cells",
		"Digestive System",
		"Embryonic Structures",
		"Endocrine System",
		"Fluids and Secretions",
		"Hemic and Immune Systems",
		"Musculoskeletal System",
		"Nervous System",
		"Reproductive System",
		"Respiratory System",
		"Sense Organs",
		"Stomatognathic System",
		"Tissue Types",
		"Urogenital System",
		NULL
		}
	},

	{"ASTRONOMICAL PARAMETERS",{
		"Abundances",
		"Bibliography",
		"Binaries",
		"Colors",
		"Cross Identifications",
		"Ephemerides",
		"Imagery",
		"Magnetic Fields",
		"Magnitudes",
		"Masses",
		"Models",
		"Morphology",
		"Object Counts",
		"Occultations",
		"Oscillations",
		"Parallaxes",
		"Photometry",
		"Polarization",
		"Positions",
		"Proper Motions",
		"Radial Velocities",
		"Reddening",
		"Redshifts",
		"Rotational Velocities",
		"Space Velocities",
		"Spectra",
		"Spectrophotometry",
		"Variability",
		NULL
		}
	},

	{"ATMOSPHERIC COMPOSITION",{
		"Aerosols",
		"Air Quality",
		"Ash",
		"Carbon Dioxide",
		"Carbon Monoxide",
		"Chlorofluorocarbons",
		"Clouds",
		"Contaminants",
		"Humidity",
		"Methane",
		"Nitric Acid",
		"Nitrous Oxide",
		"Nitrogen",
		"Nitrogen Dioxide",
		"Nitrogen Oxides",
		"Non_Methane Hydrocarbons",
		"Oxygen",
		"Ozone",
		"Trace Elements",
		"Trace Gases",
		"Tracers",
		"Water Vapor",
		NULL
		}
	},

	{"ATMOSPHERIC DYNAMICS",{
		"Altitude",
		"Atmospheric Temperature",
		"Cloud Types",
		"Evaporation",
		"Evapotranspiration",
		"Geopotential Height",
		"Heat Flux",
		"Humidity",
		"Lightning",
		"Paleoclimate Indices",
		"Precipitation",
		"Pressure",
		"Solar Radiation",
		"Storms",
		"Visibility",
		"Winds",
		NULL
		}
	},

	{"BIOLOGICAL ENTITIES",{
		"Birds",
		"Domesticated Animals",
		"Domesticated Plants",
		"Endangered Species",
		"Land Wildlife",
		"Microorganisms",
		"Minor Species",
		"Ocean Vegetation",
		"Ocean Wildlife",
		"Surface Vegetation",
		NULL
		}
	},

	{"BIOTECHNOLOGY",{
		"Cells",
		"Cell Cultures",
		"Crystal Growth",
		"Proteins",
		"Separations",
		NULL
		}
	},

	{"CHARGED PARTICLES",{
		"Alpha Particles",
		"Composition",
		"Density",
		"Differential Flux",
		"Distribution Functions",
		"Electron Flux",
		"Energetic Particles",
		"Heavy Ions",
		"Proton Flux",
		"Speed",
		"Temperature",
		NULL
		}
	},

	{"DISEASES",{
		"Addiction",
		"Bacterial",
		"Cardiovascular",
		"Chronic",
		"Communicable",
		"Dermatologic",
		"Digestive System",
		"Endocrine",
		"Eye",
		"Fungal",
		"Immunologic",
		"Infection",
		"Injury",
		"Musculoskeletal",
		"Neonatal",
		"Neoplasms",
		"Nervous System",
		"Nutritional and Metabolic",
		"Occupational",
		"Ophthalmic",
		"Otorhinolaryngologic",
		"Parasitic",
		"Poisoning",
		"Pregnancy Complications",
		"Respiratory",
		"Skin",
		"Stomatognathic",
		"Urologic",
		"Virus",
		NULL
		}
	},

	{"EARTH RADIATIVE PROCESSES",{
		"Albedo",
		"Brightness Temperature",
		"Heat Flux",
		"Irradiance",
		"Radiance",
		"Solar Activity",
		"Temperature",
		"Thermal Inertia",
		NULL
		}
	},

	{"GEODYNAMIC FEATURES",{
		"Earthquakes",
		"Erosion",
		"Geodesy",
		"Geothermal",
		"Gravity Fields",
		"Magnetic Fields",
		"Polar Motion",
		"Seismic",
		"Structures",
		"Tectonophysics",
		"Terrain Elevation",
		"Volcanoes",
		NULL
		}
	},

	{"GEOGRAPHY AND LAND COVER",{
		"Albedo",
		"Cultural Features",
		"Elevation",
		"Fires",
		"Glaciers",
		"Ice",
		"Lakes",
		"Landforms",
		"Rivers",
		"Snow",
		"Soils",
		"Surface Vegetation",
		"Surface Water",
		"Topographic Data",
		"Wetlands",
		NULL
		}
	},

	{"GEOLOGICAL PARAMETERS",{
		"Age Determinations",
		"Coal",
		"Economic Minerals",
		"Geochemical Analysis",
		"Igneous and Metamorphic Rocks",
		"Lithology",
		"Mineralogy and Crystallography",
		"Paleontology",
		"Petroleum",
		"Petrology",
		"Sedimentary Rocks",
		"Soils",
		"Stratigraphy",
		NULL
		}
	},

	{"HEALTH CARE",{
		"Clinical Care",
		"Community Care",
		"Institutional Care",
		NULL
		}
	},

	{"HYDROLOGIC PARAMETERS",{
		"Contamination",
		"Deposition",
		"Erosion",
		"Evaporation",
		"Evapotranspiration",
		"Glaciers",
		"Ground Water",
		"Infiltration",
		"Oxygen Demand",
		"Precipitation",
		"Rivers",
		"Runoff",
		"Sedimentation",
		"Soil Moisture",
		"Solids",
		"Surface Water",
		"Temperature",
		"Turbidity",
		"Water Vapor",
		"Wetlands",
		NULL
		}
	},

	{"MAGNETIC AND ELECTRIC FIELDS",{
		"Activity Indices",
		"Electric Fields (DC)",
		"Electric Wave Spectra (AC)",
		"Magnetic Fields (DC)",
		"Magnetic Wave Spectra (AC)",
		NULL
		}
	},

	{"MATERIAL SCIENCE",{
		"Containerless Processing",
		"Crystal Growth",
		"Electronic & Photonic Materials",
		"Glasses and Ceramics",
		"Metals and Alloys",
		"Solidification",
		"Thermophysical Properties",
		NULL
		}
	},

	{"NEUTRAL PARTICLES",{
		"Composition",
		"Density",
		"Distribution Functions",
		"Flux",
		"Speed",
		"Temperature",
		NULL
		}
	},

	{"OCEAN COMPOSITION",{
		"Alkalinity",
		"Aquatic Plants",
		"Biomass",
		"Carbon Dioxide",
		"Chemical Tracers",
		"Chlorophyll",
		"Conductivity",
		"Dissolved Solids",
		"Light Transmission",
		"Major Elements",
		"Minor Species",
		"Nitrate",
		"Nitric Acid",
		"Nitrite",
		"Nitrogen",
		"Nitrogen Dioxide",
		"Nutrients",
		"Ocean Wildlife",
		"Organic Matter",
		"Oxygen",
		"pH",
		"Phosphates",
		"Phytoplankton",
		"Pigment Concentration",
		"Pollutants",
		"Salinity",
		"Sea Ice",
		"Sediments",
		"Silicate",
		"Suspended Solids",
		"Trace Elements",
		"Upwelling",
		"Zooplankton",
		NULL
		}
	},

	{"OCEAN DYNAMICS",{
		"Bathymetry",
		"Brightness Temperature",
		"Currents",
		"Evaporation",
		"Geopotential Height",
		"Heat Flux",
		"Pressure",
		"Primary Production",
		"Sea Ice",
		"Sea Level",
		"Sea Surface Height",
		"Sedimentation",
		"Swell",
		"Temperature",
		"Tides",
		"Turbidity",
		"Upwelling",
		"Waves",
		"Winds",
		NULL
		}
	},

	{"PHYSIOLOGICAL PARAMETERS",{
		"Adaptation",
		"Aging",
		"Body Constitution",
		"Body Temperature",
		"Cell Count",
		"Cell Physiology",
		"Chronobiology",
		"Circulatory Physiology",
		"Digestive Physiology",
		"Electrophysiology",
		"Eye Physiology",
		"Growth and Development",
		"Hematology",
		"Homeostasis",
		"Musculoskeletal Physiology",
		"Neural Physiology",
		"Oral Physiology",
		"Reproduction Physiology",
		"Respiratory Physiology",
		"Skin Physiology",
		"Urogenital Physiology",
		NULL
		}
	},

	{"PUBLIC HEALTH",{
		"Accidents",
		"Behavior",
		"Disease Outbreaks",
		"Drug Contamination",
		"Environmental Health",
		"Epidemics",
		"Epidemiologic Measurements",
		"Food Poisoning",
		"Nutrition",
		NULL
		}
	},

	{"RADIANCE AND IMAGERY",{
		"Gamma Ray",
		"Infrared",
		"Microwave",
		"Radar",
		"Radio Wave",
		"Ultraviolet",
		"Visible",
		"X-ray",
		NULL
		}
	},

	{"SOLAR PROPERTIES",{
		"Active Regions",
		"Coronal Holes",
		"Coronal Properties",
		"Events",
		"Filaments",
		"Flares",
		"Imagery",
		"Oscillations",
		"Prominences",
		"Sunspots",
		"Synoptic Maps",
		"Velocity Fields",
		NULL
		}
	},

	{"VITAL STATISTICS",{
		"Demography",
		"Morbidity",
		"Mortality",
		NULL
		}
	},

	{NULL,NULL}
	};

static int is_dif_major_parameter (char *string) {
	struct dif_parameter *dp = dif_parameter_list;
	while (dp->major) {
		if (stricmp (string,dp->major) == 0) return (1);
		dp++;
		}
	return (0);
	}

static int is_dif_minor_parameter (char *string) {
	struct dif_parameter *dp = dif_parameter_list;
	char **m;
	while (dp->major) {
		m = dp->minor;
		while (*m) {
			if (stricmp (string,*m) == 0) return (1);
			m++;
			}
		dp++;
		}
	return (0);
	}

static char *major_dif_parameter_of (char *string) {
	struct dif_parameter *dp = dif_parameter_list;
	char **m;
	while (dp->major) {
		m = dp->minor;
		while (*m) {
			if (stricmp (string,*m) == 0) return (dp->major);
			m++;
			}
		dp++;
		}
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Code to determine whether a given string is a DIF discipline, and if	|
 | it is a DIF subdiscipline, give the corresponding discipline.		|
\*----------------------------------------------------------------------*/

struct dif_discipline {
	char *major;
	char *minor[64];
	};

static struct dif_discipline dif_discipline_list[] = {
	{"ASTRONOMY",{
		"Cosmic Ray Astronomy",
		"Gamma Ray Astronomy",
		"Infrared Astronomy",
		"Microwave Astronomy",
		"Radio Astronomy",
		"Ultraviolet Astronomy",
		"Visible Astronomy",
		"X-Ray Astronomy",
		NULL
		}
	},

	{"EARTH SCIENCE",{
		"Atmosphere",
		"Interior and Crust",
		"Land",
		"Ocean",
		NULL
		}
	},

	{"LIFE SCIENCES",{
		"Anatomy",
		"Biochemistry",
		"Biology",
		"Biophysics",
		"Chemistry",
		"Medical Science",
		"Neuroscience",
		"Physiology",
		"Psychology",
		"Public Health",
		NULL
		}
	},

	{"MICROGRAVITY",{
		NULL
		}
	},

	{"PLANETARY SCIENCE",{
		"Atmospheres",
		"Fields and Particles",
		"Geosciences",
		"Rings",
		"Small Bodies",
		NULL
		}
	},

	{"SOLAR PHYSICS",{
		"Gamma-Ray Observations",
		"Infrared Observations",
		"Microwave Observations",
		"Radio Observations",
		"Ultraviolet Observations",
		"Visible Observations",
		"X-Ray Observations",
		NULL
		}
	},

	{"SPACE PHYSICS",{
		"Interplanetary Studies",
		"Ionospheric Science",
		"Magnetospheric Science",
		NULL
		}
	},
	{NULL,NULL}
	};

static int is_dif_major_discipline (char *string) {
	struct dif_discipline *dp = dif_discipline_list;
	while (dp->major) {
		if (stricmp (string,dp->major) == 0) return (1);
		dp++;
		}
	return (0);
	}

static int is_dif_minor_discipline (char *string) {
	struct dif_discipline *dp = dif_discipline_list;
	char **m;
	while (dp->major) {
		m = dp->minor;
		while (*m) {
			if (stricmp (string,*m) == 0) return (1);
			m++;
			}
		dp++;
		}
	return (0);
	}

static char *major_dif_discipline_of (char *string) {
	struct dif_discipline *dp = dif_discipline_list;
	char **m;
	while (dp->major) {
		m = dp->minor;
		while (*m) {
			if (stricmp (string,*m) == 0) return (dp->major);
			m++;
			}
		dp++;
		}
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Code to determine whether a given string is a DIF location keyword.	|
\*----------------------------------------------------------------------*/

static char *dif_location[] = {
	"Africa",
	"Antarctica",
	"Arctic Ocean",
	"Asia",
	"Asteroids",
	"Atlantic Ocean",
	"Australia",
	"Boundary Layer",
	"Chromosphere",
	"Clusters of Galaxies",
	"Comets",
	"Core",
	"Corona",
	"Crust",
	"Equatorial",
	"Europe",
	"Extended Objects (Astronomy)",
	"Galaxies",
	"Global",
	"High Latitude Magnetosphere",
	"Indian Ocean",
	"Inner Magnetosphere",
	"Interplanetary (deep space)",
	"Interplanetary (near Earth)",
	"Interstellar Medium",
	"Ionosphere",
	"Jupiter",
	"Local Group of Galaxies",
	"Magnetosphere (other)",
	"Magnetotail",
	"Mantle",
	"Mars",
	"Mediterranean Sea",
	"Mercury",
	"Mesosphere",
	"Meteoroids",
	"Mid-Latitude",
	"Milky Way Galaxy",
	"Moon (Earth)",
	"Moons (other)",
	"Neptune",
	"North America",
	"Novae",
	"Pacific Ocean",
	"Photosphere",
	"Pluto",
	"Polar",
	"Quasars",
	"Radio Sources",
	"Rings (planetary)",
	"Saturn",
	"Sea Floor",
	"Sea Surface",
	"Solar Interior",
	"South America",
	"Southern Ocean",
	"Star Clusters",
	"Stars",
	"Stratosphere",
	"Supernova Remnants",
	"Supernovae",
	"Transition Region",
	"Troposphere",
	"Uranus",
	"Venus",
	NULL
	};

static int is_dif_location (char *string) {
	char **l = dif_location;
	while (*l) {
		if (stricmp (*l,string) == 0) return (1);
		l++;
		}
	return (0);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static void parse_name (char *name, char *first, char *middle, char *last) {
	char *b,*e;

	b = name;
	if (e = strchr (name,',')) {	/* Last, First Middle */
		if (e - b < 32) {
			memcpy (last,b,e-b);
			last[e-b] = 0;
			}
		else {
			memcpy (last,b,31);
			last[31] = 0;
			}
		e++;
		b = e; while (*b &&  isspace(*b)) b++;
		e = b; while (*e && !isspace(*e)) e++;
		if (e - b < 32) {
			memcpy (first,b,e-b);
			first[e-b] = 0;
			}
		else {
			memcpy (first,b,31);
			first[31] = 0;
			}
		b = e; while (*b && isspace(*b)) b++;
		if (strlen (b) < 32) {
			strcpy (middle,b);
			}
		else {
			memcpy (middle,b,31);
			middle[31] = 0;
			}
		}
	else {							/* First Middle Last */
		e = b; while (*e && !isspace(*e)) e++;
		if (e - b < 32) {
			memcpy (first,b,e-b);
			first[e-b] = 0;
			}
		else {
			memcpy (first,b,31);
			first[31] = 0;
			}
		b = e; while (*b &&  isspace(*b)) b++;
		e = b; while (*e && !isspace(*e)) e++;
		if (e - b < 32) {
			memcpy (middle,b,e-b);
			middle[e-b] = 0;
			}
		else {
			memcpy (middle,b,31);
			middle[31] = 0;
			}
		b = e; while (*b &&  isspace(*b)) b++;
		e = b; while (*e && !isspace(*e)) e++;

		/*--------------------------------------------------------------*\
		 | If no more words, what we put in middle is really the last	|
		 | name, otherwise store in the last name what remains.			|
		 | This won't properly parse names like "John Smith Jr."		|
		\*--------------------------------------------------------------*/

		if (e - b == 0) {
			strcpy (last,middle);
			*middle = 0;
			}
		else
			if (e - b < 32) {
				memcpy (last,b,e-b);
				last[e-b] = 0;
				}
			else {
				memcpy (last,b,31);
				last[31] = 0;
				}
		}
	}

static void write_contact_info (FILE *out, struct item *p, enum dif_keyword key) {
	char lname[256],fname[64],mname[64];
	struct item *q,*r;

	if (p->key == WContact_Information) {

		fprintf (out,"Group: %s\n",dif_text_of(key));

		/* Name */

		strcpy (lname,"N/A");
		strcpy (mname,"");
		strcpy (fname,"N/A");
		if (q = find_key (p,WContact_Person)) {
			if (q = q->child)
				parse_name (q->d,fname,mname,lname);
			}
		else
			if (q = find_key (p,WContact_Organization))
				if (q = q->child) strcpy (lname,q->d);

		fprintf (out,"  %s: %s\n",dif_text_of(DLast_name),lname);
		fprintf (out,"  %s: %s\n",dif_text_of(DFirst_name),fname);
		if (*mname) fprintf (out,"  %s: %s\n",dif_text_of(DMiddle_name),mname);

		/* Email address */

		if (q = find_key (p,WContact_Electronic_Mail_Address)) {
			if (q->child)
				fprintf (out,"  %s: INTERNET>%s\n",dif_text_of(DEmail_address),q->child->d);
			while (q = find_next_key (q,WContact_Electronic_Mail_Address))
				if (q->child)
					fprintf (out,"  %s: Internet>%s\n",dif_text_of(DEmail_address),q->child->d);
			}

		/* Phone */

		if (q = find_key (p,WContact_Voice_Telephone)) {
			if (q->child)
				fprintf (out,"  %s: %s\n",dif_text_of(DPhone),q->child->d);
			if (q = find_next_key (q,WContact_Voice_Telephone))
				if (q->child)
					fprintf (out,"  %s: %s\n",dif_text_of(DPhone),q->child->d);
			}

		/* Address */

		if (q = find_key (p,WContact_Address)) {
			fprintf (out,"  Group: Address\n");
			if (r = find_key (q,WAddress)) {
				if (r = r->child)
					do {
						fprintf (out,"    %s\n",r->d);
						r = r->next;
						} while (r);
				}
			if (r = find_key (q,WCity))
				if (r = r->child)
					fprintf (out,"    %s, ",r->d);
			if (r = find_key (q,WState_or_Province))
				if (r = r->child)
					fprintf (out,"%s ",r->d);
			if (r = find_key (q,WPostal_Code))
				if (r = r->child)
					fprintf (out,"%s",r->d);
			fprintf (out,"\n");
			if (r = find_key (q,WCountry))
				if (r = r->child)
					fprintf (out,"    %s\n",r->d);
			fprintf (out,"  End_Group\n");
			}
		else {
			fprintf (out,"  Group: %s\n",dif_text_of(DAddress));
			fprintf (out,"    N/A\n");
			fprintf (out,"  End_Group\n");
			}

		fprintf (out,"End_Group\n");
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static char *one_indent = "  ";	/* Set to desired indentation indicator */
static FILE *fp = NULL;

static void write_text_item (struct item *p, int n) {
	int i;
	if (!p) return;
	if (p->key == Wblank) fprintf (fp,"\n");
	else
		if (p->key == Wunknown) {
			for (i=0; i < n; i++) fprintf (fp,one_indent);
			fprintf (fp,"%s\n",p->d);
			}
		else {
			struct item *q;

			for (i=0; i < n; i++) fprintf (fp,one_indent);
			fprintf (fp,"%s:",text_of(p->key));
			if (p->child)
				if (p->child->key == Wunknown && p->child->next == NULL) {
					fprintf (fp," %s\n",p->child->d);
					return;
					}
			fprintf (fp,"\n");
			for (q=p->child; q; q=q->next) write_text_item (q,n+1);
			}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

void write_dif (char *output_file, struct item *first) {
	FILE *out;
	struct item *p,*q;

	if (first == NULL) return;

	if (out = fopen (output_file,"w")) {

		/*--------------------------------------------------------------*\
		 | Entry_ID
		\*--------------------------------------------------------------*/

		fprintf (out,"%s: ",dif_text_of(DEntry_ID));
		fprintf (out,"(required)");
		fprintf (out,"\n");

		/*--------------------------------------------------------------*\
		 | Entry_Title
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WCitation))
			if (p = find_key (p,WTitle))
				if (p = p->child) {
					fprintf (out,"%s: ",dif_text_of(DEntry_Title));
					if (p->next) {
						fprintf (out,"\"");
						while (p) {
							fprintf (out,"%s",p->d);
							if (p->next) fprintf (out,"\n");
							p = p->next;
							}
						fprintf (out,"\"");
						}
					else
						fprintf (out,"%s",p->d);
					fprintf (out,"\n");
					}

		/*--------------------------------------------------------------*\
		 | Start_date and Stop_date
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WIdentification_Information))
			if (p = find_key (p,WTime_Period_of_Content))
				if (p = find_key (p,WTime_Period_Information))
					if (p = p->child) {
						switch (p->key) {
							case WSingle_Date_Time:
								if (p = p->child)
									if (p->key == WCalendar_Date)
										if (p = p->child) {
											fprintf (out,"%s: ",dif_text_of(DStart_date));
											fprintf (out,"%s",p->d);
											fprintf (out,"\n");
											fprintf (out,"%s: ",dif_text_of(DStop_date));
											fprintf (out,"%s",p->d);
											fprintf (out,"\n");
											}
								break;
							case WMultiple_Dates_Times:
								for (p=p->child; p; p=p->next)
									if (p->key == WCalendar_Date) {
										if (q = p->child) {
											fprintf (out,"%s: ",dif_text_of(DStart_date));
											fprintf (out,"%s",q->d);
											fprintf (out,"\n");
											}
										}
								break;
							case WRange_of_Dates_Times:
								p = p->child;
								while (p) {
									if (p->key == WBeginning_Date)
										if (q = p->child) {
											fprintf (out,"%s: ",dif_text_of(DStart_date));
											fprintf (out,"%s",q->d);
											fprintf (out,"\n");
											}
									if (p->key == WEnding_Date)
										if (q = p->child) {
											fprintf (out,"%s: ",dif_text_of(DStop_date));
											fprintf (out,"%s",q->d);
											fprintf (out,"\n");
											}
									p = p->next;
									}
								break;
							default:
								break;
							}
						}

		/*--------------------------------------------------------------*\
		 | Sensor_Name
		\*--------------------------------------------------------------*/

		fprintf (out,"%s: ",dif_text_of(DSensor_Name));
		fprintf (out,"\n");

		/*--------------------------------------------------------------*\
		 | Source_Name
		\*--------------------------------------------------------------*/

		fprintf (out,"%s: ",dif_text_of(DSource_Name));
		fprintf (out,"\n");

		/*--------------------------------------------------------------*\
		 | Investigator
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WCitation))
			if (p = find_key (p,WOriginator))
				do {
					if (q = p->child) {
						char fname[32],mname[32],lname[32];
						parse_name (q->d,fname,mname,lname);
						fprintf (out,"Group: %s\n",dif_text_of(DInvestigator));
						fprintf (out,"  %s: %s\n",dif_text_of(DLast_name),lname);
						fprintf (out,"  %s: %s\n",dif_text_of(DFirst_name),fname);
						if (*mname) fprintf (out,"  %s: %s\n",dif_text_of(DMiddle_name),mname);
						fprintf (out,"End_Group\n");
						}
					} while (p = find_next_key (p,WOriginator));

		/*--------------------------------------------------------------*\
		 | Technical_Contact
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WCitation))
			if (p = find_key (p,WPoint_of_Contact))
				if (p = p->child)
					write_contact_info (out,p,DTechnical_Contact);

		/*--------------------------------------------------------------*\
		 | Author
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WMetadata_Reference_Information))
			if (p = find_key (p,WMetadata_Contact))
				if (p = p->child)
					write_contact_info (out,p,DAuthor);

		/*--------------------------------------------------------------*\
		 | Data_Center													|
		 |																|
		 | This conversion should be expanded significantly.  The FGDC	|
		 | standard permits many sets of Distribution_Information, but	|
		 | does not nest them separately; DIF permits many Data_Center	|
		 | groups.														|
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WDistribution_Information)) {
			fprintf (out,"Group: %s\n",dif_text_of(DData_Center));

			/*----------------------------------------------------------*\
			 | A Contact_Organization will be the Data_Center.			|
			\*----------------------------------------------------------*/

			if (q = find_key (p,WContact_Organization))
				if (q = q->child)
					fprintf (out,"  %s: %s\n",dif_text_of(DData_Center_Name),q->d);
				else
					fprintf (out,"  %s: (required)\n",dif_text_of(DData_Center_Name));
			else
				fprintf (out,"  %s: (required)\n",dif_text_of(DData_Center_Name));

			/*----------------------------------------------------------*\
			 | The first Resource_Description following the Distributor	|
			 | is the Dataset_ID.										|
			\*----------------------------------------------------------*/

			if (q = find_key (p,WResource_Description))
				if (q = q->child)
					fprintf (out,"  %s: %s\n",dif_text_of(DDataset_ID),q->d);

			/*----------------------------------------------------------*\
			 | Distributor is Data_Center_Contact.						|
			\*----------------------------------------------------------*/

			if (q = find_key (p,WDistributor))
				if (q = q->child)
					write_contact_info (out,q,DData_Center_Contact);

			fprintf (out,"End_Group\n");
			}

		/*--------------------------------------------------------------*\
		 | Originating_Center
		\*--------------------------------------------------------------*/

		fprintf (out,"%s: ",dif_text_of(DOriginating_Center));
		fprintf (out,"(required)");
		fprintf (out,"\n");

		/*--------------------------------------------------------------*\
		 | Campaign_or_Project
		\*--------------------------------------------------------------*/

		fprintf (out,"%s: ",dif_text_of(DCampaign_or_Project));
		fprintf (out,"\n");

		/*--------------------------------------------------------------*\
		 | Storage_Medium
		\*--------------------------------------------------------------*/

		fprintf (out,"%s: ",dif_text_of(DStorage_Medium));
		fprintf (out,"magnetic disk");
		fprintf (out,"\n");

		/*--------------------------------------------------------------*\
		 | Coverage
		 |
		 | Arbitrarily, I have decided that
		 |		Minimum_longitude = West_Bounding_Coordinate
		 |		Maximum_longitude = East_Bounding_Coordinate
		 |		Minimum_latitude  = South_Bounding_Coordinate
		 |		Maximum_latitude  = North_Bounding_Coordinate
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WIdentification_Information))
			if (p = find_key (p,WBounding_Coordinates)) {
				fprintf (out,"Group: %s\n",dif_text_of(DCoverage));
				if (q = find_key (p,WWest_Bounding_Coordinate))
					if (q->child)
						fprintf (out,"  %s: %s\n",dif_text_of(DMinimum_longitude),q->child->d);
				if (q = find_key (p,WEast_Bounding_Coordinate))
					if (q->child)
						fprintf (out,"  %s: %s\n",dif_text_of(DMaximum_longitude),q->child->d);
				if (q = find_key (p,WNorth_Bounding_Coordinate))
					if (q->child)
						fprintf (out,"  %s: %s\n",dif_text_of(DMaximum_latitude),q->child->d);
				if (q = find_key (p,WSouth_Bounding_Coordinate))
					if (q->child)
						fprintf (out,"  %s: %s\n",dif_text_of(DMinimum_latitude),q->child->d);
				fprintf (out,"End_Group\n");
				}

		/*--------------------------------------------------------------*\
		 | Parameter, Discipline, and (General) Keyword
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WIdentification_Information))
			if (p = find_key (p,WTheme))
				if (p = find_key (p,WTheme_Keyword))
					do {
						if (q = p->child)
							if (is_dif_major_parameter (q->d)) {
								fprintf (out,"%s: ",dif_text_of(DParameter));
								fprintf (out,"%s",q->d);
								fprintf (out,"\n");
								}
							else
								if (is_dif_minor_parameter (q->d)) {
									fprintf (out,"%s: ",dif_text_of(DParameter));
									fprintf (out,"%s>",major_dif_parameter_of (q->d));
									fprintf (out,"%s",q->d);
									fprintf (out,"\n");
									}
								else
									if (is_dif_major_discipline (q->d)) {
										fprintf (out,"%s: ",dif_text_of(DDiscipline));
										fprintf (out,"%s",q->d);
										fprintf (out,"\n");
										}
									else
										if (is_dif_minor_discipline (q->d)) {
											fprintf (out,"%s: ",dif_text_of(DDiscipline));
											fprintf (out,"%s>",major_dif_discipline_of (q->d));
											fprintf (out,"%s",q->d);
											fprintf (out,"\n");
											}
										else {
											fprintf (out,"%s: ",dif_text_of(DKeyword));
											fprintf (out,"%s",q->d);
											fprintf (out,"\n");
											}
						} while (p = find_next_key (p,WTheme_Keyword));

		/*--------------------------------------------------------------*\
		 | Location and (General) Keyword
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WIdentification_Information))
			if (p = find_key (p,WPlace))
				if (p = find_key (p,WPlace_Keyword))
					do {
						if (q = p->child)
							if (is_dif_location (q->d)) {
								fprintf (out,"%s: ",dif_text_of(DLocation));
								fprintf (out,"%s",q->d);
								fprintf (out,"\n");
								}
							else {
								fprintf (out,"%s: ",dif_text_of(DKeyword));
								fprintf (out,"%s",q->d);
								fprintf (out,"\n");
								}
						} while (p = find_next_key (p,WPlace_Keyword));

		/*--------------------------------------------------------------*\
		 | Revision_Date
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WMetadata_Reference_Information))
			if (p = find_key (p,WMetadata_Date))
				if (p = p->child) {
					fprintf (out,"%s: ",dif_text_of(DRevision_Date));
					fprintf (out,"%s",p->d);
					fprintf (out,"\n");
					}

		/*--------------------------------------------------------------*\
		 | Science_Review_Date
		\*--------------------------------------------------------------*/

		fprintf (out,"%s: ",dif_text_of(DScience_Review_Date));
		fprintf (out,"\n");

		/*--------------------------------------------------------------*\
		 | Future_Review_Date
		\*--------------------------------------------------------------*/

		fprintf (out,"%s: ",dif_text_of(DFuture_Review_Date));
		fprintf (out,"\n");

		/*--------------------------------------------------------------*\
		 | Reference
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WCitation))
			if (p = find_key (p,WCitation_Information)) {
				fprintf (out,"Group: %s\n",dif_text_of(DReference));

				fp = out;
				for (p=p->child; p; p=p->next) write_text_item (p,1);

				fprintf (out,"End_Group\n");
				}

		/*--------------------------------------------------------------*\
		 | Quality
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WData_Quality_Information))
			if (p->child) {
				fprintf (out,"Group: %s\n",dif_text_of(DQuality));
				for (p=p->child; p; p=p->next)
					if (p->key != WLineage) write_text_item (p,1);
				fprintf (out,"End_Group\n");
				}

		/*--------------------------------------------------------------*\
		 | Summary
		\*--------------------------------------------------------------*/

		if (p = find_key (first,WDescription)) {
			fprintf (out,"Group: %s\n",dif_text_of(DSummary));

			if (q = find_key (p,WPurpose))
				for (q=q->child; q; q=q->next)
					fprintf (out,"  %s\n",q->d);

			if (q = find_key (p,WAbstract)) {
				fprintf (out,"  \n");
				for (q=q->child; q; q=q->next)
					fprintf (out,"  %s\n",q->d);
				}

			if (q = find_key (p,WSupplemental_Information)) {
				fprintf (out,"  \n");
				for (q=q->child; q; q=q->next)
					fprintf (out,"  %s\n",q->d);
				}

			fprintf (out,"End_Group\n");
			}

		/*--------------------------------------------------------------*\
		\*--------------------------------------------------------------*/

		fclose (out);
		}
	else {
		fprintf (stderr,"Error: could not create output file %s\n",output_file);
		return;
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
