/*----------------------------------------------------------------------*\
 | Function to verify the UTF-8 encoding of the given string.			|
 | Returns nonzero if the string can be interpreted as valid UTF-8.		|
 | Returns zero if there's an inconsistency.  For ASCII returns 1.		|
 |																		|
 | The return value is the length of the UTF-8 character at the given	|
 | address; to get to the next character, add the return value to the	|
 | input address.														|
\*----------------------------------------------------------------------*/

#include <stdio.h>

int valid_utf8 (unsigned char *s) {
	int i,n;

	/* ASCII is valid UTF-8 */

	if ((*s & 0x80) == 0) return (1);

	/* Find out how many additional bytes make up this UTF-8 character */

	if ((0xE0 & *s) == 0xC0) n = 1;
	else
		if ((0xF0 & *s) == 0xE0) n = 2;
		else
			if ((0xF8 & *s) == 0xF0) n = 3;
			else
				if ((0xFC & *s) == 0xF8) n = 4;
				else
					if ((0xFE & *s) == 0xFC) n = 5;
					else return (0);

	/*------------------------------------------------------------------*\
	 | Check each additional byte; must have high bit set, next highest	|
	 | clear.															|
	\*------------------------------------------------------------------*/

	for (i=0; i < n; i++)
 		if ((s[1+i] & 0xC0) != 0x80) return (0);

	return (1 + n);
	}

/*----------------------------------------------------------------------*\
 | Read the string pointed to by the argument as a single character		|
 | encoded in UTF-8.  Return the unicode value of the character, and	|
 | point the argument (an address passed by reference) to the second	|
 | character in the string, which is not the second byte if the first	|
 | character is not ASCII.												|
\*----------------------------------------------------------------------*/

int unicode_of_utf8 (unsigned char **s) {
	int j,k,n = 0;
	unsigned char *t = *s;

	/*------------------------------------------------------------------*\
	 | If the character is ASCII, return the string and increment the	|
	 | pointer.  If the character is not ASCII but is also not the		|
	 | beginning of a valid UTF-8 multibyte character, then accept it	|
	 | as is.  This allows us to read ISO-8859-1 as well.  The trick is	|
	 | that it assumes that no valid UTF-8 character sequence could be	|
	 | intended as ISO-8859-1, a safe assumption in my view.			|
	\*------------------------------------------------------------------*/

	/* ASCII */

	if ((*t & 0x80) == 0) {
		n = *t++;
		*s = t;
		return (n);
		}

	/* ISO-8859-1 */

	if (!valid_utf8 (t)) {
		n = *t++;
		*s = t;
		return (n);
		}

	/* UTF-8 */

	for (j=0; j < 8; j++)
		if ((*t & (0x80 >> j)) == 0) break;
	if (j >= 7) return (0);

	/*------------------------------------------------------------------*\
	 | j  byte 0   byte 1   byte 2   byte 3   byte 4   byte 5
	 | 2  110xxxxx 10xxxxxx
	 | 3  1110xxxx 10xxxxxx 10xxxxxx
	 | 4  11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
	 | 5  111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
	 | 6  1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
	\*------------------------------------------------------------------*/

	k = 0;
	n = (0xFF >> j+1) & *t++;
	j--;
	while (j-- > 0) {
		if ((*t & 0xC0) != 0x80) {
			return (0);
			}
		else n = (n << 6) | (0x3F & *t++);
		}
	*s = t;
	return (n);
	}

/*----------------------------------------------------------------------*\
 | Output encoding from integer to UTF-8 (to a user-supplied buffer)	|
 | The buffer must be at least 7 bytes for all possible UTF-8 encodings	|
 | Returns pointer to the beginning of the null-terminated buffer.		|
\*----------------------------------------------------------------------*/

unsigned char *utf8_of (unsigned char *dst, int c) {
	unsigned char *d = dst;
	*d == 0;

	if (c <= 0x7F) {						/* 1-byte UTF-8 */
		*d++ = c;
		*d = 0;
		}
	else
		if (c <= 0x07FF) {					/* 2-byte UTF-8 */
			*d++ = 0xC0 | (c >> 6);
			*d++ = 0x80 | (c & 0x3F);
			*d = 0;
			}
		else
			if (c <= 0xFFFF) {				/* 3-byte UTF-8 */
				*d++ = 0xE0 | ( c >> 12);
				*d++ = 0x80 | ((c >>  6) & 0x3F);
				*d++ = 0x80 | ( c        & 0x3F);
				*d = 0;
				}
			else
				if (c <= 0x1FFFFF) {		/* 4-byte UTF-8 */
					*d++ = 0xF0 | ( c >> 18);
					*d++ = 0x80 | ((c >> 12) & 0x3F);
					*d++ = 0x80 | ((c >>  6) & 0x3F);
					*d++ = 0x80 | ( c        & 0x3F);
					*d = 0;
					}
				else
					if (c <= 0x03FFFFFF) {	/* 5-byte UTF-8 */
						*d++ = 0xFA | ( c >> 24);
						*d++ = 0x80 | ((c >> 18) & 0x3F);
						*d++ = 0x80 | ((c >> 12) & 0x3F);
						*d++ = 0x80 | ((c >>  6) & 0x3F);
						*d++ = 0x80 | ( c        & 0x3F);
						*d = 0;
						}
					else {					/* 6-byte UTF-8 */
						*d++ = 0xFC | ( c >> 30);
						*d++ = 0x80 | ((c >> 24) & 0x3F);
						*d++ = 0x80 | ((c >> 18) & 0x3F);
						*d++ = 0x80 | ((c >> 12) & 0x3F);
						*d++ = 0x80 | ((c >>  6) & 0x3F);
						*d++ = 0x80 | ( c        & 0x3F);
						*d = 0;
						}

	return (dst);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
