/*----------------------------------------------------------------------*\
 | Program to reformat mp's error report into html to help novices find	|
 | the elements it refers to more easily.								|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "config.h"
#include "keyword.h"
#include "stricmp.h"
#include "revision.h"

extern void read_local (char *local_file);
extern char *number_of (enum fgdc_keyword key);
extern char *link_of (enum fgdc_keyword key);
extern char *iso8859_1[];

static char *base_href = "http://www.fgdc.gov/metadata/csdgm/";

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static unsigned char *buffer = NULL;
static int buffer_length = 0;

static void clear_buffer (void) {
	if (buffer) *buffer = 0;
	else
		if (buffer = (unsigned char *) malloc (4096)) {
			memset (buffer,0,4096);
			buffer_length = 4096;
			}
		else {
			fprintf (stderr,"Error: failed to allocate 4096 bytes for translation buffer\n");
			exit (1);
			}
	}

static unsigned char *add_to_buffer (unsigned char *string, int j) {
	int k,n;
	unsigned char *t;

	if (!string || j == 0) {
		t = buffer + strlen (buffer);
		return (t);
		}

	if (!buffer)
		if (buffer = (unsigned char *) malloc (4096)) {
			memset (buffer,0,4096);
			buffer_length = 4096;
			}
		else {
			fprintf (stderr,"Error: failed to allocate 4096 bytes for translation buffer\n");
			exit (1);
			}

	k = strlen (buffer);
	n = ((j < 4096) ? 4096 : j+1);

	if (k + j + 1 >= buffer_length)
		if (buffer = (unsigned char *) realloc (buffer,buffer_length+n)) {
			buffer_length += n;
			}
		else {
			fprintf (stderr,"Error: failed to reallocate %d bytes for translation buffer\n",buffer_length + n);
			exit (1);
			}

	t = buffer + k;
	memcpy (t,string,j);
	t += j;
	*t = 0;
	return (t);
	}

static unsigned char *htmlentities (unsigned char *string) {
	unsigned char *s,*t;

	clear_buffer();
	s = string;
	while (*s) {
		switch (*s) {
			case '<':
				add_to_buffer ("&lt;",4);
				s++;
				break;
			case '>':
				add_to_buffer ("&gt;",4);
				s++;
				break;
			case '&':
				add_to_buffer ("&amp;",5);
				s++;
				break;
			case '"':
				add_to_buffer ("&quot;",6);
				s++;
				break;
			default:
				add_to_buffer (s++,1);
				break;
			}
		}
	return (buffer);
	}

static unsigned char *urlencode (unsigned char *string) {
	unsigned char *s,*t;
	unsigned char hex[8];

	clear_buffer();
	s = string;
	while (*s) {
		if (*s == '_' || isalnum(*s)) add_to_buffer (s,1);
		else
			if (isspace(*s)) add_to_buffer ("+",1);
			else {
				sprintf (hex,"%%02X",*s);
				add_to_buffer (hex,3);
				}
		s++;
		}
	return (buffer);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

FILE *out;

char *read_text (char *name) {
	long n;
	FILE *fp;
	char *dst,*end;
	long size;

	if (!name) return (NULL);

	dst = NULL;
	if (fp = fopen (name,"rb")) {
		if (fseek (fp,0L,SEEK_END) == 0) {
			size = 1L + ftell (fp);
			if (size > 1L) {
				if (dst = (char *) malloc (size)) {
					rewind (fp);
					n = fread (dst,1,size,fp);
					end = dst + n;
					*end = 0;
					}
				else
					fprintf (stderr,"Error: could not allocate space for file\n");
				}
			else
				fprintf (stderr,"Error: Input file is empty.\n");
			}
		fclose (fp);
		}
	else
		fprintf (stderr,"Error: could not open input file %s\n",name);
	return (dst);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

enum msg_type {
	BANNER,
	UPGRADE,
	WARNING,
	ERROR,
	SUMMARY,
	OTHER
	};

static char *text_of_type (enum msg_type t) {
	if (t == BANNER ) return ("Banner");
	if (t == UPGRADE) return ("Upgrade");
	if (t == WARNING) return ("Warning");
	if (t == ERROR  ) return ("Error");
	if (t == SUMMARY) return ("Summary");
	if (t == OTHER  ) return ("Other");
	return ("unknown");
	};

struct message {
	enum msg_type type;
	char *text;
	int severity;
	int count;
	};

struct ref {
	int line_number;
	struct message *message;
	};

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static char *severity_color [8] = {
	"CCFFCC",		/* 0 Informative messages */
	"FFFFDD",		/* 1 Improper values */
	"FFEECC",		/* 2 Empty elements */
	"FFFFAA",		/* 3 Missing elements */
	"FFEEEE",		/* 4 Too many elements */
	"FFCCCC",		/* 5 Misplaced elements */
	"FFAAAA",		/* 6 Unrecognized elements */
	"FFDD77",		/* 7 Indentation problems */
	};

static char *severity_text [8] = {
	"Informative warnings and upgrade notes",
	"Elements with improper values",
	"Empty elements",
	"Missing elements",
	"Elements of which you have too many",
	"Misplaced elements",
	"Unrecognized elements",
	"Indentation problems"
	};

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static void get_name (char *dst, char *src) {
	char *d = dst;
	char *s = src;
	while (*s && !isspace(*s) && *s != ',') *d++ = *s++;
	*d = 0;
	}

int compare (const void *e1, const void *e2) {
	struct message *m1 = *(struct message **) e1;
	struct message *m2 = *(struct message **) e2;
	int d = m2->severity - m1->severity;
	if (d == 0) d = strcmp (m1->text,m2->text);
	return (d);
	}

static char html_declaration[] = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">";

int main (int argc, char *argv[]) {
	char *input_file;
	char *output_file = NULL;
	char *cfg_file = NULL;
	char *language = NULL;
	char *ext_list[8];
	int ext_count = 0;
	char input_file_name [1024];
	char file_name [1024];
	int spaces_are_bad = 1;
	void *opt;
	char *profile_name = NULL;

	FILE *in;
	FILE *fp;
	char *s,*t;
	char line [1024];
	char report [1024];
	char string [128];
	enum fgdc_keyword key;

	int i,j,k,m;
	char **list;
	struct ref *ref;
	struct message **sorted;
	enum msg_type msg_type;
	int msg_line;
	char *msg_text;
	int msg_severity;
	int match,found;
	int line_count;
	int ref_count;
	int sorted_count;
	char *buf;
	time_t T;
	char *href = base_href;

	#ifdef _WIN32
		spaces_are_bad = 0;
	#endif

	out = stderr;

	if (argc <= 1) {
		fprintf (stderr,"err2html %s Peter N. Schweitzer (U.S. Geological Survey)\n",revision.err2html);
		fprintf (stderr,"Usage: %s [-c config_file] [-o output_file] input_file\n",argv[0]);
		fprintf (stderr,"          config_file is same as used for mp\n");
		fprintf (stderr,"          input_file is the error output of mp\n");
		fprintf (stderr,"          output_file is the html version\n");
		exit (0);
		}

	for (i=1; i < argc; i++)
		if (memicmp (argv[i],"-o",2) == 0) {
			i++;
			if (i < argc) output_file = argv[i];
			}
		else
			if (memicmp (argv[i],"-c",2) == 0) {
				i++;
				if (i < argc) cfg_file = argv[i];
				}
			else
				if (memicmp (argv[i],"-l",2) == 0) {
					i++;
					if (i < argc) language = argv[i];
					}
				else
					if (memicmp (argv[i],"-ext",4) == 0) {
						i++;
						if (i < argc && ext_count < 8)
							ext_list [ext_count++] = strdup (argv[i]);
						}
					else
						if (memicmp (argv[i],"-u",2) == 0) {
							i++;
							if (i < argc) href = argv[i];
							}
						else
							input_file = argv[i];

	/*------------------------------------------------------------------*\
	 | First read configuration, so we can determine whether a language	|
	 | other than English was requested.								|
	\*------------------------------------------------------------------*/

	if (cfg_file) read_configuration (cfg_file);

	/*------------------------------------------------------------------*\
	 | Use requested language.  Command-line argument overrides what is |
	 | specified in the config file.									|
	\*------------------------------------------------------------------*/

	if (!language)
		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"language"))
				if (s = text_of_option (opt))
					language = s;

	/*------------------------------------------------------------------*\
	 | Use requested profile.
	\*------------------------------------------------------------------*/

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"profile"))
			if (s = text_of_option (opt))
				profile_name = s;

	use_element_names (language,profile_name);		/* in keyword.c */

	/*------------------------------------------------------------------*\
	 | Clip standard file extension off the input file name				|
	 | Standard extensions are txt,sgml,sgm,xml,text,met,bin			|
	 | or anything specified with -ext on the command line.				|
	\*------------------------------------------------------------------*/

	strcpy (input_file_name,input_file);
	s = input_file_name + strlen (input_file_name) - 1;
	while (s > input_file_name && *s != '.') s--;
	if (s > input_file_name && *s == '.')
		if (stricmp (s,".err" ) == 0) *s = 0;
		else
			for (i=0; i < ext_count; i++)
				if (*ext_list[i] != '.') {
					if (stricmp (s+1,ext_list[i]) == 0) *s = 0;
					}
				else {
					if (stricmp (s,ext_list[i]) == 0) *s = 0;
					}

	/*------------------------------------------------------------------*\
	 | If you didn't find a standard extension, look for a custom file	|
	 | extension specified by the user in the config file at input:ext	|
	\*------------------------------------------------------------------*/

	if (s > input_file_name && *s == '.')
		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"ext"))
				do {
					if (t = text_of_option (opt))
						if (stricmp (s,t) == 0) {
							*s = 0;
							break;
							}
						else
							if (stricmp (s+1,t) == 0) {
								*s = 0;
								break;
								}
					} while (opt = find_next_option (opt,"ext"));

	if (!output_file)
		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"html"))
				if (opt = find_option (opt,"err"))
					if (s = text_of_option (opt)) {
						sprintf (file_name,s,input_file_name);
						if (spaces_are_bad) for (s=file_name; *s; s++) if (isspace(*s)) *s = '_';
						if (strcmp (file_name,input_file) != 0)
							output_file = file_name;
						else
							fprintf (stderr,"Warning: requested output file %s would overwrite the input file\n",file_name);
						}


	/*------------------------------------------------------------------*\
	 | Create the output file if one was specified
	\*------------------------------------------------------------------*/

	if (output_file) {
		fp = fopen (output_file,"w");
		if (!fp) {
			fprintf (stderr,"Error: could not create output file %s, using stdout instead\n",output_file);
			fp = stdout;
			}
		}
	else
		fp = stdout;

	/*------------------------------------------------------------------*\
	 | Read and process the error messages
	\*------------------------------------------------------------------*/

	if (buf = read_text (input_file)) {

		/* First check to see if this really is an mp error file */

		if (memcmp (buf,"mp",2) != 0) {
			fprintf (stderr,"Hmmm.  This doesn't seem to be an mp error file\n");
			exit (1);
			}

		/* Count lines */

		line_count = 0;
		for (s=buf; *s; s++)
			if (*s == '\n') line_count++;

		/* Allocate an array of pointers, one for each line */

		if (list = (char **) malloc ((1 + line_count) * sizeof (char *))) {

			/* Delimit each line with NUL */

			k = 0;
			list[k++] = buf;
			s = buf;
			while (*s)
				if (*s == '\r') {
					*s++ = 0;
					if (*s == '\n') *s++ = 0;
					list [k++] = s;
					}
				else
					if (*s == '\n') {
						*s++ = 0;
						list[k++] = s;
						}
					else
						s++;

			/*----------------------------------------------------------*\
			 | Allocate an array of ref structures that relate a line	|
			 | in the input file with an error message.					|
			\*----------------------------------------------------------*/

			if (ref = (struct ref *) malloc ((1 + line_count) * sizeof (struct ref))) {

				/* For each line */

				for (k=0; k < line_count; k++) {

					/* Figure out what type of message is on this line */

					s = list[k];
					if (isdigit (*s) || memcmp (s,"No errors",9) == 0) {
						msg_type = SUMMARY;
						msg_line = 0;
						msg_text = s;
						}
					else
						if (memcmp (s,"mp",2) == 0) {
							msg_type = BANNER;
							msg_line = 0;
							msg_text = s;
							}
						else
							if (memcmp (s,"Upgrade",7) == 0) {
								msg_type = UPGRADE;
								}
							else
								if (memcmp (s,"Warning",7) == 0) {
									msg_type = WARNING;
									}
								else
									if (memcmp (s,"Error",5) == 0) {
										msg_type = ERROR;
										}
									else
										if (memcmp (s,"No Errors",9) == 0) {
											msg_type = SUMMARY;
											msg_line = 0;
											msg_text = s;
											}
										else {
											msg_type = OTHER;
											msg_line = 0;
											msg_text = s;
											}

					/*--------------------------------------------------*\
					 | If upgrade, warning, or error, find line number	|
					 | and text.										|
					\*--------------------------------------------------*/

					if (msg_type == UPGRADE || msg_type == WARNING || msg_type == ERROR) {
						if (t = strstr (s,"(line ")) {
							char *ptr;
							t += 6;
							msg_line = (int) strtol (t,&ptr,0);
							while (*t && *t != ')') t++;
							if (*t == ')') t++;
							if (*t == ':') t++;
							while (*t && isspace(*t)) t++;
							s = t;
							}
						msg_text = s;
						}

					/*--------------------------------------------------*\
					 | Categorize message severity by looking for some	|
					 | key phrases.										|
					\*--------------------------------------------------*/

					msg_severity = 0;
					if (msg_text) {
						if (strstr (msg_text,"ambiguous indentation") ||
							strstr (msg_text,"some information may be lost"))
							msg_severity = 7;
						else
						if (strstr (msg_text,"no element recognized"))
							msg_severity = 6;
						else
						if (strstr (msg_text,"is not permitted in"))
							msg_severity = 5;
						else
						if (strstr (msg_text,"too many") ||
							strstr (msg_text,"permits only one of") ||
							strstr (msg_text,"but not both"))
							msg_severity = 4;
						else
						if (strstr (msg_text,"is required in") ||
							strstr (msg_text,"requires one of") ||
							strstr (msg_text,"no fewer than") ||
							strstr (msg_text,"requires both") ||
							strstr (msg_text,"and vice versa") ||
							strstr (msg_text,"too few") ||
							strstr (msg_text,"or both") ||
							strstr (msg_text,"requires as many") ||
							strstr (msg_text,"requires at least one of") ||
							strstr (msg_text,"requires two or more of"))
							msg_severity = 3;
						else
						if (strstr (msg_text,"does not have a value"))
							msg_severity = 2;
						else
						if (strstr (msg_text,"improper value for"))
							msg_severity = 1;
						}

					/*--------------------------------------------------*\
					 | Have we seen this message text already?			|
					\*--------------------------------------------------*/

					match = -1;
					for (j=0; j < k; j++)
						if (msg_text && *msg_text)
							if (ref[j].message)
								if (ref[j].message->text)
									if (strcmp (msg_text,ref[j].message->text) == 0) {
										ref[j].message->count++;
										match = j;	/* Remember which line */
										break;
										}

					/*--------------------------------------------------*\
					 | Store in the ref structure the line number in	|
					 | the input file.									|
					\*--------------------------------------------------*/

					ref[k].line_number = msg_line;

					/*--------------------------------------------------*\
					 | If you've seen this message before, keep only	|
					 | one copy of it.									|
					\*--------------------------------------------------*/

					if (match != -1) ref[k].message = ref[match].message;
					else

						/*----------------------------------------------*\
						 | If you haven't seen this message before,		|
						 | make a new message structure.				|
						\*----------------------------------------------*/

						if (msg_text && *msg_text)
							if (ref[k].message = (struct message *) malloc (sizeof (struct message))) {
								ref[k].message->type = msg_type;
								ref[k].message->text = msg_text;
								ref[k].message->severity = msg_severity;
								ref[k].message->count = 1;
								}
					}
				ref_count = k;

				/*------------------------------------------------------*\
				 | Create an array of pointers to messages, and sort it	|
				 | according to message severity.						|
				\*------------------------------------------------------*/

				if (sorted = (struct message **) malloc (ref_count*sizeof(struct message *))) {
					memset (sorted,0,ref_count*sizeof(struct message *));
					m = 0;
					for (j=0; j < ref_count; j++) {
						found = 0;
						for (i=0; sorted[i]; i++)
							if (ref[j].message == sorted[i]) {
								found = 1;
								break;
								}
						if (!found) sorted [m++] = ref[j].message;
						}
					sorted_count = m;
					qsort (sorted,sorted_count,sizeof(struct message *),compare);

					/*--------------------------------------------------*\
					 | Output heading information
					\*--------------------------------------------------*/

					fprintf (fp,"%s\n",html_declaration);
					fprintf (fp,"<html>\n");
					fprintf (fp,"<head>\n");
					fprintf (fp,"  <title>");
					fprintf (fp,"Error report from mp");
					fprintf (fp,"</title>\n");
					fprintf (fp,"  <style type=\"text/css\">\n");
					fprintf (fp,"    <!--\n");
					fprintf (fp,"    h1 {\n");
					fprintf (fp,"        font-size: large;\n");
					fprintf (fp,"        }\n");
					fprintf (fp,"      -->\n");
					fprintf (fp,"  </style>\n");
					fprintf (fp,"</head>\n");
					fprintf (fp,"<body>\n");

					/*--------------------------------------------------*\
					 | Find banner and write it out
					\*--------------------------------------------------*/

					for (i=0; i < sorted_count; i++)
						if (sorted[i]->type == BANNER) {
							fprintf (fp,"<h1>%s</h1>\n",sorted[i]->text);
							break;
							}
					for (i=0; i < sorted_count; i++)
						if (sorted[i]->type == SUMMARY) {
							fprintf (fp,"<p>%s</p>\n",sorted[i]->text);
							break;
							}

					/*--------------------------------------------------*\
					 | Write messages in order of decreasing severity	|
					\*--------------------------------------------------*/

					fprintf (fp,"<table border=\"1\" cellpadding=\"4\" cellspacing=\"0\" summary=\"Errors and warnings are listed in decreasing order of severity\">\n");
					fprintf (fp,"  <tr valign=\"top\" bgcolor=\"#cceeff\">");
					fprintf (fp,"    <th align=\"left\" scope=\"col\">Type</th>");
					fprintf (fp,"    <th align=\"left\" scope=\"col\">Description<br>or line numbers</th>");
					fprintf (fp,"    <th align=\"left\" scope=\"col\">Line(s)<br>(or count)</th>");
					fprintf (fp,"  </tr>\n");
					msg_severity = -1;
					for (i=0; i < sorted_count; i++)
						if ((sorted[i]->type != BANNER) && (sorted[i]->type != SUMMARY)) {
							if (sorted[i]->severity != msg_severity) {
								fprintf (fp,"  <tr bgcolor=\"#CCCCCC\">\n");
								fprintf (fp,"    <td colspan=\"3\">");
								fprintf (fp,"Severity %d: %s",sorted[i]->severity,severity_text[sorted[i]->severity]);
								fprintf (fp,"</td>\n");
								fprintf (fp,"  </tr>\n");
								msg_severity = sorted[i]->severity;
								}
							fprintf (fp,"  <tr bgcolor=\"#%s\">\n",severity_color[sorted[i]->severity]);
							fprintf (fp,"    <td valign=\"top\">%s</td>\n",text_of_type (sorted[i]->type));

							/*------------------------------------------*\
							 | Link element names in the message text	|
							 | to appropriate network addresses.		|
							\*------------------------------------------*/

							fprintf (fp,"    <td valign=\"top\">");
							s = sorted[i]->text;
							while (*s) {
								char string [128];
								get_name (string,s);
								key = key_of (string);
								if (key == Wunknown || key == Wblank) {
									unsigned char c = (unsigned char) *s;
									if (iso8859_1[c])
										fprintf (fp,"&%s;",iso8859_1[c]);
									else
										if (c > 127)
											fprintf (fp,"&#%u;",c);
										else
											fprintf (fp,"%c",c);
									s++;
									}
								else {
									fprintf (fp,"<a href=\"%s%s\">",href,link_of (key));
									fprintf (fp,"%s",text_of (key));
									fprintf (fp,"</a>");
									t = number_of (key);
									if (*t) fprintf (fp," (%s)",t);
									s += strlen (text_of(key));
									}
								}
							fprintf (fp,"</td>\n");

							fprintf (fp,"    <td valign=\"top\">");
							if (sorted[i]->count > 8)
								fprintf (fp,"(%d)",sorted[i]->count);
							else
								for (j=0; j < ref_count; j++)
									if (ref[j].message == sorted[i]) fprintf (fp,"%d<br/>",ref[j].line_number);
							fprintf (fp,"</td>\n");
							fprintf (fp,"  </tr>\n");

							if (sorted[i]->count > 8) {
								fprintf (fp,"  <tr bgcolor=\"#%s\">\n",severity_color[sorted[i]->severity]);
								fprintf (fp,"    <td>&nbsp;</td>\n");
								fprintf (fp,"    <td valign=\"top\">");
								k = 0;
								for (j=0; j < ref_count; j++)
									if (ref[j].message == sorted[i]) {
										if (k == 0) k++; else fprintf (fp,", ");
										fprintf (fp,"%d",ref[j].line_number);
										}
								fprintf (fp,"</td>\n");
								fprintf (fp,"    <td>&nbsp;</td>\n");
								fprintf (fp,"  </tr>\n");
								}
							}
					fprintf (fp,"</table>\n");
					fprintf (fp,"<br/>\n");

					fprintf (fp,"<hr/>\n");
					fprintf (fp,"Generated by <a href=\"http://geology.usgs.gov/tools/metadata/tools/doc/err2html.html\">err2html</a> %s\n",revision.err2html);

					T = time (&T);
					t = ctime (&T);
					s = t + strlen (t) - 1;
					if (*s == '\n') *s-- = 0;
					if (*s == '\r') *s-- = 0;
					fprintf (fp,"%s\n",t);

					fprintf (fp,"</body>\n");
					fprintf (fp,"</html>\n");
					}

				}
			else
				fprintf (stderr,"Memory allocation error (2), line_count = %d\n",line_count);
			}
		else
			fprintf (stderr,"Memory allocation error (1), line_count = %d\n",line_count);
		}
	if (fp) fclose (fp);
	exit (0);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
