/*----------------------------------------------------------------------*\
 | Functions to read syntactical characteristics of local extensions	|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "local.h"
#include "stricmp.h"

extern char *read_text (char *input_file);

/*----------------------------------------------------------------------*\

Local extension file syntax:

Summary:

The local extension file is encoded like the metadata; hierarchical
structure is indicated using indentation.

local:
	name <element name>
	sgml <tag name>
	z3950 <number>
	parent <element name>
	child <element name>

Compound types:

local = name + (sgml) + 1{parent}n + 1{child}n

Scalar types:

name
	The name of the element as it will appear in the metadata.  This
	must consist only of upper- and lower-case letters, numbers, and
	the underscore character.

sgml
	The element tag to be used for output in SGML.  If omitted, the
	full name is used.

z3950
	The numeric tag to be used for searching in Z39.50 systems.  Don't
	go putting just anything in here; these values have to be allocated
	by the Z39.50 Interest Group (ZIG).

parent
	The name of the element under which this element may appear.
	These names may be standard elements or local extensions.

child
	The name of an element which may appear under this element.

\*----------------------------------------------------------------------*/

enum local_word {
	EExtension_Information,
	EExtended_Element_Name,
	EShort_Name,
	EType,
	EDomain,
	ERule,
	EOptionality,
	ERepeatability,
	EDefinition,
	ERationale,
	ESource,
	Elocal,
	Ename,
	Esgml,
	Ez3950,
	Ez3950_tag,
	Eparent,
	Echild,
	Eunknown,
	Eempty,
	EBLANK,
	Enull
	};

struct keyword_list {
	enum local_word key;
	char *text;
	};

static struct keyword_list keyword_list[] = {
	{Elocal,					"Local"},
	{Ename,						"name"},
	{Esgml,						"sgml"},
	{Ez3950,					"z3950"},
	{Ez3950_tag,				"Z39.50_Tag"},
	{Eparent,					"parent"},
	{Echild,					"child"},
	{EExtension_Information,	"Extension_Information"},
	{EExtended_Element_Name,	"Extended_Element_Name"},
	{EShort_Name,				"Short_Name"},
	{EType,						"Type"},
	{EDomain,					"Domain"},
	{ERule,						"Rule"},
	{EOptionality,				"Optionality"},
	{ERepeatability,			"Repeatability"},
	{EDefinition,				"Definition"},
	{ERationale,				"Rationale"},
	{ESource,					"Source"},
	{EBLANK,					"BLANK"},
	{Eunknown,					NULL}
	};

static int keyword_list_length = 0;

/*----------------------------------------------------------------------*/

static int compare_local (const void *e1, const void *e2) {
	struct keyword_list *p = (struct keyword_list *) e1;
	struct keyword_list *q = (struct keyword_list *) e2;
	return (stricmp (p->text,q->text));
	}

static enum local_word local_key_of (char *text) {
	int i,n;
	struct keyword_list k,*key;
	char string[128];

	if (keyword_list_length == 0) {
		for (i=0; keyword_list[i].text; i++);
		keyword_list_length = i;
		qsort (keyword_list,keyword_list_length,sizeof(struct keyword_list),compare_local);
		}
	if (!text) return (Enull);
	if (*text) {
		n = strcspn (text," \t\n\v\f\r:=");
		if (n > 127) n = 127;
		memcpy (string,text,n);
		string[n] = 0;
		k.text = string;
		key = (struct keyword_list *) bsearch (&k,keyword_list,keyword_list_length,sizeof(struct keyword_list),compare_local);
		if (key) return (key->key);
		else return (Eunknown);
		}
	else return (Eempty);
	}

static char *local_text_of (enum local_word key) {
	int i;

	if (keyword_list_length == 0) {
		for (i=0; keyword_list[i].text; i++);
		keyword_list_length = i;
		qsort (keyword_list,keyword_list_length,sizeof(struct keyword_list),compare_local);
		}
	if (key == Enull) return ("(null)");
	if (key == Eempty) return ("(empty)");
	for (i=0; i < keyword_list_length; i++)
		if (keyword_list[i].key == key) return (keyword_list[i].text);
	return ("(unknown)");
	}

/*----------------------------------------------------------------------*\
 | Error reporting mechanism											|
 |																		|
 | Since the extension file may be processed before any information		|
 | has been presented to the user, we should store warnings or error	|
 | messages in a text buffer that can be retrieved by the calling		|
 | program (typically mp or tkme) and presented to the user in some		|
 | appropriate way.														|
 |																		|
 | The function error_ptr is given a buffer size needed for the next	|
 | error message.  It returns the address of a buffer big enough for	|
 | the message, or NULL if space cannot be allocated.					|
 |																		|
 | The function ext_errors simply returns a pointer to the buffer		|
 | that contains error messages.  This is NULL if no errors occurred.	|
\*----------------------------------------------------------------------*/

static char *error_buffer = NULL;
static int error_buffer_size = 0;
static int error_buffer_idx = 0;
#define GRANULARITY 4096

static char *error_ptr (int size) {
	int n;
	char *s;
	if (error_buffer_size - error_buffer_idx < 1 + size) {
		if (1 + size < GRANULARITY) n = GRANULARITY;
		else n = (1 + (1 + size)/GRANULARITY) * GRANULARITY;

		if (error_buffer)
			if (s = (char *) realloc (error_buffer,error_buffer_size + n)) {
				error_buffer = s;
				error_buffer_size += n;
				}
			else {
				free (error_buffer);
				error_buffer = NULL;
				error_buffer_size = 0;
				error_buffer_idx = 0;
				}
		else
			if (error_buffer = (char *) malloc (n))
				error_buffer_size = n;
		}
	return (error_buffer + error_buffer_idx);
	}

char *ext_errors (void) {
	return (error_buffer);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

struct item {
	struct item *next;
	struct item *prev;
	struct item *parent;
	struct item *child;
	int indent;			/* number of leading whitespaces before text */
	int line_number;	/* line in source file */
	enum local_word key;
	int original;		/* zero if member d can be released thru free() */
	int len;			/* length of bloc pointed to by d */
	char *d;			/* text of item */
	};

static struct item *allocate_item (int fail) {
	struct item *p = NULL;

	if (p = (struct item *) malloc (sizeof(struct item)))
		memset (p,0,sizeof(struct item));
	else {
		fprintf (stderr,"Warning: could not allocate item\n");
		if (fail) exit (1);
		}
	return (p);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

#define UNKNOWN_TEXT	Eunknown
#define BLANK_LINE		EBLANK
#define TEXT_OF(x)		local_text_of(x)
#define KEY_OF(x)		local_key_of(x)

static int keep_blank_lines = 0;
static int allow_indented_text = 1;

static struct item *decode_text (char *dst) {
	char *end;
	char *b,*e,*s;
	struct item *p;
	struct item *first = NULL;
	struct item *current;
	int line_count;

	if (!dst) return (NULL);

	b = e = dst;
	end = dst + strlen (dst);

	/*--------------------------------------------------*\
	 | Start with an empty item							|
	\*--------------------------------------------------*/

	if (first == NULL) {
		first = allocate_item (1);
		current = first;
		}
	else {
		for (p=first; p->next; p=p->next);
		p->next = allocate_item (1);
		p->next->prev = p;
		current = p->next;
		}

	line_count = 0;
	while (b < end) {

		/*----------------------------------------------*\
		 | Find the first CR, LF, or EOF.  Zero the		|
		 | end-of-line character(s) if present.			|
		\*----------------------------------------------*/

		while (e < end) {
			if (*e == '\r') {
				*e = 0;
				if (*(e+1) == '\n') {
					e++;
					*e = 0;
					}
				break;
				}
			else
				if (*e == '\n') {
					*e = 0;
					break;
					}
			e++;
			}

		/*----------------------------------------------*\
		 | Clear trailing whitespace from the line.		|
		\*----------------------------------------------*/

		s = e;
		while (s >= b)
			if (*s == 0) s--;
			else
				if (isspace (*s)) *s-- = 0;
				else break;

		/*----------------------------------------------*\
		 | b points to the beginning of the line.		|
		 |												|
		 | If the first non-blank character in the line	|
		 | is '#', ignore the whole line.				|
		\*----------------------------------------------*/

		if (*b || keep_blank_lines) {
			for (s=b; *s && isspace(*s); s++);
			if (*s != '#') {
				current->original = 1;
				current->len = e - b + 1;
				current->d = b;
				current->line_number = 1 + line_count;
				current->next = allocate_item (1);
				current->next->prev = current;
				current = current->next;
				}
			}

		b = e + 1;
		line_count++;
		}
	if (first->next) {
		for (p=first; p->next->d; p=p->next);
		free (p->next);
		p->next = NULL;
		}
	else
		if (first->d == NULL) first = NULL;

	return (first);
	}

/*----------------------------------------------------------------------*\
 | Rather crude function that attempts to fix indented textual values.	|
 | Specifically, this routine looks for textual values that are nested	|
 | beneath other textual values and makes these siblings rather than	|
 | descendants.  It does not attempt to reclassify standard elements	|
 | as textual values; instead, it issues a warning in this case.		|
\*----------------------------------------------------------------------*/

static void equalize_indented_scalars (struct item *p) {
	struct item *q;

	while (p) {
		if (p->child)
			if (p->key == UNKNOWN_TEXT || p->key == BLANK_LINE) {
				if (p->child->key == UNKNOWN_TEXT || p->child->key == BLANK_LINE) {

					/*--------------------------------------------------*\
					 | Both p and p->child are either UNKNOWN_TEXT or	|
					 | BLANK_LINE.										|
					 |													|
					 | Convert p->child and all of its siblings to be	|
					 | siblings, not children, of p.					|
					\*--------------------------------------------------*/

					p->child->prev = p;			/* was NULL */

					p->child->parent = p->parent;	/* was p */

					for (q=p->child; q->next; q=q->next) {
						q->parent = p->parent;	/* was p */
						if (q->child) equalize_indented_scalars (q->child);
						}

					q->next = p->next;			/* was NULL */
					if (p->next) p->next->prev = q;
					p->next = p->child;
					p->child = NULL;
					}
				else {

					/*--------------------------------------------------*\
					 | p is of type UNKNOWN_TEXT but p->child is not.	|
					 | Don't second-guess at this point, but do issue	|
					 | a warning that some information might be lost in	|
					 | subsequent handling of this line.				|
					\*--------------------------------------------------*/

					char *fmt = "Warning (line %d of extension file): element %s found in textual value, some information may be lost.\n";
					int m = 4 + strlen (fmt) + strlen (TEXT_OF (p->child->key));
					char *dst;
					if (dst = error_ptr (m))
						sprintf (dst,fmt,p->child->line_number,TEXT_OF (p->child->key));
					}
				}
			else
				equalize_indented_scalars (p->child);

		p = p->next;
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static struct item *decode_tree (struct item *first) {
	int i,j,k;
	char *s,*t;
	struct item *p,*q,*r;
	int item_count;
	struct item **list;

	if (!first) return (NULL);

	item_count = 0;
	for (p=first; p; p=p->next) {
		for (s=p->d,i=0; *s && isspace (*s); s++,i++);
		p->indent = i;
		p->d = s;
		p->key = KEY_OF (s);
		t = TEXT_OF (p->key);

		/*--------------------------------------------------------------*\
		 | If the current item can have a child on this line, determine	|
		 | whether it does, and store the child's information.  The		|
		 | rules prohibit such a child from having children of its own.	|
		\*--------------------------------------------------------------*/

		if (p->key != UNKNOWN_TEXT && p->key != BLANK_LINE) {
			s = p->d + strlen (t);
			if (isspace (*s) || *s == ':' || *s == '=') {
				*s++ = 0;
				p->len = strlen (p->d);
				}
			while (*s && isspace (*s)) s++;
			if (*s == ':' || *s == '=') s++;
			while (*s && isspace (*s)) s++;
			if (*s) {
				q = allocate_item (1);
				q->d = s;
				q->len = strlen (s);
				q->original = 1;
				q->indent = p->indent + 1;
				q->line_number = p->line_number;
				q->key = UNKNOWN_TEXT;
				q->parent = p;
				p->child = q;
				}
			}
		item_count++;
		}

	/*------------------------------------------------------------------*\
	 | Estimate the proper indentation to be assigned to blank lines.	|
	 | At this point the metadata are stored as one doubly-linked list,	|
	 | so you can go forward and backward using p->next and p->prev.	|
	\*------------------------------------------------------------------*/

	for (p=first; p; p=p->next)
		if (p->key == BLANK_LINE) {

			/* Find the most recent element that is not blank */

			for (r=p->prev; r; r=r->prev)
				if (r->key != BLANK_LINE) break;

			/* Find the next element that is not blank */

			for (q=p->next; q; q=q->next)
				if (q->key != BLANK_LINE) break;

			/*----------------------------------------------------------*\
			 | Blanks should be indented like the more indented of		|
			 | (previous nonblank) and (next nonblank).  If either of	|
			 | these nodes does not exist, the proper indentation is	|
			 | whichever does exist.  If neither exist, the file		|
			 | contains only blank lines, so zero will do.				|
			\*----------------------------------------------------------*/

			if (r && q)
				p->indent = ((r->indent > q->indent) ? r->indent : q->indent);
			else
				if (r) p->indent = r->indent;
				else
					if (q) p->indent = q->indent;
					else
						p->indent = 0;
			}

	/*------------------------------------------------------------------*\
	 | Figure out the topology of the file.  Create a list of items		|
	 | as an array of pointers.  For each item, determine whether it	|
	 | has a child and what its parent is.								|
	\*------------------------------------------------------------------*/

	if (list = (struct item **) malloc (item_count * sizeof(struct item *))) {
		memset (list,0,item_count*sizeof(struct item *));
		i = 0;
		for (p=first; p; p=p->next) list[i++] = p;

		for (i=0; i < item_count-1; i++) {
			p = list[i];
			q = list[i+1];

			/*----------------------------------------------------------*\
			 | The relationship between successive items in the file is	|
			 | determined by comparing their levels of indentation.		|
			\*----------------------------------------------------------*/

			j = q->indent - p->indent;

			/*----------------------------------------------------------*\
			 | If q is indented more than p, q is a child of p, and p	|
			 | is parent of q.											|
			 |															|
			 | Modified 30-Jan-1997 to not discard scalar child already	|
			 | in p (can only happen as text following element name.	|
			\*----------------------------------------------------------*/

			if (j > 0) {
				if (allow_indented_text &&
					p->key == UNKNOWN_TEXT && q->key == UNKNOWN_TEXT)
					q->parent = p->parent;
				else {
					if (p->child && (q->key == UNKNOWN_TEXT || q->key == BLANK_LINE)) {

						/*----------------------------------------------*\
						 | Modified 20-Feb-1998 to handle cases where a	|
						 | string immediately follows the name and a	|
						 | blank line follows them.  If the next non-	|
						 | blank is UNKNOWN_TEXT, keep the string that	|
						 | was placed on the same line as the element.	|
						 | If the next non-blank is a recognized		|
						 | element, the string after the element name	|
						 | was not permitted; discard it with a warning.|
						\*----------------------------------------------*/

						if (q->key == BLANK_LINE) {
							/* Point r at the next non-blank element */
							r = NULL;
							for (k=i+1; k < item_count; k++) {
								r = list[k];
								if (r->key != BLANK_LINE) break;
								}
							/* if r is an element, discard the text */
							if (r)
								if (r->key != UNKNOWN_TEXT && r->key != BLANK_LINE) {
									char *fmt = "Warning (line %d of extension file): Extraneous text following element %s ignored\n";
									int m = 4 + strlen (fmt) + strlen (TEXT_OF (p->key));
									char *dst;
									if (dst = error_ptr (m))
										sprintf (dst,fmt,p->line_number,TEXT_OF (p->key));
									p->child = q;
									q->prev = NULL;
									}
								else {
									/* otherwise include the text */
									for (r=p->child; r->next; r=r->next);
									r->next = q;
									q->prev = r;
									}
							}
						else {
							/* q->key == UNKNOWN_TEXT */
							for (r=p->child; r->next; r=r->next);
							r->next = q;
							q->prev = r;
							}
						}
					else {
						if (p->child) {
							/* Should this refer to p->child->key? */
							char *fmt = "Warning (line %d of extension file): Extraneous text following element %s ignored\n";
							int m = 4 + strlen (fmt) + strlen (TEXT_OF (p->key));
							char *dst;
							if (dst = error_ptr (m))
								sprintf (dst,fmt,p->line_number,TEXT_OF (p->key));
							}
						p->child = q;
						q->prev = NULL;
						}
					q->parent = p;
					p->next = NULL;
					}
				}

			/*----------------------------------------------------------*\
			 | If q is indented the same as p, q is just the next part	|
			 | of the compound type (or text field) of which p is a		|
			 | part.  The parent of q is that of p.						|
			\*----------------------------------------------------------*/

			if (j == 0) {
				q->parent = p->parent;
				}

			/*----------------------------------------------------------*\
			 | If q is indented less than p, q is the next element in a	|
			 | compound type begun earlier.  Find the previous element	|
			 | by searching backwards through the list for the nearest	|
			 | element q that was indented the same as or less than q;	|
			 | store q into its next pointer, set the parent of q to	|
			 | that of r, and sever the link from p to q.				|
			\*----------------------------------------------------------*/

			if (j < 0) {
				for (k=i; k >= 0; k--)
					if (list[k]->indent <= q->indent) {
						r = list[k];
						if (r->indent < q->indent) {
							char *fmt = "Warning: ambiguous indentation at line %d of extension file\n";
							int m = 4 + strlen (fmt);
							char *dst;
							if (dst = error_ptr (m))
								sprintf (dst,fmt,q->line_number);
							}
						if (r->next == NULL) {
							r->next = q;
							q->prev = r;
							q->parent = r->parent;
							p->next = NULL;
							}
						else {
							struct item *rr;
							for (rr=r; rr->next && rr->next != q; rr=rr->next);
							rr->next = q;
							q->prev = rr;
							q->parent = rr->parent;
							if (rr != p) p->next = NULL;
							}
						break;
						}

				/*------------------------------------------------------*\
				 | If you fell through that loop, there is an error in	|
				 | the file's indentation that cannot easily be over-	|
				 | looked.  This indicates that a line somewhere in the	|
				 | middle of the file is indented less than all other	|
				 | lines in the file; the structure is unclear.			|
				\*------------------------------------------------------*/

				if (k < 0) {
					char *fmt = "Error: too little indentation at line %d of extension file\n";
					int m = 4 + strlen (fmt);
					char *dst;
					if (dst = error_ptr (m))
						sprintf (dst,fmt,q->line_number);
					}
				}
			}

		/*--------------------------------------------------------------*\
		 | Traverse the tree, looking for situations in which scalar	|
		 | values have children that are themselves scalar values; in	|
		 | such cases, make them siblings, not children.				|
		\*--------------------------------------------------------------*/

		if (allow_indented_text) equalize_indented_scalars (first);

		free (list);
		}
	else {
		char *fmt = "Error: could not allocate array of pointers to items\n";
		int m = strlen (fmt);
		char *dst;
		if (dst = error_ptr (m))
			sprintf (dst,fmt);
		exit (1);
		}
	return (first);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

struct item *parse_local (char *input_file) {
	char *txt = read_text (input_file);
	if (txt == NULL) {
		char *fmt = "Error: could not open extension file %s\n";
		int m = strlen (fmt) + strlen (input_file);
		char *dst;
		if (dst = error_ptr (m))
			sprintf (dst,fmt,input_file);
		return (NULL);
		}
	else
		return (decode_tree (decode_text (txt)));
	}

/*----------------------------------------------------------------------*\
 | Locate a node in the tree by matching the key.  The first argument	|
 | is the root node of the tree to be searched, the second is the key	|
 | to be matched.														|
\*----------------------------------------------------------------------*/

static struct item *find_local_key (struct item *p, enum local_word key) {
	struct item *q;
	while (p) {
		if (p->key == key) return (p);
		if (p->child)
			if (q = find_local_key (p->child,key)) return (q);
		p = p->next;
		}
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Locate the next node in a list that matches a key.  First argument	|
 | is the current node; it will not be considered in the search.  This	|
 | routine searches only nodes at the same level in the tree as the		|
 | node given by the first argument.  It does not search children.		|
\*----------------------------------------------------------------------*/

static struct item *find_next_local_key (struct item *p, enum local_word key) {
	p = p->next;
	while (p) {
		if (p->key == key) return (p);
		p = p->next;
		}
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Locate the first child node in a list that matches a key.  The first	|
 | argument is the current node; its child list will searched.  This	|
 | routine searches only nodes that are children of the first argument.	|
\*----------------------------------------------------------------------*/

static struct item *find_child_local_key (struct item *p, enum local_word key) {
	p = p->child;
	while (p) {
		if (p->key == key) return (p);
		p = p->next;
		}
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Recode elements as those we can use.
\*----------------------------------------------------------------------*/

static void translate_local (struct item *p) {
	while (p) {
		if (p->key == EExtension_Information) p->key = Elocal;
		if (p->key == EExtended_Element_Name) p->key = Ename;
		if (p->key == EShort_Name)            p->key = Esgml;
		if (p->key == Ez3950_tag)             p->key = Ez3950;
		if (p->child) translate_local (p->child);
		p = p->next;
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static struct item *first = NULL;
static void build_extension_list (void);

static void list_children (struct item *p) {
	while (p) {
		fprintf (stderr,"%s ",local_text_of(p->key));
		if (p->parent) fprintf (stderr,"(%s) ",local_text_of(p->parent->key));
		else fprintf (stderr,"(root) ");
		if (p->key == Eunknown)
			if (p->d) fprintf (stderr,"= \"%s\"\n",p->d);
			else fprintf (stderr,"= NULL\n");
		else fprintf (stderr,"\n");
		if (p->child) list_children (p->child);
		p = p->next;
		}
	}

void read_local (char *local_file) {
	first = parse_local (local_file);
	translate_local (first);
	build_extension_list();
	}

/*----------------------------------------------------------------------*\
 | four external functions are defined here:							|
 |																		|
 | p = find_local (q,"local_word");										|
 |																		|
 |	which uses q = first if q == NULL on entry,							|
 |	and returns q if q->key == key_of_local_word ("local_word")			|
 |																		|
 | p = find_next_local (q,"local_word");								|
 |																		|
 |	which skips q and looks only along the list q=q->next...			|
 |																		|
 | p = find_child_local (q,"local_word");								|
 |																		|
 |	which searches only the direct children of q.						|
 |																		|
 | s = text_of_local (p);												|
 |																		|
 |	which returns p->child->d, not p->d.								|
\*----------------------------------------------------------------------*/

static struct item *find_local (struct item *q, char *s) {
	enum local_word key;
	struct item *p;

	/* If you don't have a local data base, return NULL */

	if (!first) return (NULL);

	/* Find the key corresponding to the second argument */

	key = local_key_of (s);
	if (key == Enull || key == Eunknown) return (NULL);

	/* Find the node corresponding to the first argument */

	if (!q) q = first;

	/* Find the first node in the subtree headed by q whose key matches s */

	p = find_local_key (q,key);

	return (p);
	}

static struct item *find_child_local (struct item *q, char *s) {
	enum local_word key;
	struct item *p;

	/* If you don't have a local data base, return NULL */

	if (!first) return (NULL);

	/* Find the key corresponding to the second argument */

	key = local_key_of (s);
	if (key == Enull || key == Eunknown) return (NULL);

	/* Find the node corresponding to the first argument */

	if (!q) q = first;

	/* Find the first node in the subtree headed by q whose key matches s */

	p = find_child_local_key (q,key);

	return (p);
	}

static struct item *find_next_local (struct item *q, char *s) {
	enum local_word key;
	struct item *p;

	/* If you don't have a local data base, return NULL */

	if (!first) return (NULL);

	/* Find the key corresponding to the second argument */

	key = local_key_of (s);
	if (key == Enull || key == Eunknown) return (NULL);

	/* Find the node corresponding to the first argument */

	if (!q) q = first;

	/* Find the next node in the subtree headed by q whose key matches s */

	p = find_next_local_key (q,key);

	return (p);
	}

static char *text_of_local (struct item *p) {
	if (p && p->child) return (p->child->d);
	else return (NULL);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

struct member {
	struct member *next;
	int key;
	};

struct extension_list {
	int key;
	char *text;
	char *sgml;
	int z3950;
	int compound;
	struct member *parent;
	struct member *child;
	};

static int compare_ext (const void *e1, const void *e2) {
	struct extension_list *p = (struct extension_list *) e1;
	struct extension_list *q = (struct extension_list *) e2;
	return (stricmp (p->text,q->text));
	}

static struct extension_list *extension_list = NULL;
static int extension_list_size = 0;
static int extension_list_length = 0;

/*----------------------------------------------------------------------*/

#define BASE			10000
#define UNRECOGNIZED	BASE-1

int Unrecognized_Extension = UNRECOGNIZED;

int extension_of (const char *text) {
	struct extension_list k,*key;
	char string[128];

	if (extension_list_length == 0) return (UNRECOGNIZED);

	if (!text) return (UNRECOGNIZED);
	if (*text) {
		char *s = string;
		const char *t = text;
		while (*t && !isspace(*t) && *t != ':' && *t != '=' && (s - string < 127)) *s++ = *t++;
		*s = 0;
		k.text = string;
		key = (struct extension_list *) bsearch (&k,extension_list,extension_list_length,sizeof(struct extension_list),compare_ext);
		if (key) return (key->key);
		else return (UNRECOGNIZED);
		}
	else return (UNRECOGNIZED);
	}

int extension_of_sgml (const char *text) {
	int i;
	char string[64];

	if (extension_list_length == 0) return (UNRECOGNIZED);

	if (!text) return (UNRECOGNIZED);
	if (*text) {
		char *s = string;
		const char *t = text;
		while (*t && !isspace(*t) && (s - string < 63)) *s++ = *t++;
		*s = 0;
		for (i=0; i < extension_list_length; i++)
			if (strcmp (extension_list[i].sgml,string) == 0) break;
		if (i < extension_list_length)
			return (extension_list[i].key);
		else
			return (UNRECOGNIZED);
		}
	else return (UNRECOGNIZED);
	}

char *text_of_extension (int key) {
	int i;

	for (i=0; i < extension_list_length; i++)
		if (extension_list[i].key == key) return (extension_list[i].text);
	return ("(unknown)");
	}

char *sgml_of_extension (int key) {
	int i;

	for (i=0; i < extension_list_length; i++)
		if (extension_list[i].key == key) return (extension_list[i].sgml);
	return ("");
	}

int z3950_of_extension (int key) {
	int i;

	for (i=0; i < extension_list_length; i++)
		if (extension_list[i].key == key) return (extension_list[i].z3950);
	return (0);
	}

int is_extension (int key) {
	if (key >= BASE && key - BASE < extension_list_length) return (1);
	else return (0);
	}

int parent_child (int parent, int child) {
	int i;
	struct member *m;

	/*------------------------------------------------------------------*\
	 | First check to see if child is an extension; if so, look through	|
	 | its parent list to see if the given parent is legal.				|
	\*------------------------------------------------------------------*/

	for (i=0; i < extension_list_length; i++)
		if (extension_list[i].key == child)
			for (m=extension_list[i].parent; m; m=m->next)
				if (m->key == parent) return (1);

	/*------------------------------------------------------------------*\
	 | Next check to see if parent is an extension; if so, look through	|
	 | its child list to see if the given child is legal.				|
	\*------------------------------------------------------------------*/

	if (i >= extension_list_length)
		for (i=0; i < extension_list_length; i++)
			if (extension_list[i].key == parent)
				for (m=extension_list[i].child; m; m=m->next)
					if (m->key == child) return (1);

	/* If no association exists, return zero */

	return (0);
	}

int extension_is_compound (int key) {
	int i;

	for (i=0; i < extension_list_length; i++)
		if (extension_list[i].key == key)
			return (extension_list[i].compound);

	return (0);
	}

#undef GRANULARITY
#define GRANULARITY		64

static void build_extension_list (void) {
	int i;
	struct item *p,*q;
	char *s;
	int key;
	struct member *m;

	/*------------------------------------------------------------------*\
	 | Build a list relating extended keys with their text names and	|
	 | grammatical information.  First get a list of names and keys.	|
	\*------------------------------------------------------------------*/

	i = extension_list_length;

	if (extension_list_size == 0)
		if (extension_list = (struct extension_list *) malloc (GRANULARITY * sizeof(struct extension_list)))
			extension_list_size = GRANULARITY;
		else
			fprintf (stderr,"Error: could not allocate extension table\n");

	if (p = find_local (NULL,"local"))
		do {
			if (q = find_child_local (p,"name"))
				if (s = text_of_local (q)) {

					/*--------------------------------------------------*\
					 | Inject a little leniency here.  If someone has a	|
					 | definition for an existing element, we used to	|
					 | complain and insist that they must have made a	|
					 | grievous error.  With the inclusion into mp of	|
					 | extensions from the official profiles and ESRI,	|
					 | we are more likely to encounter this situation	|
					 | in innocence.  So if we already know the element	|
					 | that is named in an extension file, just don't	|
					 | add it to the list of extensions, and hope for	|
					 | the best.  This means we will produce unexpected	|
					 | behavior if somebody is using as an extension	|
					 | an element matching one in the standard or in	|
					 | any of the included profiles and extensions.		|
					 | PNS 20031028										|
					\*--------------------------------------------------*/

					if (key_of (s) == Wunknown) {
						extension_list[i].key = BASE + i;
						extension_list[i].text = s;
						extension_list[i].sgml = s;
						extension_list[i].z3950 = 0;
						extension_list[i].parent = NULL;
						extension_list[i].child = NULL;
						extension_list[i].compound = 0;
						if (q = find_child_local (p,"sgml"))
							if (s = text_of_local (q))
								extension_list[i].sgml = s;
						if (q = find_child_local (p,"z3950"))
							if (s = text_of_local (q))
								extension_list[i].z3950 = atoi(s);
						i++;
						if (i >= extension_list_size)
							if (extension_list = (struct extension_list *) realloc (extension_list,(extension_list_size + GRANULARITY) * sizeof (struct extension_list)))
								extension_list_size += GRANULARITY;
							else
								fprintf (stderr,"Error: could not enlarge extension table\n");
						}
					}
				else
					fprintf (stderr,"Warning (line %d): In extensions file, keyword Name must have an associated value\n",
						q->line_number);
			else
				fprintf (stderr,"Warning (line %d): In extensions file, keyword Local must have an associated Name\n",
					p->line_number);
			} while (p = find_next_local (p,"local"));

	/*------------------------------------------------------------------*\
	 | Sort the extension list on the text for faster lookup.			|
	\*------------------------------------------------------------------*/

	extension_list_length = i;
	qsort (extension_list,extension_list_length,sizeof(struct extension_list),compare_ext);

	/*------------------------------------------------------------------*\
	 | Having built the list, there now exist keys such that			|
	 | key_of (any text that appears in a parent or child statement)	|
	 | will return a legitimate value.  If not, generate an error msg.	|
	\*------------------------------------------------------------------*/

	if (p = find_local (NULL,"local"))
		do {
			if (q = find_child_local (p,"name"))
				if (s = text_of_local (q)) {
					key = key_of (s);

					/*--------------------------------------------------*\
					 | The list was sorted alphabetically, so given a	|
					 | numeric key, we must do a linear search.			|
					\*--------------------------------------------------*/

					for (i=0; i < extension_list_length; i++)
						if (extension_list[i].key == key) break;
					if (i >= extension_list_length) {
						fprintf (stderr,"Error: internal inconsistency in extension handling\n");
						fprintf (stderr,"  Details: Failed to find an extension's information\n");
						fprintf (stderr,"           Looking for element code %d\n",key);
						fprintf (stderr,"           which I know as \"%s\"\n",s);
						fprintf (stderr,"           defined on line %d of the extension file.\n",q->line_number);
						if (strchr (s,' ')) {
							fprintf (stderr,"           Element names cannot contain blanks.\n");
							fprintf (stderr,"  Action:  Edit the name to remove any blanks.\n");
							}
						exit (1);
						}

					/*--------------------------------------------------*\
					 | Create a list of parents under which this element|
					 | is permitted.									|
					\*--------------------------------------------------*/

					if (q = find_child_local (p,"parent")) {
						if (s = text_of_local (q))
							if ((key = key_of (s)) != Wnull)
								if (m = (struct member *) malloc (sizeof (struct member))) {
									m->next = NULL;
									m->key = key;
									extension_list[i].parent = m;
									while (q = find_next_local (q,"parent")) {
										if (s = text_of_local (q))
											if ((key = key_of (s)) != Wnull)
												if (m->next = (struct member *) malloc (sizeof (struct member))) {
													m = m->next;
													m->next = NULL;
													m->key = key;
													}
												else
													return;
											else
												fprintf (stderr,"Error (line %d): In extensions file, unrecognized element name %s\n",
													q->line_number,s);
										else
											fprintf (stderr,"Warning (line %d): In extensions file, keyword Parent must have an associated value\n",
												q->line_number);
										}
									}
								else
									return;
							else
								fprintf (stderr,"Error (line %d): In extensions file, unrecognized element name %s\n",
									q->line_number,s);
						else
							fprintf (stderr,"Warning (line %d): In extensions file, keyword Parent must have an associated value\n",
								q->line_number);
						}

					/*--------------------------------------------------*\
					 | Create a list of children permitted under this	|
					 | element.  Always include Wblank.					|
					\*--------------------------------------------------*/

					if (m = (struct member *) malloc (sizeof (struct member))) {
						m->next = NULL;
						m->key = Wblank;
						extension_list[i].child = m;
						}
					else
						return;

					if (q = find_child_local (p,"child")) {
						if (s = text_of_local (q))
							if ((key = key_of (s)) != Wnull)
								if (m->next = (struct member *) malloc (sizeof (struct member))) {
									extension_list[i].compound = 1;
									m = m->next;
									m->next = NULL;
									m->key = key;
									while (q = find_next_local (q,"child")) {
										if (s = text_of_local (q))
											if ((key = key_of (s)) != Wnull)
												if (m->next = (struct member *) malloc (sizeof (struct member))) {
													m = m->next;
													m->next = NULL;
													m->key = key;
													}
												else
													return;
											else
												fprintf (stderr,"Error (line %d): In extensions file, unrecognized element name %s\n",
													q->line_number,s);
										else
											fprintf (stderr,"Warning (line %d): In extensions file, keyword Child must have an associated value\n",
												q->line_number);
										}
									}
								else
									return;
							else
								fprintf (stderr,"Error (line %d): In extensions file, unrecognized element name %s\n",
									q->line_number,s);
						else
							fprintf (stderr,"Warning (line %d): In extensions file, keyword Child must have an associated value\n",
								q->line_number);
						}
					else

						/*----------------------------------------------*\
						 | If no children were specified, then include	|
						 | Wunknown in the list as well.				|
						\*----------------------------------------------*/

						if (m->next = (struct member *) malloc (sizeof (struct member))) {
							m = m->next;
							m->next = NULL;
							m->key = Wunknown;
							}
						else
							return;
					}
			} while (p = find_next_local (p,"local"));
	}

/*----------------------------------------------------------------------*\
 | When parsing XML, we are increasingly likely to encounter extensions	|
 | that we don't know about ahead of time.  We can ignore them, by not	|
 | including them or their children in the parse tree.  But that policy	|
 | works against the spirit of well-formed XML, in which elements that	|
 | we don't know about, if written properly and nested properly, should	|
 | be tolerated (if not acted upon).  So we need a mechanism to add an	|
 | extension, with minimal additional information, when we find such a	|
 | thing in the XML file.  This function creates an extension element	|
 | using just only the tag name and the key of the element in which it	|
 | was encountered.  This allows editors to display the element and its	|
 | substructure, and allows mp to re-express it, albeit strangely, in	|
 | its HTML output.														|
\*----------------------------------------------------------------------*/

int add_extension (const char *tag, enum fgdc_keyword parent) {
	int i,j,k;
	struct member *m;
	char *s;

	if (extension_list_size == 0)
		if (extension_list = (struct extension_list *) malloc (GRANULARITY * sizeof(struct extension_list)))
			extension_list_size = GRANULARITY;
		else
			fprintf (stderr,"Error: could not allocate extension table\n");

	i = extension_list_length;

	if (s = malloc (1 + strlen (tag))) {
		strcpy (s,tag);
		}

	k = BASE + i;
	extension_list[i].key = k;
	extension_list[i].text = s;
	extension_list[i].sgml = s;
	extension_list[i].z3950 = 0;
	extension_list[i].parent = NULL;
	extension_list[i].child = NULL;
	extension_list[i].compound = 0;

	/*------------------------------------------------------------------*\
	 | Create a linked list of possible parents (will contain only the	|
	 | parent key given in this function call).							|
	\*------------------------------------------------------------------*/

	if (m = (struct member *) malloc (sizeof (struct member))) {
		m->next = NULL;
		m->key = parent;
		extension_list[i].parent = m;
		}

	/*------------------------------------------------------------------*\
	 | Create a linked list of possible children, which begins with		|
	 | Wunknown.														|
	\*------------------------------------------------------------------*/

	if (m = (struct member *) malloc (sizeof (struct member))) {
		m->next = NULL;
		m->key = Wunknown;
		extension_list[i].child = m;
		}

	/*------------------------------------------------------------------*\
	 | Find the parent's extension element (if there is one).  Add this	|
	 | element to its list of permitted children, and infer that the	|
	 | parent is compound.												|
	\*------------------------------------------------------------------*/

	for (j=0; j < extension_list_length; j++)
		if (extension_list[j].key == parent) {
			extension_list[j].compound = 1;
			if (m = extension_list[j].child) {
				while (m->next)
					if (m->key = Wunknown) {
						m->key = k;
						break;
						}
					else
						m = m->next;

				if (m->key != k)
					if (m->next = (struct member *) malloc (sizeof (struct member))) {
						m = m->next;
						m->next = NULL;
						m->key = k;
						}
				}
			else
				if (m = (struct member *) malloc (sizeof (struct member))) {
					m->next = NULL;
					m->key = k;
					extension_list[j].child = m;
					}
			break;
			}

	i++;
	if (i >= extension_list_size)
		if (extension_list = (struct extension_list *) realloc (extension_list,(extension_list_size + GRANULARITY) * sizeof (struct extension_list)))
			extension_list_size += GRANULARITY;
		else
			fprintf (stderr,"Error: could not enlarge extension table\n");

	extension_list_length = i;
	qsort (extension_list,extension_list_length,sizeof(struct extension_list),compare_ext);

	return (k);
	}

/*----------------------------------------------------------------------*\
 | To support the editor, we need peculiar functions that walk through	|
 | the extension list and gives info about each extension.				|
\*----------------------------------------------------------------------*/

int get_extension_count (void) {
	return (extension_list_length);
	}

struct extension_info {
	int key;
	enum fgdc_keyword parent [1 + MAX_COMPOUND_COUNT];
	enum fgdc_keyword child [1 + MAX_COMPONENT_COUNT];
	};

static struct extension_info extension_info;

struct extension_info *get_extension_info (int i) {
	struct extension_info *e = NULL;
	struct member *m;
	int j;

	if (i < extension_list_length) {
		extension_info.key = extension_list[i].key;

		j = 0;
		for (m=extension_list[i].parent; m; m=m->next) {
			extension_info.parent[j] = m->key;
			if (j < MAX_COMPOUND_COUNT) j++;
			else break;
			}
		extension_info.parent[j] = Wnull;

		j = 0;
		for (m=extension_list[i].child; m; m=m->next) {
			extension_info.child[j] = m->key;
			if (j < MAX_COMPONENT_COUNT) j++;
			else break;
			}
		extension_info.child[j] = Wnull;

		e = &extension_info;
		}
	return (e);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
