/*----------------------------------------------------------------------*\
 | Functions to identify an element from its name.						|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "stricmp.h"

struct alias_list {
	enum fgdc_keyword key;
	char *text;
	};

extern struct alias_list *get_keyword_list (void);

static struct alias_list *alias_list = NULL;
static int alias_count = 0;
static int alias_limit = 0;

/*----------------------------------------------------------------------*\
 | different returns 0 if the string at text begins with that given by	|
 | name, not considering letter case and allowing ' ' in text where '_'	|
 | appears in name.														|
\*----------------------------------------------------------------------*/

static int different (char *name, char *text) {
	char *s = name;
	char *t = text;
	int d = 0;

	while (*s && !d) {
		d = toupper(*s) - toupper(*t);
		if (*s == '_' && (*t == ' ' || *t == '_')) d = 0;
		s++;
		t++;
		}
	return (d);
	}

/*----------------------------------------------------------------------*\
 | best_match finds the longest element name that can be found intact	|
 | in the given text string, using different() to discard mismatches.	|
 | This function returns the key of the matching element, and fills the	|
 | pointer end with the address of the first character that is not part	|
 | of the element name.													|
\*----------------------------------------------------------------------*/

enum fgdc_keyword best_match (char *string, char **end) {
	int i,j,m = 0,n;

	if (!string) return (Wnull);
	if (!*string) return (Wunknown);

	if (!alias_list) {
		alias_list = get_keyword_list();
		for (i=0; alias_list[i].text; i++);
		alias_count = alias_limit = i;

#ifdef DEBUG
		fprintf (stderr,"best_match(\"%s\")\n",string);
		fprintf (stderr,"alias_list = 0x%X\n",alias_list);
		fprintf (stderr,"alias_count = %d\n",alias_count);
		for (i=0; i < alias_count; i++) {
			if (alias_list[i].text)
				fprintf (stderr,"%3d: %s\n    [%s]\n",i,text_of(alias_list[i].key),alias_list[i].text);
			else
				fprintf (stderr,"%3d: %s\n    [(null)] ERROR!\n",i,text_of(alias_list[i].key));
			}
#endif
		}

	n = 0;

	for (i=0; i < alias_count; i++)
		if (!different(alias_list[i].text,string))
			if ((j = strlen (alias_list[i].text)) > n) {
				m = i;
				n = j;
				}
	*end = string + n;
	if (n) return (alias_list[m].key);
	else return (Wunknown);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

#define GRANULARITY	64

static void add_one_alias (enum fgdc_keyword key, char *text) {
	int i,n;
	char *s;
	struct alias_list *keyword_list;

	if (!alias_list) {
		keyword_list = get_keyword_list();
		for (i=0; keyword_list[i].text; i++);
		alias_count = i;
		alias_limit = (1 + alias_count/GRANULARITY)*GRANULARITY;
		if (alias_list = (struct alias_list *) malloc (alias_limit * sizeof (struct alias_list)))
			memcpy (alias_list,keyword_list,alias_count*sizeof (struct alias_list));
		else {
			fprintf (stderr,"Error: could not allocate memory for standard aliases\n");
			exit (1);
			}
		}

	if (alias_count + 1 >= alias_limit) {
		n = alias_limit + GRANULARITY;
		if (alias_list = (struct alias_list *) realloc (alias_list,n * sizeof (struct alias_list)))
			alias_limit = n;
		else {
			fprintf (stderr,"Error: could not allocate memory for new aliases\n");
			exit (1);
			}
		}

	if (s = (char *) malloc (1 + strlen (text))) {
		strcpy (s,text);
		alias_list[alias_count].key = key;
		alias_list[alias_count].text = s;
		alias_count++;
		alias_list[alias_count].key = Wnull;
		alias_list[alias_count].text = NULL;
		}
	else {
		fprintf (stderr,"Error: could not allocate memory for new alias text\n");
		exit (1);
		}

	}

#define MAX_LINE_LENGTH	1024

extern int get_extension_count (void);
extern int *get_extension_info (int i);

void add_aliases (char *alias_file) {
	FILE *in;
	char line [MAX_LINE_LENGTH];
	char *s,*t;
	enum fgdc_keyword key;
	int line_number = 0;
	int i,n;

	if (n = get_extension_count())
		for (i=0; i < n; i++) {
			key = *get_extension_info(i);
			if (s = text_of(key))
				if (*s) add_one_alias (key,s);
			}

	if (alias_file)
		if (in = fopen (alias_file,"r")) {
			while (fgets (line,MAX_LINE_LENGTH,in)) {
				line_number++;
				s = line + strlen (line) - 1;
				if (*s == '\n') *s-- = 0;
				if (*s == '\r') *s-- = 0;
				while (s >= line && isspace(*s)) *s-- = 0;
				if (s >= line) {
					for (s=line; *s && isspace(*s); s++);
					key = key_of (s);
					t = s;
					while (*s && !isspace(*s)) s++;
					if (*s) *s++ = 0;
					while (*s &&  isspace(*s)) s++;
					if (key == Wunknown)
						if (stricmp (t,"unknown") != 0)
							fprintf (stderr,"Warning (alias file line %d): alias maps to unknown key \"%s\" the text \"%s\"\n",line_number,line,s);
					if (*s) add_one_alias (key,s);
					else
						fprintf (stderr,"Warning: no alias for %s at line %d of %s\n",text_of(key),line_number,alias_file);
					}
				}
			fclose (in);


#ifdef DUMP_ALIAS_TABLE
			for (i=0; i < alias_count; i++) {
				if (alias_list[i].text)
					fprintf (stderr,"%3d: %s\n    [%s]\n",i,text_of(alias_list[i].key),alias_list[i].text);
				else
					fprintf (stderr,"%3d: %s\n    [(null)] ERROR!\n",i,text_of(alias_list[i].key));
				}
#endif
			}
		else
			fprintf (stderr,"Warning: could not open alias file %s\n",alias_file);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
