/*----------------------------------------------------------------------*\
 | MetaMerge provides the ability to merge two partial metadata files	|
 | into one FGDC compliant metadata file.  The purpose of this is to	|
 | facilitate the combining of a template file with a file that fills	|
 | in specific information.  This is useful when creating metadata for	|
 | a data set for which the difference between each datum is well		|
 | defined.																|
 |																		|
 | Doug Bakewell (Canadian Forest Service, Natural Resources Canada)	|
 | 506 W. Burnside Road, Victoria, BC  V8Z 1M5  Canada					|
 | (250)363-6031 - dbakewel@nrcan.gc.ca									|
 | http://nfis-metadata.cfs.nrcan.gc.ca									|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* MP includes */
#include "keyword.h"
#include "item.h"
#include "config.h"
#include "revision.h"

#define REVISION "2.3"

FILE *out = stderr;
int check_text_values = 1;

/* from modified text.c */
static int allow_indented_text = 1;
static int keep_blank_lines = 1;

extern void equalize_indented_scalars (struct item *p);
extern struct item *decode_text (char *dst);
extern char *read_text (char *name);
/*end*/

/* from tree.c */
extern struct item *copy_item (struct item *p);
/* end */

extern int stricmp (const char *s, const char *t);

extern void read_local (char *local_file);

extern struct item *parse_text (char *name, char **buffer);

extern void upgrade (struct item *p);
extern int check_syntax (struct item *p);
extern void summarize_errors (FILE *fp);

extern void write_text (char *name, struct item *p);

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static struct item *mm_decode_tree (struct item *first) {
	int i,j,k;
	char *s,*t;
	struct item *p,*q,*r;
	int item_count;
	struct item **list;

	if (!first) return (NULL);

	item_count = 0;
	for (p=first; p; p=p->next) {
		for (s=p->d,i=0; *s && isspace (*s); s++,i++);
		p->indent = i;
		p->d = s;
		p->key = key_of (s);
		t = text_of (p->key);

		/*--------------------------------------------------------------*\
		 | If the current item can have a child on this line, determine	|
		 | whether it does, and store the child's information.  The		|
		 | rules prohibit such a child from having children of its own.	|
		\*--------------------------------------------------------------*/

		if (p->key != Wunknown && p->key != Wblank) {
			s = p->d + strlen (t);
			if (isspace (*s) || *s == ':' || *s == '=') {
				*s++ = 0;
				p->len = strlen (p->d);
			}
			while (*s && isspace (*s)) s++;
			if (*s == ':' || *s == '=') s++;
			while (*s && isspace (*s)) s++;
			if (*s) {
				q = allocate_item (1);
				q->d = s;
				q->len = strlen (s);
				q->original = 1;
				q->indent = p->indent + 1;
				q->line_number = p->line_number;
				q->key = Wunknown;
				q->parent = p;
				p->child = q;
			}
		}
		item_count++;
	}

	/*------------------------------------------------------------------*\
	 | Estimate the proper indentation to be assigned to blank lines.	|
	 | At this point the metadata are stored as one doubly-linked list,	|
	 | so you can go forward and backward using p->next and p->prev.	|
	\*------------------------------------------------------------------*/

	for (p=first; p; p=p->next)
		if (p->key == Wblank) {

			/* Find the most recent element that is not blank */

			for (r=p->prev; r; r=r->prev)
				if (r->key != Wblank) break;

			/* Find the next element that is not blank */

			for (q=p->next; q; q=q->next)
				if (q->key != Wblank) break;

			/*----------------------------------------------------------*\
			 | Blanks should be indented like the more indented of		|
			 | (previous nonblank) and (next nonblank).  If either of	|
			 | these nodes does not exist, the proper indentation is	|
			 | whichever does exist.  If neither exist, the file		|
			 | contains only blank lines, so zero will do.			|
			\*----------------------------------------------------------*/

			if (r && q)
				p->indent = ((r->indent > q->indent) ? r->indent : q->indent);
			else
				if (r) p->indent = r->indent;
				else
					if (q) p->indent = q->indent;
					else
						p->indent = 0;
			}

	/*------------------------------------------------------------------*\
	 | Figure out the topology of the file.  Create a list of items		|
	 | as an array of pointers.  For each item, determine whether it	|
	 | has a child and what its parent is.					|
	\*------------------------------------------------------------------*/

	if (list = (struct item **) malloc (item_count * sizeof(struct item *))) {
		memset (list,0,item_count*sizeof(struct item *));
		i = 0;
		for (p=first; p; p=p->next) list[i++] = p;

		for (i=0; i < item_count-1; i++) {
			p = list[i];
			q = list[i+1];

			/*----------------------------------------------------------*\
			 | The relationship between successive items in the file is	|
			 | determined by comparing their levels of indentation.		|
			\*----------------------------------------------------------*/

			j = q->indent - p->indent;

			/*----------------------------------------------------------*\
			 | If q is indented more than p, q is a child of p, and p	|
			 | is parent of q.						|
			 |								|
			 | Modified 30-Jan-1997 to not discard scalar child already	|
			 | in p (can only happen as text following element name.	|
			\*----------------------------------------------------------*/

			if (j > 0) {
				if (allow_indented_text &&
					p->key == Wunknown && q->key == Wunknown)
					q->parent = p->parent;
				else {
					if (p->child && (q->key == Wunknown || q->key == Wblank)) {

						/*----------------------------------------------*\
						 | Modified 20-Feb-1998 to handle cases where a	|
						 | string immediately follows the name and a	|
						 | blank line follows them.  If the next non-	|
						 | blank is Wunknown, keep the string that was	|
						 | placed on the same line as the element.  If	|
						 | the next non-blank is a recognized element,	|
						 | the string after the element name was not	|
						 | permitted; discard it with a warning.	|
						\*----------------------------------------------*/

						if (q->key == Wblank) {
							r = NULL;
							/* Point r at the next non-blank element */
							for (k=i+1; k < item_count; k++) {
								r = list[k];
								if (r->key != Wblank) break;
							}
							/* if r is an element, discard the text */
							if (r)
								if (r->key != Wunknown && r->key != Wblank) {
									fprintf (out,"Warning 1 (line %d): Extraneous text following element %s ignored\n",
										p->line_number,
										text_of(p->key)
										);
									p->child = q;
									q->prev = NULL;
								} else {
									/* otherwise include the text */
									for (r=p->child; r->next; r=r->next);
									r->next = q;
									q->prev = r;
								}
						} else {
							/* q->key == Wunknown */
							for (r=p->child; r->next; r=r->next);
							r->next = q;
							q->prev = r;
						}
					} else {
						/*DOUG'S BIG CHANGE TO ALLOW ATTRIBUTES TO HAVE A VALUE WITH THEM*/
						if (p->child) {
/*							fprintf (out,"Warning 2 (line %d): Extraneous text following element %s ignored\n", p->line_number, text_of(p->key));*/
							p->child->next = q;
						    	q->prev = p->child;
						} else {
							p->child = q;
							q->prev = NULL;
						}
					}
					q->parent = p;
					p->next = NULL;
				}
/*\
						if (p->child) {
							fprintf (out,"Warning (line %d): Extraneous text following element %s ignored\n", p->line_number, text_of(p->key));
							p->child->next = q;
							q->prev = p->child;
							p->child->prev = NULL;
							p->child->child = NULL;
							}
						else {
							p->child = q;
							q->prev = NULL;
							}
						q->parent = p;
						p->next = NULL;
					}
\*/
			}

			/*----------------------------------------------------------*\
			 | If q is indented the same as p, q is just the next part	|
			 | of the compound type (or text field) of which p is a		|
			 | part.  The parent of q is that of p.						|
			\*----------------------------------------------------------*/

			if (j == 0) {
				q->parent = p->parent;
				}

			/*----------------------------------------------------------*\
			 | If q is indented less than p, q is the next element in a	|
			 | compound type begun earlier.  Find the previous element	|
			 | by searching backwards through the list for the nearest	|
			 | element q that was indented the same as or less than q;	|
			 | store q into its next pointer, set the parent of q to	|
			 | that of r, and sever the link from p to q.				|
			\*----------------------------------------------------------*/

			if (j < 0) {
				for (k=i; k >= 0; k--)
					if (list[k]->indent <= q->indent) {
						r = list[k];
						if (r->indent < q->indent)
							fprintf (out,"Warning: ambiguous indentation at line %d\n",q->line_number);
						if (r->next == NULL) {
							r->next = q;
							q->prev = r;
							q->parent = r->parent;
							p->next = NULL;
							}
						else {
							struct item *rr;
							for (rr=r; rr->next && rr->next != q; rr=rr->next);
							rr->next = q;
							q->prev = rr;
							q->parent = rr->parent;
							if (rr != p) p->next = NULL;
							}
						break;
						}

				/*------------------------------------------------------*\
				 | If you fell through that loop, there is an error in	|
				 | the file's indentation that cannot easily be over-	|
				 | looked.  This indicates that a line somewhere in the	|
				 | middle of the file is indented less than all other	|
				 | lines in the file; the structure is unclear.			|
				\*------------------------------------------------------*/

				if (k < 0)
					fprintf (out,"Error: too little indentation at line %d\n",q->line_number);
				}
			}

		/*--------------------------------------------------------------*\
		 | Traverse the tree, looking for situations in which scalar	|
		 | values have children that are themselves scalar values; in	|
		 | such cases, make them siblings, not children.				|
		\*--------------------------------------------------------------*/

		if (allow_indented_text) equalize_indented_scalars (first);

		free (list);
		}
	else {
		fprintf (out,"Error: could not allocate array of pointers to items\n");
		exit (1);
		}
	return (first);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

struct item *mm_parse_text (char *input_file) {
	void *opt;

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"indent"))
			if (stricmp (text_of_option (opt),"strict") == 0)
				allow_indented_text = 0;

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"blanks"))
			if (stricmp (text_of_option (opt),"ignore") == 0)
				keep_blank_lines = 0;

	return (mm_decode_tree (decode_text (read_text (input_file))));
}

struct item * mm_mkMetadata(struct item *m)
{
	if (m->key != WMetadata) {
		struct item *p = allocate_item (1);
		p->next = NULL;
		p->parent = NULL;
		p->child = m;
		p->indent = -1;
		p->line_number = -1;
		p->key = WMetadata;
		p->original = 0;
		p->len = 1 + strlen (text_of(WMetadata));
		p->d = (char *) malloc (p->len);
		strcpy (p->d,text_of(WMetadata));
		m = p;
		for (p=m->child; p; p=p->next) p->parent = m;
	}
	return(m);
}

void mm_errors(int err_level, struct item *node, int language) {
	if (language == 0) {
		if (err_level == 0) {
			fprintf (stderr,"\nERROR: Cannot find location in template file as specified in merge file. Directive or value may be missing. Offense occurred at line number %d of ""merge""\n",node->line_number);
		} else if (err_level == 1) {
			fprintf (stderr,"ERROR: Directive to skip could not be performed because attribute to skip was not found in the current hierarchy and level of the template. Offense occurred at line number %d of ""merge""\n",node->line_number);
		}
	} else if (language == 1) {
		if (err_level == 0) {
			fprintf (stderr,"\nERREUR: NOCHILD a %d\n",node->line_number);
		} else if (err_level == 1) {
			fprintf (stderr,"\nERREUR: NOSKIP a %d\n",node->line_number);
		}
	} else {
		fprintf (stderr,"Non Indication Langue/No Language Specified\nErreur/Error %d",err_level);
	}
	exit(1);
}

void mm_warning(int i, struct item *node, int language) {
	if (language == 0) {
		if (i == 0) {
			fprintf (stderr, "\nNo Child to Add at %d\n",node->parent->line_number);
		}
	} else if (language == 1) {
		if (i == 0) {
			fprintf (stderr, " \nLes Enfants Ne Faire Pas une Addition a %d\n",node->parent->line_number);
		}
	} else {
		fprintf (stderr,"\nNon Indication Langue/No Language Specified\nAlerte/Warning %d",i);
	}
}

/***
MERGE PHASE
MetaMerge accomplishes the merge phase by manipulating the trees created by
the load phase. In the normal-case the result creates no output except for
the merged tree which is used by the output phase. The structure and
directives of the merge file drive the merge. Figure 1 shows the FSM for
the merge process. Exceptions are given by ID and are expanded in Table 2.
***/
/*** Transitions and Outputs ***/
struct item *mm_merge(struct item *t, struct item *m) {

/* This LANGUAGE definition is only temporary until proper language exceptions are implemented */
#define language 0

	/*** o:= t ***/
	struct item *o = t;

	/*** me := first element of m ***/
	struct item *me = m;

	/*** oe := first element of o ***/
	struct item *oe = o;

	/* tmpe and tmpe2 used for manipulating the tree structure */
	struct item *tmpe, *tmpe2;

	/*** while me still points at a valid element ***/
	while (me) {

		/*** perform one of the following cases (evaluate in order): ***/
		/*** if me has as its first child a value and (the value does not start with "#MM" or the value is "#MMADD#") then ***/
		if ((me->child && me->child->key == 335) && (strncmp(me->child->d, "#MM", 3) != 0 || strcmp(me->child->d, "#MMADD#") == 0)) {

			/*** if the value is "#MMADD# then ***/
			if (strcmp(me->child->d, "#MMADD#") == 0) {

				/*** remove the "#MMADD# from me ***/
				if (me->child->next) {
					me->child->next->prev = me->child->prev; /* Should be NULL always */
					me->child = me->child->next;
				} else {
					mm_warning(0,me,language); /* There are no children to add */
					me->child = NULL;
				}
			}

			/* me is copied to tmpe so that me is unchanged */
			tmpe = copy_item(me);

			/*** if oe is a valid element then ***/
			if (oe) {

				/*** insert the entire me hierarchy as the element previous to oe ***/
				tmpe2 = tmpe;
				tmpe2->parent = oe->parent;
				tmpe2->next = oe;
				tmpe2->prev = oe->prev;
				oe->prev = tmpe2;

				/* if tmpe2 is not the first element for the root level of o then */
				if (tmpe2->prev) {
					tmpe2->prev->next = tmpe2;

				/* else set o to be the first element of the subtree */
				} else {
					o = tmpe2;
				}

			/*** else ***/
			} else {

				/*** insert the entire me hierarchy as the last element of o ***/
				if (!o) {

					/* Will o ever be NULL? */
					o = tmpe;
					o->next = NULL;
					o->prev = NULL;
					/*note parent is NULL but handled after in recursive call*/
				} else {
					tmpe2 = o;
					while(tmpe2->next) {
						tmpe2 = tmpe2->next;
					}
					tmpe2->next = tmpe;
					tmpe->parent = tmpe2->parent;
					tmpe->next = NULL;
					tmpe->prev = tmpe2;
				}
			}

		/*** if me has as its first child a value and the value is "#MMSKIP" then ***/
		} else if ((me->child && me->child->key == 335) && (strcmp(me->child->d, "#MMSKIP#") == 0 )) {

			/*** search oe and the elements following oe at the root level of o and find the first element whose name matches me ***/
			while (oe != NULL) {
				if (strcmp(me->d, oe->d) == 0) {
					break;
				}
				oe = oe->next;
			}

			/*** oe := the next element after the found element else invalid ***/
			if (oe) {
				oe = oe->next;

			/*** if oe is invalid then exception(NOSKIP) ***/
			} else {
				mm_errors(1,me,language);
			}

		/*** if me has as its first child a value and the value is "#MMDEL#" then ***/
		} else if ((me->child && me->child->key == 335) && (strcmp(me->child->d, "#MMDEL#") == 0 )) {

			/*** search oe and the elements following oe at the root level of o and find the first element whose name matches me ***/
			while (oe != NULL) {
				if (strcmp(me->d, oe->d) == 0) {
					break;
				}
				oe = oe->next;
			}

			/*** if oe is a valid element then ***/
			if (oe) {

				/*** Remove oe from o ***/
				if (oe->next) {
					oe->next->prev = oe->prev;
				}
				if (oe->prev) {
					oe->prev->next = oe->next;
				}
				if (o == oe) {
					o = o->next;
				}

				/*** oe := the next element after the deleted element else oe points nowhere ***/
				oe = oe->next;

			/*** if oe is invalid then exception(NODEL) ***/
			} else {

				/* mm_errors(2,me,language); */
				fprintf (stderr,"Nothing to Delete\n");
				exit(1);
			}

		/*** otherwise ***/
		} else {

			/*** search oe and the elements following oe at the root level of o and find the first element whose name matches me ***/
			while (oe != NULL) {
				if (strcmp(me->d, oe->d) == 0) {
					break;
				}
				oe = oe->next;
			}

			/*** oe := the found item else invalid ***/
			/*** if oe is valid then ***/
			if (oe) {

				/*** replace oe's children with the output tree from merge(oe,me) ***/
				oe->child = mm_merge(oe->child,me->child);

				/* sets all of oe's children's parent to oe (case when adding me to a NULL o) */
				if (oe->child) {
					oe->child->parent = oe;
					tmpe = oe->child;
					while(tmpe->next) {
						tmpe->next->parent = oe;
						tmpe = tmpe->next;
					}
				} /* else The Merge and the Template had NULL children */
				/*** oe:= the element next after oe else invalid ***/
				oe = oe->next;
			/*** else exception(NOCHILD) ***/
			} else {
				mm_errors(0,me,language);
			}
		}

		/*** me := the element next after me else invalid ***/
		me = me->next;
 	}

	/*** return o ***/
	return(o);
}


int main (int argc, char *argv[]) {
	int i;
	int stop = 0;
	char *template_file  = NULL;
	char *merge_file = NULL;
	char *metadata_file = NULL;
	char *cfg_file   = NULL;
	char *ext_file   = NULL;

	struct item *merge = NULL;
	struct item *template = NULL;
	struct item *metadata = NULL;
	void *opt;
	int do_upgrade = 1;

	for (i=1; i < argc; i++)
		if (memcmp (argv[i],"-m",2) == 0) {
			i++;
			merge_file = argv[i];
			}
		else
			if (memcmp (argv[i],"-t",2) == 0) {
				i++;
				template_file = argv[i];
				}
			else
				if (memcmp (argv[i],"-o",2) == 0) {
					i++;
					metadata_file = argv[i];
					}
				else
					if (memcmp (argv[i],"-c",2) == 0) {
						i++;
						cfg_file = argv[i];
						}

	if (!template_file || !merge_file || !metadata_file) {
		fprintf (stderr,"MetaMerge %s by Doug Bakewell (Natural Resources Canada, Canadian Forest Service)\n",REVISION);
		fprintf (stderr,"http://nfis-metadata.cfs.nrcan.gc.ca/");
		fprintf (stderr,"Based on mp version %s - Peter N. Schweitzer (U.S. Geological Survey)\n",revision.mp);
		fprintf (stderr,"Usage: %s -c config -t template -m merge -o metadata\n", argv[0]);
		fprintf (stderr,"      Where: config is the mp config file\n");
		fprintf (stderr,"             template is the metadata template file\n");
		fprintf (stderr,"             merge is the merge file\n");
		fprintf (stderr,"             metadata is where the merged output will be written\n");
		exit (0);
		}

	out = stderr;

	if (cfg_file) read_configuration (cfg_file);
	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"extensions")) {
			if (ext_file = text_of_option (opt))
				read_local (ext_file);
			while (opt = find_next_option (opt,"extensions"))
				if (ext_file = text_of_option (opt))
					read_local (ext_file);
			}

	fprintf (out,"MetaMerge %s - Doug Bakewell (Pacific Forestry Centre - Natural Resources Canada)\n",REVISION);

	/*------------------------------------------------------------------*\
	 | Parse the input file.  Check the extension to see how.			|
	\*------------------------------------------------------------------*/

	if(!(merge = mm_parse_text (merge_file))) {
		printf("e1\n");
		stop = 1;
		}
	if(!(template = parse_text (template_file, NULL))) {
		printf("e2\n");
		stop = 1;
		}

	if(stop) exit(1);

	merge = mm_mkMetadata(merge);
	/*
	{
		FILE *tf;
		if (!(tf = fopen ("merge_tree","w"))) tf = stdout;
		list_children (tf,merge);
		if (tf != stdout) fclose (tf);
	}
	*/

	template = mm_mkMetadata(template);
	/*
	{
		FILE *tf;
		if (!(tf = fopen ("template_tree","w"))) tf = stdout;
		list_children (tf,template);
		if (tf != stdout) fclose (tf);
	}
	*/

	/*---------------------DO MERGE HERE--------------------------------*/
	metadata = mm_mkMetadata(mm_merge(template,merge));
	/*------------------------------------------------------------------*/

	if (metadata) {
		/*
		{
			FILE *tf;
			if (!(tf = fopen ("output_tree","w"))) tf = stdout;
			list_children (tf,metadata);
			if (tf != stdout) fclose (tf);
		}
		*/

		if (opt = find_option (NULL,"upgrade"))
			if (stricmp (text_of_option (opt),"no") == 0) do_upgrade = 0;

		if (do_upgrade) upgrade (metadata);

		check_syntax (metadata);
		summarize_errors (out);

		write_text (metadata_file,metadata);

	}
	exit (0);
}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
