/*----------------------------------------------------------------------*\
 | Program to parse formal metadata to determine whether it conforms	|
 | to the FGDC Content Standards for Digital Geospatial Metadata.		|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#ifndef _MSC_VER
#include <unistd.h>
#else
#define getcwd _getcwd
#endif

#include "keyword.h"
#include "item.h"
#include "config.h"
#include "revision.h"
#include "stricmp.h"
#include "local.h"

/*----------------------------------------------------------------------*\
 | main routine decodes command line arguments, parses input, checks	|
 | the parse tree, and creates output.									|
\*----------------------------------------------------------------------*/

FILE *out = NULL;
int check_text_values = 1;

extern struct item *parse_text (char *name, char **buffer);
extern struct item *parse_sgml (char *name);
extern struct item *parse_xml  (char *name);
extern struct item *parse_dif  (char *name);

extern void use_element_names (char *language, char *profile);
extern void use_profile (char *name);

extern void fixdoc (struct item *p);
extern void upgrade (struct item *p);
extern int check_syntax (struct item *p);
extern void summarize_errors (FILE *fp);

extern void write_text (char *name, struct item *p);
extern void write_sgml (char *name, struct item *p);
extern void write_xml  (char *name, struct item *p);
extern void write_html (char *name, struct item *p);
extern void write_html_faq (char *name, struct item *p);
extern void write_dif  (char *name, struct item *p);

static void add_related_file (char *type, char *name);
static void dump_related_files (FILE *fp);
char *related_file (char *type);
char *profile_name = NULL;

static char *es_credit = "  Spanish-language element names kindly provided by Clearinghouse Nacional\n  de Datos Geogrficos, Uruguay <http://www.clearinghouse.gub.uy/>\n";
static char *ca_credit = "  Catalan-language element names kindly provided by Clearinghouse Nacional\n  de Datos Geogrficos, Uruguay <http://www.clearinghouse.gub.uy/>\n";
static char *id_credit = "  Indonesian-language element names kindly provided by the Indonesian\n  National Coordination Agency for Surveys and Mapping BAKOSURTANAL\n";
static char *fr_credit = "  French-language element names kindly provided by Environment Canada (John Cree)\n";
static char *de_credit = "  German-language element names kindly provided by Peter Korduan (University of Rostock)\n";
static char *pt_credit = "  Portuguese-language element names kindly provided by Luis Cavalcanti Bahiana (IBGE, Brazil)\n";

static char *usage_text = \
"\n"
"Usage: mp [options] input-file\n"
"\n"
"Parse FGDC metadata, report structural errors and generate useful\n"
"re-expressions of the information.\n"
"\n"
"input-file is indented text or sgml or xml\n"
"\n"
"Options:\n"
"  -c config-file      Read supplied config-file for more options\n"
"  -l language-code    Use element names in the language specified\n"
"  -e error-file       Write errors to the named error-file\n"
"  -t text-file        Write indented text to the named text-file\n"
"  -h html-file        Write outline-style HTML to the named html-file\n"
"  -f faq-file         Write FAQ-style HTML to the named html-file\n"
"  -s sgml-file        Write SGML to the named sgml-file\n"
"  -x xml-file         Write XML to the named xml-file\n"
"  -d dif-file         Write DIF (NASA, v6) to the named dif-file\n"
"  -fixdoc             Run special clean-up on DOCUMENT.aml output\n"
"\n"
"Language codes are 2-letter abbreviations\n"
"  en  English (default)\n"
"  es  Spanish\n"
"  ca  Catalan\n"
"  id  Indonesian\n"
"  fr  French\n"
"  de  German\n"
"  pt  Portuguese\n"
"\n"
"Further information at <http://geology.usgs.gov/tools/metadata/>\n"
;

/*----------------------------------------------------------------------*\
 | Handle character encoding
\*----------------------------------------------------------------------*/

static char character_encoding[256] = "ISO-8859-1";

void set_character_encoding (char *new_value) {
	if (strlen (new_value) < 256)
		strcpy (character_encoding,new_value);
	else {
		memcpy (character_encoding,new_value,255);
		character_encoding[255] = 0;
		}
	}

char *get_character_encoding (void) {
	return (character_encoding);
	}

/*----------------------------------------------------------------------*\
 | main program
\*----------------------------------------------------------------------*/

int main (int argc, char *argv[]) {
	int i;
	char *input_file = NULL;
	char *error_file = NULL;
	char *text_file  = NULL;
	char *html_file  = NULL;
	char *faq_file   = NULL;
	char *sgml_file  = NULL;
	char *xml_file   = NULL;
	char *dif_file   = NULL;
	char *cfg_file   = NULL;
	char *ext_file   = NULL;
	char *language   = NULL;
	char file_name [1024];
	char input_file_name [1024];
	char *s,*t;
	struct item *first = NULL;
	void *opt;
	int do_upgrade     = 1;
	int do_fixdoc      = 0;
	int do_prune       = 0;
	int spaces_are_bad = 1;
	int do_timestamp   = 1;
	char *text_buffer = NULL;

	#ifdef _WIN32
		spaces_are_bad = 0;
	#endif

	out = stderr;

	for (i=1; i < argc; i++)
		if (memcmp (argv[i],"-e",2) == 0) {
			i++;
			error_file = argv[i];
			}
		else
			if (memcmp (argv[i],"-t",2) == 0) {
				i++;
				text_file = argv[i];
				}
			else
				if (memcmp (argv[i],"-h",2) == 0) {
					i++;
					html_file = argv[i];
					}
				else
					if (memcmp (argv[i],"-s",2) == 0) {
						i++;
						sgml_file = argv[i];
						}
					else
						if (memcmp (argv[i],"-d",2) == 0) {
							i++;
							dif_file = argv[i];
							}
						else
							if (memcmp (argv[i],"-c",2) == 0) {
								i++;
								cfg_file = argv[i];
								}
							else
								if (memcmp (argv[i],"-fixdoc",7) == 0)
									do_fixdoc = 1;
								else
									if (memcmp (argv[i],"-f",2) == 0) {
										i++;
										faq_file = argv[i];
										}
									else
										if (memcmp (argv[i],"-x",2) == 0) {
											i++;
											xml_file = argv[i];
											}
										else
											if (memcmp (argv[i],"-l",2) == 0) {
												i++;
												language = argv[i];
												}
											else
												input_file = argv[i];

	/*------------------------------------------------------------------*\
	 | First read configuration, so we can determine whether a language	|
	 | other than English was requested.								|
	\*------------------------------------------------------------------*/

	if (cfg_file) read_configuration (cfg_file);

	/*------------------------------------------------------------------*\
	 | If no input file, output a usage message.						|
	\*------------------------------------------------------------------*/

	if (!input_file) {
		fprintf (stderr,"mp %s - Peter N. Schweitzer (U.S. Geological Survey)\n",revision.mp);

		if (!language)
			if (opt = find_option (NULL,"input"))
				if (opt = find_option (opt,"language"))
					if (s = text_of_option (opt))
						language = s;

		if (language) {
			if (strcmp (language,"es") == 0) fprintf (stderr,es_credit);
			if (strcmp (language,"ca") == 0) fprintf (stderr,ca_credit);
			if (strcmp (language,"id") == 0) fprintf (stderr,id_credit);
			if (strcmp (language,"fr") == 0) fprintf (stderr,fr_credit);
			if (strcmp (language,"de") == 0) fprintf (stderr,de_credit);
			if (strcmp (language,"pt") == 0) fprintf (stderr,pt_credit);
			}

		fprintf (stderr,usage_text);

		if (s = config_errors()) {
			fprintf (out,"Errors occurred when processing the configuration file:\n");
			fprintf (out,"%s\n",s);
			}

		if (s = ext_errors()) {
			fprintf (out,"Errors occurred when processing the extension files:\n");
			fprintf (out,"%s\n",s);
			}

		exit (0);
		}

	/*------------------------------------------------------------------*\
	 | Clip standard file extension off the input file name				|
	 | Standard extensions are txt,sgml,sgm,xml,text,met,bin			|
	\*------------------------------------------------------------------*/

	strcpy (input_file_name,input_file);
	s = input_file_name + strlen (input_file_name) - 1;
	while (s > input_file_name && *s != '.') s--;
	if (s > input_file_name && *s == '.')
		if (stricmp (s,".txt" ) == 0) *s = 0;
		else
			if (stricmp (s,".sgml") == 0) *s = 0;
			else
				if (stricmp (s,".sgm" ) == 0) *s = 0;
				else
					if (stricmp (s,".xml" ) == 0) *s = 0;
					else
						if (stricmp (s,".text") == 0) *s = 0;
						else
							if (stricmp (s,".met" ) == 0) *s = 0;

	/*------------------------------------------------------------------*\
	 | If you didn't find a standard extension, look for a custom		|
	 | extension specified by the user in the config file under			|
	 | input:ext														|
	\*------------------------------------------------------------------*/

	if (s > input_file_name && *s == '.')
		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"ext"))
				do {
					if (t = text_of_option (opt))
						if (stricmp (s,t) == 0) {
							*s = 0;
							break;
							}
						else
							if (stricmp (s+1,t) == 0) {
								*s = 0;
								break;
								}
					} while (opt = find_next_option (opt,"ext"));

	/*------------------------------------------------------------------*\
	 | Get error file name and create the error file.					|
	\*------------------------------------------------------------------*/

	if (!error_file)
		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"errors"))
				if (s = text_of_option (opt)) {
					sprintf (file_name,s,input_file_name);
					if (spaces_are_bad) for (s=file_name; *s; s++) if (isspace(*s)) *s = '_';
					if (strcmp (file_name,input_file) != 0)
						error_file = file_name;
					else
						fprintf (stderr,"Warning: requested error file %s would overwrite the input file\n",file_name);
					}
	if (error_file)
		if (!(out = fopen (error_file,"w"))) {
			fprintf (stderr,"Error: could not open error file %s\n",error_file);
			out = stderr;
			}

	/*------------------------------------------------------------------*\
	 | Print a banner message											|
	\*------------------------------------------------------------------*/

	fprintf (out,"mp %s - Peter N. Schweitzer (U.S. Geological Survey)\n",revision.mp);

	if (s = config_errors()) {
		fprintf (out,"Errors occurred when processing the configuration file:\n");
		fprintf (out,"%s\n",s);
		}

	if (s = ext_errors()) {
		fprintf (out,"Errors occurred when processing the extension files:\n");
		fprintf (out,"%s\n",s);
		}

	fprintf (out,"Info: input file = %s\n",input_file);

	if (do_timestamp) {
		time_t now;
		struct tm *tm;
		char procdate[10];
		char proctime[10];

		now = time (NULL);
		tm = localtime (&now);

		strftime (procdate,10,"%Y%m%d",  tm);
		strftime (proctime,10,"%H:%M:%S",tm);

		fprintf (out,"Info: process date = %s\n",procdate);
		fprintf (out,"Info: process time = %s\n",proctime);
		}

	/*------------------------------------------------------------------*\
	 | Use requested language.  Command-line argument overrides what is |
	 | specified in the config file.									|
	\*------------------------------------------------------------------*/

	if (!language)
		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"language"))
				if (s = text_of_option (opt))
					language = s;

	/*------------------------------------------------------------------*\
	 | Use requested profile.
	\*------------------------------------------------------------------*/

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"profile"))
			if (s = text_of_option (opt))
				profile_name = s;

	use_element_names (language,profile_name);		/* in keyword.c */
	if (profile_name) use_profile (profile_name);	/* in actions.c */

	if (language) {
		if (strcmp (language,"es") == 0) fprintf (out,es_credit);
		if (strcmp (language,"ca") == 0) fprintf (out,ca_credit);
		if (strcmp (language,"id") == 0) fprintf (out,id_credit);
		if (strcmp (language,"fr") == 0) fprintf (out,fr_credit);
		if (strcmp (language,"de") == 0) fprintf (out,de_credit);
		if (strcmp (language,"pt") == 0) fprintf (out,pt_credit);
		}

	/*------------------------------------------------------------------*\
	 | Read local extensions											|
	\*------------------------------------------------------------------*/

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"extensions")) {
			if (ext_file = text_of_option (opt)) read_local (ext_file);
			while (opt = find_next_option (opt,"extensions"))
				if (ext_file = text_of_option (opt)) read_local (ext_file);
			}

	/*------------------------------------------------------------------*\
	 | Parse the input file.  Check the extension to see how.			|
	\*------------------------------------------------------------------*/

	if (s = strrchr (input_file,'.'))
		if (stricmp (s,".sgml") == 0 || stricmp (s,".sgm") == 0)
			first = parse_sgml (input_file);
		else
			if (stricmp (s,".xml") == 0) {
				first = parse_xml (input_file);
				if (first) do_upgrade = 0;
				}

	/* If not SGML or XML and not binary or you got nothing that way, try text */

	if (!first) first = parse_text (input_file, &text_buffer);

	/* Remember the full name of the input file */

	add_related_file ("original",input_file);

	/* Remember the name of the input file, but not its path and not its extension */

	if (s = strrchr (input_file_name,'/')) add_related_file ("input",s);
	else
		if (s = strrchr (input_file_name,'\\')) add_related_file ("input",s);
		else
			add_related_file ("input",input_file_name);

	/*------------------------------------------------------------------*\
	 | If you got data from the file, proceed, otherwise exit			|
	\*------------------------------------------------------------------*/

	if (first) {

		/*--------------------------------------------------------------*\
		 | Users typically won't want to dump the parse tree, but the	|
		 | facility is there in case the parser needs to be debugged.	|
		\*--------------------------------------------------------------*/

		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"tree")) {
				FILE *tf;
				if (!(tf = fopen (text_of_option (opt),"w"))) tf = stdout;
				list_children (tf,first);
				if (tf != stdout) fclose (tf);
				}

		/*--------------------------------------------------------------*\
		 | If the first node is not Metadata, make it so.  This ensures	|
		 | that required elements in Metadata will be checked.			|
		\*--------------------------------------------------------------*/

		if (first->key != WMetadata) {
			struct item *p = allocate_item (1);
			p->next = NULL;
			p->parent = NULL;
			p->child = first;
			p->indent = -1;
			p->line_number = -1;
			p->key = WMetadata;
			p->original = 0;
			p->len = 1 + strlen (text_of(WMetadata));
			p->d = (char *) malloc (p->len);
			strcpy (p->d,text_of(WMetadata));
			first = p;
			for (p=first->child; p; p=p->next) p->parent = first;
			}

		/*--------------------------------------------------------------*\
		 | Modifications to the metadata that are supported are fixdoc,	|
		 | an unusual set of changes designed specifically for altering	|
		 | the output of DOCUMENT.aml; upgrade, which changes the		|
		 | structure of the metadata to conform to CSDGM version 2; and	|
		 | prune, which deletes all empty subtrees.  Of these the only	|
		 | one that is on by default is upgrade; the others must be		|
		 | requested (with care) explicitly by the user.				|
		\*--------------------------------------------------------------*/

		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"upgrade")) {
				if (stricmp (text_of_option (opt),"no")  == 0) do_upgrade = 0;
				if (stricmp (text_of_option (opt),"yes") == 0) do_upgrade = 1;
				}

		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"prune")) do_prune = 1;

		if (do_fixdoc) fixdoc (first);
		if (do_upgrade) upgrade (first);
		if (do_prune) prune (first);

		/*--------------------------------------------------------------*\
		 | Check the syntax and summarize errors at the end.			|
		\*--------------------------------------------------------------*/

		check_syntax (first);
		summarize_errors (out);

		/*--------------------------------------------------------------*\
		 | Create output file names from input file stem or the name	|
		 | specified by the user on the command line.					|
		\*--------------------------------------------------------------*/

		if (!text_file)
			if (opt = find_option (NULL,"output"))
				if (opt = find_option (opt,"text"))
					if (opt = find_option (opt,"file"))
						if (s = text_of_option (opt)) {
							sprintf (file_name,s,input_file_name);
							if (spaces_are_bad) for (s=file_name; *s; s++) if (isspace(*s)) *s = '_';
							if (strcmp (file_name,input_file) != 0)
								text_file = file_name;
							else
								fprintf (stderr,"Warning: requested text file %s would overwrite the input file\n",file_name);
							}
		if (text_file) add_related_file ("text",text_file);

		if (!sgml_file)
			if (opt = find_option (NULL,"output"))
				if (opt = find_option (opt,"sgml"))
					if (opt = find_option (opt,"file"))
						if (s = text_of_option (opt)) {
							sprintf (file_name,s,input_file_name);
							if (spaces_are_bad) for (s=file_name; *s; s++) if (isspace(*s)) *s = '_';
							if (strcmp (file_name,input_file) != 0)
								sgml_file = file_name;
							else
								fprintf (stderr,"Warning: requested sgml file %s would overwrite the input file\n",file_name);
							}
		if (sgml_file) add_related_file ("sgml",sgml_file);

		if (!xml_file)
			if (opt = find_option (NULL,"output"))
				if (opt = find_option (opt,"xml"))
					if (opt = find_option (opt,"file"))
						if (s = text_of_option (opt)) {
							sprintf (file_name,s,input_file_name);
							if (spaces_are_bad) for (s=file_name; *s; s++) if (isspace(*s)) *s = '_';
							if (strcmp (file_name,input_file) != 0)
								xml_file = file_name;
							else
								fprintf (stderr,"Warning: requested xml file %s would overwrite the input file\n",file_name);
							}
		if (xml_file) add_related_file ("xml",xml_file);

		if (!html_file)
			if (opt = find_option (NULL,"output"))
				if (opt = find_option (opt,"html"))
					if (opt = find_option (opt,"file"))
						if (s = text_of_option (opt)) {
							sprintf (file_name,s,input_file_name);
							if (spaces_are_bad) for (s=file_name; *s; s++) if (isspace(*s)) *s = '_';
							if (strcmp (file_name,input_file) != 0)
								html_file = file_name;
							else
								fprintf (stderr,"Warning: requested html file %s would overwrite the input file\n",file_name);
							}
		if (html_file) add_related_file ("html",html_file);

		if (!faq_file)
			if (opt = find_option (NULL,"output"))
				if (opt = find_option (opt,"html"))
					if (opt = find_option (opt,"faq"))
						if (s = text_of_option (opt)) {
							sprintf (file_name,s,input_file_name);
							if (spaces_are_bad) for (s=file_name; *s; s++) if (isspace(*s)) *s = '_';
							if (strcmp (file_name,input_file) != 0)
								faq_file = file_name;
							else
								fprintf (stderr,"Warning: requested faq file %s would overwrite the input file\n",file_name);
							}
		if (faq_file) add_related_file ("faq",faq_file);

		if (!dif_file)
			if (opt = find_option (NULL,"output"))
				if (opt = find_option (opt,"dif"))
					if (opt = find_option (opt,"file"))
						if (s = text_of_option (opt)) {
							sprintf (file_name,s,input_file_name);
							if (spaces_are_bad) for (s=file_name; *s; s++) if (isspace(*s)) *s = '_';
							if (strcmp (file_name,input_file) != 0)
								dif_file = file_name;
							else
								fprintf (stderr,"Warning: requested DIF file %s would overwrite the input file\n",file_name);
							}
		if (dif_file) add_related_file ("dif",dif_file);

		/* dump_related_files (stderr); */

		if (s = related_file ("sgml")) write_sgml     (s,first);
		if (s = related_file ("html")) write_html     (s,first);
		if (s = related_file ("faq" )) write_html_faq (s,first);
		if (s = related_file ("xml" )) write_xml      (s,first);
		if (s = related_file ("dif" )) write_dif      (s,first);
		if (s = related_file ("text")) write_text     (s,first);
		}

	/* Close the error file */

	if (out) {
		fclose (out);
		out = NULL;
		}

	/*------------------------------------------------------------------*\
	 | Free buffers used for text.
	\*------------------------------------------------------------------*/

	if (text_buffer) {
		free (text_buffer);
		text_buffer = NULL;
		}

	exit (0);
	}

/*----------------------------------------------------------------------*\
 | Code to keep track of related files (input and output) in a simple	|
 | database keyed to the function of the files.							|
 | These are used to build hypertext links among the different formats	|
 | that are created as output.											|
\*----------------------------------------------------------------------*/

struct file_index {
	char type[16];
	char name[1024];
	};

static struct file_index file_index[16];
static int file_index_length = 0;

static void add_related_file (char *type, char *name) {
	if (file_index_length < 16) {
		char cwd[1024];
		if (getcwd (cwd, 1024)) {
			char *s = cwd + strlen (cwd) - 1;
			if (*s != '/') {
				*s++ = '/';
				*s = 0;
				}
			if (memcmp (name,cwd,strlen(cwd)) == 0)
				name += strlen (cwd);
			}

		if ((strlen (type) < 16) &&
			(strlen (name) < 1024)) {
			strcpy (file_index[file_index_length].type,type);
			strcpy (file_index[file_index_length].name,name);
			file_index_length++;
			}
		}
	}

char *related_file (char *type) {
	int i;
	for (i=0; i < file_index_length; i++)
		if (stricmp (file_index[i].type,type) == 0)
			return (file_index[i].name);
	return (NULL);
	}

void dump_related_files (FILE *fp) {
	int i;
	for (i=0; i < file_index_length; i++) {
		fprintf (fp,"%d:\n",i);
		fprintf (fp,"  Type: %s\n",file_index[i].type);
		fprintf (fp,"  Name: %s\n",file_index[i].name);
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
