/*----------------------------------------------------------------------*\
 | mq: exposes mp's capabilities to Tcl/Tk.								|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <tcl.h>

#ifdef USE_TCL_STUBS
#include "mq.h"
#else
#include "mqinit.h"
#endif

#include "revision.h"
#include "keyword.h"
#include "item.h"
#include "config.h"
#include "stricmp.h"
#include "actions.h"

#define PATH_SEPARATOR  '/'

extern void read_local (char *local_file);
extern struct item *parse_text (char *file_name, char **buffer);
extern struct item *parse_sgml (char *file_name);
extern struct item *parse_xml (char *file_name);
extern struct item *parse_string (char *string);
extern struct item *parse_xml_string (char *string);
extern int check_syntax (struct item *p);
extern char *ps8_tag_of (enum fgdc_keyword key);
extern char *number_of (enum fgdc_keyword key);
extern enum fgdc_keyword key_of_ps8_tag (char *tag);

/*----------------------------------------------------------------------*\
 | Public data															|
\*----------------------------------------------------------------------*/

FILE *out = NULL;

/*----------------------------------------------------------------------*\
 | Static data															|
\*----------------------------------------------------------------------*/

static struct item *root = NULL;
static char cmdline_file [FILENAME_MAX] = "";
static char name_of_file [FILENAME_MAX] = "";
static int node_count = 0;
static int changed = 0;
static char *empty = "";
static char *failure = "0";
static char *success = "1";
static int record_count = 0;

static char message[512];

/*----------------------------------------------------------------------*\
 | Handle character encoding
\*----------------------------------------------------------------------*/

static char character_encoding[256] = "ISO-8859-1";

void set_character_encoding (char *new_value) {
	if (strlen (new_value) < 256)
		strcpy (character_encoding,new_value);
	else {
		memcpy (character_encoding,new_value,255);
		character_encoding[255] = 0;
		}
	}

char *get_character_encoding (void) {
	return (character_encoding);
	}

/*----------------------------------------------------------------------*\
 | Here the original item structure created by the text parser is fixed |
 | so that the indent member holds the level of nesting, not the number |
 | of whitespaces that preceded the element name in the input file.		|
\*----------------------------------------------------------------------*/

static void adjust_indent (struct item *p, int n) {
	struct item *q;

	if (p) {
		if (p->key != Wblank) p->indent = n;
		for (q=p->child; q; q=q->next) adjust_indent (q,n+1);
		}
	}

/*----------------------------------------------------------------------*\
 | config_cmd
 |
 | config read config_file
 | config find_first option
 | config find_next address ?option?
 | config find_in address option
 | config value_of address
 | config
\*----------------------------------------------------------------------*/

int config_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *cmd;
	char *config_file;
	char *element;
	char *address;
	char *option;
	void *opt;
	void *current;
	char *e = NULL;
	char *value;
	char *s;
	char *language = NULL;
	char *profile_name = NULL;

	if (argc > 1) {
		cmd = argv[1];

		/*--------------------------------------------------------------*\
		 | read config_file
		\*--------------------------------------------------------------*/

		if (stricmp (cmd,"read") == 0) {
			if (argc > 2) {
				config_file = argv[2];

				if (config_file) {
					void *v;
					read_configuration (config_file);

					/*--------------------------------------------------*\
					 | Use language requested in config file if possible.
					\*--------------------------------------------------*/

					if (opt = find_option (NULL,"input"))
						if (opt = find_option (opt,"language"))
							if (s = text_of_option (opt))
								language = s;

					/*--------------------------------------------------*\
					 | Use profile requested in config file if possible.
					\*--------------------------------------------------*/

					if (opt = find_option (NULL,"input"))
						if (opt = find_option (opt,"profile"))
							if (s = text_of_option (opt))
								profile_name = s;

					use_element_names (language,profile_name);		/* in keyword.c */
					if (profile_name) use_profile (profile_name);	/* in actions.c */

					/*--------------------------------------------------*\
					 | Read local extensions
					\*--------------------------------------------------*/

					if (v = find_option (NULL,"input")) {
						if (opt = find_option (v,"extensions")) {
							char *ext_file;
							if (ext_file = text_of_option (opt)) read_local (ext_file);
							while (opt = find_next_option (opt,"extensions"))
								if (ext_file = text_of_option (opt)) read_local (ext_file);
							}
						}
					}

				Tcl_SetResult (interp,success,TCL_STATIC);
				return (TCL_OK);
				}
			else {
				Tcl_SetResult (interp,"config read requires a file name",TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | find_first option
		\*--------------------------------------------------------------*/

		if (stricmp (cmd,"find_first") == 0) {
			if (argc > 2) {
				option = argv[2];

				opt = find_option (NULL,option);
				sprintf (message,"0x%X",opt);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_OK);
				}
			else {
				sprintf (message,"config %s requires an option name",cmd);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | find_next address ?option?
		\*--------------------------------------------------------------*/

		if (stricmp (cmd,"find_next") == 0) {
			if (argc > 2) {
				address = argv[2];

				current = (void *) strtoul (address,&e,0);
				if (e != (char *) current) {
					if (in_config (current)) {
						if (argc > 3) option = argv[3];
						else option = name_of_option (current);

						opt = find_next_option (current,option);

						sprintf (message,"0x%X",opt);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_OK);
						}
					else {
						sprintf (message,"config %s: address %s does not match a current config node",cmd,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					sprintf (message,"config %s: argument %s cannot be a node address",cmd,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				sprintf (message,"config %s requires aconfig node address",cmd);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | find_in address option
		\*--------------------------------------------------------------*/

		if (stricmp (cmd,"find_in") == 0) {
			if (argc > 2) {
				address = argv[2];

				current = (void *) strtoul (address,&e,0);
				if (e != (char *) current) {
					if (in_config (current)) {
						if (argc > 3) {
							option = argv[3];

							opt = find_child_option (current,option);

							sprintf (message,"0x%X",opt);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_OK);
							}
						else {
							sprintf (message,"config %s requires target option name to seek ",cmd);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}
						}
					else {
						sprintf (message,"config %s: address %s does not match a current config node",cmd,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					sprintf (message,"config %s: argument %s cannot be a node address",cmd,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				sprintf (message,"config %s requires aconfig node address",cmd);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | value_of address
		\*--------------------------------------------------------------*/

		if (stricmp (cmd,"value_of") == 0) {
			if (argc > 2) {
				address = argv[2];

				current = (void *) strtoul (address,&e,0);
				if (e != (char *) current) {
					if (in_config (current)) {
						value = text_of_option (current);

						Tcl_SetResult (interp,value,TCL_STATIC);
						return (TCL_OK);
						}
					else {
						sprintf (message,"config %s: address %s does not match a current config node",cmd,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					sprintf (message,"config %s: argument %s cannot be a node address",cmd,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				sprintf (message,"config %s requires aconfig node address",cmd);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}
		}
	else {

		/*--------------------------------------------------------------*\
		 | With no arguments, return 1 if config file has been read, 0	|
		 | if no config information is available.  Test this by looking	|
		 | for "input" and "output"; if either one is present, assume	|
		 | that a config tree is present.								|
		\*--------------------------------------------------------------*/

		if (!(opt = find_option (NULL,"input"))) opt = find_option (NULL,"output");
		Tcl_SetResult (interp,(opt ? success : failure),TCL_STATIC);
		return (TCL_OK);
		}
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | Function to decide whether a given address matches the address of a	|
 | node in the tree.													|
\*----------------------------------------------------------------------*/

static int in_tree (struct item *p, struct item *target) {
	struct item *q;
	int found = 0;

	if (p)
		if (p == target) found = 1;
		else
			for (q=p->child; q; q=q->next)
				if (in_tree (q,target)) {
					found = 1;
					break;
					}
	return (found);
	}

/*----------------------------------------------------------------------*\
 | Function to get the full text of an element and store it in a buffer	|
 | allocated through Tcl												|
\*----------------------------------------------------------------------*/

static char *full_text (struct item *p, int nonewline) {
	char *s;
	struct item *q;
	char *buffer;

	if (!p       ) return (NULL);
	if (!p->child) return (NULL);

	for (q=p->child; q; q=q->next)
		if (q->key != Wblank) break;

	if (q->key == Wunknown) {
		int n = 0;
		for (q=p->child; q; q=q->next)
			if (q->d) n += 1 + strlen (q->d);

		if (buffer = (char *) Tcl_Alloc (1+n)) {
			s = buffer;
			for (q=p->child; q; q=q->next)
				if (q->d) {
					strcpy (s,q->d);
					s += strlen (q->d);
					if (q->next)
						if (nonewline) *s++ = ' ';
						else *s++ = '\n';
					}
			*s = 0;
			return (buffer);
			}
		else
			return (NULL);
		}
	else
		return (NULL);
	}

/*----------------------------------------------------------------------*\
 | count_item returns the number of nodes in the tree.					|
\*----------------------------------------------------------------------*/

static int count_item (struct item *p) {
	struct item *q;
	int count = 0;

	if (p) {
		count = 1;
		for (q=p->child; q; q=q->next) count += count_item (q);
		}
	return (count);
	}

/*----------------------------------------------------------------------*\
 | delete_metadata
\*----------------------------------------------------------------------*/

static void delete_metadata (ClientData clientData) {
	root = (struct item *) clientData;
	if (root) {
		forget_buffer (root);
		delete_item (root);
		root = NULL;
		}
	}

/*----------------------------------------------------------------------*\
 | add_child_in_order
\*----------------------------------------------------------------------*/

static struct item *add_child_in_order (struct item *p, enum fgdc_keyword key) {
	int i,k;
	struct item *q = NULL;
	struct item *r;
	enum fgdc_keyword *e;

	/*------------------------------------------------------*\
	 | First look to see whether there are any children of	|
	 | the given node p.  If there are none, then simply	|
	 | add the new node as p's first child.					|
	 |														|
	 | If p has children, try to insert the new child in an	|
	 | appropriate place.  First, look to see whether		|
	 | there are any children with the same key.  If one or	|
	 | more of p's children have the same key as the		|
	 | child to be added, then insert the new child after	|
	 | the last child that has the same key.				|
	 |														|
	 | If none of p's children have the key that you		|
	 | want to add, then add the new child before the first	|
	 | existing child whose key falls after the target key	|
	 | in the list of possible child elements.				|
	\*------------------------------------------------------*/

	if (!p->child) q = add_child (p);
	else {
		/* Point r at the last child of current */

		for (r=p->child; r->next; r=r->next);

		/*--------------------------------------------------*\
		 | Walk backwards towards p->child, looking for a	|
		 | node whose key matches the element that we want	|
		 | to add.											|
		\*--------------------------------------------------*/

		while (r)
			if (r->key == key) break;
			else r = r->prev;

		/*--------------------------------------------------*\
		 | If r is not NULL, then r points to the last child|
		 | of current whose key matches the element we want	|
		 | to add; insert a new node after r.				|
		\*--------------------------------------------------*/

		if (r) {
			q = insert_item_after (r);
			q->indent = r->indent;
			}
		else {

			/*----------------------------------------------*\
			 | If r is NULL, then this will be the first	|
			 | child of current to have this key.			|
			\*----------------------------------------------*/

			if (e = element_list_of (p->key)) {

				/* What is the index of key in e? */

				for (k=0; e[k] != Wnull; k++)
					if (e[k] == key) break;

				/* If q is NULL, we haven't added the child yet */

				q = NULL;

				/* For each child of current */

				for (r = p->child; r; r=r->next)
					if (r->key != Wblank) {	/* skip blank lines */

						/* What is the index of r's key in e? */

						for (i=0; e[i] != Wnull; i++)
							if (e[i] == r->key) break;

						/*--------------------------------------*\
						 | If the key of r comes after k in e,	|
						 | insert the child before r and stop	|
						 | looking at the children of current.	|
						\*--------------------------------------*/

						if (i > k) {
							q = insert_item_before (r);
							q->indent = r->indent;
							break;
							}
						}

				/*------------------------------------------*\
				 | If you didn't find a child of current	|
				 | whose key follows the target key in the	|
				 | list of possible children of current,	|
				 | just add a child to current at the end.	|
				\*------------------------------------------*/

				if (!q) q = add_child (p);
				}
			else {

				/*------------------------------------------*\
				 | This should never happen.  All compound	|
				 | elements should have an element list.	|
				 |											|
				 | Something is very wrong, but we can just	|
				 | add the child anyway and hope it's okay.	|
				\*------------------------------------------*/

				q = add_child (p);
				}
			}
		}

	if (q) {
		q->key = key;
		q->d = text_of (key);
		q->original = 1;
		q->len = 1 + strlen (q->d);
		}

	return (q);
	}

/*----------------------------------------------------------------------*\
 | attach_subtree_in_order
\*----------------------------------------------------------------------*/

static void attach_subtree_in_order (struct item *p, struct item *u) {
	int i,k;
	struct item *q = NULL;
	struct item *r;
	enum fgdc_keyword key = u->key;
	enum fgdc_keyword *e;

	/*------------------------------------------------------*\
	 | First look to see whether there are any children of	|
	 | the given node p.  If there are none, then simply	|
	 | add the new node as p's first child.					|
	 |														|
	 | If p has children, try to insert the new child in an	|
	 | appropriate place.  First, look to see whether		|
	 | there are any children with the same key.  If one or	|
	 | more of p's children have the same key as the		|
	 | child to be added, then insert the new child after	|
	 | the last child that has the same key.				|
	 |														|
	 | If none of p's children have the key that you		|
	 | want to add, then add the new child before the first	|
	 | existing child whose key falls after the target key	|
	 | in the list of possible child elements.				|
	\*------------------------------------------------------*/

	if (!p->child) {
		p->child = u;
		u->parent = p;
		adjust_indent (u,1 + p->indent);
		}
	else {
		/* Point r at the last child of current */

		for (r=p->child; r->next; r=r->next);

		/*--------------------------------------------------*\
		 | Walk backwards towards p->child, looking for a	|
		 | node whose key matches the element that we want	|
		 | to add.											|
		\*--------------------------------------------------*/

		while (r)
			if (r->key == key) break;
			else r = r->prev;

		/*--------------------------------------------------*\
		 | If r is not NULL, then r points to the last child|
		 | of current whose key matches the element we want	|
		 | to add; insert a new node after r.				|
		\*--------------------------------------------------*/

		if (r) {
			u->next = r->next;
			u->prev = r;
			u->parent = p;
			if (r->next) r->next->prev = u;
			r->next = u;
			adjust_indent (u,r->indent);
			}
		else {

			/*----------------------------------------------*\
			 | If r is NULL, then this will be the first	|
			 | child of current to have this key.			|
			\*----------------------------------------------*/

			if (e = element_list_of (p->key)) {

				/* What is the index of key in e? */

				for (k=0; e[k] != Wnull; k++)
					if (e[k] == key) break;

				/* If q is NULL, we haven't added the child yet */

				q = NULL;

				/* For each child of current */

				for (r = p->child; r; r=r->next)
					if (r->key != Wblank) {	/* skip blank lines */

						/* What is the index of r's key in e? */

						for (i=0; e[i] != Wnull; i++)
							if (e[i] == r->key) break;

						/*--------------------------------------*\
						 | If the key of r comes after k in e,	|
						 | insert the child before r and stop	|
						 | looking at the children of current.	|
						\*--------------------------------------*/

						if (i > k) {
							u->next = r;
							u->prev = r->prev;
							if (r->prev) r->prev->next = u;
							r->prev = u;
							u->parent = r->parent;
							adjust_indent (u,r->indent);
							q = u;
							break;
							}
						}

				/*------------------------------------------*\
				 | If you didn't find a child of current	|
				 | whose key follows the target key in the	|
				 | list of possible children of current,	|
				 | just add a child to current at the end.	|
				\*------------------------------------------*/

				if (!q) {
					for (r=p->child; r->next; r=r->next);
					r->next = u;
					u->prev = r;
					u->parent = p;
					adjust_indent (u,r->indent);
					}
				}
			else {

				/*------------------------------------------*\
				 | This should never happen.  All compound	|
				 | elements should have an element list.	|
				 |											|
				 | Something is very wrong, but we can just	|
				 | add the child anyway and hope it's okay.	|
				\*------------------------------------------*/

				for (r=p->child; r->next; r=r->next);
				r->next = u;
				u->prev = r;
				u->parent = p;
				adjust_indent (u,r->indent);
				}
			}
		}
	}

/*----------------------------------------------------------------------*\
 | process_metadata_cmd
\*----------------------------------------------------------------------*/

int process_metadata_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i,k,n;
	enum fgdc_keyword key;
	struct item *p,*q,*r;
	char *b,*e,*s;
	char *address;
	char *element;
	char *value;

	root = (struct item *) clientData;

	if (argc > 1) {
		char *command = argv[1];

		/*--------------------------------------------------------------*\
		 | find_first <element>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"find_first") == 0)
			if (argc > 2) {
				element = argv[2];
				key = key_of (element);

				if (key == Wunknown || key == Wblank || key == Wnull) {
					sprintf (message,"%s: %s is not a valid element name",command,element);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				if (!root) {
					Tcl_SetResult (interp,"find_first: cannot find; no metadata has been read",TCL_STATIC);
					return (TCL_ERROR);
					}

				if (p = find_key (root,key)) {
					sprintf (message,"0x%X",p);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_OK);
					}
				else {
					Tcl_SetResult (interp,failure,TCL_STATIC);
					return (TCL_OK);
					}
				}
			else {
				Tcl_SetResult (interp,"find_first expects the name of an element to seek",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | find_in <address> <element>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"find_in") == 0)
			if (argc > 3) {

				/* Decode first argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {

					element = argv[3];
					key = key_of (element);

					if (key == Wunknown || key == Wblank || key == Wnull) {
						sprintf (message,"%s: %s is not a valid element name",command,element);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}

					if (q = find_key (p,key)) {
						sprintf (message,"0x%X",q);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_OK);
						}
					else {
						Tcl_SetResult (interp,failure,TCL_STATIC);
						return (TCL_OK);
						}
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"find_in expects an element address and the name of an element to seek",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | find_next <address> ?<element>?
		\*--------------------------------------------------------------*/

		if (strcmp (command,"find_next") == 0)
			if (argc > 2) {

				/* Decode first argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {

					if (argc > 3)
						element = argv[3];
					else
						element = text_of (p->key);

					key = key_of (element);

					if (key == Wunknown || key == Wblank || key == Wnull) {
						sprintf (message,"%s: %s is not a valid element name",command,element);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}

					for (q=p->next; q; q=q->next)
						if (q->key == key) break;

					if (q) {
						sprintf (message,"0x%X",q);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_OK);
						}
					else {
						Tcl_SetResult (interp,failure,TCL_STATIC);
						return (TCL_OK);
						}
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"find_in expects an element address and the name of an element to seek",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | insert ?-before | -after | -child? <address> ?<element>?
		\*--------------------------------------------------------------*/

		if (strcmp (command,"insert") == 0)
			if (argc > 2) {
				struct item *(*insert)(struct item *);
				char *placement = "-child";

				/* Decode placement option, if present */

				k = 2;
				if (strcmp (argv[k],"-before") == 0) {
					placement = argv[k];
					insert = insert_item_before;
					k++;
					}
				else
					if (strcmp (argv[k],"-after") == 0) {
						placement = argv[k];
						insert = insert_item_after;
						k++;
						}
					else
						if (strcmp (argv[k],"-child") == 0) {
							placement = argv[k];
							insert = NULL;
							k++;
							}
						else
							insert = NULL;

				if (argc > k) {

					/* Decode address argument; make sure it's a number */

					address = argv[k++];
					p = (struct item *) strtoul (address,&s,0);
					if (s == address) {
						sprintf (message,"%s: %s is not a valid element address",command,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}

					/* Check to see whether the address matches that of a node in the tree */

					if (in_tree (root,p)) {

						/* Decode element name argument, if present */

						if (argc > k)
							element = argv[k];
						else
							element = text_of (p->key);

						key = key_of (element);

						if (key == Wunknown || key == Wblank || key == Wnull) {
							sprintf (message,"%s: %s is not a valid element name",command,element);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}

						if (insert) {
							if (q = insert (p)) {
								q->key = key;
								q->d = text_of (key);
								q->original = 1;
								q->len = 1 + strlen (q->d);
								}
							}
						else q = add_child_in_order (p,key);

						if (q) {
							sprintf (message,"0x%X",q);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_OK);
							}
						else {
							sprintf (message,"%s: failed to insert %s %s %s at %s",command,element,placement,text_of(p->key),address);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}
						}
					else {
						sprintf (message,"%s: %s is not a valid element address",command,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					Tcl_SetResult (interp,"insert expects an element address and the name of an element to insert",TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"insert expects an element address and the name of an element to insert",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | attach ?-before | -after | -child? <address> <loose>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"attach") == 0)
			if (argc > 2) {
				struct item *(*insert)(struct item *);
				char *placement = "-child";
				char *loose;

				/* Decode placement option, if present */

				k = 2;
				if (strcmp (argv[k],"-before") == 0) {
					placement = argv[k];
					k++;
					}
				else
					if (strcmp (argv[k],"-after") == 0) {
						placement = argv[k];
						k++;
						}
					else
						if (strcmp (argv[k],"-child") == 0) {
							placement = argv[k];
							k++;
							}

				if (argc > k) {

					/* Decode address argument; make sure it's a number */

					address = argv[k++];
					p = (struct item *) strtoul (address,&s,0);
					if (s == address) {
						sprintf (message,"%s: %s is not a valid element address",command,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}

					if (argc > k) {

						/* Decode snippet address argument; make sure it's a number */

						loose = argv[k];
						q = (struct item *) strtoul (loose,&s,0);
						if (s == loose || q == NULL) {
							sprintf (message,"%s: %s is not a valid element address",command,loose);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}

						/* Check to see whether the address matches that of a node in the tree */

						if (in_tree (root,p)) {

							/* The loose item cannot have prev, next, or parent links */

							if (q->prev || q->next || q->parent) {
								sprintf (message,"%s: %s is not a detached subtree",command,loose);
								Tcl_SetResult (interp,message,TCL_STATIC);
								return (TCL_ERROR);
								}

							/* The loose item cannot be unknown, blank, or null */

							if (q->key == Wunknown || q->key == Wblank || q->key == Wnull) {
								sprintf (message,"%s: %s is not a valid element",command,loose);
								Tcl_SetResult (interp,message,TCL_STATIC);
								return (TCL_ERROR);
								}

							/* Insert the loose item as child, after, or before p */

							if (stricmp (placement,"-child") == 0)
								attach_subtree_in_order (p,q);
							else
								if (stricmp (placement,"-after") == 0)
									if (p->next) {
										p->next->prev = q;
										q->next = p->next;
										q->prev = p;
										p->next = q;
										q->parent = p->parent;
										}
									else {
										p->next = q;
										q->prev = p;
										q->parent = p->parent;
										}
								else
									if (stricmp (placement,"-before") == 0)
										if (p->prev) {
											p->prev->next = q;
											q->prev = p->prev;
											q->next = p;
											p->prev = q;
											q->parent = p->parent;
											}
										else {
											p->prev = q;
											q->next = p;
											q->parent = p->parent;
											p->parent->child = q;
											}
							adjust_indent (q,1+p->indent);

							Tcl_ResetResult (interp);
							return (TCL_OK);
							}
						else {
							sprintf (message,"%s: %s is not a valid element address",command,address);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}
						}
					else {
						Tcl_SetResult (interp,"attach expects both an element address and the address of a loose element or subtree to attach",TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					Tcl_SetResult (interp,"attach expects both an element address and the address of a loose element or subtree to attach",TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"attach expects an element address and the address of a loose element or subtree to attach",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | detach ?<address>?
		\*--------------------------------------------------------------*/

		if (strcmp (command,"detach") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {
					detach_item (p);

					Tcl_SetResult (interp,success,TCL_STATIC);
					return (TCL_OK);
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"detach expects the address of an element to separate from the tree",TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | delete ?<address>?
		\*--------------------------------------------------------------*/

		if (strcmp (command,"delete") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {
					delete_item (p);

					Tcl_SetResult (interp,success,TCL_STATIC);
					return (TCL_OK);
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"delete expects the address of an element to remove",TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | copy ?<address>?
		\*--------------------------------------------------------------*/

		if (strcmp (command,"copy") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {
					if (q = copy_item (p)) {
						sprintf (message,"0x%X",q);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_OK);
						}
					else {
						sprintf (message,"%s: %s inexplicably could not be copied",command,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"copy expects the address of an element to copy",TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | paste ?-before | -after | -child? <address> <text>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"paste") == 0)
			if (argc > 2) {
				struct item *(*insert)(struct item *);
				char *placement = "-child";

				/* Decode placement option, if present */

				k = 2;
				placement = argv[k];
				k++;

				if (argc > k) {

					/* Decode address argument; make sure it's a number */

					address = argv[k++];
					p = (struct item *) strtoul (address,&s,0);
					if (s == address) {
						sprintf (message,"%s: %s is not a valid element address",command,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}

					/* Check to see whether the address matches that of a node in the tree */

					if (in_tree (root,p)) {

						/* The attachment point cannot be a Wunknown or Wblank */

						if (p->key == Wunknown || p->key == Wblank) {
							sprintf (message,"%s: %s is not a compound element",command,address);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}

						/* Decode source text */

						if (argc > k) {
							char *source;

							/* First make a copy of the text */

							n = 1 + strlen (argv[k]);
							if (source = (char *) malloc (n)) {
								strcpy (source,argv[k]);
								if (*source == '<')
									q = parse_xml_string (source);
								else
									q = parse_string (source);

								if (!q) {
									free (source);
									sprintf (message,"%s: unrecognizable text given to paste",command);
									Tcl_SetResult (interp,message,TCL_STATIC);
									return (TCL_ERROR);
									}

								if (check_syntax (q) != 0) {
									Tcl_SetResult (interp,"Error: Mixed content not allowed in elements.",TCL_STATIC);
									return (TCL_ERROR);
									}

								/* Attach the subtree */

								if (strcmp (placement,"-before") == 0) {
									if (p->prev) p->prev->next = q;
									q->prev = p->prev;
									q->next = p;
									p->prev = q;
									q->parent = p->parent;
									}
								else
									if (strcmp (placement,"-after") == 0) {
										if (p->next) p->next->prev = q;
										q->next = p->next;
										p->next = q;
										q->prev = p;
										q->parent = p->parent;
										}
									else
										attach_subtree_in_order (p,q);

								adjust_indent (q,1+p->indent);

								/* Don't free(source); you're still using it! */
								Tcl_SetResult (interp,success,TCL_STATIC);
								return (TCL_OK);
								}
							else {
								sprintf (message,"%s: Unable to allocate %d bytes to copy the text",command,n);
								Tcl_SetResult (interp,message,TCL_STATIC);
								return (TCL_ERROR);
								}
							}
						else {
							sprintf (message,"%s: text of subtree to paste is required",command);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}
						}
					else {
						sprintf (message,"%s: %s is not a valid element address",command,address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					Tcl_SetResult (interp,"insert expects an element address and the name of an element to insert",TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"insert expects an element address and the name of an element to insert",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | prune ?<address>?
		\*--------------------------------------------------------------*/

		if (strcmp (command,"prune") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {
					k = prune (p);
					sprintf (message,"%d",k);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_OK);
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"prune expects the address of a subtree",TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | value_of ?-nonewline | -list? <address>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"value_of") == 0)
			if (argc > 2) {
				int nonewline = 0;
				int do_list = 0;

				p = NULL;
				for (i=2; i < argc; i++)
					if (strcmp (argv[i],"-nonewline") == 0) nonewline = 1;
					else
						if (strcmp (argv[i],"-list") == 0) do_list = 1;
						else address = argv[i];

				if (nonewline && do_list) {
					sprintf (message,"%s: specify -nonewline or -list or neither, but not both",command);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Decode argument; make sure it's a number */

				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p))

					if (do_list) {
						Tcl_Obj *obj;
						Tcl_ResetResult (interp);

						/* Verify that this element contains only text */

						for (q=p->child; q; q=q->next)
							if (q->key != Wblank) break;
						if (q && q->key != Wunknown) {
							sprintf (message,"%s: %s (%s) is compound and cannot contain text",command,text_of(p->key),address);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}

						/* Store each line as an element of a list returned to the user */

						obj = Tcl_NewListObj (0,NULL);

						for (q=p->child; q; q=q->next)
							if (q->d)
								Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (q->d,strlen (q->d)));
							else
								Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (empty,0));

						Tcl_SetObjResult (interp,obj);
						return (TCL_OK);
						}
					else	/* Return the value as a single string */

						/* Get the text belonging to this element */

						if (s = full_text (p,nonewline)) {
							Tcl_SetResult (interp,s,TCL_DYNAMIC);
							return (TCL_OK);
							}
						else

							/* The element has no data.  Check to see if there is an error. */

							if (p->child) {
								for (q=p->child; q; q=q->next)
									if (q->key != Wblank) break;
								if (q)
									if (q->key != Wunknown) {
										sprintf (message,"%s: %s (%s) is compound and cannot contain text",command,text_of(p->key),address);
										Tcl_SetResult (interp,message,TCL_STATIC);
										return (TCL_ERROR);
										}
									else {
										sprintf (message,"%s: This can't happen (%s) %s:%d",command,address,__FILE__,__LINE__);
										Tcl_SetResult (interp,message,TCL_STATIC);
										return (TCL_ERROR);
										}
								else {	/* No error; p->child is empty */
									Tcl_SetResult (interp,empty,TCL_STATIC);
									return (TCL_OK);
									}
								}
							else {	/* No error; p->child is NULL */
								Tcl_SetResult (interp,empty,TCL_STATIC);
								return (TCL_OK);
								}

				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"value_of expects the address of an element",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | value_set <address> <text>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"value_set") == 0)
			if (argc > 2) {
				p = NULL;
				address = argv[2];

				/* Decode argument; make sure it's a number */

				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {

					/* Make sure this node is a data element that can be assigned a value */

					if (element_list_of (p->key) == NULL) {

						/* Some value must be specified, even if it's an empty string */

						if (argc > 3) {
							value = argv[3];

							delete_children (p);

							/* Rebuild the value one line at a time */

							b = value;
							while (*b) {
								e = b;
								while (*e && *e != '\n') e++;

								s = (char *) malloc (1 + e - b);
								memcpy (s,b,e - b);
								s [e-b] = 0;

								q = add_child (p);
								q->key = ((e > b) ? Wunknown : Wblank);
								q->d = s;
								q->original = 0;
								q->len = e - b + 1;
								b = e;
								if (*b) b++;
								}
							Tcl_SetResult (interp,success,TCL_STATIC);
							return (TCL_OK);
							}
						else {
							sprintf (message,"%s: %s (%s) a value must be specified",command,text_of(p->key),address);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}
						}
					else {
						sprintf (message,"%s: %s (%s) is compound and cannot contain text",command,text_of(p->key),address);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"value_set expects the address of an element",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | count ?<address>?
		\*--------------------------------------------------------------*/

		if (strcmp (command,"count") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {
					sprintf (message,"%d",count_item (p));
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_OK);
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				sprintf (message,"%d",count_item (root));
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_OK);
				}
			}

		/*--------------------------------------------------------------*\
		 | next <address>
		 | prev <address>
		 | parent <address>
		 | child <address>
		\*--------------------------------------------------------------*/

		if ((strcmp (command,"next") == 0) ||
			(strcmp (command,"prev") == 0) ||
			(strcmp (command,"parent") == 0) ||
			(strcmp (command,"child") == 0))
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Proceed only if p is not NULL */

				if (p != 0) {

					if (strcmp (command,"next")   == 0) q = p->next;
					if (strcmp (command,"prev")   == 0) q = p->prev;
					if (strcmp (command,"parent") == 0) q = p->parent;
					if (strcmp (command,"child")  == 0) q = p->child;

					sprintf (message,"0x%X",q);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_OK);
					}
				else {
					sprintf (message,"%s: element address is NULL",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				sprintf (message,"%s expects the address of an element",command);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | name <address>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"name") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* If p is NULL, return (null) */

				if (p == 0) strcpy (message,"(null)");
				else
					sprintf (message,"%s",text_of(p->key));

				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_OK);
				}
			else {
				sprintf (message,"%s requires address of an element ",command);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | is_compound <address>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"is_compound") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p)) {
					if (element_list_of (p->key))
						Tcl_SetResult (interp,success,TCL_STATIC);
					else
						Tcl_SetResult (interp,failure,TCL_STATIC);

					return (TCL_OK);
					}
				else {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				sprintf (message,"%s requires address of an element ",command);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | contains <address>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"contains") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/* Check to see whether the address matches that of a node in the tree */

				if (in_tree (root,p))
					Tcl_SetResult (interp,success,TCL_STATIC);
				else
					Tcl_SetResult (interp,failure,TCL_STATIC);
				return (TCL_OK);
				}
			else {
				sprintf (message,"%s requires address of an element",command);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | line <address>
		\*--------------------------------------------------------------*/

		if (strcmp (command,"line") == 0) {
			if (argc > 2) {

				/* Decode argument; make sure it's a number */

				address = argv[2];
				p = (struct item *) strtoul (address,&s,0);
				if (s == address) {
					sprintf (message,"%s: %s is not a valid element address",command,address);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				/*------------------------------------------------------*\
				 | Previously, I checked to see whether the address		|
				 | matched that of a node in the tree.  But since you	|
				 | can detach a subtree and use it later as a snippet,	|
				 | there might be good uses for the line number even if	|
				 | the node is no longer part of the parse tree.  It's	|
				 | a form of historical metadata for the node.			|
				 |														|
				 | So generate an error only if p is NULL.				|
				\*------------------------------------------------------*/

				if (p) {
					sprintf (message,"%d",p->line_number);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_OK);
					}
				else {
					sprintf (message,"%s: cannot get the line number of a null element",command);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				sprintf (message,"%s requires address of an element",command);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | write ?-format <format>? filespec
		\*--------------------------------------------------------------*/

		if (strcmp (command,"write") == 0) {
			char *output_file;
			if (argc > 2) {
				char *format = "text";
				if (strcmp (argv[2],"-format") == 0) {
					if (argc > 3) {
						format = argv[3];
						if (argc > 4)
							output_file = argv[4];
						else {
							sprintf (message,"%s requires output file name",command);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}
						}
					else {
						sprintf (message,"%s -format option used but no format given",command);
						Tcl_SetResult (interp,message,TCL_STATIC);
						return (TCL_ERROR);
						}
					}
				else {
					output_file = argv[2];

					/*--------------------------------------------------*\
					 | No -format specified.  If the output file ends	|
					 | with .xml or .sgml or .sgm, write xml or sgml,	|
					 | otherwise write indented text.					|
					\*--------------------------------------------------*/

					if (s = strrchr (output_file,'.'))
						if (stricmp (s,".xml") == 0) format = "xml";
						else
							if ((stricmp (s,".sgml") == 0) ||
								(stricmp (s,".sgm" ) == 0)) format = "sgml";
							else
								format = "text";
					}

				/* Use the format and the file name to write the metadata */

				if (stricmp (format,"text") == 0)
					write_text (output_file,root);
				else
					if (stricmp (format,"sgml") == 0)
						write_sgml (output_file,root);
					else
						if (stricmp (format,"xml") == 0)
							write_xml (output_file,root);
						else {
							sprintf (message,"%s: format %s not supported",command,format);
							Tcl_SetResult (interp,message,TCL_STATIC);
							return (TCL_ERROR);
							}

				Tcl_SetResult (interp,success,TCL_STATIC);
				return (TCL_OK);
				}
			else {
				sprintf (message,"%s requires output file name",command);
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_ERROR);
				}
			}

		/*--------------------------------------------------------------*\
		 | forget
		\*--------------------------------------------------------------*/

		if (strcmp (command,"forget") == 0) {
			if (root) {
				forget_buffer (root);
				delete_item (root);
				Tcl_DeleteCommand (interp,argv[0]);
				Tcl_SetResult (interp,success,TCL_STATIC);
				}
			else
				Tcl_SetResult (interp,failure,TCL_STATIC);
			return (TCL_OK);
			}

		}
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | metadata id -parse input_file
\*----------------------------------------------------------------------*/

int metadata_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *variable_name;
	char *input_file;
	char id [64];
	char *s;
	char *text_buffer = NULL;
	int i;

	Tcl_ResetResult (interp);

	if (argc > 1) {
		variable_name = argv[1];
		if (argc > 1) {

			/*----------------------------------------------------------*\
			 | -parse input_file
			\*----------------------------------------------------------*/

			if (strcmp (argv[2],"-parse") == 0) {
				if (argc > 2) {
					input_file = argv[3];

					/*--------------------------------------------------*\
					 | Parse the input file.  File extension tells how.	|
					\*--------------------------------------------------*/

					root = NULL;
					if (input_file)
						if (s = strrchr (input_file,'.'))
							if (stricmp (s,".sgml") == 0 || stricmp (s,".sgm") == 0)
								root = parse_sgml (input_file);
							else
								if (stricmp (s,".xml") == 0)
									root = parse_xml (input_file);

					if (!root) root = parse_text (input_file, &text_buffer);

					/*--------------------------------------------------*/

					if (root) {
						strcpy (name_of_file,input_file);

						/*--------------------------------------------------------------*\
						 | The parser classifies all nodes by their keys.  If a valid	|
						 | element name appears in a textual value (at the beginning of	|
						 | a line), the parser classifies that node according to the	|
						 | element name.  But the node should be treated as text.  To	|
						 | correct this problem requires an understanding of the syntax	|
						 | of the standard.  Consequently it is necessary to minimally	|
						 | check the syntax to catch and recode elements that are found	|
						 | in textual values.											|
						\*--------------------------------------------------------------*/

						if (check_syntax (root) != 0) {
							Tcl_SetResult (interp,"Error: Mixed content is not permitted in elements.  Structural information will be lost if you continue.",TCL_STATIC);
							forget_buffer (root);
							delete_item (root);
							root = NULL;
							return (TCL_ERROR);
							}

						/*--------------------------------------------------------------*\
						 | If the top node isn't Metadata, then you won't be able to	|
						 | add any of the immediate children of Metadata to the record.	|
						 | To solve this problem, create a Metadata node at the top if	|
						 | one is not present.											|
						 |																|
						 | More ecumenically, we should allow people to nest Metadata   |
						 | in their extensions.  So we only plant a Metadata node at	|
						 | the root if there isn't a Metadata node somewhere in the		|
						 | tree; it could be buried several levels down.				|
						\*--------------------------------------------------------------*/

						if (find_key (root,WMetadata) == NULL) {
							struct item *p,*q;

							p = allocate_item (1);
							p->key = WMetadata;
							p->d = text_of (p->key);
							p->original = 1;
							p->len = strlen (p->d);
							p->indent = 0;
							if (root->indent == p->indent)
								for (q=root; q; q=q->next) indent (q);
							p->child = root;
							root->parent = p;
							root = p;
							for (q=root->child; q; q=q->next) q->parent = root;
							}
						adjust_indent (root,0);

						remember_buffer (root,text_buffer);

						/*----------------------------------------------*\
						 | If successful parse, create an identifier,	|
						 | set it as the the Tcl variable's value, and	|
						 | create a new Tcl command to facilitate		|
						 | processing of this metadata record.			|
						\*----------------------------------------------*/

						sprintf (id,"metadata.%04X",record_count++);
						Tcl_SetVar (interp,variable_name,id,0);
						Tcl_CreateCommand (interp,id,(Tcl_CmdProc *)process_metadata_cmd,(ClientData)root, (Tcl_CmdDeleteProc *)delete_metadata);
						Tcl_SetResult (interp,success,TCL_STATIC);
						}
					else
						Tcl_SetResult (interp,failure,TCL_STATIC);

					return (TCL_OK);
					}
				else {
					sprintf (message,"Error: no input file name given");
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}

			/*----------------------------------------------------------*\
			 | -create ?name_of_file?
			\*----------------------------------------------------------*/

			if (strcmp (argv[2],"-create") == 0) {
				struct item *p,*q;

				p = allocate_item (1);
				p->key = WMetadata;
				p->d = text_of (p->key);
				p->original = 1;
				p->len = strlen (p->d);
				p->indent = 0;
				root = p;

				/*----------------------------------------------*\
				 | If successful parse, create an identifier,	|
				 | set it as the the Tcl variable's value, and	|
				 | create a new Tcl command to facilitate		|
				 | processing of this metadata record.			|
				\*----------------------------------------------*/

				sprintf (id,"metadata.%04X",record_count++);
				Tcl_SetVar (interp,variable_name,id,0);
				Tcl_CreateCommand (interp,id,(Tcl_CmdProc *)process_metadata_cmd,(ClientData)root, (Tcl_CmdDeleteProc *)delete_metadata);
				Tcl_SetResult (interp,success,TCL_STATIC);
				return (TCL_OK);
				}

			}
		else {
			sprintf (message,"Error: got variable name %s, no -parse",argv[1]);
			Tcl_SetResult (interp,message,TCL_STATIC);
			return (TCL_ERROR);
			}
		}
	else {
		Tcl_SetResult (interp,"metadata expects variable name",TCL_STATIC);
		return (TCL_ERROR);
		}
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | element provides information about metadata elements
 |
 | element <name or tag>
 |   returns 1 if recognized name, 2 if recognized tag, 0 if not recognized
 | element <name or tag> -name
 |   returns long name
 | element <name or tag> -tag
 |   returns XML tag
 | element <name or tag> -source
 |   returns standard, profile, or extension
 | element <name or tag> -type
 |   returns compound or data
 | element <name or tag> -children
 |   returns a list of permitted child elements by name
 | element <name or tag> -childtags
 |   returns a list of permitted child elements by tag
\*----------------------------------------------------------------------*/

int element_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *element;
	char *option;
	char *tag;
	char *name;
	int recognized;
	enum fgdc_keyword key;

	if (argc > 1) {
		element = argv[1];

		recognized = 0;
		key = key_of (element);
		if (key == Wunknown) {
			key = key_of_ps8_tag (element);
			if (key != Wunknown) recognized = 2;
			}
		else recognized = 1;

		if (argc > 2) {
			option = argv[2];

			if (strcmp (option,"-name") == 0) {
				name = text_of (key);
				Tcl_SetResult (interp,name,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-number") == 0) {
				tag = number_of (key);
				Tcl_SetResult (interp,tag,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-tag") == 0) {
				tag = ps8_tag_of (key);
				Tcl_SetResult (interp,tag,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-source") == 0) {
				if (key < 500)
					strcpy (message,"standard");
				else
					if (key < 700)
						strcpy (message,"biological data profile");
					else
						if (key < 800)
							strcpy (message,"shoreline profile");
						else
							if (key < 1500)
								strcpy (message,"remote-sensing profile");
							else
								strcpy (message,"extension");
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-section") == 0) {
				strcpy (message,number_of(key));
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-type") == 0) {
				if (element_list_of (key))
					strcpy (message,"compound");
				else
					strcpy (message,"data");
				Tcl_SetResult (interp,message,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-parents") == 0) {
				enum fgdc_keyword *e;

				if (e = parents_of (key)) {
					Tcl_Obj *obj = Tcl_NewListObj (0,NULL);
					while (*e != Wnull) {
						name = text_of (*e);
						Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (name,strlen (name)));
						e++;
						}
					Tcl_SetObjResult (interp,obj);
					}
				else
					Tcl_SetResult (interp,empty,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-parenttags") == 0) {
				enum fgdc_keyword *e;

				if (e = parents_of (key)) {
					Tcl_Obj *obj = Tcl_NewListObj (0,NULL);
					while (*e != Wnull) {
						tag = ps8_tag_of (*e);
						Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (tag,strlen (tag)));
						e++;
						}
					Tcl_SetObjResult (interp,obj);
					}
				else
					Tcl_SetResult (interp,empty,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-children") == 0) {
				enum fgdc_keyword *e;

				if (e = element_list_of (key)) {
					Tcl_Obj *obj = Tcl_NewListObj (0,NULL);
					while (*e != Wnull) {
						name = text_of (*e);
						Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (name,strlen (name)));
						e++;
						}
					Tcl_SetObjResult (interp,obj);
					}
				else
					Tcl_SetResult (interp,empty,TCL_STATIC);
				return (TCL_OK);
				}

			if (strcmp (option,"-childtags") == 0) {
				enum fgdc_keyword *e;

				if (e = element_list_of (key)) {
					Tcl_Obj *obj = Tcl_NewListObj (0,NULL);
					while (*e != Wnull) {
						tag = ps8_tag_of (*e);
						Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (tag,strlen (tag)));
						e++;
						}
					Tcl_SetObjResult (interp,obj);
					}
				else
					Tcl_SetResult (interp,empty,TCL_STATIC);
				return (TCL_OK);
				}
			}
		else {
			if (key == Wunknown)
				Tcl_SetResult (interp,"0",TCL_STATIC);
			else
				Tcl_SetResult (interp,"1",TCL_STATIC);
			return (TCL_OK);
			}
		}
	else {
		Tcl_SetResult (interp,"element expects an element name",TCL_STATIC);
		return (TCL_ERROR);
		}
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | version_cmd returns the version of mq
\*----------------------------------------------------------------------*/

int version_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	Tcl_SetResult (interp,revision.mq,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | Call Tcl_CreateCommand for application-specific commands, if they	|
 | weren't already created by the init procedures called above.			|
\*----------------------------------------------------------------------*/

int Mq_Init (Tcl_Interp *interp) {
	if (Tcl_InitStubs (interp,"8.1",0) == NULL)
		return (TCL_ERROR);

	if (Tcl_PkgRequire (interp,"Tcl","8.0",0) == NULL)
		return (TCL_ERROR);

	if (Tcl_PkgProvide (interp,"mq",revision.mq) != TCL_OK)
		return (TCL_ERROR);

	Tcl_CreateCommand (interp,"config",   (Tcl_CmdProc *)config_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"element",  (Tcl_CmdProc *)element_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"metadata", (Tcl_CmdProc *)metadata_cmd, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"version",  (Tcl_CmdProc *)version_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	out = stderr;

	return (TCL_OK);
	}

int Et_AppInit (Tcl_Interp *interp) {
	Tcl_CreateCommand (interp,"config",   (Tcl_CmdProc *)config_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"element",  (Tcl_CmdProc *)element_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"metadata", (Tcl_CmdProc *)metadata_cmd, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"version",  (Tcl_CmdProc *)version_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

	out = stderr;

	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
