/*----------------------------------------------------------------------*\
 | Data and functions to retrieve the section number corresponding to a	|
 | keyword of the Content Standards for Digital Geospatial Metadata.	|
 |																		|
 | Revised 2004-04-12 to reflect the 1998 revision and profiles			|
 | Note that section numbers now do not all refer to the same document	|
 | are not exactly the same as the 1994 standard's numbers, and one		|
 | element (Online_Linkage) now has two different numbers.				|
 | The biological and remote-sensing profiles do not have numbers.		|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "keyword.h"

struct section_number {
	char *text;
	enum fgdc_keyword key;
	};

static struct section_number number_table[] = {
	{"0",				WMetadata},
	{"1",				WIdentification_Information},
	{"1.1",				WCitation},
	{"1.2",				WDescription},
	{"1.2.1",			WAbstract},
	{"1.2.2",			WPurpose},
	{"1.2.3",			WSupplemental_Information},
	{"1.3",				WTime_Period_of_Content},
	{"1.3.1",			WCurrentness_Reference},
	{"1.4",				WStatus},
	{"1.4.1",			WProgress},
	{"1.4.2",			WMaintenance_and_Update_Frequency},
	{"1.5",				WSpatial_Domain},
	{"1.5.1",			WBounding_Coordinates},
	{"1.5.1.1",			WWest_Bounding_Coordinate},
	{"1.5.1.2",			WEast_Bounding_Coordinate},
	{"1.5.1.3",			WNorth_Bounding_Coordinate},
	{"1.5.1.4",			WSouth_Bounding_Coordinate},
	{"1.5.2",			WData_Set_G_Polygon},
	{"1.5.2.1",			WData_Set_G_Polygon_Outer_G_Ring},
	{"1.5.2.1.1",		WG_Ring_Point},
	{"1.5.2.1.1.1",		WG_Ring_Latitude},
	{"1.5.2.1.1.2",		WG_Ring_Longitude},
	{"1.5.2.1.2",		WG_Ring},
	{"1.5.2.2",			WData_Set_G_Polygon_Exclusion_G_Ring},
	{"1.6",				WKeywords},
	{"1.6.1",			WTheme},
	{"1.6.1.1",			WTheme_Keyword_Thesaurus},
	{"1.6.1.2",			WTheme_Keyword},
	{"1.6.2",			WPlace},
	{"1.6.2.1",			WPlace_Keyword_Thesaurus},
	{"1.6.2.2",			WPlace_Keyword},
	{"1.6.3",			WStratum},
	{"1.6.3.1",			WStratum_Keyword_Thesaurus},
	{"1.6.3.2",			WStratum_Keyword},
	{"1.6.4",			WTemporal},
	{"1.6.4.1",			WTemporal_Keyword_Thesaurus},
	{"1.6.4.2",			WTemporal_Keyword},
	{"1.7",				WAccess_Constraints},
	{"1.8",				WUse_Constraints},
	{"1.9",				WPoint_of_Contact},
	{"1.10",			WBrowse_Graphic},
	{"1.10.1",			WBrowse_Graphic_File_Name},
	{"1.10.2",			WBrowse_Graphic_File_Description},
	{"1.10.3",			WBrowse_Graphic_File_Type},
	{"1.11",			WData_Set_Credit},
	{"1.12",			WSecurity_Information},
	{"1.12.1",			WSecurity_Classification_System},
	{"1.12.2",			WSecurity_Classification},
	{"1.12.3",			WSecurity_Handling_Description},
	{"1.13",			WNative_Data_Set_Environment},
	{"1.14",			WCross_Reference},
	{"2",				WData_Quality_Information},
	{"2.1",				WAttribute_Accuracy},
	{"2.1.1",			WAttribute_Accuracy_Report},
	{"2.1.2",			WQuantitative_Attribute_Accuracy_Assessment},
	{"2.1.2.1",			WAttribute_Accuracy_Value},
	{"2.1.2.2",			WAttribute_Accuracy_Explanation},
	{"2.2",				WLogical_Consistency_Report},
	{"2.3",				WCompleteness_Report},
	{"2.4",				WPositional_Accuracy},
	{"2.4.1",			WHorizontal_Positional_Accuracy},
	{"2.4.1.1",			WHorizontal_Positional_Accuracy_Report},
	{"2.4.1.2",			WQuantitative_Horizontal_Positional_Accuracy_Assessment},
	{"2.4.1.2.1",		WHorizontal_Positional_Accuracy_Value},
	{"2.4.1.2.2",		WHorizontal_Positional_Accuracy_Explanation},
	{"2.4.2",			WVertical_Positional_Accuracy},
	{"2.4.2.1",			WVertical_Positional_Accuracy_Report},
	{"2.4.2.2",			WQuantitative_Vertical_Positional_Accuracy_Assessment},
	{"2.4.2.2.1",		WVertical_Positional_Accuracy_Value},
	{"2.4.2.2.2",		WVertical_Positional_Accuracy_Explanation},
	{"2.5",				WLineage},
	{"2.5.1",			WSource_Information},
	{"2.5.1.1",			WSource_Citation},
	{"2.5.1.2",			WSource_Scale_Denominator},
	{"2.5.1.3",			WType_of_Source_Media},
	{"2.5.1.4",			WSource_Time_Period_of_Content},
	{"2.5.1.4.1",		WSource_Currentness_Reference},
	{"2.5.1.5",			WSource_Citation_Abbreviation},
	{"2.5.1.6",			WSource_Contribution},
	{"2.5.2",			WProcess_Step},
	{"2.5.2.1",			WProcess_Description},
	{"2.5.2.2",			WSource_Used_Citation_Abbreviation},
	{"2.5.2.3",			WProcess_Date},
	{"2.5.2.4",			WProcess_Time},
	{"2.5.2.5",			WSource_Produced_Citation_Abbreviation},
	{"2.5.2.6",			WProcess_Contact},
	{"2.6",				WCloud_Cover},
	{"3",				WSpatial_Data_Organization_Information},
	{"3.1",				WIndirect_Spatial_Reference},
	{"3.2",				WDirect_Spatial_Reference_Method},
	{"3.3",				WPoint_and_Vector_Object_Information},
	{"3.3.1",			WSDTS_Terms_Description},
	{"3.3.1.1",			WSDTS_Point_and_Vector_Object_Type},
	{"3.3.1.2",			WPoint_and_Vector_Object_Count},
	{"3.3.2",			WVPF_Terms_Description},
	{"3.3.2.1",			WVPF_Topology_Level},
	{"3.3.2.2",			WVPF_Point_and_Vector_Object_Information},
	{"3.3.2.2.1",		WVPF_Point_and_Vector_Object_Type},
	{"3.4",				WRaster_Object_Information},
	{"3.4.1",			WRaster_Object_Type},
	{"3.4.2",			WRow_Count},
	{"3.4.3",			WColumn_Count},
	{"3.4.4",			WVertical_Count},
	{"4",				WSpatial_Reference_Information},
	{"4.1",				WHorizontal_Coordinate_System_Definition},
	{"4.1.1",			WGeographic},
	{"4.1.1.1",			WLatitude_Resolution},
	{"4.1.1.2",			WLongitude_Resolution},
	{"4.1.1.3",			WGeographic_Coordinate_Units},
	{"4.1.2",			WPlanar},
	{"4.1.2.1",			WMap_Projection},
	{"4.1.2.1.1",		WMap_Projection_Name},
	{"4.1.2.1.2",		WAlbers_Conical_Equal_Area},
	{"4.1.2.1.3",		WAzimuthal_Equidistant},
	{"4.1.2.1.4",		WEquidistant_Conic},
	{"4.1.2.1.5",		WEquirectangular},
	{"4.1.2.1.6",		WGeneral_Vertical_Near_sided_Perspective},
	{"4.1.2.1.7",		WGnomonic},
	{"4.1.2.1.8",		WLambert_Azimuthal_Equal_Area},
	{"4.1.2.1.9",		WLambert_Conformal_Conic},
	{"4.1.2.1.10",		WMercator},
	{"4.1.2.1.11",		WModified_Stereographic_for_Alaska},
	{"4.1.2.1.12",		WMiller_Cylindrical},
	{"4.1.2.1.13",		WOblique_Mercator},
	{"4.1.2.1.14",		WOrthographic},
	{"4.1.2.1.15",		WPolar_Stereographic},
	{"4.1.2.1.16",		WPolyconic},
	{"4.1.2.1.17",		WRobinson},
	{"4.1.2.1.18",		WSinusoidal},
	{"4.1.2.1.19",		WSpace_Oblique_Mercator_Landsat},
	{"4.1.2.1.20",		WStereographic},
	{"4.1.2.1.21",		WTransverse_Mercator},
	{"4.1.2.1.22",		Wvan_der_Grinten},
	{"4.1.2.1.23",		WMap_Projection_Parameters},
	{"4.1.2.1.23.1",	WStandard_Parallel},
	{"4.1.2.1.23.2",	WLongitude_of_Central_Meridian},
	{"4.1.2.1.23.3",	WLatitude_of_Projection_Origin},
	{"4.1.2.1.23.4",	WFalse_Easting},
	{"4.1.2.1.23.5",	WFalse_Northing},
	{"4.1.2.1.23.6",	WScale_Factor_at_Equator},
	{"4.1.2.1.23.7",	WHeight_of_Perspective_Point_Above_Surface},
	{"4.1.2.1.23.8",	WLongitude_of_Projection_Center},
	{"4.1.2.1.23.9",	WLatitude_of_Projection_Center},
	{"4.1.2.1.23.10",	WScale_Factor_at_Center_Line},
	{"4.1.2.1.23.11",	WOblique_Line_Azimuth},
	{"4.1.2.1.23.11.1",	WAzimuthal_Angle},
	{"4.1.2.1.23.11.2",	WAzimuth_Measure_Point_Longitude},
	{"4.1.2.1.23.12",	WOblique_Line_Point},
	{"4.1.2.1.23.12.1",	WOblique_Line_Latitude},
	{"4.1.2.1.23.12.2",	WOblique_Line_Longitude},
	{"4.1.2.1.23.13",	WStraight_Vertical_Longitude_from_Pole},
	{"4.1.2.1.23.14",	WScale_Factor_at_Projection_Origin},
	{"4.1.2.1.23.15",	WLandsat_Number},
	{"4.1.2.1.23.16",	WPath_Number},
	{"4.1.2.1.23.17",	WScale_Factor_at_Central_Meridian},
	{"4.1.2.1.23.18",	WOther_Projections_Definition},
	{"4.1.2.2",			WGrid_Coordinate_System},
	{"4.1.2.2.1",		WGrid_Coordinate_System_Name},
	{"4.1.2.2.2",		WUniversal_Transverse_Mercator},
	{"4.1.2.2.2.1",		WUTM_Zone_Number},
	{"4.1.2.2.3",		WUniversal_Polar_Stereographic},
	{"4.1.2.2.3.1",		WUPS_Zone_Identifier},
	{"4.1.2.2.4",		WState_Plane_Coordinate_System},
	{"4.1.2.2.4.1",		WSPCS_Zone_Identifier},
	{"4.1.2.2.5",		WARC_Coordinate_System},
	{"4.1.2.2.5.1",		WARC_System_Zone_Identifier},
	{"4.1.2.2.6",		WOther_Grid_Systems_Definition},
	{"4.1.2.3",			WLocal_Planar},
	{"4.1.2.3.1",		WLocal_Planar_Description},
	{"4.1.2.3.2",		WLocal_Planar_Georeference_Information},
	{"4.1.2.4",			WPlanar_Coordinate_Information},
	{"4.1.2.4.1",		WPlanar_Coordinate_Encoding_Method},
	{"4.1.2.4.2",		WCoordinate_Representation},
	{"4.1.2.4.2.1",		WAbscissa_Resolution},
	{"4.1.2.4.2.2",		WOrdinate_Resolution},
	{"4.1.2.4.3",		WDistance_and_Bearing_Representation},
	{"4.1.2.4.3.1",		WDistance_Resolution},
	{"4.1.2.4.3.2",		WBearing_Resolution},
	{"4.1.2.4.3.3",		WBearing_Units},
	{"4.1.2.4.3.4",		WBearing_Reference_Direction},
	{"4.1.2.4.3.5",		WBearing_Reference_Meridian},
	{"4.1.2.4.4",		WPlanar_Distance_Units},
	{"4.1.3",			WLocal},
	{"4.1.3.1",			WLocal_Description},
	{"4.1.3.2",			WLocal_Georeference_Information},
	{"4.1.4",			WGeodetic_Model},
	{"4.1.4.1",			WHorizontal_Datum_Name},
	{"4.1.4.2",			WEllipsoid_Name},
	{"4.1.4.3",			WSemi_major_Axis},
	{"4.1.4.4",			WDenominator_of_Flattening_Ratio},
	{"4.2",				WVertical_Coordinate_System_Definition},
	{"4.2.1",			WAltitude_System_Definition},
	{"4.2.1.1",			WAltitude_Datum_Name},
	{"4.2.1.2",			WAltitude_Resolution},
	{"4.2.1.3",			WAltitude_Distance_Units},
	{"4.2.1.4",			WAltitude_Encoding_Method},
	{"4.2.2",			WDepth_System_Definition},
	{"4.2.2.1",			WDepth_Datum_Name},
	{"4.2.2.2",			WDepth_Resolution},
	{"4.2.2.3",			WDepth_Distance_Units},
	{"4.2.2.4",			WDepth_Encoding_Method},
	{"5",				WEntity_and_Attribute_Information},
	{"5.1",				WDetailed_Description},
	{"5.1.1",			WEntity_Type},
	{"5.1.1.1",			WEntity_Type_Label},
	{"5.1.1.2",			WEntity_Type_Definition},
	{"5.1.1.3",			WEntity_Type_Definition_Source},
	{"5.1.2",			WAttribute},
	{"5.1.2.1",			WAttribute_Label},
	{"5.1.2.2",			WAttribute_Definition},
	{"5.1.2.3",			WAttribute_Definition_Source},
	{"5.1.2.4",			WAttribute_Domain_Values},
	{"5.1.2.4.1",		WEnumerated_Domain},
	{"5.1.2.4.1.1",		WEnumerated_Domain_Value},
	{"5.1.2.4.1.2",		WEnumerated_Domain_Value_Definition},
	{"5.1.2.4.1.3",		WEnumerated_Domain_Value_Definition_Source},
	{"5.1.2.4.2",		WRange_Domain},
	{"5.1.2.4.2.1",		WRange_Domain_Minimum},
	{"5.1.2.4.2.2",		WRange_Domain_Maximum},
	{"5.1.2.4.3",		WCodeset_Domain},
	{"5.1.2.4.3.1",		WCodeset_Name},
	{"5.1.2.4.3.2",		WCodeset_Source},
	{"5.1.2.4.4",		WUnrepresentable_Domain},
	{"5.1.2.5",			WAttribute_Units_of_Measure},
	{"5.1.2.6",			WAttribute_Measurement_Resolution},
	{"5.1.2.7",			WBeginning_Date_of_Attribute_Values},
	{"5.1.2.8",			WEnding_Date_of_Attribute_Values},
	{"5.1.2.9",			WAttribute_Value_Accuracy_Information},
	{"5.1.2.9.1",		WAttribute_Value_Accuracy},
	{"5.1.2.9.2",		WAttribute_Value_Accuracy_Explanation},
	{"5.1.2.10",		WAttribute_Measurement_Frequency},
	{"5.2",				WOverview_Description},
	{"5.2.1",			WEntity_and_Attribute_Overview},
	{"5.2.2",			WEntity_and_Attribute_Detail_Citation},
	{"6",				WDistribution_Information},
	{"6.1",				WDistributor},
	{"6.2",				WResource_Description},
	{"6.3",				WDistribution_Liability},
	{"6.4",				WStandard_Order_Process},
	{"6.4.1",			WNon_digital_Form},
	{"6.4.2",			WDigital_Form},
	{"6.4.2.1",			WDigital_Transfer_Information},
	{"6.4.2.1.1",		WFormat_Name},
	{"6.4.2.1.2",		WFormat_Version_Number},
	{"6.4.2.1.3",		WFormat_Version_Date},
	{"6.4.2.1.4",		WFormat_Specification},
	{"6.4.2.1.5",		WFormat_Information_Content},
	{"6.4.2.1.6",		WFile_Decompression_Technique},
	{"6.4.2.1.7",		WTransfer_Size},
	{"6.4.2.2",			WDigital_Transfer_Option},
	{"6.4.2.2.1",		WOnline_Option},
	{"6.4.2.2.1.1",		WComputer_Contact_Information},
	{"6.4.2.2.1.1.1",	WNetwork_Address},
	{"6.4.2.2.1.1.1.1",	WNetwork_Resource_Name},
	{"6.4.2.2.1.1.2",	WDialup_Instructions},
	{"6.4.2.2.1.1.2.1",	WLowest_BPS},
	{"6.4.2.2.1.1.2.2",	WHighest_BPS},
	{"6.4.2.2.1.1.2.3",	WNumber_DataBits},
	{"6.4.2.2.1.1.2.4",	WNumber_StopBits},
	{"6.4.2.2.1.1.2.5",	WParity},
	{"6.4.2.2.1.1.2.6",	WCompression_Support},
	{"6.4.2.2.1.1.2.7",	WDialup_Telephone},
	{"6.4.2.2.1.1.2.8",	WDialup_File_Name},
	{"6.4.2.2.1.2",		WAccess_Instructions},
	{"6.4.2.2.1.3",		WOnline_Computer_and_Operating_System},
	{"6.4.2.2.2",		WOffline_Option},
	{"6.4.2.2.2.1",		WOffline_Media},
	{"6.4.2.2.2.2",		WRecording_Capacity},
	{"6.4.2.2.2.2.1",	WRecording_Density},
	{"6.4.2.2.2.2.2",	WRecording_Density_Units},
	{"6.4.2.2.2.3",		WRecording_Format},
	{"6.4.2.2.2.4",		WCompatibility_Information},
	{"6.4.3",			WFees},
	{"6.4.4",			WOrdering_Instructions},
	{"6.4.5",			WTurnaround},
	{"6.5",				WCustom_Order_Process},
	{"6.6",				WTechnical_Prerequisites},
	{"6.7",				WAvailable_Time_Period},
	{"7",				WMetadata_Reference_Information},
	{"7.1",				WMetadata_Date},
	{"7.2",				WMetadata_Review_Date},
	{"7.3",				WMetadata_Future_Review_Date},
	{"7.4",				WMetadata_Contact},
	{"7.5",				WMetadata_Standard_Name},
	{"7.6",				WMetadata_Standard_Version},
	{"7.7",				WMetadata_Time_Convention},
	{"7.8",				WMetadata_Access_Constraints},
	{"7.9",				WMetadata_Use_Constraints},
	{"7.10",			WMetadata_Security_Information},
	{"7.10.1",			WMetadata_Security_Classification_System},
	{"7.10.2",			WMetadata_Security_Classification},
	{"7.10.3",			WMetadata_Security_Handling_Description},
	{"7.11",			WMetadata_Extensions},
	{"7.11.2",			WProfile_Name},
	{"8",				WCitation_Information},
	{"8.1",				WOriginator},
	{"8.2",				WPublication_Date},
	{"8.3",				WPublication_Time},
	{"8.4",				WTitle},
	{"8.5",				WEdition},
	{"8.6",				WGeospatial_Data_Presentation_Form},
	{"8.7",				WSeries_Information},
	{"8.7.1",			WSeries_Name},
	{"8.7.2",			WIssue_Identification},
	{"8.8",				WPublication_Information},
	{"8.8.1",			WPublication_Place},
	{"8.8.2",			WPublisher},
	{"8.9",				WOther_Citation_Details},
	{"8.10",			WOnline_Linkage},	/* also 7.11.1 */
	{"8.11",			WLarger_Work_Citation},
	{"9",				WTime_Period_Information},
	{"9.1",				WSingle_Date_Time},
	{"9.1.1",			WCalendar_Date},
	{"9.1.2",			WTime_of_Day},
	{"9.2",				WMultiple_Dates_Times},
	{"9.3",				WRange_of_Dates_Times},
	{"9.3.1",			WBeginning_Date},
	{"9.3.2",			WBeginning_Time},
	{"9.3.3",			WEnding_Date},
	{"9.3.4",			WEnding_Time},
	{"10",				WContact_Information},
	{"10.1",			WContact_Person_Primary},
	{"10.1.1",			WContact_Person},
	{"10.1.2",			WContact_Organization},
	{"10.2",			WContact_Organization_Primary},
	{"10.3",			WContact_Position},
	{"10.4",			WContact_Address},
	{"10.4.1",			WAddress_Type},
	{"10.4.2",			WAddress},
	{"10.4.3",			WCity},
	{"10.4.4",			WState_or_Province},
	{"10.4.5",			WPostal_Code},
	{"10.4.6",			WCountry},
	{"10.5",			WContact_Voice_Telephone},
	{"10.6",			WContact_TDD_TTY_Telephone},
	{"10.7",			WContact_Facsimile_Telephone},
	{"10.8",			WContact_Electronic_Mail_Address},
	{"10.9",			WHours_of_Service},
	{"10.10",			WContact_Instructions},

	/* Biological data profile */

	{"1.5.3",			WDescription_of_Geographic_Extent},	/* numbered in the shoreline profile */
	{"",				WBounding_Altitudes},
	{"",				WAltitude_Minimum},
	{"",				WAltitude_Maximum},
	{"",				WTaxonomy},
	{"",				WKeywords_Taxon},
	{"",				WTaxonomic_Keyword_Thesaurus},
	{"",				WTaxonomic_Keywords},
	{"",				WTaxonomic_System},
	{"",				WClassification_System_Authority},
	{"",				WClassification_System_Citation},
	{"",				WClassification_System_Modifications},
	{"",				WIdentification_Reference},
	{"",				WIdentifier},
	{"",				WTaxonomic_Procedures},
	{"",				WTaxonomic_Completeness},
	{"",				WVouchers},
	{"",				WSpecimen},
	{"",				WRepository},
	{"",				WGeneral_Taxonomic_Coverage},
	{"",				WTaxonomic_Classification},
	{"",				WTaxon_Rank_Name},
	{"",				WTaxon_Rank_Value},
	{"",				WApplicable_Common_Name},
	{"",				WAnalytical_Tool},
	{"",				WAnalytical_Tool_Description},
	{"",				WTool_Access_Information},
	{"",				WTool_Contact},
	{"",				WTool_Citation},
	{"",				WTool_Access_Instructions},
	{"",				WTool_Computer_and_Operating_System},
	{"",				WMethodology},
	{"",				WMethodology_Type},
	{"",				WMethodology_Identifier},
	{"",				WMethodology_Keyword_Thesaurus},
	{"",				WMethodology_Keyword},
	{"",				WMethodology_Description},
	{"",				WMethodology_Citation},
	{"",				WASCII_File_Structure},
	{"",				WRecord_Delimiter},
	{"",				WNumber_Header_Lines},
	{"",				WDescription_of_Header_Content},
	{"",				WOrientation},
	{"",				WCase_Sensitive},
	{"",				WAuthentication},
	{"",				WQuote_Character},
	{"",				WData_Field},
	{"",				WData_Field_Name},
	{"",				WMissing_Value_Code},
	{"",				WData_Field_Width_Delimiter},
	{"",				WData_Field_Width},
	{"",				WBeginning_Geologic_Age},
	{"",				WEnding_Geologic_Age},
	{"",				WGeologic_Age},
	{"",				WGeologic_Time_Scale},
	{"",				WGeologic_Age_Estimate},
	{"",				WGeologic_Age_Uncertainty},
	{"",				WGeologic_Age_Explanation},
	{"",				WGeologic_Citation},

	/* Shoreline metadata profile */

	{"2.5.2.7",			WProcess_Step_Citation},
	{"2.7",				WTidal_Information},
	{"2.7.1",			WType_of_Tide},
	{"2.7.2",			WTime_of_Tide},
	{"2.7.2.1",			WTime_of_Low_Tide},
	{"2.7.2.2",			WTime_of_High_Tide},
	{"2.7.2.3",			WTidal_Datum},
	{"2.7.2.4",			WRange_of_Tide},
	{"2.7.3",			WTide_Table_Reference},
	{"2.7.4",			WSupplemental_Tidal_Information},
	{"2.8",				WMarine_Weather_Condition},
	{"2.8.1",			WWind_Speed},
	{"2.8.2",			WWind_Direction},
	{"2.8.3",			WWave_Height},
	{"2.8.4",			WBarometric_Pressure},
	{"2.9",				WEnvironmental_Event},

	/* Remote-sensing profile */

	{"",				WAcquisition_Information},
	{"",				WAerial_Film_Speed},
	{"",				WAerotriangulation_Reference},
	{"",				WAffine_Distortion_X_Prime_Coefficient},
	{"",				WAffine_Distortion_Y_Prime_Coefficient},
	{"",				WAggregation_Criteria},
	{"",				WAggregation_Information},
	{"",				WAggregation_Member_ID},
	{"",				WAlgorithm_Change_Description},
	{"",				WAlgorithm_Change_History},
	{"",				WAlgorithm_Change_Reference},
	{"",				WAlgorithm_Description},
	{"",				WAlgorithm_Identifiers},
	{"",				WAlgorithm_Information},
	{"",				WAlgorithm_Peer_Review_Description},
	{"",				WAlgorithm_Peer_Review_Information},
	{"",				WAlgorithm_Peer_Review_Reference},
	{"",				WAlgorithm_Reference},
	{"",				WAlgorithm_Text_Description},
	{"",				WAltitude_Distance_Layer},
	{"",				WAncillary_Dataset},
	{"",				WAncillary_Dataset_Description},
	{"",				WAncillary_Dataset_Identifier},
	{"",				WAncillary_Description},
	{"",				WAncillary_Reference},
	{"",				WAngle_Dependent_Distortion},
	{"",				WAngle_Distortion_Value},
	{"",				WArea_Weighted_Average_Resolution},
	{"",				WArgument_of_Perigee},
	{"",				WAscending_Descending_Indicator},
	{"",				WAttitude_Angular_Units},
	{"",				WAvailability_of_Element_Locations},
	{"",				WAxes},
	{"",				WAxis_Rotation_Convention},
	{"",				WBand_Boundary_Definition},
	{"",				WBand_Description},
	{"",				WBand_ID},
	{"",				WBand_Identification},
	{"",				WBand_Measurement_Mode_ID},
	{"",				WBand_Quality},
	{"",				WCalibrated_Detector_Positions},
	{"",				WCalibrated_Focal_Length},
	{"",				WCamera_Identifier},
	{"",				WCamera_Type},
	{"",				WCell_Value_Type},
	{"",				WClock_Time_Drift},
	{"",				WCollection_Type},
	{"",				WColumn_Delta_X},
	{"",				WColumn_Delta_Y},
	{"",				WCommand_Line_Processing_Parameter},
	{"",				WComponent_Information},
	{"",				WContainer_Packet_ID},
	{"",				WControl_Point_Authority},
	{"",				WControl_Point_Column},
	{"",				WControl_Point_Earth_Location},
	{"",				WControl_Point_Horizontal_X_Accuracy},
	{"",				WControl_Point_Horizontal_Y_Accuracy},
	{"",				WControl_Point_ID},
	{"",				WControl_Point_Identification},
	{"",				WControl_Point_Origin},
	{"",				WControl_Point_Raster_Position},
	{"",				WControl_Point_Row},
	{"",				WControl_Point_Type},
	{"",				WControl_Point_Use_Flag},
	{"",				WControl_Point_Vertical_Accuracy},
	{"",				WControl_Point_x_Value},
	{"",				WControl_Point_y_Value},
	{"",				WControl_Point_z_Value},
	{"",				WCoordinate_Point},
	{"",				WCoordinate_System},
	{"",				WCoordinate_x_Value},
	{"",				WCoordinate_XY_Units},
	{"",				WCoordinate_y_Value},
	{"",				WCoordinate_Z_Units},
	{"",				WCoordinate_z_Value},
	{"",				WCross_Track_Axis},
	{"",				WCross_Track_Description},
	{"",				WCross_Track_Direction},
	{"",				WCross_Track_Extent_Angle},
	{"",				WCross_Track_Fixed_Angle},
	{"",				WCross_Track_Motion},
	{"",				WCross_Track_Start_Angle},
	{"",				WCross_Track_Step_Angle},
	{"",				WCross_Track_Sweep},
	{"",				WCross_Track_Zero},
	{"",				WData_Dictionary_Reference},
	{"",				WData_Scaling_Information},
	{"",				WDataset_Identifier},
	{"",				WDate_of_Last_Calibration},
	{"",				WDeveloping_Institution},
	{"",				WDimension_Count},
	{"",				WDimension_Description},
	{"",				WDimension_Properties},
	{"",				WDistance_Dependent_Distortion},
	{"",				WDistance_Distortion_Value},
	{"",				WDistortion},
	{"",				WDistortion_Type_Affine},
	{"",				WDistortion_Type_Radial_Asymmetrical},
	{"",				WDistortion_Type_Radial_Symmetrical},
	{"",				WDocumentation},
	{"",				WEccentricity},
	{"",				WEffective_Aerial_Film_Speed},
	{"",				WElevation_Axis},
	{"",				WElevation_Description},
	{"",				WElevation_Direction},
	{"",				WElevation_Extent_Angle},
	{"",				WElevation_Fixed_Angle},
	{"",				WElevation_Start_Angle},
	{"",				WElevation_Step_Angle},
	{"",				WElevation_Sweep},
	{"",				WElevation_Zero},
	{"",				WEphemeris},
	{"",				WExposure_Time},
	{"",				WExterior_Orientation_Accuracy},
	{"",				WFiducial},
	{"",				WFiducial_Center},
	{"",				WFilm_Type},
	{"",				WFilter_on_Camera},
	{"",				WFilter_on_Camera_Indicator},
	{"",				WFilter_Type},
	{"",				WFlight_Protocol},
	{"",				WFlying_Height},
	{"",				WForward_Motion_Compensation},
	{"",				WFrame_Area},
	{"",				WFrame_Area_Units},
	{"",				WFrame_Area_Value},
	{"",				WFrame_Camera},
	{"",				WFrame_Geometric_Properties},
	{"",				WFrame_Hardware},
	{"",				WFrame_Operation},
	{"",				WFrame_Optics},
	{"",				WFrame_Radiometric_Calibration},
	{"",				WFrame_Radiometric_Properties},
	{"",				WFrame_Spectral_Information},
	{"",				WFrame_Spectral_Properties},
	{"",				WGeorectified_Raster},
	{"",				WGeoreferenceable_Raster},
	{"",				WGeoreferencing_Description},
	{"",				WGeoreferencing_Information},
	{"",				WGPS_Information_System_Availability},
	{"",				WGrid_First_Element},
	{"",				WGrid_First_Element_Map_X_Coordinate},
	{"",				WGrid_First_Element_Map_Y_Coordinate},
	{"",				WGrid_Orientation},
	{"",				WGround_Control_Point_Description},
	{"",				WGround_Control_Point_Information},
	{"",				WGround_Control_Point_Organization},
	{"",				WGround_Control_Point_Position},
	{"",				WGround_Shape},
	{"",				WID_Overlapping_Dataset},
	{"",				WIFOV_Units},
	{"",				WIFOV_x_Definition},
	{"",				WIFOV_x_Value},
	{"",				WIFOV_y_Definition},
	{"",				WIFOV_y_Value},
	{"",				WImage_Quality},
	{"",				WImage_Size},
	{"",				WImage_Size_x_Value},
	{"",				WImage_Size_y_Value},
	{"",				WInclination},
	{"",				WIndividual_Band_Identification},
	{"",				WInput_Dataset_Identifier},
	{"",				WInput_Description},
	{"",				WInput_Level},
	{"",				WInput_Reference},
	{"",				WINS_Reading_Availability},
	{"",				WInstantaneous_Field_of_View},
	{"",				WInstitution_of_Last_Calibration},
	{"",				WInstrument_Description},
	{"",				WInstrument_Full_Name},
	{"",				WInstrument_Information},
	{"",				WInstrument_Properties_Description},
	{"",				WInstrument_Reference},
	{"",				WInstrument_Short_Name},
	{"",				WInstrument_Specific_Georeferencing},
	{"",				WInstrument_Type},
	{"",				WKeplerian_Orbit},
	{"",				WLast_Calibration},
	{"",				WLayer_Name},
	{"",				WLens},
	{"",				WLens_Identifier},
	{"",				WLens_Type},
	{"",				WLight_Drop},
	{"",				WLocation_Information},
	{"",				WMagazine_Identifier},
	{"",				WMaximum_Wavelength},
	{"",				WMethod_of_Last_Calibration},
	{"",				WMinimum_Wavelength},
	{"",				WMission_Completion},
	{"",				WMission_Description},
	{"",				WMission_History},
	{"",				WMission_Information},
	{"",				WMission_Name},
	{"",				WMission_Significant_Event},
	{"",				WMission_Start_Date},
	{"",				WMultiple_Image_Alignment},
	{"",				WNadir},
	{"",				WNadir_Latitude},
	{"",				WNadir_Longitude},
	{"",				WName_of_Dimension},
	{"",				WNominal_Geostationary_Position},
	{"",				WNominal_Spatial_Resolution},
	{"",				WNon_Polynomial_Scaling},
	{"",				WNumber_of_Angle_Distortion_Values},
	{"",				WNumber_of_Bands},
	{"",				WNumber_of_Cross_Track_Samples},
	{"",				WNumber_of_Data_Dimensions},
	{"",				WNumber_of_Denominator_Terms},
	{"",				WNumber_of_Distance_Distortion_Values},
	{"",				WNumber_of_Elevation_Samples},
	{"",				WNumber_of_Numerator_Terms},
	{"",				WNumber_of_Points},
	{"",				WNumber_of_Profile_Samples},
	{"",				WNumber_of_Resolution_Values},
	{"",				WNumber_of_Thematic_Layers},
	{"",				WNumber_of_Wavelength_Bands},
	{"",				WOperational_Mode},
	{"",				WOrbit_Angle_Units},
	{"",				WOrbit_Period},
	{"",				WOrbit_Period_Units},
	{"",				WOrbit_Period_Value},
	{"",				WOther_Collector_Description},
	{"",				WOther_Georeferencing_Description},
	{"",				WPath},
	{"",				WPeak_Wavelength},
	{"",				WPercentage_of_Overlap},
	{"",				WPerigee_Passage_Time},
	{"",				WPhotographic_Resolving_Power},
	{"",				WPitch},
	{"",				WPitch_Accuracy},
	{"",				WPixel_Cross_Track_Size},
	{"",				WPixel_Description},
	{"",				WPixel_Elevation_Size},
	{"",				WPixel_Height_Above_Ellipsoid},
	{"",				WPixel_Point_Spread_Function},
	{"",				WPixel_Profile_Size},
	{"",				WPixel_Resolution},
	{"",				WPlatform_and_Instrument_Identification},
	{"",				WPlatform_and_Mission_Information},
	{"",				WPlatform_Description},
	{"",				WPlatform_Full_Name},
	{"",				WPlatform_Information},
	{"",				WPlatform_Nominal_Altitude},
	{"",				WPlatform_Nominal_Altitude_Units},
	{"",				WPlatform_Nominal_Altitude_Value},
	{"",				WPlatform_Nominal_Longitude},
	{"",				WPlatform_Orbit},
	{"",				WPlatform_Serial_Identifier},
	{"",				WPlatform_Short_Name},
	{"",				WPlatform_Sponsor},
	{"",				WPlatform_Start_Date},
	{"",				WPoint_Position_In_Pixel},
	{"",				WPolarization_Characteristics},
	{"",				WPolynomial_Coefficient},
	{"",				WPolynomial_Denominator},
	{"",				WPolynomial_Function},
	{"",				WPolynomial_Numerator},
	{"",				WPositional_Information},
	{"",				WPrincipal_Point_of_Autocollimation},
	{"",				WPrincipal_Point_of_Symmetry},
	{"",				WProcessing_Change_History},
	{"",				WProcessing_Documentation},
	{"",				WProcessing_Environment},
	{"",				WProcessing_Identifiers},
	{"",				WProcessing_Information},
	{"",				WProcessing_Input_Dataset},
	{"",				WProcessing_Level},
	{"",				WProcessing_Level_Authority},
	{"",				WProcessing_Level_Identifier},
	{"",				WProcessing_Procedure},
	{"",				WProcessing_Procedure_Description},
	{"",				WProcessing_Run_History},
	{"",				WProcessing_Software},
	{"",				WProcessing_Software_Description},
	{"",				WProcessing_Software_Reference},
	{"",				WProfile_Description},
	{"",				WProfile_Extent},
	{"",				WProfile_Fixed},
	{"",				WProfile_Properties},
	{"",				WProfile_Sounding},
	{"",				WProfile_Start},
	{"",				WProfile_Step},
	{"",				WProfiling_Direction},
	{"",				WProjection_Center_X_Position},
	{"",				WProjection_Center_Y_Position},
	{"",				WProjection_Center_Z_Position},
	{"",				WQuality_of_Autocollimation_Principal_Point_x_Value},
	{"",				WQuality_of_Autocollimation_Principal_Point_y_Value},
	{"",				WQuality_of_Focal_Length},
	{"",				WQuality_of_Principal_Point_of_Autocollimation},
	{"",				WQuality_of_Principal_Point_of_Symmetry},
	{"",				WQuality_of_Symmetry_Principal_Point_x_Value},
	{"",				WQuality_of_Symmetry_Principal_Point_y_Value},
	{"",				WRadial_Asymmetrical_Coefficient_B1},
	{"",				WRadial_Asymmetrical_Coefficient_B2},
	{"",				WRadial_Symmetrical_Angle_Interval},
	{"",				WRadial_Symmetrical_Distance_Interval},
	{"",				WRadial_Symmetrical_Distortion_Polynomial},
	{"",				WReceiver_Polarization},
	{"",				WReferenced_Coordinate_System},
	{"",				WReferencing_Polynomial},
	{"",				WRelative_Aperture},
	{"",				WRelative_Azimuth},
	{"",				WReseau},
	{"",				WResolution_Value_Set},
	{"",				WResolving_Angle},
	{"",				WResolving_Value_Radial},
	{"",				WResolving_Value_Tangential},
	{"",				WRight_Ascension_of_Ascending_Node},
	{"",				WRoll},
	{"",				WRoll_Accuracy},
	{"",				WRotation_Description},
	{"",				WRotation_Sequence},
	{"",				WRow},
	{"",				WRow_Delta_X},
	{"",				WRow_Delta_Y},
	{"",				WSample_Angle_Units},
	{"",				WSample_Description_Units},
	{"",				WSample_Profile_Units},
	{"",				WSample_Properties},
	{"",				WSatellite_Local_Azimuth_Angle},
	{"",				WSatellite_Local_Zenith_Angle},
	{"",				WScan},
	{"",				WScan_Angle_Units},
	{"",				WScan_Cross_Track_Properties},
	{"",				WScan_Distance_Units},
	{"",				WScan_Duration},
	{"",				WScan_Elevation_Properties},
	{"",				WScan_Geometric_Properties},
	{"",				WScan_Radiometric_Properties},
	{"",				WScan_Repeat_Time},
	{"",				WScan_Spectral_Properties},
	{"",				WScan_Start_Time},
	{"",				WScan_Step_Time},
	{"",				WScan_Time_Units},
	{"",				WScan_Timing},
	{"",				WScience_Paper},
	{"",				WSemimajor_Axis},
	{"",				WSender_Polarization},
	{"",				WSensor_Element_Location},
	{"",				WSensor_Grid},
	{"",				WSensor_Orientation},
	{"",				WSensor_System},
	{"",				WSolar_Azimuth_Angle},
	{"",				WSolar_Zenith_Angle},
	{"",				WSource_of_Element_Locations},
	{"",				WSpatial_Resolution_Units},
	{"",				WSpatial_Resolution_Value},
	{"",				WSpectral_Information},
	{"",				WSpectral_Limit},
	{"",				WStabilized_Mount},
	{"",				WStorage_Order},
	{"",				WSwath_Track_Information},
	{"",				WThematic_Layer_Identification},
	{"",				WTranslation_Description},
	{"",				WType_of_Overlap},
	{"",				WUnreferenced_Coordinate_System},
	{"",				WUsers_Guide},
	{"",				WWavelength_Band_Properties},
	{"",				WWavelength_Region},
	{"",				WWavelength_Units},
	{"",				WWorldwide_Reference_System},
	{"",				WX_Axis_Definition},
	{"",				WX_Position_Accuracy},
	{"",				WY_Axis_Definition},
	{"",				WY_Position_Accuracy},
	{"",				WYaw},
	{"",				WYaw_Accuracy},
	{"",				WZ_Axis_Definition},
	{"",				WZ_Position_Accuracy},

	{NULL,Wnull}
	};
static int number_table_length = 0;

static int compare (const void *e1, const void *e2) {
	struct section_number *p = (struct section_number *) e1;
	struct section_number *q = (struct section_number *) e2;
	return (p->key - q->key);
	}

char *number_of (enum fgdc_keyword key) {
	int i;
	struct section_number N,*pN;

	if (number_table_length == 0) {
		for (i=0; number_table[i].text; i++);
		number_table_length = i;
		qsort (number_table,number_table_length,sizeof(struct section_number),compare);
		}

	if (key == Wnull) return ("");
	if (key == Wunknown) return ("");

	N.key = key;
	pN = (struct section_number *) bsearch (&N,number_table,number_table_length,sizeof(struct section_number),compare);
	if (pN) return (pN->text);
	else return ("");
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
