# Script to read a tab-delimited text file with headings and
# write out, for each column, an Attribute_Domain_Values

# This is accomplished by first reading the input file (named
# on the command line), taking the first line to contain the
# names of the fields, and for each field, storing the values
# in a list.  The lists are then sorted, and for those fields
# that are numeric, the numerical range is determined.  For
# those that are not numeric, the alphabetized values are
# listed in the output so the metadata author can explain what
# each value means.

if {$argc > 0} {
	set input_file [lindex $argv 0]
	set in [open $input_file r]
	set line [gets $in]
	set names [split $line "\t"]
	while {[eof $in] == 0} {
		set line [gets $in]
		set values [split $line "\t"]
		set i 0
		foreach item $values {
			set name [lindex $names $i]
			# You might want to trim quote marks; if so, then
			# remove the pound sign from the next line
			#set item [string trim $item "\""]
			lappend value($name) $item
			incr i
			}
		}
	close $in

	# Compose a skeletal Attribute section for each column
	# containing the Attribute_Label and Attribute_Domain_Values

	foreach name [array names value] {

		puts "Attribute:"
		puts "  Attribute_Label: $name"
		puts "  Attribute_Domain_Values:"

		# Get unique values in order

		set sorted [lsort -unique -dictionary $value($name)]

		# If any are empty, remember this but remove the empty element

		set empty 0
		set i [lsearch -exact $sorted {}]
		if {$i != -1} {
			set empty 1
			set sorted [lreplace $sorted $i $i]
			}

		# If the values are integer or real, report the range
		set is_integer 1
		foreach unique_value $sorted {
			if {[string is integer $unique_value] == 0} {
				set is_integer 0
				break
				}
			}
		if {$is_integer} {

			# But if there are only a few integer values, then
			# report these individually (note this is just a guess
			# that the values might be codes that really mean
			# something other than their nominal value

			if {[llength $sorted] < 11} {
				foreach unique_value $sorted {
					puts "    Enumerated_Domain:"
					puts "      Enumerated_Domain_Value: $unique_value"
					puts "      Enumerated_Domain_Value_Definition:"
					}
				} \
			else {
				set minimum [lindex $sorted 0]
				set maximum [lindex $sorted end]
				puts "    Range_Domain:"
				puts "      Range_Domain_Minimum: $minimum"
				puts "      Range_Domain_Maximum: $maximum"
				}
			} \
		else {
			set is_double 1
			foreach unique_value $sorted {
				if {[string is double $unique_value] == 0} {
					set is_double 0
					break
					}
				}
			if {$is_double} {
				set minimum [lindex $sorted 0]
				set maximum [lindex $sorted end]
				puts "    Range_Domain:"
				puts "      Range_Domain_Minimum: $minimum"
				puts "      Range_Domain_Maximum: $maximum"
				} \
			else {

				# For character data, just list the possible values

				foreach unique_value $sorted {
					puts "    Enumerated_Domain:"
					puts "      Enumerated_Domain_Value: $unique_value"
					puts "      Enumerated_Domain_Value_Definition:"
					}
				}
			}

		# If any values were empty, note this in an Enumerated_Domain
		# so the metadata author can explain what this means.

		if {$empty} {
			puts "    Enumerated_Domain:"
			puts "      Enumerated_Domain_Value: (empty)"
			puts "      Enumerated_Domain_Value_Definition:"
			}

		}
	}
