/*----------------------------------------------------------------------*\
 | Code to read and write SGML versions of formal metadata.				|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "item.h"
#include "config.h"
#include "stricmp.h"

extern int is_extension (int key);

extern FILE *out;

extern void unescape_string (char *s);

extern char *iso8859_1[];
extern char *dos_codeset[];
extern char *mac_codeset[];

/*\
For CSDGM version 1, use this doctype:
#define DOCTYPE_DECLARATION	"<!DOCTYPE METADATA PUBLIC \"-//FGDC//DTD METADATA 1.0//EN\">"
\*/
#define DOCTYPE_DECLARATION	"<!DOCTYPE METADATA PUBLIC \"-//FGDC//DTD METADATA 2.0//EN\">"

/*----------------------------------------------------------------------*/

extern const char *astm_tag_of (enum fgdc_keyword key);
extern int code_of (enum fgdc_keyword key);
extern enum fgdc_keyword key_of_astm_tag (const char *astm_tag);

extern const char *ps8_tag_of (enum fgdc_keyword key);
extern enum fgdc_keyword key_of_ps8_tag (const char *ps8_tag);

const char *(*sgml_tag_of)(enum fgdc_keyword key);
enum fgdc_keyword (*key_of_sgml_tag) (const char *sgml_tag);

void use_sgml_tags (char *name) {
	if (stricmp (name,"ASTM") == 0) {
		sgml_tag_of = astm_tag_of;
		key_of_sgml_tag = key_of_astm_tag;
		}
	else {
		sgml_tag_of = ps8_tag_of;
		key_of_sgml_tag = key_of_ps8_tag;
		}
	}

/*----------------------------------------------------------------------*\
 | Simple entity decoder, recognizes only ISO 8859-1 entities.			|
\*----------------------------------------------------------------------*/

static void decode_entities (char *d, char *s) {
	int i,n;
	char **e;

	while (*s) {
		if (*s == '&') {
			s++;
			e = iso8859_1;
			for (i=0; i < 256; i++) {
				if (*e) {
					n = strlen (*e);
					if ((memcmp (s,*e,n) == 0) && *(s+n) == ';') {
						*d++ = i;
						s += 1 + n;
						break;
						}
					}
				e++;
				}
			if (i == 256) *d++ = *s;
			}
		else *d++ = *s;
		s++;
		}
	*d = 0;
	}

/*----------------------------------------------------------------------*\
 | Ridiculously primitive SGML parser.  This reads SGML into the tree	|
 | structure used by mp.  The input SGML must be perfect or it won't	|
 | work.  Specifically, no end-tags may be omitted, the nesting of		|
 | elements must be correct, and there can be no attributes in the		|
 | element tags, although there may be character entities in the text.	|
 | No element tag is broken at the end of a line.						|
 |																		|
 | If you need more flexibility, use a real SGML translator.			|
\*----------------------------------------------------------------------*/

enum tag_type_t {
	TAG_DECLARATION,
	TAG_START,
	TAG_END
	};

#ifdef USE_OLD_SGML_PARSER

struct item *parse_sgml (char *input_file) {
	void *opt;
	FILE *in;
	int i,n;
	char *line,*next_line;
	char *output = NULL;
	int output_size = 0;
	char *dst,*end;
	char *d,*s,*t;
	enum fgdc_keyword key;
	struct item *root = NULL;
	struct item *current = NULL;
	struct item *p;
	int line_number;
	int insert_blank = 0;

	if (!input_file) return (NULL);

	use_sgml_tags ("ps8");
	if (opt = find_option (NULL,"output"))
		if (opt = find_option (opt,"sgml"))
			if (opt = find_option (opt,"tags"))
				if (stricmp (text_of_option (opt),"astm") == 0)
					use_sgml_tags ("ASTM");

	if (in = fopen (input_file,"r")) {
		line_number = 0;

		if (output = (char *) malloc (8192)) {
			output_size = 8192;
			}
		else {
			fprintf (stderr,"Error: could not allocate parse text buffer\n");
			exit (1);
			}

		/* Snarf the whole file at once */

		if (fseek (in,0L,SEEK_END) == 0) {
			long size = 1L + ftell (in);
			if (size > 1L) {
				if (dst = (char *) malloc (size)) {
					rewind (in);
					n = fread (dst,1,size,in);
					end = dst + n;
					*end = 0;
					}
				else
					fprintf (out,"Error: could not allocate space for file\n");
				}
			else
				fprintf (out,"Error: Input file is empty.\n");
			}
		fclose (in);

		/* Process the data line by line.  Start at the beginning. */

		d = output;
		*d = 0;
		line = dst;

		/*--------------------------------------------------------------*\
		 | Find the next line.  Zero-out the EOL (LF, CR, or CRLF)		|
		\*--------------------------------------------------------------*/

		while (line < end) {
			for (s=line; s < end; s++)
				if (*s == '\r') {
					*s++ = 0;
					next_line = s;
					if (s < end && *s == '\n') {
						*s++ = 0;
						next_line++;
						break;
						}
					}
				else
					if (*s == '\n') {
						*s++ = 0;
						next_line = s;
						break;
						}
			line_number++;

			/*----------------------------------------------------------*\
			 | line now points to the beginning of this line, and it is	|
			 | null-terminated.											|
			\*----------------------------------------------------------*/

			s = line;
			while (*s) {

				/*------------------------------------------------------*\
				 | When you find a tag, determine whether it is a		|
				 | start tag, an end tag, or a declaration.				|
				\*------------------------------------------------------*/

				if (*s == '<') {
					enum tag_type_t tag_type = TAG_START;
					s++;
					if (*s == '/') {
						s++;
						tag_type = TAG_END;
						}
					else
						if (*s == '!') {
							s++;
							tag_type = TAG_DECLARATION;
							}

					/*--------------------------------------------------*\
					 | Find the end of the tag.							|
					\*--------------------------------------------------*/

					for (t=s; *t; t++)
						if (*t == '>') break;
					if (*t == '>') {
						*t++ = 0;
						if (tag_type != TAG_DECLARATION)

							/*------------------------------------------*\
							 | When you find a start tag, create an		|
							 | element of this type and make it the		|
							 | current node.							|
							\*------------------------------------------*/

							if ((key = key_of_sgml_tag (s)) != Wunknown) {
								if (tag_type == TAG_START) {
									if (d > output) {
										char *w;
										for (w=output; isspace (*w); w++);
										if (*w)
											fprintf (stderr,"Error (line %3d): extraneous text in SGML ignored: \"%s\"\n",
												line_number,output);
										d = output;
										*d = 0;
										}
									if (!current) {
										root = allocate_item (1);
										current = root;
										current->key = key;
										current->line_number = line_number;
										}
									else {
										p = add_child (current);
										current = p;
										current->key = key;
										current->line_number = line_number;
										}
									}
								if (tag_type == TAG_END)
									if (key == current->key) {
										char *w;

										/*------------------------------*\
										 | If only whitespace in output	|
										 | then consider it empty.		|
										\*------------------------------*/

										for (w=output; isspace(*w); w++);
										if (*w == 0) *output = 0;

										if (strlen (output) == 0) {
											if (current->parent) current = current->parent;
											}
										else
											if (p = add_child (current)) {
												p->key = Wunknown;
												p->line_number = line_number;
												if (p->d = (char *) malloc (1 + strlen(output))) {
													decode_entities (p->d,output);
													d = output;
													*d = 0;
													if (current->parent) current = current->parent;
													}
												else {
													fprintf (stderr,"Error (line %3d): could not allocate space for text of %s\n",
														line_number,text_of(current->key));
													exit (1);
													}
												}
											else {
												fprintf (stderr,"Error (line %3d): could not allocate scalar child of %s\n",
													line_number,text_of(current->key));
												exit (1);
												}
										}
									else {
										fprintf (stderr,"Error (line %3d): expected end-tag for %s, got %s\n",
											line_number,text_of(current->key),text_of(key));
										exit (1);
										}
								}
							else {
								fprintf (stderr,"Error (line %3d): unrecognized %s%s>\n",
									line_number,
									((tag_type == TAG_START) ? "start tag <" : "end tag </"),
									s
									);
								/* exit (1); */
								}
						else
							fprintf (stderr,"Warning (line %3d): declaration ignored in SGML input: %s\n",
								line_number,s);
						s = t;
						}
					else {
						fprintf (stderr,"Error (line %3d): unmatched '<'\n",line_number);
						exit (1);
						}
					}
				else

					/*--------------------------------------------------*\
					 | Do entity substitution if you can find one.		|
					 | Copy the byte if you cannot.						|
					\*--------------------------------------------------*/

					if (*s == '&') {
						for (i=0; i < 256; i++)
							if (t = iso8859_1[i])
								if (memcmp (s+1,t,strlen(t)) == 0)
									if (*(s+1+strlen(t)) == ';') {
										*d++ = i;
										*d = 0;
										s += 2 + strlen (t);
										break;
										}
						if (memcmp (s+1,"break;",6) == 0) {
							s += 7;
							insert_blank = 1;
							}
						else
							if (i >= 256) {
								*d++ = *s++;
								*d = 0;
								}
						}
					else {
						*d++ = *s++;
						*d = 0;
						}

				/*------------------------------------------------------*\
				 | If you're about to write past the end of the output	|
				 | buffer, don't.  Reallocate it for more space.		|
				\*------------------------------------------------------*/

				if (d - output >= output_size - 1) {
					n = d - output;
					if (output = (char *) realloc (output,4096+output_size)) {
						d = output + n;
						output_size += 4096;
						}
					else {
						fprintf (stderr,"Error (line %3d): could not reallocate parse text buffer\n",
							line_number);
						exit (1);
						}
					}
				}

			/*----------------------------------------------------------*\
			 | At the end of a line, if there's anything in the output	|
			 | buffer, create a Wunknown to hold it.  If there isn't	|
			 | any text on the line but you're in a textual value,		|
			 | create a blank element to match the blank line.			|
			\*----------------------------------------------------------*/

			if (strlen (output) > 0)
				if (p = add_child (current)) {
					p->key = Wunknown;
					p->line_number = line_number;
					n = strlen (output);
					if (p->d = (char *) malloc (1 + n)) {
						strcpy (p->d,output);
						p->len = 1 + n;
						p->original = 0;
						d = output;
						*d = 0;
						}
					else {
						fprintf (stderr,"Error (line %3d): could not allocate space for text of %s\n",
							line_number,text_of(current->key));
						exit (1);
						}
					}
				else {
					fprintf (stderr,"Error (line %3d): could not allocate scalar child of %s\n",
						line_number,text_of(current->key));
					exit (1);
					}
			else
				if (current)
					if (current->child)
						if (current->child->key == Wunknown)
							insert_blank = 1;

			if (insert_blank) {
				if (p = add_child (current)) {
					p->key = Wblank;
					p->line_number = line_number;
					}
				insert_blank = 0;
				}

			line = next_line;
			}
		}
	else {
		fprintf (stderr,"Error: could not open input file %s\n",input_file);
		exit (1);
		}
	return (root);
	}

#endif

/*----------------------------------------------------------------------*\
 | Keep track of those elements the user has asked be omitted from the	|
 | output file.  Their keys are stored in a static array of size 1024,	|
 | which is consulted whenever an element is to be output.				|
\*----------------------------------------------------------------------*/

static enum fgdc_keyword omit_key[1024];
static int omit_key_count = -1;

static int omit (enum fgdc_keyword key) {
	int i,result = 0;
	if (omit_key_count > 0)
		for (i=0; i < omit_key_count; i++)
			if (key == omit_key[i]) {
				result = 1;
				break;
				}
	return (result);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static char **entity = iso8859_1;
static int output_extensions = 1;
static char *what_to_print_for_Wblank = "\n";
static char *one_indent = "";
static int indent_len = 0;
static char *empty = "";

static FILE *fp = NULL;

static void write_sgml_text (unsigned char *s) {
	char *t;
	fflush (fp);
	while (*s) {
		if (t = entity[*s]) {
			fputc ('&',fp);
			while (*t) fputc (*t++,fp);
			fputc (';',fp);
			}
		else
			fputc (*s,fp);
		s++;
		}
	fflush (fp);
	}

static void write_sgml_item (struct item *p, int n) {
	int i;
	if (!p) return;
	if (omit (p->key)) return;
	if (!output_extensions && is_extension (p->key)) return;
	if (p->key == Wblank) {
		if (p->next && p->next->key == Wunknown)
			fprintf (fp,what_to_print_for_Wblank);
		}
	else
		if (p->key == Wunknown) {
			if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);
			write_sgml_text ((unsigned char *) p->d);
			fprintf (fp,"\n");
			}
		else {
			struct item *q;

			if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);
			fprintf (fp,"<%s>",sgml_tag_of(p->key));
			if (p->child)
				if (p->child->key == Wunknown && p->child->next == NULL)
					write_sgml_text ((unsigned char *)p->child->d);
				else {
					fprintf (fp,"\n");
					for (q=p->child; q; q=q->next) write_sgml_item (q,n+1);
					if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);
					}
			fprintf (fp,"</%s>\n",sgml_tag_of(p->key));
			}
	}

void write_sgml (char *output_file, struct item *first) {
	int i = 0;
	struct item *p;
	int do_astm = 0;
	void *opt;

	if (first == NULL) return;

	if (fp = fopen (output_file,"w")) {

		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"sgml"))
				if (opt = find_option (opt,"tags"))
					if (stricmp (text_of_option (opt),"astm") == 0)
						do_astm = 1;

		if (do_astm) {
			use_sgml_tags ("ASTM");

			fprintf (fp,"<!SGML \"ISO 8879:1986\"\n");
			fprintf (fp,"CHARSET\n");
			fprintf (fp,"BASESET \"ISO 646-1983//CHARSET International Reference Version (IRV)//ESC 2/5 4/0\"\n");
			fprintf (fp,"DESCSET     0   9   UNUSED\n");
			fprintf (fp,"       9    2   9\n");
			fprintf (fp,"      11    2   UNUSED\n");
			fprintf (fp,"      13    1   13\n");
			fprintf (fp,"      14   18   UNUSED\n");
			fprintf (fp,"      32   95   32\n");
			fprintf (fp,"     127    1   UNUSED\n");
			fprintf (fp,"CAPACITY     PUBLIC \"ISO 8879:1986//CAPACITY Reference//EN\"\n");
			fprintf (fp,"SCOPE          DOCUMENT\n");
			fprintf (fp,"SYNTAX\n");
			fprintf (fp,"SHUNCHAR CONTROLS 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17\n");
			fprintf (fp,"         18 19 20 21 22 23 24 25 26 27 28 29 30 31 127 255\n");
			fprintf (fp,"BASESET \"ISO 646-1983//CHARSET International Reference Version (IRV)//ESC 2/5 4/0\"\n");
			fprintf (fp,"DESCSET    0          128         0\n");
			fprintf (fp,"FUNCTION   RE                    13\n");
			fprintf (fp,"           RS                    10\n");
			fprintf (fp,"           SPACE                 32\n");
			fprintf (fp,"           TAB        SEPCHAR     9\n");
			fprintf (fp,"NAMING     LCNMSTRT   \"\"\n");
			fprintf (fp,"           UCNMSTRT   \"\"\n");
			fprintf (fp,"           LCNMCHAR   \"-.\"\n");
			fprintf (fp,"           UCNMCHAR   \"-.\"\n");
			fprintf (fp,"NAMECASE   GENERAL    YES\n");
			fprintf (fp,"           ENTITY     NO\n");
			fprintf (fp,"DELIM      GENERAL    SGMLREF\n");
			fprintf (fp,"           SHORTREF   SGMLREF\n");
			fprintf (fp,"NAMES      SGMLREF\n");
			fprintf (fp,"QUANTITY   SGMLREF\n");
			fprintf (fp,"NAMELEN 10\n");
			fprintf (fp,"FEATURES\n");
			fprintf (fp,"MINIMIZE   DATATAG   NO   OMITTAG    YES          RANK       NO     SHORTTAG YES\n");
			fprintf (fp,"LINK       SIMPLE    NO   IMPLICIT   NO           EXPLICIT   NO\n");
			fprintf (fp,"OTHER      CONCUR    NO   SUBDOC     NO           FORMAL     NO\n");
			fprintf (fp,"APPINFO    NONE>\n");

			}
		else
			use_sgml_tags ("ps8");

		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"sgml")) {
				void *v;
				if (v = find_option (opt,"indent")) {
					char *s = text_of_option (v);
					if (!s) s = empty;
					if (*s) unescape_string (s);
					one_indent = s;
					indent_len = strlen (one_indent);
					}
				if (v = find_option (opt,"blanks")) {
					char *s = text_of_option (v);
					if (!s) s = empty;
					if (*s) unescape_string (s);
					what_to_print_for_Wblank = s;
					}
				if (opt = find_option (opt,"skip_extensions"))
					output_extensions = 0;
				}

		/*------------------------------------------------------------------*\
		 | For up to 1024 keys, the user can specify elements that will not	|
		 | be included in the output.										|
		\*------------------------------------------------------------------*/

		if (omit_key_count == -1)
			if (opt = find_option (NULL,"output"))
				if (opt = find_option (opt,"sgml"))
					if (opt = find_option (opt,"omit"))
						do {
							enum fgdc_keyword key;
							char *s = text_of_option (opt);
							if (*s) unescape_string (s);
							if ((key = key_of (s)) != Wunknown) {
								if (omit_key_count == -1) omit_key_count = 0;
								if (omit_key_count < 1024)
									omit_key[omit_key_count++] = key;
								}
							} while (opt = find_next_option (opt,"omit"));

		/*--------------------------------------------------------------*\
		 | Use appropriate entity substitution depending on the codeset	|
		 | of the source text.  Default is ISO8859-1.					|
		\*--------------------------------------------------------------*/

		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"codeset")) {
				char *s = text_of_option (opt);
				if (stricmp (s,"ISO8859-1") == 0) entity = iso8859_1;
				if (stricmp (s,"DOS"      ) == 0) entity = dos_codeset;
				if (stricmp (s,"MAC"      ) == 0) entity = mac_codeset;
				}

		fprintf (fp,"%s\n",DOCTYPE_DECLARATION);
		for (p=first; p; p=p->next) write_sgml_item (p,i);
		fclose (fp);
		}
	else {
		fprintf (stderr,"Error: could not create output file %s\n",output_file);
		return;
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
