/*----------------------------------------------------------------------*\
 | Reads shapefile header and prints it to stdout.						|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static char *shapetype[] = {
	"Null Shape",
	"Point",
	"reserved",
	"PolyLine",
	"reserved",
	"Polygon",
	"reserved",
	"reserved",
	"MultiPoint",
	"reserved",
	"reserved",
	"PointZ",
	"reserved",
	"PolyLineZ",
	"reserved",
	"PolygonZ",
	"reserved",
	"reserved",
	"MultiPointZ",
	"reserved",
	"reserved",
	"PointM",
	"reserved",
	"PolyLineM",
	"reserved",
	"PolygonM",
	"reserved",
	"reserved",
	"MultiPointM",
	"reserved",
	"reserved",
	"MultiPatch",
	NULL
	};

struct shape_header {
	int FileCode;
	int Unused04;
	int Unused08;
	int Unused12;
	int Unused16;
	int Unused20;
	int FileLength;
	int Version;
	int ShapeType;
	double Xmin;
	double Ymin;
	double Xmax;
	double Ymax;
	double Zmin;
	double Zmax;
	double Mmin;
	double Mmax;
	};

/*----------------------------------------------------------------------*\
 | Code to handle byte-order variations									|
\*----------------------------------------------------------------------*/

static int byte_order;

#ifndef LSBFirst
#define LSBFirst	0
#endif

#ifndef MSBFirst
#define MSBFirst	1
#endif

static void get_byte_order (void) {
	short w = 255;
	unsigned char *s = (unsigned char *) &w;
	if (*s) byte_order = LSBFirst;
	else byte_order = MSBFirst;
	}

void swab_dword (unsigned char *s) {
	unsigned char t;
	t = *(s+0); *(s+0) = *(s+3); *(s+3) = t;
	t = *(s+1); *(s+1) = *(s+2); *(s+2) = t;
	}

void swab_qword (unsigned char *s) {
	unsigned char t;
	t = *(s+0); *(s+0) = *(s+7); *(s+7) = t;
	t = *(s+1); *(s+1) = *(s+6); *(s+6) = t;
	t = *(s+2); *(s+2) = *(s+5); *(s+5) = t;
	t = *(s+3); *(s+3) = *(s+4); *(s+4) = t;
	}

static int stricmp (const char *s, const char *t) {
	int d = 0;
	do {
		d = toupper(*s) - toupper(*t);
		} while (*s++ && *t++ && !d);
	return (d);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

int main (int argc, char *argv[]) {
	char input_file[1024];
	FILE *in;
	long n;
	unsigned char buffer[100];
	unsigned char *s;
	int i,j;
	struct shape_header header;
	int do_hex = 0;
	int do_fgdc = 0;

	*input_file = 0;
	get_byte_order();

	if (argc > 1)
		for (i=1; i < argc; i++)
			if (memcmp (argv[i],"-hex",4) == 0) do_hex = 1;
			else
				if (memcmp (argv[i],"-fgdc",5) == 0) do_fgdc = 1;
				else
					strcpy (input_file,argv[i]);

	if (*input_file) {
		if (s = strrchr (input_file,'.')) {
			if (stricmp (s,".shp") != 0) strcat (input_file,".shp");
			}
		else strcat (input_file,".shp");

		if (in = fopen (input_file,"rb")) {
			n = fread (buffer,1,100,in);
			if (n < 100) {
				fprintf (stderr,"Error: only got %d bytes, asked for 100\n",n);
				exit (1);
				}

			if (byte_order == LSBFirst) {
				/* This machine is LSBFirst */
				/* swab first 7 dwords */
				for (i=0, s = buffer; i < 7; i++, s+=4) swab_dword (s);
				}
			else {
				/* This machine is MSBFirst */
				/* swab dwords 8 and 9, then swab 8 qwords */
				for (i=0, s = buffer + 28; i < 2; i++, s+=4) swab_dword (s);
				for (i=0; i < 8; i++, s+=8) swab_qword (s);
				}

			if (do_hex) {
				printf ("Shapefile header (hexadecimal):\n");
				s = buffer;
				for (i=0; i < 9; i++) {
					printf ("  ");
					for (j=3; j >= 0; j--)
						printf ("%02X",*(s+j));
					printf ("\n");
					s += 4;
					}
				for (i=0; i < 8; i++) {
					printf ("  ");
					for (j=7; j >= 0; j--)
						printf ("%02X",*(s+j));
					printf ("\n");
					s += 8;
					}
				}
			else {

				/*------------------------------------------------------*\
				 | The header structure will normally be padded with 4	|
				 | bytes after the ShapeType member in order to make	|
				 | Xmin occur on an address that is a multiple of 8.	|
				 | So we copy the buffer into the structure in two		|
				 | steps, the first 9 dwords to the beginning, and the	|
				 | remaining 8 qwords starting at Xmin.					|
				\*------------------------------------------------------*/

				memcpy (&header,buffer,36);
				memcpy (&header.Xmin,buffer+36,64);

				/*------------------------------------------------------*\
				 | Don't do FGDC-style snippet as output if the file's	|
				 | bounding coordinates do not look like they are given	|
				 | in latitude and longitude							|
				\*------------------------------------------------------*/

				if (do_fgdc) {
					if (header.Xmin < -180.0 || header.Xmin > 180.0) do_fgdc = 0;
					if (header.Xmax < -180.0 || header.Xmax > 180.0) do_fgdc = 0;
					if (header.Ymin <  -90.0 || header.Ymin >  90.0) do_fgdc = 0;
					if (header.Ymax <  -90.0 || header.Ymax >  90.0) do_fgdc = 0;
					if (do_fgdc == 0) printf ("-- bounding coordinates extend beyond geographic range\n");
					}

				/*------------------------------------------------------*\
				 | FGDC-style snippet is just the bounding coordinates	|
				\*------------------------------------------------------*/

				if (do_fgdc) {
					printf ("Spatial_Domain:\n");
					printf ("  Bounding_Coordinates:\n");
					printf ("    West_Bounding_Coordinate:  %lf\n",header.Xmin);
					printf ("    East_Bounding_Coordinate:  %lf\n",header.Xmax);
					printf ("    North_Bounding_Coordinate: %lf\n",header.Ymax);
					printf ("    South_Bounding_Coordinate: %lf\n",header.Ymin);
					}
				else {
					printf ("Shapefile name: %s\n",input_file);
					printf ("  FileCode:   %d\n",header.FileCode);
					printf ("  Unused:     %d %d %d %d %d\n",header.Unused04,header.Unused08,header.Unused12,header.Unused16,header.Unused20);
					printf ("  FileLength: %d\n",2*header.FileLength);
					printf ("  Version:    %d\n",header.Version);
					if (header.ShapeType > 0 && header.ShapeType < 32)
						printf ("  ShapeType:  %s\n",shapetype[header.ShapeType]);
					else
						printf ("  ShapeType:  %d\n",header.ShapeType);
					printf ("  Xmin:       %lf\n",header.Xmin);
					printf ("  Ymin:       %lf\n",header.Ymin);
					printf ("  Xmax:       %lf\n",header.Xmax);
					printf ("  Ymax:       %lf\n",header.Ymax);
					printf ("  Zmin:       %lf\n",header.Zmin);
					printf ("  Zmax:       %lf\n",header.Zmax);
					printf ("  Mmin:       %lf\n",header.Mmin);
					printf ("  Mmax:       %lf\n",header.Mmax);
					}
				}

			fclose (in);
			}
		else {
			fprintf (stderr,"Error: could not open shapefile %s\n",input_file);
			exit (1);
			}
		}
	else {
		fprintf (stderr,"shpinfo 1.1 (Peter N. Schweitzer, U.S. Geological Survey, Reston VA 20192)\n");
		fprintf (stderr,"Usage: %s [-hex] [-fgdc] shapefile\n",argv[0]);
		fprintf (stderr,"  -hex   print the file header in hexadecimal (for debugging)\n");
		fprintf (stderr,"  -fgdc  produce a snippet containing the bounding coordinates\n");
		exit (0);
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
