/*----------------------------------------------------------------------*\
 | CGI program to create Source_Information from bibliography entries.	|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "formio.h"
#include "buffer.h"

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static char header[] =
"<table border=\"0\" bgcolor=\"#99CCFF\" width=\"100%\" cellpadding=\"4\"><tr><td>"
"<A HREF=\"http://www.usgs.gov/\"><img src=\"/icons/000000/usgs-mini.gif\" width=\"80\" height=\"21\" border=\"0\" alt=\"U.S. Geological Survey\"></A>"
"</td></tr></table>\n"
"<hr><p>";

static char *maintainer_line = "Maintained by <A HREF=\"/peter/\">Peter Schweitzer</A><br>";

static char *empty = "(empty)";

static char *style = NULL;
static char *stuff = NULL;

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

#ifndef HAVE_STRICMP

int stricmp (char *s, char *t) {
	int d = 0;
	do {
		d = toupper(*s) - toupper(*t);
		} while (*s++ && *t++ && !d);
	return (d);
	}

#endif

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

void decode_stuff (char *string) {
	if (string && *string) stuff = string;
	}

void decode_style (char *string) {
	if (string && *string) style = string;
	}

/*----------------------------------------------------------------------*/

struct keyword_table keyword[] = {
	{"stuff",		decode_stuff},
	{"style",		decode_style},
	{NULL,			NULL}
	};
static int keyword_count = 0;

#define MAX_LINE_LENGTH	4095

static char *static_list [64];

/*----------------------------------------------------------------------*\
 | Rough parser for names in a list.
 | Expects Last1, First, Last2, F.M., ..., and Lastn, F.M.
\*----------------------------------------------------------------------*/

static char **parse_authors (char *string) {
	char **list = NULL;
	char **p;
	char *s,*t;

	if (string && *string) {
		p = static_list;
		s = string;

		while (*s) {
			while (*s && isspace (*s)) s++;
			if (memcmp (s,"and ",4) == 0) s += 4;
			*p++ = s;
			while (*s && *s != ',') s++;
			if (*s == ',') s++;
			while (*s && *s != ',') s++;
			if (*s == ',') {
				t = s + 1;
				while (*t && isspace (*t)) t++;
				if ((memcmp (t,"Jr.",3) == 0) ||
					(memcmp (t,"III",3) == 0)) s = t+3;
				}
			if (*s == ',') *s++ = 0;
			if ((strcmp (*(p-1),"ed." ) == 0) ||
				(strcmp (*(p-1),"eds.") == 0) ||
				(strcmp (*(p-1),"."   ) == 0)) p--;
			if (p - static_list > 62) break;
			}
		*p = NULL;
		list = static_list;
		}

	return (list);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static char *alt_date[] = {
	"in press",
	"in preparation",
	"in prep.",
	"in prep",
	"unpublished",
	NULL
	};

static void parse_reference (char *line) {
	int i,n;
	char *s,*t;
	char *authors,*year,*title,*venue;
	char **originator;
	char denominator [16];
	char *issue;
	char abbrev[64];
	int found;

	if (! line) return;
	if (!*line) return;

	s = line;

	authors = year = title = venue = NULL;

	/*--------------------------------------------------*\
	 | Assume originator names come first, followed by	|
	 | date of publication, title terminated by colon,	|
	 | then publication venue.							|
	\*--------------------------------------------------*/

	authors = line;
	s = line;
	while (*s && !isdigit (*s) && (memcmp (s,"in pre",6) != 0) && (memcmp (s,"unpublished",11) != 0)) s++;

	/* Null-terminate the author list */

	t = s - 1;
	while (t >= line && isspace(*t)) *t-- = 0;
	if (*t == ',') *t-- = 0;

	/* Decode the year */

	if (!isdigit (*s)) {
		year = s;
		for (i=0; alt_date[i]; i++)
			if (memcmp (s,alt_date[i],strlen(alt_date[i])) == 0) {
				s += strlen (alt_date[i]);
				break;
				}
		}
	else {
		n = strtol (s,&t,0);
		if (n < 1800 || n > 2050) {
			fprintf (stderr,"Warning: Number %d presumed to be year of publication may not be correct\n",n);
			}
		year = s;
		while (*s && isdigit (*s)) s++;
		if (isalpha (*s)) *s++ = 0;		/* e.g. 1970a */
		}
	if (*s == ',') *s++ = 0;
	if (*s == '.') *s++ = 0;
	if (*s == ':') *s++ = 0;

	while (*s && isspace (*s)) *s++ = 0;
	title = s;
	while (*s && *s != ':') s++;
	if (*s == ':') *s++ = 0;

	while (*s && isspace (*s)) s++;
	venue = s;

	/* Separate authors */

	if (authors && *authors)
		originator = parse_authors (authors);

	/* Source_Scale_Denominator */

	*denominator = 0;
	if (strlen (venue) > 5) {
		s = venue + strlen (venue) - 5;
		found = 0;
		while (s >= venue && !found) {
			if ((memcmp (s,"scale",5) == 0) ||
				(memcmp (s,"Scale",5) == 0)) {
				found = 1;
				*s = 0;
				s += 5;
				while (*s && !isdigit (*s)) s++;
				if (memcmp (s,"1:",2) == 0) s += 2;
				t = denominator;
				while (*s && !isalpha (*s)) {
					if (isdigit (*s) || *s == '.') *t++ = *s;
					if (t - denominator > 15) break;
					s++;
					}
				if (t > denominator && *(t-1) == '.') t--;
				*t = 0;
				}
			else
				s--;
			}
		}

	/* Can you do anything with venue? */

	s = venue;
	while (*s && !isdigit(*s)) s++;
	t = s;
	found = 0;
	while (t > venue && !found)
		if (memcmp (t,"v.",2) == 0) found = 1;
		else t--;
	if (found) s = t;
	issue = s--;
	while (s >= venue && isspace(*s)) *s-- = 0;
	if (*s == ',') *s-- = 0;

	/* Clean up venue and issue if you can */

	if (issue) {
		s = issue - 1;
		while (s > venue && !isspace(*s)) s--;
		if (isspace(*s)) {
			issue = s+1;
			while (s > venue && isspace(*s)) *s-- = 0;
			if (s > venue && *s == ',') *s-- = 0;
			}

		s = issue + strlen (issue) - 1;
		while (s > issue && isspace(*s)) *s-- = 0;
		if (s > issue && *s == ',') *s-- = 0;
		}

	/* Source_Citation_Abbreviation */

	n = 0;
	if (originator) {
		for (n=0; originator[n]; n++);
		switch (n) {
			case 1:
				strcpy (abbrev,originator[0]);
				if (s = strchr (abbrev,',')) *s = 0;
				break;
			case 2:
				strcpy (abbrev,originator[0]);
				if (s = strchr (abbrev,',')) *s = 0;
				strcat (abbrev," and ");
				strcat (abbrev,originator[1]);
				if (s = strchr (abbrev,',')) *s = 0;
				break;
			default:
				strcpy (abbrev,originator[0]);
				if (s = strchr (abbrev,',')) *s = 0;
				strcat (abbrev," and others");
				break;
			}
		}
	else
		strcpy (abbrev,authors);

	strcat (abbrev," (");
	strcat (abbrev,year);
	strcat (abbrev,")");

	/*--------------------------------------------------*\
	 | Output as an FGDC Source_Information				|
	\*--------------------------------------------------*/

	printf ("Source_Information:\n");
	printf ("  Source_Citation:\n");
	printf ("    Citation_Information:\n");
	if (originator)
		for (i=0; originator[i]; i++)
			printf ("      Originator: %s\n",entity_substitution(originator[i]));
	else
		printf ("      Originator: %s\n",entity_substitution(authors));
	printf ("      Title: %s\n",entity_substitution(title));
	printf ("      Publication_Date: %s\n",entity_substitution(year));
	printf ("      Series_Information:\n");
	printf ("        Series_Name: %s\n",entity_substitution(venue));
	printf ("        Issue_Identification: %s\n",entity_substitution(issue));
	printf ("      Other_Citation_Details: \n");
	if (*denominator)
		printf ("  Source_Scale_Denominator: %s\n",entity_substitution(denominator));
	printf ("  Type_of_Source_Media: paper\n");
	printf ("  Source_Time_Period_of_Content: \n");
	printf ("    Time_Period_Information: \n");
	printf ("      Single_Date/Time: \n");
	printf ("        Calendar_Date: %s\n",entity_substitution(year));
	printf ("    Source_Currentness_Reference: publication date\n");
	printf ("  Source_Citation_Abbreviation: %s\n",entity_substitution(abbrev));
	printf ("  Source_Contribution: \n");
	printf ("\n");
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

/* static char line [1 + MAX_LINE_LENGTH]; */
static char paragraph [1 + MAX_LINE_LENGTH];

#ifdef NOT_CGI
static char *get_line (FILE *in) {
	char *s,*t;

	if (fgets (line,MAX_LINE_LENGTH,in)) {
		s = line + strlen (line) - 1;
		if (*s == '\n') *s-- = 0;
		if (*s == '\r') *s-- = 0;
		while (s >= line && isspace(*s)) *s-- = 0;
		for (t=line; *t && isspace(*t); t++);
		if (t > line) strcpy (line,t);
		return (line);
		}
	else return (NULL);
	}
#endif

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

main (int argc, char *argv[]) {
	int para = 0;
	char *b,*e,*s,*t;
	FILE *in = stdin;

	get_form_input();

	printf ("Content-type: text/html\n");
	printf ("\n");

	printf ("<html>\n");
	printf ("<head>\n");
	printf ("<title>");
	printf ("Source_Information helper results");
	printf ("</title>\n");
	printf ("</head>\n");
	printf ("<body bgcolor=\"#FFFFFF\">\n");
	printf ("%s\n",header);

	printf ("<h3>Source_Information helper results</h3>\n");

	if (stricmp (style,"para") == 0) para = 1;

	/* Parse stuff */

	if (stuff) {

		printf ("Edit the text below to fix any mistakes, then select the text and paste into your metadata editor.<p>\n");

		printf ("<form method=\"POST\" action=\"/cgi-bin/dm/echo\">\n");
		printf ("<textarea name=\"results\" rows=\"30\" cols=\"60\">\n");

		b = stuff;
		*paragraph = 0;
		if (para) {
			while (*b) {
				/* Find first non-blank and non-tab */

				while (*b && (*b == ' ' || *b == '\t')) b++;

				/* Find end of the line or file */

				e = b;
				while (*e && *e != '\n') e++;

				/* If \n is preceded by \r, delete the \r */

				t = e - 1;
				if (*t == '\r') *t-- = 0;

				/* Strip trailing blanks from the line */

				while (isspace(*t)) *t-- = 0;

				/* If e points to \n, zero it and point e to next */

				if (*e) *e++ = 0;

				/* If b points to a non-blank line, add to the paragraph. */
				/* If b points to a blank line, decode the paragraph. */

				if (*b) {
					if (*paragraph) strcat (paragraph," ");
					strcat (paragraph,b);
					}
				else {
					parse_reference (paragraph);
					*paragraph = 0;
					}

				/* Point b at beginning of next line */

				b = e;
				}
			if (*paragraph) parse_reference (paragraph);
			}
		else
			while (*b) {
				while (*b && isspace(*b)) b++;
				e = b;
				while (*e && *e != '\n') e++;
				for (t=e-1; t > b && isspace(*t); t--) *t = 0;
				if (*e) *e++ = 0;
				parse_reference (b);
				b = e;
				}

		printf ("</textarea>\n");
		printf ("</form>\n");
		}

	printf ("<hr>\n");
	printf ("This page is the result of a web form entry.<br>\n");
	printf ("%s\n",maintainer_line);
	printf ("</body>\n");
	printf ("</html>\n");
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
