/*----------------------------------------------------------------------*\
 | SGML to MP translator.  This converts simple SGML to the indented	|
 | text format used by mp.  The input SGML must be perfect or it won't	|
 | work.  Specifically, no end-tags may be omitted, the nesting of		|
 | elements must be correct, and there can be no attributes in the		|
 | element tags, although there may be character entities in the text.	|
 | No element tag is broken at the end of a line.						|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "config.h"
#include "stricmp.h"

extern void read_local (char *local_file);
extern void use_sgml_tags (char *tagset_name);
extern char *(*sgml_tag_of)(enum fgdc_keyword key);
extern enum fgdc_keyword (*key_of_sgml_tag) (char *sgml_tag);

extern char *iso8859_1[];

static int line_number;

/*----------------------------------------------------------------------*\
 | Dummy procedure definitions to avoid having to link in the world.	|
\*----------------------------------------------------------------------*/

struct item *allocate_item (int fail) {return (NULL);}
struct item *add_child (struct item *p) {return (NULL);}
void unescape_string (char *s) {}

/*----------------------------------------------------------------------*\
 | Stack operations
\*----------------------------------------------------------------------*/

#define MAX_DEPTH	1023

static enum fgdc_keyword stack [1+MAX_DEPTH];
static enum fgdc_keyword *top = NULL;
static enum fgdc_keyword *limit = NULL;

static void init (void) {
	limit = stack + MAX_DEPTH;
	top = stack;
	}

static void push (enum fgdc_keyword key) {
	if (top < limit) {
		top++;
		*top = key;
		}
	else {
		fprintf (stderr,"Error (line %3d): tag stack overflow on %s\n",line_number,text_of(key));
		exit (1);
		}
	}

static enum fgdc_keyword pop (void) {
	enum fgdc_keyword key = Wnull;
	if (top >= stack) key = *top--;
	else {
		fprintf (stderr,"Error (line %3d): tag stack underflow\n",line_number);
		exit (1);
		}
	return (key);
	}

static enum fgdc_keyword peek (int i) {
	enum fgdc_keyword *ptr = top - i;
	if (ptr >= stack) return (*ptr);
	else return (Wnull);
	}

/*----------------------------------------------------------------------*\
 | Functions to handle output buffer									|
\*----------------------------------------------------------------------*/

#define BUFFER_SIZE	8192

static char buffer [BUFFER_SIZE];
static char *buffer_ptr = buffer;

static int buffer_length (void) {
	return (buffer_ptr - buffer);
	}

static void clear_buffer (void) {
	buffer_ptr = buffer;
	*buffer_ptr = 0;
	}

static void output_buffer (void) {
	printf ("%s",buffer);
	}

static void add_string (char *string) {
	int n = strlen (string);

	if (buffer_ptr - buffer + n + 1 > BUFFER_SIZE) {
		output_buffer ();
		clear_buffer();
		}

	strcpy (buffer_ptr,string);
	buffer_ptr += n;
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

#define LINE_LEN	4096

enum tag_type_t {
	TAG_DECLARATION,
	TAG_START,
	TAG_END
	};

int main (int argc, char *argv[]) {
	int i;
	char *cfg_file = NULL;
	char *ext_file = NULL;
	char *input_file = NULL;
	FILE *in;
	void *opt;

	if (argc > 1) {
		for (i=1; i < argc; i++)
			if (memcmp (argv[i],"-c",2) == 0) {
				i++;
				cfg_file = argv[i];
				}
			else
				input_file = argv[i];

		/*--------------------------------------------------------------*\
		 | Handle configuration and extensions.							|
		\*--------------------------------------------------------------*/

		if (cfg_file) read_configuration (cfg_file);

		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"extensions")) {
				if (ext_file = text_of_option (opt)) read_local (ext_file);
				while (opt = find_next_option (opt,"extensions"))
					if (ext_file = text_of_option (opt)) read_local (ext_file);
				}

		use_sgml_tags ("ps8");
		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"sgml"))
				if (opt = find_option (opt,"tags"))
					if (stricmp (text_of_option (opt),"astm") == 0)
						use_sgml_tags ("ASTM");

		/*--------------------------------------------------------------*\
		\*--------------------------------------------------------------*/

		if (in = fopen (input_file,"r")) {
			int j,k;
			char line [LINE_LEN];
			char output [LINE_LEN];
			int indent = 0;
			char *d,*s,*t;
			enum fgdc_keyword key;

			init();

			line_number = 0;
			while (fgets (line,LINE_LEN,in)) {
				s = line + strlen (line) - 1;
				if (*s == '\n') *s-- = 0;
				if (*s == '\r') *s-- = 0;
				line_number++;

				s = line;
				d = output;
				while (*s) {

					/*--------------------------------------------------*\
					 | When you find a tag, determine whether it is a	|
					 | start tag, an end tag, or a declaration.			|
					\*--------------------------------------------------*/

					if (*s == '<') {
						enum tag_type_t tag_type = TAG_START;
						s++;
						if (*s == '/') {
							s++;
							tag_type = TAG_END;
							}
						else
							if (*s == '!') {
								s++;
								tag_type = TAG_DECLARATION;
								}

						/*----------------------------------------------*\
						 | Find the end of the tag.						|
						\*----------------------------------------------*/

						for (t=s; *t; t++)
							if (*t == '>') break;
						if (*t == '>') {
							*t++ = 0;
							if (tag_type != TAG_DECLARATION)

								/*--------------------------------------*\
								 | When you find a start tag, output	|
								 | the element name, increment indent,	|
								 | and push the element's key onto the	|
								 | tag stack.							|
								\*--------------------------------------*/

								if (key = key_of_sgml_tag (s)) {
									if (tag_type == TAG_START) {
										push (key);
										if (d > output) {
											*d = 0;

											/*--------------------------*\
											 | Don't complain if it's	|
											 | only blank space.		|
											\*--------------------------*/

											for (k=0; output[k]; k++)
												if (!isspace(output[k])) break;
											if (output[k])
												fprintf (stderr,"Error: extraneous text ignored at line %d: \"%s\"\n",line_number,output);
											d = output;
											}

										/*------------------------------*\
										 | An element name must be at	|
										 | the beginning of the line,	|
										 | so output the buffer if it	|
										 | isn't empty.					|
										\*------------------------------*/

										if (buffer_length()) {
											add_string ("\n");
											output_buffer();
											clear_buffer();
											}

										for (k=0; k < indent; k++) add_string ("  ");
										add_string (text_of(key));
										add_string (": ");
										indent++;
										}
									if (tag_type == TAG_END)
										if (key == pop()) {
											*d = 0;
											add_string (output);
											d = output;
											if (indent > 0) indent--;
											}
										else {
											enum fgdc_keyword token = peek(0);
											for (j=1; token != Wnull; j++)
												if (key != (token = peek(j)))
													fprintf (stderr,"Error (line %3d): expected end-tag for %s, got %s\n",
														line_number,text_of(key),text_of(token));
											exit (1);
											}
									}
								else {
									fprintf (stderr,"Error: unrecognized element %s at line %d\n",s,line_number);
									exit (1);
									}
							s = t;
							}
						else {
							fprintf (stderr,"Error: unmatched '<' on line %d\n",line_number);
							exit (1);
							}
						}
					else

						/*----------------------------------------------*\
						 | Do entity substitution if you can find one.	|
						 | Copy the byte if you cannot.					|
						\*----------------------------------------------*/

						if (*s == '&') {
							for (i=0; i < 256; i++)
								if (t = iso8859_1[i])
									if (memcmp (s+1,t,strlen(t)) == 0)
										if (*(s+1+strlen(t)) == ';') {
											*d++ = i;
											s += 2 + strlen (t);
											break;
											}
							if (i >= 256) *d++ = *s++;
							}
						else
							*d++ = *s++;
					}

				/*------------------------------------------------------*\
				 | End of line reached.  If you found any data, indent	|
				 | and output it as text.								|
				\*------------------------------------------------------*/

				if (d > output) {
					for (k=0; k < indent; k++) add_string ("  ");
					*d = 0;
					add_string (output);
					d = output;
					}
				if (buffer_length()) {
					add_string ("\n");
					output_buffer();
					clear_buffer();
					}
				}

			fclose (in);
			}
		else {
			fprintf (stderr,"Error: could not open input file %s\n",input_file);
			exit (1);
			}
		}
	else {
		fprintf (stderr,"stomp 2.2.0\n");
		fprintf (stderr,"Usage: %s [-t tag_file] sgml_file\n",argv[0]);
		}
	return (0);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
