/*----------------------------------------------------------------------*\
 | Program to parse formal metadata to determine whether it conforms	|
 | to the FGDC Content Standards for Digital Geospatial Metadata.		|
 |																		|
 | Input file must be encoded using the following scheme:				|
 |																		|
 |	Terms:																|
 |		tab := ASCII 9													|
 |		keyword := A sequence of bytes consisting of alphanumeric		|
 |			characters, the underscore, hyphen, and forward slash.  It	|
 |			is implied that this sequence is one of the formal names	|
 |			given to metadata entities in the standard.					|
 |			Examples:													|
 |				Citation:												|
 |				Identification_Information								|
 |				Data_Set_G-Polygon_Outer_G-Ring							|
 |				Range_of_Dates/Times									|
 |		value := A text string associated  with a keyword by the		|
 |			originator of the metadata.									|
 |																		|
 |	Arrangement:														|
 |		1.	Metadata files contain only ASCII characters; lines may be	|
 |			terminated using CR, LF, or CR followed by LF.				|
 |		2.	The number of characters per line is not limited.			|
 |		3.	Indentation is accomplished using tab characters only.		|
 |		4.	Blank lines may occur anywhere in the file.					|
 |		5.	Keywords are spelled out in the metadata file exactly as in	|
 |			the syntax rules of the metadata content standard.			|
 |		6.	A single colon or equal sign may follow each keyword.		|
 |		7.	Whitespace may occur between keyword and colon or equal		|
 |			sign, and may occur after the colon or equal sign.			|
 |		8.	Values are associated with a keyword in one of two ways:	|
 |			(a)	The value begins at the first nonblank following the	|
 |				keyword (or following colon or equal sign) and extends	|
 |				to the end of the line.									|
 |			(b)	The value begins on the line following the keyword.  It	|
 |				is indented one tab more than the keyword, i.e. there	|
 |				is exactly one more tab preceding the value than		|
 |				precede the keyword.									|
 |		9.	Values of compound types occur on successive lines using	|
 |			the same degree of indentation.								|
 |			Example:													|
 |				Citation_Information:									|
 |					Originator:											|
 |					Publication_Date:									|
 |					Publication_Time:									|
 |					Title:												|
 |					Type_of_Map:										|
 |					Serial_Information:									|
 |						Serial_Name:									|
 |						Issue_Identification:							|
 |					Publication_Information:							|
 |						Publication_Place:								|
 |						Publisher:										|
 |					Other_Citation_Details:								|
 |					Online_Linkage:										|
 |					Larger_Work_Citation:								|
 |		10.	Values specified using the format described in 8(b) may		|
 |			extend onto subsequent lines.  All lines used to contain	|
 |			the same value are indented to the same degree.				|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "item.h"
#include "config.h"
#include "stricmp.h"

static int allow_indented_text = 1;
static int keep_blank_lines = 1;

extern int unicode_of_utf8 (char **s);			/* in encoding.c */
extern char *get_character_encoding (void);		/* in xml.c */
static int utf8 = 0;

extern FILE *out;

/*----------------------------------------------------------------------*\
 | Code to manage large chunks of memory used for text buffers.			|
\*----------------------------------------------------------------------*/

struct buffer_list {
	struct buffer_list *next;
	struct item *root;
	void *buffer;
	};
static struct buffer_list *buffer_list = NULL;

void remember_buffer (struct item *root, void *address) {
	struct buffer_list *b,*bb;
	if (root && address) {
		if (b = (struct buffer_list *) malloc (sizeof (struct buffer_list))) {
			b->next = NULL;
			b->root = root;
			b->buffer = address;
			}
		if (buffer_list) {
			for (bb=buffer_list; bb->next; bb=bb->next);
			bb->next = b;
			}
		else
			buffer_list = b;
		}
	}

void forget_buffer (struct item *root) {
	struct buffer_list *b,*prev,*next;

	if (root) {
		for (b = buffer_list; b; b=b->next)
			if (b->root == root) break;
		if (b) {
			next = b->next;
			if (b == buffer_list) prev = NULL;
			else
				for (prev = buffer_list; prev; prev=prev->next)
					if (prev->next == b) break;

			if (prev) prev->next = next;
			else buffer_list = next;

			if (b->buffer) free (b->buffer);
			free (b);
			}
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

void unescape_string (char *s) {
	char *t;
	t = s;
	if (*s == '"') {
		s++;
		while (*s && *s != '"') {
			switch (*s) {
				case '\\':
					s++;
					switch (*s) {
						case 'a': *t++ = '\a'; break;
						case 'b': *t++ = '\b'; break;
						case 'f': *t++ = '\f'; break;
						case 'n': *t++ = '\n'; break;
						case 'r': *t++ = '\r'; break;
						case 't': *t++ = '\t'; break;
						case 'v': *t++ = '\v'; break;
						default:  *t++ = *s;   break;
						}
					s++;
					break;
				default:
					*t++ = *s++;
					break;
				}
			}
		}
	else
		while (*s) {
			switch (*s) {
				case '\\':
					s++;
					switch (*s) {
						case 'a': *t++ = '\a'; break;
						case 'b': *t++ = '\b'; break;
						case 'f': *t++ = '\f'; break;
						case 'n': *t++ = '\n'; break;
						case 'r': *t++ = '\r'; break;
						case 't': *t++ = '\t'; break;
						case 'v': *t++ = '\v'; break;
						default:  *t++ = *s;   break;
						}
					s++;
					break;
				default:
					*t++ = *s++;
					break;
				}
			}
	*t = 0;
	}

/*----------------------------------------------------------------------*\
 | Get metadata from a file by first reading in the entire file, then	|
 | decoding the contents.												|
\*----------------------------------------------------------------------*/

struct item *decode_text (char *dst) {
	char *end;
	char *b,*e,*s;
	struct item *p;
	struct item *first = NULL;
	struct item *current = NULL;
	int line_count;

	if (!dst) return (NULL);

	b = e = dst;
	end = dst + strlen (dst);

	line_count = 0;
	while (b < end) {

		/*----------------------------------------------*\
		 | Find the first CR, LF, or EOF.  Zero the		|
		 | end-of-line character(s) if present.			|
		\*----------------------------------------------*/

		while (e < end) {
			if (*e == '\r') {
				*e = 0;
				if (*(e+1) == '\n') {
					e++;
					*e = 0;
					}
				break;
				}
			else
				if (*e == '\n') {
					*e = 0;
					break;
					}
			e++;
			}

		/*----------------------------------------------*\
		 | Clear trailing whitespace from the line.		|
		\*----------------------------------------------*/

		s = e;
		while (s >= b)
			if (*s == 0) s--;
			else
				if (isspace (*s)) *s-- = 0;
				else break;

		/*----------------------------------------------*\
		 | b points to the beginning of the line.		|
		\*----------------------------------------------*/

		if (*b || keep_blank_lines)
			if (p = allocate_item (1)) {
				p->original = 1;
				p->len = s - b + 1;
				p->d = b;
				p->line_number = 1 + line_count;
				if (current) {
					current->next = p;
					p->prev = current;
					}
				else
					first = p;
				current = p;
				}

		b = e + 1;
		line_count++;
		}

	return (first);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

char *read_text (char *name) {
	long n;
	FILE *fp;
	char *dst,*end;
	long size;

	if (!name) return (NULL);

	dst = NULL;
	if (fp = fopen (name,"rb")) {
		if (fseek (fp,0L,SEEK_END) == 0) {
			size = 1L + ftell (fp);
			if (size > 1L) {
				if (dst = (char *) malloc (size)) {
					rewind (fp);
					n = fread (dst,1,size,fp);
					end = dst + n;
					*end = 0;
					}
				else
					fprintf (out,"Error: could not allocate space for file\n");
				}
			else
				fprintf (out,"Error: Input file is empty.\n");
			}
		fclose (fp);
		}
	else
		fprintf (out,"Error: could not open input file %s\n",name);
	return (dst);
	}

/*----------------------------------------------------------------------*\
 | Rather crude function that attempts to fix indented textual values.	|
 | Specifically, this routine looks for textual values that are nested	|
 | beneath other textual values and makes these siblings rather than	|
 | descendants.  It does not attempt to reclassify standard elements	|
 | as textual values; instead, it issues a warning in this case.		|
\*----------------------------------------------------------------------*/

void equalize_indented_scalars (struct item *p) {
	struct item *q;

	while (p) {
		if (p->child)
			if (p->key == Wunknown || p->key == Wblank) {
				if (p->child->key == Wunknown || p->child->key == Wblank) {

					/*--------------------------------------------------*\
					 | Both p and p->child are Wunknown or Wblank.		|
					 |													|
					 | Convert p->child and all of its siblings to be	|
					 | siblings, not children, of p.					|
					\*--------------------------------------------------*/

					p->child->prev = p;			/* was NULL */

					p->child->parent = p->parent;	/* was p */

					for (q=p->child; q->next; q=q->next) {
						q->parent = p->parent;	/* was p */
						q->indent = p->indent;	/* was 1 + p->indent */
						if (q->child) equalize_indented_scalars (q->child);
						}

					q->parent = p->parent;		/* was p */
					q->next = p->next;			/* was NULL */
					q->indent = p->indent;		/* was 1 + p->indent */
					if (p->next) p->next->prev = q;
					p->next = p->child;
					p->child = NULL;
					}
				else {

					/*--------------------------------------------------*\
					 | p is of type Wunknown but p->child is something	|
					 | else.  Don't second-guess at this point, but do	|
					 | issue a warning that some information might be	|
					 | lost in subsequent handling of this line.		|
					\*--------------------------------------------------*/

					fprintf (out,"Warning (line %d): element %s found in textual value, some information may be lost.\n",
						p->child->line_number,text_of(p->child->key));
					}
				}
			else
				equalize_indented_scalars (p->child);

		p = p->next;
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static struct item *decode_tree (struct item *first) {
	int i,j,k;
	char *s,*t;
	struct item *p,*q,*r;
	int item_count;
	struct item **list;

	if (!first) return (NULL);

	item_count = 0;
	for (p=first; p; p=p->next) {
		for (s=p->d,i=0; *s && isspace (*s); s++,i++);
		p->indent = i;
		p->d = s;
		p->key = key_of (s);
		t = text_of (p->key);

		/*--------------------------------------------------------------*\
		 | If the current item can have a child on this line, determine	|
		 | whether it does, and store the child's information.  The		|
		 | rules prohibit such a child from having children of its own.	|
		\*--------------------------------------------------------------*/

		if (p->key != Wunknown && p->key != Wblank) {
			s = p->d + strlen (t);
			if (isspace (*s) || *s == ':' || *s == '=') {
				*s++ = 0;
				p->len = strlen (p->d);
				}
			while (*s && isspace (*s)) s++;
			if (*s == ':' || *s == '=') s++;
			while (*s && isspace (*s)) s++;
			if (*s) {
				q = allocate_item (1);
				q->d = s;
				q->len = strlen (s);
				q->original = 1;
				q->indent = p->indent + 1;
				q->line_number = p->line_number;
				q->key = Wunknown;
				q->parent = p;
				p->child = q;
				}
			}
		item_count++;
		}

	/*------------------------------------------------------------------*\
	 | Estimate the proper indentation to be assigned to blank lines.	|
	 | At this point the metadata are stored as one doubly-linked list,	|
	 | so you can go forward and backward using p->next and p->prev.	|
	\*------------------------------------------------------------------*/

	for (p=first; p; p=p->next)
		if (p->key == Wblank) {

			/* Find the most recent element that is not blank */

			for (r=p->prev; r; r=r->prev)
				if (r->key != Wblank) break;

			/* Find the next element that is not blank */

			for (q=p->next; q; q=q->next)
				if (q->key != Wblank) break;

			/*----------------------------------------------------------*\
			 | Blanks should be indented like the more indented of		|
			 | (previous nonblank) and (next nonblank).  If either of	|
			 | these nodes does not exist, the proper indentation is	|
			 | whichever does exist.  If neither exist, the file		|
			 | contains only blank lines, so zero will do.				|
			\*----------------------------------------------------------*/

			if (r && q)
				p->indent = ((r->indent > q->indent) ? r->indent : q->indent);
			else
				if (r) p->indent = r->indent;
				else
					if (q) p->indent = q->indent;
					else
						p->indent = 0;
			}

	/*------------------------------------------------------------------*\
	 | Figure out the topology of the file.  Create a list of items		|
	 | as an array of pointers.  For each item, determine whether it	|
	 | has a child and what its parent is.								|
	\*------------------------------------------------------------------*/

	if (list = (struct item **) malloc (item_count * sizeof(struct item *))) {
		memset (list,0,item_count*sizeof(struct item *));
		i = 0;
		for (p=first; p; p=p->next) list[i++] = p;

		for (i=0; i < item_count-1; i++) {
			p = list[i];
			q = list[i+1];

			/*----------------------------------------------------------*\
			 | The relationship between successive items in the file is	|
			 | determined by comparing their levels of indentation.		|
			\*----------------------------------------------------------*/

			j = q->indent - p->indent;

			/*----------------------------------------------------------*\
			 | If q is indented more than p, q is a child of p, and p	|
			 | is parent of q.											|
			 |															|
			 | Modified 30-Jan-1997 to not discard scalar child already	|
			 | in p (can only happen as text following element name.	|
			\*----------------------------------------------------------*/

			if (j > 0) {
				if (allow_indented_text &&
					p->key == Wunknown && q->key == Wunknown)
					q->parent = p->parent;
				else {
					if (p->child && (q->key == Wunknown || q->key == Wblank)) {

						/*----------------------------------------------*\
						 | Modified 20-Feb-1998 to handle cases where a	|
						 | string immediately follows the name and a	|
						 | blank line follows them.  If the next non-	|
						 | blank is Wunknown, keep the string that was	|
						 | placed on the same line as the element.  If	|
						 | the next non-blank is a recognized element,	|
						 | the string after the element name was not	|
						 | permitted; discard it with a warning.		|
						\*----------------------------------------------*/

						if (q->key == Wblank) {
							/* Point r at the next non-blank element */
							r = NULL;
							for (k=i+1; k < item_count; k++) {
								r = list[k];
								if (r->key != Wblank) break;
								}
							/* if r is an element, discard the text */
							if (r)
								if (r->key != Wunknown && r->key != Wblank) {
									fprintf (out,"Warning (line %d): Extraneous text following element %s ignored\n",
										p->line_number,
										text_of(p->key)
										);
									p->child = q;
									q->prev = NULL;
									}
								else {
									/* otherwise include the text */
									for (r=p->child; r->next; r=r->next);
									r->next = q;
									q->prev = r;
									}
							}
						else {
							/* q->key == Wunknown */
							for (r=p->child; r->next; r=r->next);
							r->next = q;
							q->prev = r;
							}
						}
					else {
						if (p->child)
							fprintf (out,"Warning (line %d): Extraneous text following element %s ignored.  Check indentation of line %d.\n",
								p->line_number,
								text_of(p->key),
								q->line_number
								);
						p->child = q;
						q->prev = NULL;
						}
					q->parent = p;
					p->next = NULL;
					}
				}

			/*----------------------------------------------------------*\
			 | If q is indented the same as p, q is just the next part	|
			 | of the compound type (or text field) of which p is a		|
			 | part.  The parent of q is that of p.						|
			\*----------------------------------------------------------*/

			if (j == 0) {
				q->parent = p->parent;
				}

			/*----------------------------------------------------------*\
			 | If q is indented less than p, q is the next element in a	|
			 | compound type begun earlier.  Find the previous element	|
			 | by searching backwards through the list for the nearest	|
			 | element q that was indented the same as or less than q;	|
			 | store q into its next pointer, set the parent of q to	|
			 | that of r, and sever the link from p to q.				|
			\*----------------------------------------------------------*/

			if (j < 0) {
				for (k=i; k >= 0; k--)
					if (list[k]->indent <= q->indent) {
						r = list[k];
						if (r->indent < q->indent)
							fprintf (out,"Warning: ambiguous indentation at line %d\n",q->line_number);
						if (r->next == NULL) {
							r->next = q;
							q->prev = r;
							q->parent = r->parent;
							p->next = NULL;
							}
						else {
							struct item *rr;
							for (rr=r; rr->next && rr->next != q; rr=rr->next);
							rr->next = q;
							q->prev = rr;
							q->parent = rr->parent;
							if (rr != p) p->next = NULL;
							}
						break;
						}

				/*------------------------------------------------------*\
				 | If you fell through that loop, there is an error in	|
				 | the file's indentation that cannot easily be over-	|
				 | looked.  This indicates that a line somewhere in the	|
				 | middle of the file is indented less than all other	|
				 | lines in the file; the structure is unclear.			|
				\*------------------------------------------------------*/

				if (k < 0)
					fprintf (out,"Error: too little indentation at line %d\n",q->line_number);
				}
			}

		/*--------------------------------------------------------------*\
		 | Traverse the tree, looking for situations in which scalar	|
		 | values have children that are themselves scalar values; in	|
		 | such cases, make them siblings, not children.				|
		\*--------------------------------------------------------------*/

		if (allow_indented_text) equalize_indented_scalars (first);

		free (list);
		}
	else {
		fprintf (out,"Error: could not allocate array of pointers to items\n");
		exit (1);
		}
	return (first);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

struct item *parse_text (char *input_file, char **text_buffer) {
	void *opt;
	struct item *p;
	char *b;

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"indent"))
			if (stricmp (text_of_option (opt),"strict") == 0)
				allow_indented_text = 0;

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"blanks"))
			if (stricmp (text_of_option (opt),"ignore") == 0)
				keep_blank_lines = 0;

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"encoding"))
			if (stricmp (text_of_option (opt),"UTF-8") == 0)
				set_character_encoding ("UTF-8");

	if (b = read_text (input_file))
		if (text_buffer) *text_buffer = b;

	return (decode_tree (decode_text (b)));
	}

struct item *parse_string (char *string) {
	void *opt;

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"indent"))
			if (stricmp (text_of_option (opt),"strict") == 0)
				allow_indented_text = 0;

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"blanks"))
			if (stricmp (text_of_option (opt),"ignore") == 0)
				keep_blank_lines = 0;

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"encoding"))
			if (stricmp (text_of_option (opt),"UTF-8") == 0)
				set_character_encoding ("UTF-8");

	/* Skip forward past any newline characters */

	while (*string == '\n') string++;

	return (decode_tree (decode_text (string)));
	}

/*----------------------------------------------------------------------*\
 | Keep track of those elements the user has asked be omitted from the	|
 | output file.  Their keys are stored in a static array of size 1024,	|
 | which is consulted whenever an element is to be output.				|
\*----------------------------------------------------------------------*/

static enum fgdc_keyword omit_key[1024];
static int omit_key_count = -1;

static int omit (enum fgdc_keyword key) {
	int i,result = 0;
	if (omit_key_count > 0)
		for (i=0; i < omit_key_count; i++)
			if (key == omit_key[i]) {
				result = 1;
				break;
				}
	return (result);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static int skip_top_level = 1;	/* Set to 0 to output top-level Metadata */
static char *one_indent = "  ";
static char *element_prefix = "";

static FILE *fp = NULL;

static void write_text_item (struct item *p, int n) {
	int i,k;
	char *s;
	if (!p) return;
	if (omit (p->key)) return;
	if (p->key == Wblank) fprintf (fp,"\n");
	else
		if (p->key == Wunknown) {
			for (i=0; i < n; i++) fprintf (fp,one_indent);
			if (utf8 == 0) fprintf (fp,"%s\n",p->d);
			else {
				s = p->d;
				while (*s)
					if ((k = unicode_of_utf8 (&s)) < 256) fprintf (fp,"%c",k);
					else { /* skip unicode characters that cannot be rendered as ISO-8859-1 */ }
				fprintf (fp,"\n");
				}
			}
		else {
			struct item *q;

			for (i=0; i < n; i++) fprintf (fp,one_indent);
			if (*element_prefix) fprintf (fp,"%s",element_prefix);
			fprintf (fp,"%s:",text_of(p->key));

			/*----------------------------------------------------------*\
			 | Check for two special cases: a single scalar child or a	|
			 | scalar child followed only by one blank line.  Render	|
			 | those cases in the simplest format.						|
			\*----------------------------------------------------------*/

			if (p->child)
				if (p->child->key == Wunknown)
					if (p->child->next == NULL) {
						if (utf8 == 0) fprintf (fp," %s\n",p->child->d);
						else {
							s = p->child->d;
							fprintf (fp," ");
							while (*s)
								if ((k = unicode_of_utf8 (&s)) < 256) fprintf (fp,"%c",k);
								else { /* skip unicode characters that cannot be rendered as ISO-8859-1 */ }
							fprintf (fp,"\n");
							}
						return;
						}
					else
						if (p->child->next->key == Wblank &&
							p->child->next->next == NULL) {
							if (utf8 == 0) fprintf (fp," %s\n",p->child->d);
							else {
								s = p->child->d;
								fprintf (fp," ");
								while (*s)
									if ((k = unicode_of_utf8 (&s)) < 256) fprintf (fp,"%c",k);
									else { /* skip unicode characters that cannot be rendered as ISO-8859-1 */ }
								fprintf (fp,"\n");
								}
							fprintf (fp,"\n");
							return;
							}

			/*----------------------------------------------------------*\
			 | If there are no children, more than two scalar children,	|
			 | or non-scalar children, render them indented.			|
			\*----------------------------------------------------------*/

			fprintf (fp,"\n");
			for (q=p->child; q; q=q->next) write_text_item (q,n+1);
			}
	}

void write_text (char *output_file, struct item *first) {
	int i = 0;
	struct item *p;
	void *opt;
	char *s;

	if (first == NULL) return;

	if (opt = find_option (NULL,"output"))
		if (opt = find_option (opt,"text"))
			if (opt = find_option (opt,"top_level"))
				if (s = text_of_option (opt))
					if (stricmp (s,"skip") != 0)
						skip_top_level = 0;

	if (opt = find_option (NULL,"output"))
		if (opt = find_option (opt,"text"))
			if (opt = find_option (opt,"indent")) {
				char *s = text_of_option (opt);
				if (*s) unescape_string (s);
				one_indent = s;
				}

	if (opt = find_option (NULL,"output"))
		if (opt = find_option (opt,"text"))
			if (opt = find_option (opt,"prefix")) {
				char *s = text_of_option (opt);
				if (*s) unescape_string (s);
				element_prefix = s;
				}

	if (opt = find_option (NULL,"output"))
		if (opt = find_option (opt,"text"))
			if (opt = find_option (opt,"wrap")) {
				int width;
				char *t;
				char *s = text_of_option (opt);
				if (*s) unescape_string (s);
				width = (int) strtol (s,&t,0);
				if (t == s) width = 0;
				if (width > 0) wrap_subtree (first,width);
				}

	/*------------------------------------------------------------------*\
	 | For up to 1024 keys, the user can specify elements that will not	|
	 | be included in the output.										|
	\*------------------------------------------------------------------*/

	if (omit_key_count == -1)
		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"text"))
				if (opt = find_option (opt,"omit"))
					do {
						enum fgdc_keyword key;
						s = text_of_option (opt);
						if (*s) unescape_string (s);
						if ((key = key_of (s)) != Wunknown) {
							if (omit_key_count == -1) omit_key_count = 0;
							if (omit_key_count < 1024)
								omit_key[omit_key_count++] = key;
							}
						} while (opt = find_next_option (opt,"omit"));

	/*------------------------------------------------------------------*\
	 | If we read XML, the encoding is going to be UTF-8, but we want	|
	 | the text output to be ISO-8859-1 if possible.					|
	\*------------------------------------------------------------------*/

	if (strcmp (get_character_encoding(),"UTF-8") == 0) utf8 = 1;

	/* Create the output file and write the metadata */

	if (fp = fopen (output_file,"w")) {
		p = first;
		if (skip_top_level && p->key == WMetadata) p = p->child;
		while (p) {
			write_text_item (p,i);
			p=p->next;
			}
		fclose (fp);
		}
	else {
		fprintf (out,"Error: could not create output file %s\n",output_file);
		return;
		}
	}

/*----------------------------------------------------------------------*\
 | Code to maintain a list of item addresses, tied to line numbers in	|
 | the output window of Xtme and Tkme.									|
\*----------------------------------------------------------------------*/

static struct item **item_list = NULL;
static int item_list_count = 0;
static int item_list_limit = 0;

static void remember_item (struct item *p) {
	if (item_list_count >= item_list_limit) {
		struct item **q;
		if (q = (struct item **) realloc (item_list,(item_list_limit + 1024)* sizeof(struct item *))) {
			item_list_limit += 1024;
			item_list = q;
			}
		else return;
		}
	item_list [item_list_count++] = p;
	}

struct item *item_of (int line_number) {
	if (line_number < 0) return (NULL);
	if (line_number >= item_list_count) return (NULL);

	return (item_list [line_number]);
	}

/*----------------------------------------------------------------------*\
 | Encode the subtree beginning at p in a dynamically-allocated string.	|
\*----------------------------------------------------------------------*/

static int number_lines = 0;
static int line_number = 0;

static char *encode_line_number (char *s) {
	line_number++;
	sprintf (s,"%5d ",line_number);
	s += 6;
	return (s);
	}

static int length_of (struct item *p, int n) {
	int k = 0;

	if (p)
		if (p->key == Wblank) {
			if (number_lines) k += 6;
			k++;
			}
		else {
			if (number_lines) k += 6;
			k += n*strlen (one_indent);
			if (p->key == Wunknown) k += 1 + strlen (p->d);
			else {
				struct item *q;

				k += 1 + strlen(text_of(p->key));
				if (p->child)
					if (p->child->key == Wunknown && p->child->next == NULL) {
						k += 2 + strlen (p->child->d);
						return (k);
						}
				k++;
				for (q=p->child; q; q=q->next) k += length_of (q,n+1);
				}
			}
	return (k);
	}

static char *encode_item (struct item *p, char *s, int n) {
	int i,j;

	if (p)
		if (p->key == Wblank) {

			if (number_lines) s = encode_line_number (s);

			/* Encode blank line */

			remember_item (p);
			*s++ = '\n';
			*s = 0;
			}
		else {

			if (number_lines) s = encode_line_number (s);

			/* Encode indentation */

			j = strlen (one_indent);
			for (i=0; i < n; i++) {
				memcpy (s,one_indent,j);
				s += j;
				}

			if (p->key == Wunknown) {

				/* Encode scalar value */

				j = strlen (p->d);
				memcpy (s,p->d,j);
				s += j;
				remember_item (p);
				*s++ = '\n';
				*s = 0;
				}
			else {
				struct item *q;
				char *t;

				/* Encode keyword */

				t = text_of(p->key);
				j = strlen (t);
				memcpy (s,t,j);
				s += j;
				*s++ = ':';

				/*----------------------------------------------------------*\
				 | Check for two special cases: a single scalar child or a	|
				 | scalar child followed only by one blank line.  Render	|
				 | those cases in the simplest format.						|
				\*----------------------------------------------------------*/

				if (p->child)
					if (p->child->key == Wunknown)
						if (p->child->next == NULL) {
							*s++ = ' ';
							t = p->child->d;
							j = strlen (t);
							memcpy (s,t,j);
							s += j;
							remember_item (p);
							*s++ = '\n';
							*s = 0;
							return (s);
							}
						else
							if (p->child->next->key == Wblank &&
								p->child->next->next == NULL) {
								*s++ = ' ';
								t = p->child->d;
								j = strlen (t);
								memcpy (s,t,j);
								s += j;
								remember_item (p);
								*s++ = '\n';
								if (number_lines) s = encode_line_number (s);
								remember_item (p);
								*s++ = '\n';
								*s = 0;
								return (s);
								}

				/*----------------------------------------------------------*\
				 | If there are no children, more than two scalar children,	|
				 | or non-scalar children, render them indented.			|
				\*----------------------------------------------------------*/

				remember_item (p);
				*s++ = '\n';
				*s = 0;
				for (q=p->child; q; q=q->next) s = encode_item (q,s,n+1);
				}
			}
	return (s);
	}

char *encode_text (struct item *p) {
	int n;
	char *string,*s;

	number_lines = 0;

	item_list_count = 0;
	n = length_of (p,0);
	if (string = (char *) malloc (1 + n)) {
		s = encode_item (p,string,0);
		if (s - string > n) fprintf (out,"Aaaack!  You blew it, Schweitzer!\nStand by for catastrophic failure!\n");
		}
	return (string);
	}

char *encode_text_numbered (struct item *p) {
	int n;
	char *string,*s;

	number_lines = 1;
	line_number = 0;

	item_list_count = 0;
	n = length_of (p,0);
	if (string = (char *) malloc (1 + n)) {
		s = encode_item (p,string,0);
		if (s - string > n) fprintf (out,"Aaaack!  You blew it, Schweitzer!\nStand by for catastrophic failure!\n");
		}
	return (string);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

