/*----------------------------------------------------------------------*\
 | Tkme: an editor for formal metadata based on Xtme and Tcl/Tk			|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <tcl.h>
#include <tk.h>
#include "appinit.h"

#include "keyword.h"
#include "item.h"
#include "config.h"
#include "actions.h"
#include "stricmp.h"
#include "local.h"

#define PATH_SEPARATOR  '/'

extern int Et_Init (int argc, char **argv);

/* extern void read_local (char *local_file); */
extern void read_tips (char *tip_file);
extern char *tip_text (char *parent_name, char *child_name);
extern struct item *parse_text (char *file_name, char **buffer);
extern struct item *parse_sgml (char *file_name);
extern struct item *parse_xml (char *file_name);
extern struct item *parse_string (char *string);
extern struct item *parse_xml_string (char *string);
extern int check_syntax (struct item *p);
extern void write_text (char *file_name, struct item *p);
extern void write_sgml (char *file_name, struct item *p);
extern void write_xml (char *file_name, struct item *p);
extern char *encode_text (struct item *p);
extern char *encode_text_numbered (struct item *p);
extern struct item *item_of (int n);

extern void read_help (char *help_file);
extern char *advice_about (enum fgdc_keyword key);
extern char *about_text[];

/*----------------------------------------------------------------------*\
 | Public data															|
\*----------------------------------------------------------------------*/

struct item *root = NULL;
struct item *current = NULL;
char *one_indent = "  ";
FILE *out = NULL;

/*----------------------------------------------------------------------*\
 | Static data															|
\*----------------------------------------------------------------------*/

static char cmdline_file [FILENAME_MAX] = "";
static char name_of_file [FILENAME_MAX] = "";
static int node_count = 0;
static char *language = NULL;
static char *profile_name = NULL;

static char *default_tip = "Tkme: Tk Metadata Editor by Peter Schweitzer (USGS)";

static char message[512];
static struct item *clipboard = NULL;
static char *clipboard_text = NULL;

static char *output_string = NULL;
static int output_string_length = 0;

/*----------------------------------------------------------------------*\
 | Handle character encoding
\*----------------------------------------------------------------------*/

static char character_encoding[256] = "ISO-8859-1";

void set_character_encoding (char *new_value) {
	if (strlen (new_value) < 256)
		strcpy (character_encoding,new_value);
	else {
		memcpy (character_encoding,new_value,255);
		character_encoding[255] = 0;
		}
	}

char *get_character_encoding (void) {
	return (character_encoding);
	}

/*----------------------------------------------------------------------*\
 | Create an error message box using Tk									|
\*----------------------------------------------------------------------*/

static void error_message (Tcl_Interp *interp, char *msg) {
	char cmd [4096];
	char *s,*d,*end;

	if (interp && msg) {
		strcpy (cmd,"tk_messageBox -type ok -icon error -title \"Tkme error\" -message \"");

		s = msg;
		d = cmd + strlen (cmd);
		end = cmd + 4095;
		while (*s && d < end) {
			if (*s == '\"') *d++ = '\\';
			*d++ = *s++;
			}
		*d++ = '\"';
		*d = 0;
		Tcl_Eval (interp,cmd);
		}
	}

/*----------------------------------------------------------------------*\
 | Here the original item structure created by the text parser is fixed |
 | so that the indent member holds the level of nesting, not the number |
 | of whitespaces that preceded the element name in the input file.		|
\*----------------------------------------------------------------------*/

static void adjust_indent (struct item *p, int n) {
	struct item *q;

	if (p) {
		if (p->key != Wblank) p->indent = n;
		for (q=p->child; q; q=q->next) adjust_indent (q,n+1);
		}
	}

/*----------------------------------------------------------------------*\
 | Build a linear array of pointers to items that will be displayed.	|
\*----------------------------------------------------------------------*/

static struct item **line = NULL;
static int line_limit = 0;
static int line_count = 0;

static void build_list (struct item *p) {
	struct item *q;

	if (p) {
		if (p->key != Wunknown && p->key != Wblank && p->indent >= 0) {

			/*----------------------------------------------------------*\
			 | The line list is built as a dynamically-allocated array	|
			\*----------------------------------------------------------*/

			if (line_limit == 0) {
				line = (struct item **) calloc (1024,sizeof(struct item *));
				line_limit = 1024;
				line_count = 0;
				}

			line [line_count++] = p;

			if (line_count >= line_limit) {
				line = (struct item **) realloc (line,(line_limit + 1024)*sizeof(struct item *));
				memset (line+line_limit,0,1024*sizeof(struct item *));
				line_limit += 1024;
				}

			for (q=p->child; q; q=q->next) build_list (q);
			}
		}
	}

/*----------------------------------------------------------------------*/

typedef char * String;
static String *element = NULL;
static int element_count = 0;

static void build_element_list (void) {
	struct item *q;

	if (element_count == 0) {
		if (element = (String *) calloc (line_count,sizeof(String *))) {
			int i,n;
			char *s;

			for (i=0; i < line_count; i++) {
				n = line[i]->indent * 2;
				s = text_of(line[i]->key);
				element[i] = (String) malloc (1 + n + strlen(s));
				memset (element[i],' ',n);
				strcpy (element[i]+n,s);
				if (q = line[i]->child)
					if (q->key != Wunknown &&
						q->key != Wblank &&
						q->indent < 0) *element[i] = '+';
				}
			}
		element_count = line_count;
		}
	}

/*----------------------------------------------------------------------*/

static void rebuild_lists (void) {
	int i;
	if (element) {
		for (i=0; i < element_count; i++)
			if (element[i]) {
				free (element[i]);
				element[i] = NULL;
				}
		free (element);
		element = NULL;
		element_count = 0;
		}
	line_count = 0;
	build_list (root);
	build_element_list ();
	}

/*----------------------------------------------------------------------*\
 | parse_text_cmd
\*----------------------------------------------------------------------*/

int parse_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *input_file;
	char *s;
	int i;
	Tcl_Obj *obj;
	char *text_buffer = NULL;

	if (argc > 1) {
		int parsed_as_text = 0;
		input_file = argv[1];

		/*------------------------------------------------------------------*\
		 | Parse the input file.  Check the extension to see how.			|
		\*------------------------------------------------------------------*/

		root = NULL;
		if (input_file)
			if (s = strrchr (input_file,'.'))
				if (stricmp (s,".sgml") == 0 || stricmp (s,".sgm") == 0)
					root = parse_sgml (input_file);
				else
					if (stricmp (s,".xml") == 0)
						root = parse_xml (input_file);

		if (!root) {
			root = parse_text (input_file, &text_buffer);
			parsed_as_text = 1;
			}

		/*------------------------------------------------------------------*/

		if (root)
			strcpy (name_of_file,input_file);
		else {
			struct item *p;

			/*--------------------------------------------------------------*\
			 | If for whatever reason there are no nodes to begin with,		|
			 | create a top node of type Metadata to hang everything on.	|
			\*--------------------------------------------------------------*/

			p = allocate_item (1);
			p->key = WMetadata;
			p->d = text_of (p->key);
			p->original = 1;
			p->len = strlen (p->d);
			p->indent = 0;
			root = p;
			if (input_file)
				strcpy (name_of_file,input_file);
			else
				strcpy (name_of_file,"Untitled");
			}

		/*--------------------------------------------------------------*\
		 | A couple of special conditions can occur in text files; they	|
		 | need to be corrected only if we parsed indented text.		|
		\*--------------------------------------------------------------*/

		if (root && parsed_as_text) {

			/*--------------------------------------------------------------*\
			 | The parser classifies all nodes by their keys.  If a valid	|
			 | element name appears in a textual value (at the beginning of	|
			 | a line), the parser classifies that node according to the	|
			 | element name.  But the node should be treated as text.  To	|
			 | correct this problem requires an understanding of the syntax	|
			 | of the standard.  Consequently it is necessary to minimally	|
			 | check the syntax to catch and recode elements that are found	|
			 | in textual values.											|
			 |																|
			 | Mixed content in XML input cannot be reliably fixed up.  If	|
			 | an element with mixed content is encountered in XML, this	|
			 | should generate a Tcl error message.							|
			\*--------------------------------------------------------------*/

			if (check_syntax (root) != 0) {
				Tcl_SetResult (interp,"Error: This file cannot be read.  Mixed content is not permitted in elements.  Structural information would be lost if you continued.",TCL_STATIC);
				forget_buffer (root);
				delete_item (root);
				root = NULL;
				current = NULL;
				return (TCL_ERROR);
				}

			/*--------------------------------------------------------------*\
			 | If the top node isn't Metadata, then you won't be able to	|
			 | add any of the immediate children of Metadata to the record.	|
			 | To solve this problem, create a Metadata node at the top if	|
			 | one is not present.											|
			 |																|
			 | More ecumenically, we should allow people to nest Metadata   |
			 | in their extensions.  So we only plant a Metadata node at	|
			 | the root if there isn't a Metadata node somewhere in the		|
			 | tree; it could be buried several levels down.				|
			\*--------------------------------------------------------------*/

			if (find_key (root,WMetadata) == NULL) {
				struct item *p,*q;

				p = allocate_item (1);
				p->key = WMetadata;
				p->d = text_of (p->key);
				p->original = 1;
				p->len = strlen (p->d);
				p->indent = 0;
				if (root->indent == p->indent)
					for (q=root; q; q=q->next) indent (q);
				p->child = root;
				root->parent = p;
				root = p;
				for (q=root->child; q; q=q->next) q->parent = root;
				}
			}

		remember_buffer (root,text_buffer);

		node_count = item_count (root);
		adjust_indent (root,0);
		build_list (root);
		build_element_list ();
		current = root;

#ifdef DEBUG
		/* Debug output */

		if (fp = fopen ("tkme.out","w")) {
			for (i=0; i < element_count; i++)
				fprintf (fp,"%s\n",element[i]);
			fclose (fp);
			fprintf (stderr,"File parsed successfully; %d elements\n",element_count);
			}
		else
			fprintf (stderr,"Error: could not open output file tkme.out\n");
#endif

		/* Actual output */

		Tcl_ResetResult (interp);
		obj = Tcl_NewListObj (0,NULL);
		for (i=0; i < element_count; i++) {
			Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
			}
		Tcl_SetObjResult (interp,obj);

		return (TCL_OK);
		}
	else {
		Tcl_SetResult (interp,"Error: parse_text called without input_file",TCL_STATIC);
		return (TCL_ERROR);
		}
	}

/*----------------------------------------------------------------------*\
 | Get the list of elements
\*----------------------------------------------------------------------*/

int get_list_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	Tcl_Obj *obj;

	rebuild_lists();

	Tcl_ResetResult (interp);
	obj = Tcl_NewListObj (0,NULL);
	for (i=0; i < element_count; i++) {
		Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
		}
	Tcl_SetObjResult (interp,obj);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | write_text_cmd
\*----------------------------------------------------------------------*/

int write_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	char *s;
	struct item *p;
	char *name,*ext;

	if (argc > 2) {
		i = (int) strtol (argv[1],&s,0);
		if (s == argv[1]) {
			sprintf (message,"write_text expects a number first, got %s",argv[1]);
			Tcl_SetResult (interp,message,TCL_STATIC);
			return (TCL_ERROR);
			}
		name = argv[2];
		s = name + strlen (name) - 1;
		while (s > name && *s != '.') s--;
		if (*s == '.') ext = s; else ext = NULL;
		if (line_count > 0)
			if (p = line[i])
				if (ext)
					if (stricmp (ext,".sgml") == 0 || stricmp (ext,".sgm") == 0)
						write_sgml (name,p);
					else
						if (stricmp (ext,".xml") == 0)
							write_xml (name,p);
						else
							write_text (name,p);
				else
					write_text (name,p);

		Tcl_SetResult (interp,NULL,TCL_STATIC);
		return (TCL_OK);
		}
	else
		Tcl_SetResult (interp,"write_text expects a number and a file name",TCL_STATIC);
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | tip_text_cmd
\*----------------------------------------------------------------------*/

int tip_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *s = NULL;
	char *menu = NULL;
	char *choice = NULL;

	if (argc > 2) {
		menu   = argv[1];
		choice = argv[2];

		if (menu == NULL || choice == NULL) {
			sprintf (message,"tip_text_cmd (menu = 0x%X, choice = 0x%X)",menu,choice);
			Tcl_SetResult (interp,message,TCL_STATIC);
			return (TCL_ERROR);
			}

		if (stricmp (menu,"File") == 0) {
			if (stricmp (choice,"Open") == 0)
				s = "Start work on an existing metadata record";
			if (stricmp (choice,"Save") == 0)
				s = "Writes current file to disk";
			if (stricmp (choice,"Save As") == 0)
				s = "Prompts user for a new name, then writes file";
			if (stricmp (choice,"Close") == 0)
				s = "Saves file and exits program";
			if (stricmp (choice,"Configure") == 0)
				s = "Specify and read a configuration file";
			if (stricmp (choice,"Recent") == 0)
				s = "Select one of your most recently opened files";
			if (stricmp (choice,"Quit") == 0)
				s = "Exits, prompting for confirmation if file changed.";
			}
		if (stricmp (menu,"Edit") == 0) {
			if (stricmp (choice,"Cut") == 0)
				s = "Detaches a subtree, remembering it for paste";
			if (stricmp (choice,"Copy") == 0)
				s = "Remembers a subtree for later paste";
			if (stricmp (choice,"Paste") == 0)
				s = "Places cut or copied info into the metadata as text or as a subtree";
			if (stricmp (choice,"Clear") == 0)
				s = "Removes the text belonging to a subtree";
			if (stricmp (choice,"Duplicate") == 0)
				s = "Makes and places a copy of the current element";
			if (stricmp (choice,"Swap") == 0)
				s = "Switches the order of the current subtree with the preceding sibling";
			if (stricmp (choice,"Prune") == 0)
				s = "Deletes all subtrees that do not contain text";
			}
		if (stricmp (menu,"View") == 0) {
			if (stricmp (choice,"Hide") == 0)
				s = "Hides children of the current element";
			if (stricmp (choice,"Show") == 0)
				s = "Shows children of the current node";
			if (stricmp (choice,"All levels") == 0)
				s = "Shows the entire hierarchy";
			if (stricmp (choice,"1 level") == 0)
				s = "Hides all but the major metadata sections";
			if (stricmp (choice,"2 levels") == 0)
				s = "Shows the major sections plus one level down";
			if (stricmp (choice,"3 levels") == 0)
				s = "Shows major sections plus two levels down";
			if (stricmp (choice,"Wrap") == 0)
				s = "Turns on word-wrap in the value window";
			if (stricmp (choice,"No Wrap") == 0)
				s = "Turns off word-wrap in the value window";
			}
		if (stricmp (menu,"Help") == 0) {
			if (stricmp (choice,"Version") == 0)
				s = "Show version information and general hints";
			if (stricmp (choice,"Element") == 0)
				s = "Show specific information about the current element";
			if (stricmp (choice,"Output") == 0)
				s = "Show the output as it will appear";
			if (stricmp (choice,"Fonts") == 0)
				s = "Change style and typefaces of menus and windows";
			}
		if (stricmp (menu,"Add") == 0)
			if (current)
				s = tip_text (text_of (current->key),choice);
			else
				if (stricmp (choice,text_of(WMetadata)) == 0)
					s = "Root of the current document";
		if (stricmp (menu,"Snippets") == 0) {
			if (stricmp (choice,"Save as snippet") == 0)
				s = "Store the current subtree for later recall";
			}
		if (stricmp (menu,"Tree") == 0) {
			if (stricmp (choice,"Cut") == 0)
				s = "Detaches a subtree, remembering it for paste";
			if (stricmp (choice,"Copy") == 0)
				s = "Remembers a subtree for later paste";
			if (stricmp (choice,"Paste") == 0)
				s = "Places cut or copied info into the metadata as text or as a subtree";
			if (stricmp (choice,"Clear") == 0)
				s = "Removes the text belonging to a subtree";
			if (stricmp (choice,"Duplicate") == 0)
				s = "Makes and places a copy of the current element";
			if (stricmp (choice,"Prune") == 0)
				s = "Deletes all subtrees that do not contain text";
			if (stricmp (choice,"Save as snippet") == 0)
				s = "Store the current subtree for later recall";
			}
		if (stricmp (menu,"Value") == 0) {
			if (stricmp (choice,"Cut") == 0)
				s = "Cuts selected text from the value";
			if (stricmp (choice,"Copy") == 0)
				s = "Copies selected text for later paste";
			if (stricmp (choice,"Paste") == 0)
				s = "Places cut or copied text into the value as text";
			if (stricmp (choice,"Revert") == 0)
				s = "Replaces value with most recently stored copy";
			if (stricmp (choice,"Wrap") == 0)
				s = "Turns on word-wrap in the value window";
			if (stricmp (choice,"No Wrap") == 0)
				s = "Turns off word-wrap in the value window";
			if (stricmp (choice,"HardWrap") == 0)
				s = "Breaks lines in this text to fit an 80-character page";
			}
		}
	if (!s) s = default_tip;
	Tcl_SetResult (interp,s,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | get_current_cmd
\*----------------------------------------------------------------------*/

int get_current_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;

	if (line_count == 0) {
		Tcl_SetResult (interp,NULL,TCL_STATIC);
		return (TCL_OK);
		}

	for (i=0; i < line_count; i++)
		if (line[i] == current) break;
	if (i >= line_count) {
		current = root;
		i = 0;
		}
	sprintf (message,"%d",i);
	Tcl_SetResult (interp,message,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | set_text_cmd
\*----------------------------------------------------------------------*/

static int buffer_granularity = 16384;
static char *buffer = NULL;
static int buffer_size = 0;

int set_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	struct item *p;
	int i,k;
	char *s = NULL;
	char *string = NULL;

	if (line_count == 0) {
		Tcl_SetResult (interp,NULL,TCL_STATIC);
		return (TCL_OK);
		}

	if (argc > 1) {
		i = (int) strtol (argv[1],&s,0);
		if (s == argv[1]) {
			sprintf (message,"set_text expects a number, got %s",argv[1]);
			Tcl_SetResult (interp,message,TCL_STATIC);
			return (TCL_ERROR);
			}

		if (p = line[i]) {
			int compound = (element_list_of (p->key) ? 1 : 0);

			if (compound == 0)
				if (p->child)
					if (p->child->key != Wunknown) compound = 1;

			if (compound) string = text_of (p->key);
			else {
				if (p->child) {
					struct item *q;
					int n = 0;

					/*--------------------------------------------------*\
					 | How much space will be needed for the value?		|
					\*--------------------------------------------------*/

					for (q=p->child; q; q=q->next)
						if (q->key == Wblank) n++;
						else
							if (q->d) n += 1 + strlen (q->d);

					/*--------------------------------------------------*\
					 | Make sure we have enough space in the buffer.	|
					\*--------------------------------------------------*/

					if (n > buffer_size) {
						k = n / buffer_granularity;
						n = (k + 1) * buffer_granularity;
						if (s = (char *) malloc (n)) {
							free (buffer);
							buffer = s;
							memset (buffer,0,n);
							buffer_size = n;
							}
						else {
							sprintf (message,"Error: could not allocate %d bytes for value\n",n);
							error_message (interp,message);
							}
						}

					/*--------------------------------------------------*\
					 | Fill the buffer with the data.					|
					\*--------------------------------------------------*/

					if (buffer) {
						char *s = buffer;
						char *end = buffer + buffer_size;
						for (q=p->child; q; q=q->next)
							if (q->key == Wblank) {
								*s++ = '\n';
								*s = 0;
								}
							else
								if (q->d) {
									k = strlen (q->d);
									if (s + k < end) {
										strcpy (s,q->d);
										s += strlen (q->d);
										*s++ = '\n';
										*s = 0;
										}
									}
						string = buffer;
						}
					else string = "Error: could not allocate sufficient space";
					}
				else
					if (buffer == NULL) {
						if (buffer = (char *) malloc (buffer_granularity)) {
							memset (buffer,0,buffer_granularity);
							string = buffer;
							}
						}
					else {
						string = buffer;
						*buffer = 0;
						}
				}
			current = p;
			}
		else
			sprintf (string,"Error: no element %d found",i);
		Tcl_SetResult (interp,string,TCL_STATIC);
		return (TCL_OK);
		}
	else {
		Tcl_SetResult (interp,"set_text requires an integer argument",TCL_STATIC);
		return (TCL_ERROR);
		}
	}

static void print_current (void) {
	fprintf (stderr,"current = 0x%X\n",current);
	fprintf (stderr,"         next = 0x%X\n",current->next);
	fprintf (stderr,"         prev = 0x%X\n",current->prev);
	fprintf (stderr,"       parent = 0x%X\n",current->parent);
	fprintf (stderr,"        child = 0x%X\n",current->child);
	fprintf (stderr,"       indent = %d\n",current->indent);
	fprintf (stderr,"  line_number = %d\n",current->line_number);
	fprintf (stderr,"          key = %s\n",text_of(current->key));
	fprintf (stderr,"     original = %d\n",current->original);
	fprintf (stderr,"          len = %d\n",current->len);
	fprintf (stderr,"            d = 0x%X\n",current->d);
	}

/*----------------------------------------------------------------------*\
 | get_text_cmd
\*----------------------------------------------------------------------*/

static int is_different (struct item *p, char *text) {
	struct item *q;
	char *s,*t;

	if (!p) return (1);
	if (element_list_of (p->key) != NULL) return (1);

	t = text;
	for (q=p->child; q; q=q->next)
		if (s = q->d) {
			while (*s)
				if (*s == *t) {
					s++;
					t++;
					}
				else return (1);
			if (*t == '\n') t++;
			else return (1);
			}
		else
			if (*t != 0) return (1);
	if (*t != 0) return (1);

	return (0);
	}

int get_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *b,*e,*s,*t;
	struct item *p,*q;
	int different = 0;

	if (argc > 1) {
		if (p = current) {

			if (element_list_of (p->key) == NULL) {

				/* Remove the extra newline that Tk adds */

				b = argv[1];
				if (*b) {
					e = b + strlen (b) - 1;
					if (*e == '\n') *e = 0;
					}

				/* Check the value against the original version */

				if (is_different (p,argv[1])) {
					different = 1;

					/*--------------------------------------------------*\
					 | To replace the value of an element, you must		|
					 | delete its existing children (i.e. the items		|
					 | containing the current text), then create new	|
					 | children.										|
					\*--------------------------------------------------*/

					delete_children (p);

					b = argv[1];

					/* Rebuild the value one line at a time */

					while (*b) {
						e = b;
						while (*e && *e != '\n') e++;

						s = (char *) malloc (1 + e - b);
						memcpy (s,b,e - b);
						s [e-b] = 0;

						q = add_child (p);
						q->key = ((e > b) ? Wunknown : Wblank);
						q->d = s;
						q->original = 0;
						q->len = e - b + 1;
						b = e;
						if (*b) b++;
						}
					}
				}
			sprintf (message,"%d",different);
			Tcl_SetResult (interp,message,TCL_STATIC);
			return (TCL_OK);
			}
		else {  /* This is ok, means quitting with no metadata. */
			Tcl_SetResult (interp,"0",TCL_STATIC);
			return (TCL_OK);
			}
		}
	else
		Tcl_SetResult (interp,"get_text expects an argument",TCL_STATIC);
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | add_menu_items_cmd
\*----------------------------------------------------------------------*/

int add_menu_items_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	enum fgdc_keyword *e;
	enum fgdc_keyword key;
	char *s;
	Tcl_Obj *obj;

	if (argc > 1) {
		s = argv[1];
		if (*s == '+') s++;
		while (*s && isspace (*s)) s++;
		key = key_of (s);
		if (key != Wunknown && key != Wblank) {
			if (e = element_list_of (key)) {
				Tcl_ResetResult (interp);
				obj = Tcl_NewListObj (0,NULL);
				while (*e != Wnull) {
					s = text_of (*e++);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (s,strlen (s)));
					}
				Tcl_SetObjResult (interp,obj);
				return (TCL_OK);
				}
			else {
				Tcl_SetResult (interp,NULL,TCL_STATIC);
				return (TCL_OK);
				}
			}
		else {
			Tcl_SetResult (interp,NULL,TCL_STATIC);
			return (TCL_OK);
			}
		}
	else {
		Tcl_SetResult (interp,"add_menu_items requires an element name as its argument",TCL_STATIC);
		return (TCL_ERROR);
		}
	}

/*----------------------------------------------------------------------*\
 | element_add_cmd
\*----------------------------------------------------------------------*/

int element_add_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i,k;
	enum fgdc_keyword key;
	enum fgdc_keyword *e;
	struct item *q, *r;
	Tcl_Obj *obj;

	if (argc > 1) {
		key = key_of (argv[1]);
		if (key != Wunknown && key != Wnull && key != Wblank) {
			if (current) {

				/*------------------------------------------------------*\
				 | First look to see whether there are any children of	|
				 | the current node.  If there are none, then simply	|
				 | add the new node as current's first child.			|
				 |														|
				 | If current has children, try to insert the new child	|
				 | in an appropriate place.  First, look to see whether	|
				 | there are any children with the same key.  If one or	|
				 | more of current's children have the same key as the	|
				 | child to be added, then insert the new child after	|
				 | the last child that has the same key.				|
				 |														|
				 | If none of current's children have the key that you	|
				 | want to add, then add the new child before the first	|
				 | existing child whose key falls after the target key	|
				 | in the list of possible child elements.				|
				\*------------------------------------------------------*/

				if (!current->child) q = add_child (current);
				else {
					/* Point r at the last child of current */

					for (r=current->child; r->next; r=r->next);

					/*--------------------------------------------------*\
					 | Walk backwards towards current->child, looking	|
					 | for a node whose key matches the element that we	|
					 | want to add.										|
					\*--------------------------------------------------*/

					while (r)
						if (r->key == key) break;
						else r = r->prev;

					/*--------------------------------------------------*\
					 | If r is not NULL, then r points to the last child|
					 | of current whose key matches the element we want	|
					 | to add; insert a new node after r.				|
					\*--------------------------------------------------*/

					if (r) {
						q = insert_item_after (r);
						q->indent = r->indent;
						}
					else {

						/*----------------------------------------------*\
						 | If r is NULL, then this will be the first	|
						 | child of current to have this key.			|
						\*----------------------------------------------*/

						if (e = element_list_of (current->key)) {

							/* What is the index of key in e? */

							for (k=0; e[k] != Wnull; k++)
								if (e[k] == key) break;

							/* If q is NULL, we haven't added the child yet */

							q = NULL;

							/* For each child of current */

							for (r = current->child; r; r=r->next)
								if (r->key != Wblank) {	/* skip blank lines */

									/* What is the index of r's key in e? */

									for (i=0; e[i] != Wnull; i++)
										if (e[i] == r->key) break;

									/*--------------------------------------*\
									 | If the key of r comes after k in e,	|
									 | insert the child before r and stop	|
									 | looking at the children of current.	|
									\*--------------------------------------*/

									if (i > k) {
										q = insert_item_before (r);
										q->indent = r->indent;
										break;
										}
									}

							/*------------------------------------------*\
							 | If you didn't find a child of current	|
							 | whose key follows the target key in the	|
							 | list of possible children of current,	|
							 | just add a child to current at the end.	|
							\*------------------------------------------*/

							if (!q) q = add_child (current);
							}
						else {

							/*------------------------------------------*\
							 | This should never happen.  All compound	|
							 | elements should have an element list.	|
							 |											|
							 | Something is very wrong, but we can just	|
							 | add the child anyway and hope it's okay.	|
							\*------------------------------------------*/

							q = add_child (current);
							}
						}
					}
				}
			else {
				q = allocate_item (1);
				q->key = WMetadata;
				q->d = text_of (WMetadata);
				q->original = 1;
				q->len = 1 + strlen (q->d);
				q->indent = 0;
				root = q;
				}

			if (q) {
				q->key = key;
				q->d = text_of (key);
				q->original = 1;
				q->len = 1 + strlen (q->d);

				/*------------------------------------------------------*\
				 | Whenever you add an element, ensure that all of its	|
				 | siblings are visible.								|
				\*------------------------------------------------------*/

				if (q->parent)
					for (r=q->parent->child; r; r=r->next)
						if (r->indent < 0) r->indent = -r->indent;

				rebuild_lists();
				current = q;

				/* Output the list of elements */

				Tcl_ResetResult (interp);
				obj = Tcl_NewListObj (0,NULL);
				for (i=0; i < element_count; i++) {
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
					}
				Tcl_SetObjResult (interp,obj);
				}
			else
				Tcl_SetResult (interp,NULL,TCL_STATIC);
			return (TCL_OK);
			}
		else {
			if (key == Wunknown)
				sprintf (message,"element_add expected element name, got \"%s\"",argv[1]);
			else
				if (key == Wblank || key == Wnull)
					sprintf (message,"element_add expected element name, got %s",text_of (key));

			Tcl_SetResult (interp,message,TCL_STATIC);
			return (TCL_ERROR);
			}
		}
	Tcl_SetResult (interp,"element_add expects element name as argument, got none",TCL_STATIC);
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | add_children_cmd
\*----------------------------------------------------------------------*/

int add_children_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	enum fgdc_keyword *e;
	struct item *q;
	Tcl_Obj *obj;

	if (current) {
		if (e = element_list_of (current->key)) {
			while (*e != Wnull) {
				if (q = add_child (current)) {
					q->key = *e;
					q->d = text_of (*e);
					q->original = 1;
					q->len = 1 + strlen (q->d);
					}
				e++;
				}

			rebuild_lists();
			Tcl_SetVar (interp,"changed","1",TCL_GLOBAL_ONLY);

			/* Output the list of elements */

			Tcl_ResetResult (interp);
			obj = Tcl_NewListObj (0,NULL);
			for (i=0; i < element_count; i++) {
				Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
				}
			Tcl_SetObjResult (interp,obj);
			return (TCL_OK);
			}
		else {
			Tcl_SetResult (interp,NULL,TCL_STATIC);
			return (TCL_OK);
			}
		}
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | element_cut_cmd
\*----------------------------------------------------------------------*/

int element_cut_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	struct item *q,*p;
	Tcl_Obj *obj;

	if (q = current)
		if (p = q->parent) {
			if (q->prev) current = q->prev;
			else current = q->parent;
			detach_item (q);

			if (clipboard) delete_item (clipboard);
			clipboard = q;
			if (clipboard_text) {
				free (clipboard_text);
				clipboard_text = NULL;
				}
			clipboard_text = encode_text (clipboard);

			Tcl_SetVar (interp,"changed","1",TCL_GLOBAL_ONLY);

			rebuild_lists();

			/* Output the list of elements */

			Tcl_ResetResult (interp);
			obj = Tcl_NewListObj (0,NULL);
			for (i=0; i < element_count; i++) {
				Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
				}
			Tcl_SetObjResult (interp,obj);
			}
		else
			Tcl_SetResult (interp,NULL,TCL_STATIC);
	else
		Tcl_SetResult (interp,NULL,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | element_copy_cmd
\*----------------------------------------------------------------------*/

int element_copy_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	if (current) {
		if (clipboard) delete_item (clipboard);
		clipboard = copy_item (current);
		if (clipboard_text) {
			free (clipboard_text);
			clipboard_text = NULL;
			}
		clipboard_text = encode_text (clipboard);
		}
	Tcl_SetResult (interp,NULL,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | clipped_text_cmd
\*----------------------------------------------------------------------*/

int clipped_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	Tcl_SetResult (interp,clipboard_text,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | encode_current_cmd
\*----------------------------------------------------------------------*/

int encode_current_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *s,*t;

	if (current) {
		s = encode_text (current);
		if (t = Tcl_Alloc (1 + strlen (s))) {
			strcpy (t,s);
			free (s);
			Tcl_SetResult (interp,t,TCL_DYNAMIC);
			}
		else
			Tcl_SetResult (interp,s,TCL_STATIC);
		return (TCL_OK);
		}
	Tcl_SetResult (interp,NULL,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | duplicate_cmd
\*----------------------------------------------------------------------*/

int duplicate_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	struct item *p,*q;
	Tcl_Obj *obj;

	if (p = current)
		if (q = insert_item_after (p)) {
			q->key = p->key;
			q->indent = p->indent;

			q->d = text_of (q->key);
			q->original = 1;
			q->len = 1 + strlen (q->d);

			rebuild_lists();
			current = q;

			/* Output the list of elements */

			Tcl_ResetResult (interp);
			obj = Tcl_NewListObj (0,NULL);
			for (i=0; i < element_count; i++) {
				Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
				}
			Tcl_SetObjResult (interp,obj);
			return (TCL_OK);
			}
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | swap_cmd
\*----------------------------------------------------------------------*/

int swap_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	struct item *p,*q;
	Tcl_Obj *obj;

	if (p = current)
		if (q = p->prev) {
			if (q->prev) q->prev->next = p;
			if (p->next) p->next->prev = q;
			p->prev = q->prev;
			q->next = p->next;
			p->next = q;
			q->prev = p;
			if (p->parent)
				if (p->parent->child == q) p->parent->child = p;

			rebuild_lists();
			current = q;

			/* Output the list of elements */

			Tcl_ResetResult (interp);
			obj = Tcl_NewListObj (0,NULL);
			for (i=0; i < element_count; i++) {
				Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
				}
			Tcl_SetObjResult (interp,obj);
			return (TCL_OK);
			}
		else {
			Tcl_ResetResult (interp);
			return (TCL_OK);
			}
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | prune_cmd
\*----------------------------------------------------------------------*/

int prune_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	struct item *p;
	Tcl_Obj *obj;

	if (p = current) {
		prune (p);
		rebuild_lists();

		/* Output the list of elements */

		Tcl_ResetResult (interp);
		obj = Tcl_NewListObj (0,NULL);
		for (i=0; i < element_count; i++) {
			Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
			}
		Tcl_SetObjResult (interp,obj);
		return (TCL_OK);
		}
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | hide_element_cmd
\*----------------------------------------------------------------------*/

int hide_element_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	struct item *p,*q;
	Tcl_Obj *obj;

	if (p = current) {
		if (p != root)
			if (p->child) {
				for (q=p->child; q; q=q->next)
					if (q->indent > 0) q->indent = -q->indent;
				rebuild_lists();
				}

		/* Output the list of elements */

		Tcl_ResetResult (interp);
		obj = Tcl_NewListObj (0,NULL);
		for (i=0; i < element_count; i++) {
			Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
			}
		Tcl_SetObjResult (interp,obj);
		}
	else
		Tcl_SetResult (interp,NULL,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | show_element_cmd
\*----------------------------------------------------------------------*/

int show_element_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	struct item *p,*q;
	Tcl_Obj *obj;

	if (p = current) {
		if (p != root)
			if (p->child) {
				for (q=p->child; q; q=q->next)
					if (q->indent < 0) q->indent = -q->indent;
				rebuild_lists();
				}

		/* Output the list of elements */

		Tcl_ResetResult (interp);
		obj = Tcl_NewListObj (0,NULL);
		for (i=0; i < element_count; i++) {
			Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
			}
		Tcl_SetObjResult (interp,obj);
		}
	else
		Tcl_SetResult (interp,NULL,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | Detail callback                                                                                                              |
\*----------------------------------------------------------------------*/

static int current_level = 0;

static void reveal_level (struct item *p, int target) {
	struct item *q;
	if (p) {
		int level = abs (p->indent);
		p->indent = ((level <= target) ? level : -level);
		for (q=p->child; q; q=q->next) reveal_level (q,target);
		}
	}

int collapse_list_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	Tcl_Obj *obj;

	if (line_count == 0) {
		Tcl_SetResult (interp,NULL,TCL_STATIC);
		return (TCL_OK);
		}

	if (current_level == 0) current_level = 128;
	if (current_level == 128) current_level = 4;
	if (current_level > 1) {
		current_level--;
		reveal_level (root,current_level);
		rebuild_lists();

		if (current->indent < 0) {
			while (current->indent < 0)
				if (current->parent) current = current->parent;
				else break;
			}
		}

	/* Output the list of elements */

	Tcl_ResetResult (interp);
	obj = Tcl_NewListObj (0,NULL);
	for (i=0; i < element_count; i++) {
		Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
		}
	Tcl_SetObjResult (interp,obj);
	return (TCL_OK);
	}

int expand_list_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	Tcl_Obj *obj;

	if (line_count == 0) {
		Tcl_SetResult (interp,NULL,TCL_STATIC);
		return (TCL_OK);
		}

	if (current_level == 0) current_level = 128;
	if (current_level < 128) {
		current_level++;
		reveal_level (root,current_level);
		rebuild_lists();
		}

	/* Output the list of elements */

	Tcl_ResetResult (interp);
	obj = Tcl_NewListObj (0,NULL);
	for (i=0; i < element_count; i++) {
		Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
		}
	Tcl_SetObjResult (interp,obj);
	return (TCL_OK);
	}

int view_levels_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	char *s;
	Tcl_Obj *obj;

	if (line_count == 0) {
		Tcl_SetResult (interp,NULL,TCL_STATIC);
		return (TCL_OK);
		}

	if (argc > 1) {
		i = (int) strtol (argv[1],&s,0);
		if (s != argv[1]) {
			if (i == 0) current_level = 128;
			else current_level = i;
			reveal_level (root,current_level);
			rebuild_lists();

			if (current->indent < 0) {
				while (current->indent < 0)
					if (current->parent) current = current->parent;
					else break;
				}
			}

		/* Output the list of elements */

		Tcl_ResetResult (interp);
		obj = Tcl_NewListObj (0,NULL);
		for (i=0; i < element_count; i++) {
			Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
			}
		Tcl_SetObjResult (interp,obj);
		return (TCL_OK);
		}
	else
		Tcl_SetResult (interp,"view_levels requires an integer argument",TCL_STATIC);
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | is_scalar_cmd
\*----------------------------------------------------------------------*/

int is_scalar_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	char *s;
	struct item *p;
	enum fgdc_keyword *e;

	if (argc > 1) {
		i = (int) strtol (argv[1],&s,0);
		if (s != argv[1]) {
			int compound = 0;
			p = line [i];

			if (element_list_of (p->key)) compound = 1;
			else
				if (p->child)
					if (p->child->key != Wunknown) compound = 1;

			if (compound) Tcl_SetResult (interp,"0",TCL_STATIC);
			else          Tcl_SetResult (interp,"1",TCL_STATIC);

			return (TCL_OK);
			}
		else
			Tcl_SetResult (interp,"is_scalar requires an integer argument",TCL_STATIC);
		}
	else
		Tcl_SetResult (interp,"is_scalar requires an integer argument",TCL_STATIC);
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | clear_subtree_cmd
\*----------------------------------------------------------------------*/

int clear_subtree_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	struct item *p;

	if (p = current) {
		clear_scalars (p);
		Tcl_SetResult (interp,NULL,TCL_STATIC);
		}
	else
		Tcl_SetResult (interp,NULL,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | help_text_cmd
\*----------------------------------------------------------------------*/

int help_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	if (current)
		Tcl_SetResult (interp,advice_about (current->key),TCL_STATIC);
	else
		Tcl_SetResult (interp,NULL,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | output_text_cmd
\*----------------------------------------------------------------------*/

int output_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	if (output_string) {
		free (output_string);
		output_string = NULL;
		output_string_length = 0;
		}
	Tcl_SetResult (interp,NULL,TCL_STATIC);
	if (root)
		if (output_string = encode_text_numbered (root))
			Tcl_SetResult (interp,output_string,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | edit_line_cmd
\*----------------------------------------------------------------------*/

int edit_line_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i;
	long n;
	char *s;
	struct item *p,*q,*r;
	int do_list_update = 0;
	Tcl_Obj *obj;

	/*------------------------------------------------------------------*\
	 | Find out what line of the output text window you want to edit
	\*------------------------------------------------------------------*/

	if (argc > 1) {
		n = strtol (argv[1],&s,0);
		if (s == argv[1]) {
			Tcl_SetResult (interp,"edit_line requires an integer argument",TCL_STATIC);
			return (TCL_ERROR);
			}

		/* In the window, they start from 1; the array index starts at 0 */

		n--;

		/*--------------------------------------------------------------*\
		 | Find the item that is associated with this line.				|
		\*--------------------------------------------------------------*/

		if (p = item_of (n)) {

			if (!valid_node (p,root)) {
				while (--n >= 0)
					if (valid_node ((p = item_of(n)),root)) break;

				if (n < 0) {
					Tcl_SetResult (interp,"This element may have been edited or cut; please choose Update from the View menu.",TCL_STATIC);
					return (TCL_ERROR);
					}
				}

			/*----------------------------------------------------------*\
			 | If the line you specified in the output window is Wblank,|
			 | move backwards until you find Wunknown or an element or	|
			 | NULL.  If you end up with nothing, do nothing and return.|
			 | That should never happen, of course, but recover well	|
			 | anyway.													|
			\*----------------------------------------------------------*/

			while (p->key == Wblank)
				if (p->prev) p = p->prev;
			if (p->key == Wunknown || p->key == Wblank) p = p->parent;
			if (!p) {
				Tcl_SetResult (interp,"edit_line cannot jump to this element",TCL_STATIC);
				return (TCL_ERROR);
				}

			/*----------------------------------------------------------*\
			 | The item you're after could be hidden.  If so, unhide it,|
			 | rebuild the lists, and update the list widget's list.	|
			\*----------------------------------------------------------*/

			for (q=p; q; q=q->parent)
				if (q->indent < 0) {
					q->indent = -q->indent;
					do_list_update = 1;
					for (r=q->child; r; r=r->next)
						if (r->indent < 0) r->indent = -r->indent;
					}

			if (do_list_update) rebuild_lists();

			current = p;

			/* Output the list of elements */

			Tcl_ResetResult (interp);
			obj = Tcl_NewListObj (0,NULL);
			for (i=0; i < element_count; i++) {
				Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (element[i],strlen (element[i])));
				}
			Tcl_SetObjResult (interp,obj);

			return (TCL_OK);
			}
		else
			Tcl_SetResult (interp,"edit_line can't find the element",TCL_STATIC);
		}
	else
		Tcl_SetResult (interp,"edit_line requires an integer argument",TCL_STATIC);
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | recognized_cmd
\*----------------------------------------------------------------------*/

int recognized_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *s;
	enum fgdc_keyword key;

	if (argc > 1) {
		s = argv[1];
		while (*s && isspace (*s)) s++;
		key = key_of (s);
		if (key != Wunknown && key != Wblank && key != Wnull)
			Tcl_SetResult (interp,"1",TCL_STATIC);
		else
			Tcl_SetResult (interp,"0",TCL_STATIC);
		}
	else
		Tcl_SetResult (interp,"0",TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | paste_subtree_cmd
\*----------------------------------------------------------------------*/

static int paste_subtree (struct item *p) {
	int i,j,k;
	struct item *q,*r,*s,*u;
	enum fgdc_keyword *key_list,*e;
	int found = 0;

	if (!p) return (0);

	/*------------------------------------------------------------------*\
	 | I have been persuaded that the snippets interface could be used	|
	 | to distribute templates.  Hence in the special case that there	|
	 | is no metadata already loaded, if you paste a subtree whose top	|
	 | element is WMetadata, it should simply make that subtree the		|
	 | current tree.													|
	\*------------------------------------------------------------------*/

	if (!current)
		if (p->key == WMetadata) {
			root = p;
			current = root;
			strcpy (name_of_file,"Untitled");
			adjust_indent (p,0);

			return (1);
			}
		else
			return (0);

	/* Find the current node */

	q = current;

	/*------------------------------------------------------------------*\
	 | Try to make the subtree p a child of the current node.  If p is	|
	 | not a potential child of q, try to make p a child of q's parent.	|
	 | If that cannot be done, keep looking up the tree for a place to	|
	 | put p until you get to the root node.  If the subtree cannot be	|
	 | a child of the root node, don't put it anywhere.					|
	\*------------------------------------------------------------------*/

	while (!found && q) {
		if (key_list = element_list_of (q->key))
			for (i=0; key_list[i] != Wnull; i++)
				if (p->key == key_list[i]) {

					/* Attach the subtree */

					if (q->child) {
						r = NULL;

						/*----------------------------------------------*\
						 | If the current node is one of the children	|
						 | of q, insert the subtree after the current	|
						 | node, otherwise try to place the subtree in	|
						 | some logical position.  If all else fails,	|
						 | make it the last child of q.					|
						\*----------------------------------------------*/

						for (u=q->child; u; u=u->next)
							if (u == current) r = u;

						/*----------------------------------------------*\
						 | If there's a child of the same type as the	|
						 | head node of the subtree, place the subtree	|
						 | after the last such child.					|
						\*----------------------------------------------*/

						if (!r) {
							for (u=q->child; u->next; u=u->next);
							while (u) {
								if (u->key == p->key) {
									r = u;
									break;
									}
								u = u->prev;
								}
							}

						/*----------------------------------------------*\
						 | If the current node is not a child of q and	|
						 | there isn't an element of the same type as	|
						 | p already among the children of q, try to	|
						 | insert the subtree in the order given by the	|
						 | Standard.									|
						\*----------------------------------------------*/

						if (!r) {
							for (k=0; key_list[k] != Wnull; k++)
								if (key_list[k] == p->key) break;

							/*------------------------------------------*\
							 | Find the first child of q whose key is	|
							 | after the key of p in the list of		|
							 | possible children of q that is given in	|
							 | the Standard.							|
							\*------------------------------------------*/

							for (u=q->child; u; u=u->next)
								if (u->key != Wblank) {
									for (j=0; key_list[j] != Wnull; j++)
										if (key_list[j] == u->key) break;
									if (j >= k) {
										r = u;
										break;
										}
									}

							/*------------------------------------------*\
							 | Since you want to put it before r, place	|
							 | the subtree after the element before r.	|
							\*------------------------------------------*/

							if (r)
								if (r->prev) r = r->prev;
							}

						/*----------------------------------------------*\
						 | If all else fails, just put the subtree at	|
						 | the end of the list of children of q.		|
						\*----------------------------------------------*/

						if (!r)
							for (r=q->child; r->next; r=r->next);

						/*----------------------------------------------*\
						 | The subtree might not be headed by a single	|
						 | node.  It could be a list, and p is just the	|
						 | head of the list.  Then you need to connect	|
						 | the node after r, if there is one, to the	|
						 | last node in the list that heads the subtree	|
						 | rather than to p, the head of the list.		|
						\*----------------------------------------------*/

						for (s=p; s->next; s=s->next);
						s->next = r->next;
						if (s->next) s->next->prev = s;

						r->next = p;
						p->prev = r;
						}
					else {
						q->child = p;
						}

					/* Assign parent of the subtree */

					p->parent = q;
					if (p->next) {
						struct item *u;
						for (u=p->next; u; u=u->next) u->parent = q;
						}

					/* Adjust indentation of the subtree */

					adjust_indent (p,1+q->indent);
					if (p->next) {
						struct item *u;
						for (u=p->next; u; u=u->next) adjust_indent (u,1+q->indent);
						}

					found = 1;
					}
		q = q->parent;
		}
	return (found);
	}

int paste_subtree_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int n;
	struct item *p;
	char *s;
	int ok = 0;

	if (argc > 1) {
		n = 1 + strlen (argv[1]);
		if (s = (char *) malloc (n)) {
			strcpy (s,argv[1]);
			if (*s == '<')
				p = parse_xml_string (s);
			else
				p = parse_string (s);

			if (check_syntax (p) != 0) {
				Tcl_SetResult (interp,"This information cannot be included because an element within it has mixed content.",TCL_STATIC);
				return (TCL_ERROR);
				}

			if (ok = paste_subtree (p))
				Tcl_SetVar (interp,"changed","1",TCL_GLOBAL_ONLY);
			}
		else
			error_message (interp,"Could not allocate memory for paste string");
		}

	sprintf (message,"%d",ok);
	Tcl_SetResult (interp,message,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | read_config_cmd
\*----------------------------------------------------------------------*/

static void read_config (char *config_file) {
	void *opt,*v;
	char *ext_file = NULL;
	char *tip_file = NULL;
	char *s;

	if (config_file) {
		read_configuration (config_file);

		if (!language)
			if (opt = find_option (NULL,"input"))
				if (opt = find_option (opt,"language"))
					if (s = text_of_option (opt))
						language = s;

		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"profile"))
				if (s = text_of_option (opt))
					profile_name = s;

		use_element_names (language,profile_name);		/* in keyword.c */
		if (profile_name) use_profile (profile_name);	/* in actions.c */

		if (v = find_option (NULL,"input")) {
			if (opt = find_option (v,"extensions")) {
				if (ext_file = text_of_option (opt)) read_local (ext_file);
				while (opt = find_next_option (opt,"extensions"))
					if (ext_file = text_of_option (opt)) read_local (ext_file);
				}

			if (opt = find_option (v,"tips")) {
				if (tip_file = text_of_option (opt)) read_tips (tip_file);
				while (opt = find_next_option (opt,"tips"))
					if (tip_file = text_of_option (opt)) read_tips (tip_file);
				}
			}
		}
	}

int read_config_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	if (argc > 1) read_config (argv[1]);
	return (TCL_OK);
	}

int config_read_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	void *opt;

	strcpy (message,"0");
	if (!(opt = find_option (NULL,"input")))
		opt = find_option (NULL,"output");
	if (opt) *message = '1';

	Tcl_SetResult (interp,message,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | cmdline_file_cmd
\*----------------------------------------------------------------------*/

int cmdline_file_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	Tcl_SetResult (interp,cmdline_file,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | about_text_cmd
\*----------------------------------------------------------------------*/

static char *about_buffer = NULL;

int about_text_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	if (!about_buffer) {
		char **p;
		int n = 0;
		for (p=about_text; *p; p++) n += 1 + strlen (*p);
		if (about_buffer = (char *) malloc (1 + n)) {
			*about_buffer = 0;
			for (p=about_text; *p; p++)
				strcat (about_buffer,*p);
			}
		else
			about_buffer = "Internal error: failed to allocate space for about_buffer\n";
		}
	Tcl_SetResult (interp,about_buffer,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | ext_list_cmd
\*----------------------------------------------------------------------*/

int ext_list_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	void *opt;
	char *s,*t;
	char string[16];
	Tcl_Obj *obj;

	Tcl_ResetResult (interp);
	obj = Tcl_NewListObj (0,NULL);

	strcpy (string,".met");
	Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (string,strlen (string)));

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"ext"))
			do {
				if (s = text_of_option (opt))
					if (*s && (strlen (string) < 16)) {
						*string = 0;
						t = string;
						if (*s != '.') *t++ = '.';
						strcpy (t,s);
						t += strlen (s);
						Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (string,strlen (string)));
						}
				} while (opt = find_next_option (opt,"ext"));
	Tcl_SetObjResult (interp,obj);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | wrap_current_cmd
\*----------------------------------------------------------------------*/

int wrap_current_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	wrap_text (current,76);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | forget_cmd
\*----------------------------------------------------------------------*/

int forget_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	if (root) {
		forget_buffer (root);
		delete_item (root);
		root = NULL;
		current = NULL;
		rebuild_lists ();
		}
	Tcl_ResetResult (interp);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | rc_file_spec_cmd
\*----------------------------------------------------------------------*/

#ifdef _WIN32
	static char rc_file [1024] = "/usgs/tools/recent_files";
#else
	static char rc_file [1024] = "~/.tkme/recent_files";
#endif

int rc_file_spec_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	Tcl_SetResult (interp,rc_file,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | topology_cmd
\*----------------------------------------------------------------------*/

int topology_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	check_topology (root);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | cfg_messages_cmd
\*----------------------------------------------------------------------*/

int cfg_messages_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *s = config_errors();
	if (!s) s = "";
	Tcl_SetResult (interp,s,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | ext_messages_cmd
\*----------------------------------------------------------------------*/

int ext_messages_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *s = ext_errors();
	if (!s) s = "";
	Tcl_SetResult (interp,s,TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | root_name_cmd
\*----------------------------------------------------------------------*/

int root_name_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	Tcl_SetResult (interp,text_of(WMetadata),TCL_STATIC);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | replace_cmd
\*----------------------------------------------------------------------*/

typedef int(*compare_function)(const void *,const void *,size_t);

static int replace_in (struct item *p, char *find_text, char *repl_text, compare_function compare) {
	int change_count = 0;
	int k,n,m;
	char *s,*t;
	struct item *q;

	if (!p || !find_text || !repl_text || !compare) return (0);
	if (p->key == Wunknown || p->key == Wblank) return (0);

	n = strlen (find_text);
	m = strlen (repl_text);

	if (p->child) {

		/* Is the first non-blank child of type Wunknown? */

		for (q=p->child; q && q->key == Wblank; q=q->next);

		/* If so, search and replace here */

		if (q && q->key == Wunknown)
			for (q=p->child; q; q=q->next) {
				if (s = q->d)
					while (*s)
						if (compare (s,find_text,n) != 0) s++;	/* no match */
						else {	/* match */

							/*------------------------------------------*\
							 | How many characters are on the current	|
							 | line?  Store in k.						|
							\*------------------------------------------*/

							k = strlen (q->d);

							/*------------------------------------------*\
							 | After the substitution, you will have	|
							 | subtracted n characters and added m.		|
							 | Remember to move the nul terminator.		|
							\*------------------------------------------*/

							if (k - n + m + 1 > q->len) {

								/*--------------------------------------*\
								 | If the resulting string will be longer
								 | than the buffer is  reallocate the	|
								 | line to be at least q->len + m bytes	|
								\*--------------------------------------*/

								if (t = (char *) malloc (q->len + m)) {

									/* Copy old string to new */

									strcpy (t,q->d);

									/* Point s into new string */

									s = t + (s - q->d);

									/* Free old space if okay */

									if (!q->original) free (q->d);

									/*----------------------------------*\
									 | Update q->d, q->len, q->original	|
									\*----------------------------------*/

									q->d = t;
									q->len += m;
									q->original = 1;
									}
								}

							/*------------------------------------------*\
							 | Move the part of q->d that is after the	|
							 | find_text to position that will be after	|
							 | the repl_text then insert the repl_text	|
							 | at s.									|
							 |											|
							 | Remember to move one more byte so the	|
							 | string is nul terminated.				|
							\*------------------------------------------*/

							memmove (s+m, s+n, 1 + strlen (s+n));
							memcpy (s,repl_text,m);

							/*------------------------------------------*\
							 | Point s at the first character after the	|
							 | replacement text.						|
							\*------------------------------------------*/

							s += m;

							/* Increment change count */

							change_count++;

						/*\
							if (change_count == 1)
								fprintf (stderr,"Change made in %s at line %d\n",text_of(p->key),p->line_number);
						\*/
							}
				}
		else
			for (q=p->child; q; q=q->next)
				change_count += replace_in (q,find_text,repl_text,compare);
		}

	return (change_count);
	}

int replace_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	if (argc > 4) {
		int change_count = 0;
		char *find_text = argv[1];
		char *find_case = argv[2];
		char *repl_text = argv[3];
		char *where     = argv[4];
		int (*compare) (const void *s, const void *t, size_t n);

		if (atoi (find_case) == 0) compare = memicmp;
		else compare = memcmp;

		if (stricmp (where,"all") == 0)
			change_count = replace_in (root,find_text,repl_text,compare);
		else
			if (stricmp (where,"within") == 0)
				change_count = replace_in (current,find_text,repl_text,compare);
			else
				if (current && stricmp (where,"this") == 0) {
					struct item *q;

					/* Find the first non-blank child of current */

					for (q=current->child; q && q->key == Wblank; q=q->next);

					/*--------------------------------------------------*\
					 | If the first non-blank child is Wunknown, do the	|
					 | replace, otherwise don't.						|
					\*--------------------------------------------------*/

					if (q->key == Wunknown)
						change_count = replace_in (current,find_text,repl_text,compare);
					}

		if (change_count > 0)
			Tcl_SetVar (interp,"changed","1",TCL_GLOBAL_ONLY);

		sprintf (message,"%d",change_count);
		Tcl_SetResult (interp,message,TCL_STATIC);
		return (TCL_OK);
		}
	else
		Tcl_SetResult (interp,"replace_everywhere requires 3 arguments",TCL_STATIC);
	return (TCL_ERROR);
	}

/*----------------------------------------------------------------------*\
 | Call Tcl_CreateCommand for application-specific commands, if they	|
 | weren't already created by the init procedures called above.			|
\*----------------------------------------------------------------------*/

int Et_AppInit (Tcl_Interp *interp) {
	Tcl_CreateCommand (interp,"cmdline_file",  (Tcl_CmdProc *) cmdline_file_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"read_config",   (Tcl_CmdProc *) read_config_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"config_read",   (Tcl_CmdProc *) config_read_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"parse_text",    (Tcl_CmdProc *) parse_text_cmd,    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"get_list",      (Tcl_CmdProc *) get_list_cmd,      (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"write_text",    (Tcl_CmdProc *) write_text_cmd,    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"tip_text",      (Tcl_CmdProc *) tip_text_cmd,      (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"set_text",      (Tcl_CmdProc *) set_text_cmd,      (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"get_text",      (Tcl_CmdProc *) get_text_cmd,      (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"add_menu_items",(Tcl_CmdProc *) add_menu_items_cmd,(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"element_add",   (Tcl_CmdProc *) element_add_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"element_cut",   (Tcl_CmdProc *) element_cut_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"element_copy",  (Tcl_CmdProc *) element_copy_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"encode_current",(Tcl_CmdProc *) encode_current_cmd,(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"add_children",  (Tcl_CmdProc *) add_children_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"hide_element",  (Tcl_CmdProc *) hide_element_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"show_element",  (Tcl_CmdProc *) show_element_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"collapse_list", (Tcl_CmdProc *) collapse_list_cmd, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"expand_list",   (Tcl_CmdProc *) expand_list_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"view_levels",   (Tcl_CmdProc *) view_levels_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"get_current",   (Tcl_CmdProc *) get_current_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"is_scalar",     (Tcl_CmdProc *) is_scalar_cmd,     (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"clear_subtree", (Tcl_CmdProc *) clear_subtree_cmd, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"duplicate",     (Tcl_CmdProc *) duplicate_cmd,     (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"swap",          (Tcl_CmdProc *) swap_cmd,          (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"prune",         (Tcl_CmdProc *) prune_cmd,         (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"help_text",     (Tcl_CmdProc *) help_text_cmd,     (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"output_text",   (Tcl_CmdProc *) output_text_cmd,   (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"edit_line",     (Tcl_CmdProc *) edit_line_cmd,     (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"clipped_text",  (Tcl_CmdProc *) clipped_text_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"recognized",    (Tcl_CmdProc *) recognized_cmd,    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"paste_subtree", (Tcl_CmdProc *) paste_subtree_cmd, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"about_text",    (Tcl_CmdProc *) about_text_cmd,    (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"wrap_current",  (Tcl_CmdProc *) wrap_current_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"ext_list",      (Tcl_CmdProc *) ext_list_cmd,      (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"forget",        (Tcl_CmdProc *) forget_cmd,        (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"rc_file_spec",  (Tcl_CmdProc *) rc_file_spec_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"topology",      (Tcl_CmdProc *) topology_cmd,      (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"root_name",     (Tcl_CmdProc *) root_name_cmd,     (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"replace",       (Tcl_CmdProc *) replace_cmd,       (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

	Tcl_CreateCommand (interp,"cfg_messages",  (Tcl_CmdProc *) cfg_messages_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand (interp,"ext_messages",  (Tcl_CmdProc *) ext_messages_cmd,  (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);

	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
 | Main procedure
\*----------------------------------------------------------------------*/

int main (int argc, char *argv[]) {
	int i;
	char *config_file = NULL;
	char help_file[FILENAME_MAX] = "";
	char *s,*t;
	void *opt;

	out = stderr;

	/*------------------------------------------------------------------*\
	 | Interpret command-line arguments
	\*------------------------------------------------------------------*/

	if (argc > 0)
		for (i=1; i < argc; i++) {
			if (memcmp (argv[i],"-c",2) == 0)
				if (i+1 < argc) {
					i++;
					config_file = argv[i];
					}
				else
					fprintf (stderr,"Warning: no config file name found after command option -c\n");
			else
				if (memcmp (argv[i],"-h",2) == 0)
					if (i+1 < argc) {
						i++;
						if (strlen (argv[i]) < FILENAME_MAX)
							strcpy (help_file,argv[i]);
						}
					else
						fprintf (stderr,"Warning: no help file name found after command option -h\n");
				else
					if (memcmp (argv[i],"-r",2) == 0)
						if (i+1 < argc) {
							i++;
							strcpy (rc_file,argv[i]);
							#ifdef _WIN32
								for (s=rc_file; *s; s++) if (*s == '\\') *s = '/';
							#endif
							}
						else
							fprintf (stderr,"Warning: no recent-files-list file name found after command option -h\n");
					else
						if (memcmp (argv[i],"-l",2) == 0)
							if (i+1 < argc) {
								i++;
								language = argv[i];
								}
							else
								fprintf (stderr,"Warning: no language code found after command option -l\n");
						else
							if (*argv[i] != '-')
								strcpy (cmdline_file,argv[i]);
			}

	/*------------------------------------------------------------------*\
	 | Read the config file and local extensions
	\*------------------------------------------------------------------*/

	if (config_file) read_config (config_file);
	else
		if (language) use_element_names (language,NULL);	/* in keyword.c */

	/*------------------------------------------------------------------*\
	 | Read the help file
	 | Modified 2006-02-28 to automatically look for a help file for the
	 | language that is in use if the help file was not specified.
	\*------------------------------------------------------------------*/

	if (*help_file == 0)
		if (language) sprintf (help_file,"/usgs/tools/doc/xtme-%2s.hlp",language);
		else strcpy (help_file,"/usgs/tools/doc/xtme-en.hlp");

	read_help (help_file);

	/*------------------------------------------------------------------*\
	 | Initialize Tcl/Tk
	\*------------------------------------------------------------------*/

	Et_Init (1,argv);

	return (0);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
