/*----------------------------------------------------------------------*\
 | Code to handle parse trees.											|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "item.h"
#include "config.h"

/*----------------------------------------------------------------------*\
 | Function to check topological linkage								|
\*----------------------------------------------------------------------*/

void check_topology (struct item *p) {
	struct item *r,*q;

	if (p->child) {

		/* Check links to parent */

		q = p->child;
		for (q=p->child; q; q=q->next)
			if (q->parent != p)
				fprintf (stderr,"Topology error: child %s of %s does not link to parent\n",
					text_of(q->key),text_of(p->key));

		/* Within a list of child nodes, the head has prev == NULL */

		q = p->child;
		if (q->prev)
			fprintf (stderr,"Topology error: head child %s of %s has prev link\n",
				text_of(q->key),text_of(p->key));

		/* Check the prev links */

		if (q->next) {
			r = q->next;
			while (r) {
				if (r->prev != q)
					fprintf (stderr,"Topology error: child %s of %s lacks prev link to %s\n",
						text_of(r->key),text_of(p->key),text_of(q->key));
				r = r->next;
				q = q->next;
				}
			}

		/* Check topology of each child */

		for (q=p->child; q; q=q->next) check_topology (q);
		}
	}

/*----------------------------------------------------------------------*\
 | Print out a node in full or brief format.							|
\*----------------------------------------------------------------------*/

void print_item (struct item *p, int verbose) {
	if (verbose) {
		fprintf (stderr,"node at 0x%lX\n",p);
		if (p) {
			fprintf (stderr,"  next:   0x%lX\n",p->next);
			fprintf (stderr,"  prev:   0x%lX\n",p->prev);
			fprintf (stderr,"  parent: 0x%lX\n",p->parent);
			fprintf (stderr,"  child:  0x%lX\n",p->child);
			fprintf (stderr,"  indent: %d\n",p->indent);
			fprintf (stderr,"  line_number: %d\n",p->line_number);
			fprintf (stderr,"  key: %s\n",text_of (p->key));
			fprintf (stderr,"  original: %d\n",p->original);
			fprintf (stderr,"  len: %d\n",p->len);
			if (p->d)
				fprintf (stderr,"  d: \"%s\"\n",p->d);
			else
				fprintf (stderr,"  d: (null)\n");
			}
		}
	else {
		printf ("node at 0x%lX",p);
		if (p) {
			printf (" 0x%lX",p->next);
			printf (" 0x%lX",p->prev);
			printf (" 0x%lX",p->parent);
			printf (" 0x%lX",p->child);
			printf (" >%d",p->indent);
			printf (" #%d",p->line_number);
			printf (" %s",text_of(p->key));
			printf (" \"%s\"",p->d);
			printf ("\n");
			}
		}
	fflush (stdout);
	}

/*----------------------------------------------------------------------*\
 | Functions to allocate and deallocate items.  Trying to make these	|
 | more sophisticated by allocating in fairly large bunches.			|
\*----------------------------------------------------------------------*/

#define ITEMS_PER_CHUNK	1024

struct chunk {
	struct chunk *next;
	struct item item [ITEMS_PER_CHUNK];
	char in_use [ITEMS_PER_CHUNK];
	};

static struct chunk *chunk_list = NULL;

void chunk_status (void) {
	struct chunk *ch;
	int i,j,k;

	for (k=0,ch=chunk_list; ch; k++,ch=ch->next);
	fprintf (stderr,"Number of chunks: %d\n",k);
	j = 0;
	for (ch=chunk_list; ch; ch=ch->next) {
		k = 0;
		for (i=0; i < ITEMS_PER_CHUNK; i++)
			k += ch->in_use[i];
		fprintf (stderr,"  chunk %d at 0x%X: %d items in use (of %d)\n",j,ch,k,ITEMS_PER_CHUNK);
		j++;
		}
	}

static struct chunk *get_chunk (void) {
	struct chunk *ch;
	if (ch = (struct chunk *) malloc (sizeof (struct chunk))) {
		memset (ch,0,sizeof(struct chunk));
		}
	return (ch);
	}

struct item *allocate_item (int fail) {
	int i;
	struct chunk *ch;
	struct item *p = NULL;

	if (!chunk_list)
		if (!(chunk_list = get_chunk())) {
			fprintf (stderr,"Failed to allocate a chunk of items\n");
			exit (1);
			}
	ch = chunk_list;
	while (ch && !p) {
		for (i=0; i < ITEMS_PER_CHUNK; i++)
			if (ch->in_use[i] == 0) {
				p = &ch->item[i];
				memset (p,0,sizeof(struct item));
				ch->in_use[i] = 1;
				break;
				}
		if (!p)
			if (ch->next) ch = ch->next;
			else
				if (ch->next = get_chunk()) ch = ch->next;
				else {
					fprintf (stderr,"Failed to allocate a chunk of items\n");
					exit (1);
					}
		}
	return (p);
	}

void deallocate_item (struct item *p) {
	int j;
	int found;
	struct chunk *ch;

	if (p) {
		if (p->d && p->len > 0)
			if (!p->original) free (p->d);

		ch = chunk_list;
		found = 0;
		while (ch && !found) {
			j = p - ch->item;
			if (j >= 0 && j < ITEMS_PER_CHUNK) {
				memset (p,0,sizeof(struct item));
				ch->in_use[j] = 0;
				found = 1;
				}
			else
				ch = ch->next;
			}
		}
	}

/*----------------------------------------------------------------------*\
 | Functions to insert items before or after a given node.				|
\*----------------------------------------------------------------------*/

struct item *insert_item_after (struct item *r) {
	struct item *p;

	if (!r) return (NULL);

	if (p = allocate_item (1)) {
		p->next = r->next;
		p->prev = r;
		p->parent = r->parent;
		p->indent = r->indent;
		if (p->next) p->next->prev = p;
		r->next = p;
		}
	return (p);
	}

struct item *insert_item_before (struct item *q) {
	struct item *r;

	if (!q) return (NULL);
	if (r = allocate_item (1)) {
		r->next = q;
		r->prev = q->prev;
		r->parent = q->parent;
		r->indent = q->indent;
		if (r->prev) r->prev->next = r;
		q->prev = r;
		if (r->parent)
			if (r->parent->child == q) r->parent->child = r;
		}
	return (r);
	}

/*----------------------------------------------------------------------*\
 | Function to add a child to a node.  If there are already children of	|
 | this node, the new child is added at the end of the list of children.|
\*----------------------------------------------------------------------*/

struct item *add_child (struct item *p) {
	struct item *q,*r;

	if (!p) return (NULL);
	if (p->child) {
		for (r=p->child; r->next; r=r->next);
		q = allocate_item (1);
		r->next = q;
		q->prev = r;
		q->parent = p;
		}
	else {
		q = allocate_item (1);
		p->child = q;
		q->parent = p;
		}
	q->indent = 1 + p->indent;
	q->line_number = 0;
	return (q);
	}

/*----------------------------------------------------------------------*\
 | Function to copy a subtree.											|
\*----------------------------------------------------------------------*/

struct item *copy_item (struct item *p) {
	struct item *q,*r,*s;

	q = NULL;
	if (p) {
		q = allocate_item (1);
		q->next = NULL;
		q->prev = NULL;
		q->parent = NULL;
		if (p->child) {
			q->child = copy_item (p->child);
			q->child->parent = q;
			s = q->child;
			for (r=p->child->next; r; r=r->next) {
				s->next = copy_item (r);
				s->next->prev = s;
				s->next->parent = q;
				s = s->next;
				}
			}
		q->indent = p->indent;
		q->line_number = p->line_number;
		q->key = p->key;
		q->original = 0;
		if (p->d) {
			q->len = 1 + strlen (p->d);
			q->d = (char *) malloc (q->len);
			strcpy (q->d,p->d);
			}
		else
			if (q->d = (char *) malloc (8)) {
				q->len = 8;
				memset (q->d,0,8);
				}
			else {
				q->d = NULL;
				q->len = 0;
				}
		}
	return (q);
	}

/*----------------------------------------------------------------------*\
 | Function to detach an item, keeping its children with it.			|
\*----------------------------------------------------------------------*/

void detach_item (struct item *p) {
	if (p->prev) p->prev->next = p->next;
	if (p->next) p->next->prev = p->prev;
	if (p->parent)
		if (p->parent->child == p) p->parent->child = p->next;
	p->parent = p->prev = p->next = NULL;
	}

/*----------------------------------------------------------------------*\
 | Function to delete an item.											|
\*----------------------------------------------------------------------*/

void delete_item (struct item *p) {

	if (!p) return;
	detach_item (p);
	if (p->child) delete_children (p);
	deallocate_item (p);
	}

/*----------------------------------------------------------------------*\
 | Delete all of the children of a given node.  This is recursive since	|
 | it must delete the children of the children.							|
\*----------------------------------------------------------------------*/

void delete_children (struct item *p) {
	struct item *q,*r;

	if (!p) return;
	while (q = p->child) {
		delete_item (q);
		}

	}

/*----------------------------------------------------------------------*\
 | Clear the values of all scalar children of a node.					|
\*----------------------------------------------------------------------*/

void clear_scalars (struct item *p) {
	struct item *q;

	if (p)
		if (p->key == Wunknown || p->key == Wblank) delete_item (p);
		else
			if (q = p->child) {
				if (q->next) {
					while (q->next) q = q->next;
					while (q) {
						clear_scalars (q->next);
						q = q->prev;
						}
					}
				clear_scalars (p->child);
				}
	}

/*----------------------------------------------------------------------*\
 | Increment indentation of a node and its children.					|
\*----------------------------------------------------------------------*/

void indent (struct item *p) {
	struct item *q;

	if (p->key != Wblank) p->indent++;
	else
		for (q=p->child; q; q=q->next) indent (q);
	}

/*----------------------------------------------------------------------*\
 | Locate a node in the tree by matching the key.  The first argument	|
 | is the first node of the tree to be searched, the second is the key	|
 | to be matched.														|
\*----------------------------------------------------------------------*/

struct item *find_key (struct item *p, enum fgdc_keyword key) {
	struct item *q,*r;
	if (p->key == key) return (p);
	else
		if (q = p->child)
			while (q)
				if (r = find_key (q,key)) return (r);
				else q = q->next;
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Locate the next node in a list that matches a key.  First argument	|
 | is the current node; it will not be considered in the search.  This	|
 | routine searches only nodes at the same level in the tree as the		|
 | node given by the first argument.  It does not search children.		|
\*----------------------------------------------------------------------*/

struct item *find_next_key (struct item *p, enum fgdc_keyword key) {
	p = p->next;
	while (p) {
		if (p->key == key) return (p);
		p = p->next;
		}
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Function to remove empty subtrees of a node.  This does not remove	|
 | the node at the head of the subtree, even if all its children are	|
 | removed in the process.  The UI has an easier time that way, because	|
 | it doesn't have to (a) figure out what the current node should be	|
 | and (b) contend with the possibility that you just deleted the		|
 | entire tree.															|
\*----------------------------------------------------------------------*/

static int has_data (struct item *p) {
	int result = 0;
	struct item *q;

	if (p) {
		if (p->key == Wunknown) result++;
		if (p->child)
			for (q=p->child; q; q=q->next) result += has_data (q);
		}
	return (result);
	}

int prune (struct item *p) {
	int changes = 0;
	struct item *q,*r;

	if (p)
		if (p->child) {
			for (q=p->child; q->next; q=q->next);
			while (q)
				if (has_data (q)) {
					changes += prune (q);
					q = q->prev;
					}
				else {
					int delete_it = 1;

					/*--------------------------------------------------*\
					 | Delete the empty node UNLESS it is a blank line	|
					 | following a non-blank line in a text value.		|
					\*--------------------------------------------------*/

					r = q;
					q = q->prev;
					if (q)
						if (q->key == Wunknown &&
							(r->key == Wunknown || r->key == Wblank) &&
							has_data (q)) delete_it = 0;

					if (delete_it) {
						delete_item (r);
						changes++;
						}
					}
			}
	return (changes);
	}

/*----------------------------------------------------------------------*\
 | Function to count the number of valid elements in a tree.			|
 |																		|
 | Note that this ignores blank and unknown nodes.						|
\*----------------------------------------------------------------------*/

static int count = 0;

static void count_item (struct item *p) {
	struct item *q;

	if (!p) return;
	if (p->key == Wunknown || p->key == Wblank) return;
	count++;
	for (q=p->child; q; q=q->next) count_item (q);
	}

int item_count (struct item *p) {
	count = 0;
	count_item (p);
	return (count);
	}

/*----------------------------------------------------------------------*\
 | Program to dump the parse tree.  This could be made prettier.		|
\*----------------------------------------------------------------------*/

void list_children (FILE *tf, struct item *p) {
	while (p) {
		fprintf (tf,"%s ",text_of(p->key));
		if (p->parent) fprintf (tf,"(%s) ",text_of(p->parent->key));
		else fprintf (tf,"(root) ");
		if (p->key == Wunknown)
			if (p->d) fprintf (tf,"= \"%s\"\n",p->d);
			else fprintf (tf,"= NULL\n");
		else fprintf (tf,"\n");
		if (p->child) list_children (tf,p->child);
		p = p->next;
		}
	}

/*----------------------------------------------------------------------*\
 | Determine whether node is a descendant of root.						|
\*----------------------------------------------------------------------*/

int valid_node (struct item *node, struct item *root) {
	struct item *p,*q;

	if (!node || !root) return (0);

	p = root;
	if (p == node) return (1);
	for (q=p->child; q; q=q->next)
		if (valid_node (node,q)) return (1);
	return (0);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

int wrap_text (struct item *p, int width) {
	struct item *q,*new,*r,*u,*v;
	int n;
	char *wrap_buffer = NULL;
	int wrap_buffer_len = 0;
	char *s,*t;
	int changed = 0;

	if (!p) return (0);
	if (p->key == Wunknown || p->key == Wblank) return (0);

	if (q = p->child) {

		/* Skip blanks */

		while (q && q->key == Wblank) q = q->next;

		/* Do all this only if it's a text value */

		if (q->key == Wunknown) {

			/*----------------------------------------------------------*\
			 | Assume, perhaps incorrectly, that 2-space indentation is	|
			 | being used.												|
			\*----------------------------------------------------------*/

			new = NULL;
			width -= 2 * q->indent;
			while (q) {

				/*------------------------------------------------------*\
				 | Calculate the length of the wrappable text.  This is	|
				 | all of the lines of text up to but not including the	|
				 | next blank line or the next line beginning with the	|
				 | greater-than sign '>'.								|
				\*------------------------------------------------------*/

				n = 0;
				for (r=q; r; r=r->next)
					if (r->key == Wblank) break;
					else
						if (r->d)
							if (*r->d == '>') break;
							else n += 1 + strlen (r->d);

				/* Allocate space for the wrappable text */

				if (n > wrap_buffer_len) {
					if (wrap_buffer) free (wrap_buffer);
					if (wrap_buffer = (char *) malloc (1 + n))
						wrap_buffer_len = n;
					}

				/* Copy the wrappable text into the buffer */

				if (t = wrap_buffer) {
					for (r=q; r; r=r->next)
						if (r->key == Wblank) break;
						else
							if (r->d)
								if (*r->d == '>') break;
								else {
									strcpy (t,r->d);
									t += strlen (r->d);
									*t++ = ' ';
									*t = 0;
									}

					/*--------------------------------------------------*\
					 | Parcel out the text into lines of width or fewer	|
					 | characters.										|
					\*--------------------------------------------------*/

					s = wrap_buffer;
					while (*s) {

						/*----------------------------------------------*\
						 | Point t at line break point, s + width if	|
						 | there are more characters on the line than	|
						 | width, or the end of the line if fewer.		|
						\*----------------------------------------------*/

						if (strlen (s) > width)
							t = s + width;
						else
							t = s + strlen (s);

						/*----------------------------------------------*\
						 | Back up to find the end of the current word.	|
						 | If that takes you to the beginning of the	|
						 | line, go forward instead.					|
						\*----------------------------------------------*/

						while (t > s && !isspace(*t)) t--;
						if (t <= s)
							while (*t && !isspace(*t)) t++;

						/*----------------------------------------------*\
						 | Now s marks the beginning and t the end of a	|
						 | block of text of the appropriate size.		|
						 | Copy this text into a the next element in a	|
						 | list whose first element is "new".			|
						\*----------------------------------------------*/

						if (new) {
							for (u=new; u->next; u=u->next);
							if (v = insert_item_after (u)) {
								v->key = Wunknown;
								v->indent = q->indent;
								v->parent = p;
								if (v->d = (char *) malloc (1 + t-s)) {
									v->len = t - s;
									memcpy (v->d,s,t-s);
									v->d[t-s] = 0;
									}
								v->original = 0;
								}
							else {
								/* should never happen */
								}
							}
						else {
							v = allocate_item (1);
							v->key = Wunknown;
							v->indent = q->indent;
							v->parent = p;
							if (v->d = (char *) malloc (1 + t-s)) {
								v->len = t - s;
								memcpy (v->d,s,t-s);
								v->d[t-s] = 0;
								}
							v->original = 0;
							new = v;
							}

						/*----------------------------------------------*\
						 | Advance s to the character following t, but	|
						 | not if t points to the end of the buffer.	|
						\*----------------------------------------------*/

						s = t + (*s ? 1 : 0);
						}
					}

				/*------------------------------------------------------*\
				 | Add blank lines to the list of new lines headed by	|
				 | the pointer "new".  Also add lines that begin with	|
				 | a greater-than sign without changing their text.		|
				\*------------------------------------------------------*/

				if (r)
					while (r) {
						if (r->key == Wblank) {
							if (new) {
								for (u=new; u->next; u=u->next);
								if (v = insert_item_after (u)) {
									v->key = Wblank;
									v->indent = q->indent;
									v->parent = p;
									v->d = NULL;
									v->len = 0;
									v->original = 1;
									}
								else {
									/* should never happen */
									}
								}
							else {
								v = allocate_item (1);
								v->key = Wblank;
								v->indent = q->indent;
								v->parent = p;
								v->d = NULL;
								v->len = 0;
								v->original = 0;
								new = v;
								}
							}
						else
							if (r->d)
								if (*r->d == '>') {
									if (new) {
										for (u=new; u->next; u=u->next);
										if (v = insert_item_after (u)) {
											v->key = Wunknown;
											v->indent = q->indent;
											v->parent = p;
											if (v->d = (char *) malloc (1 + strlen (r->d))) {
												v->len = strlen (r->d);
												strcpy (v->d,r->d);
												}
											v->original = 0;
											}
										else {
											/* should never happen */
											}
										}
									else {
										v = allocate_item (1);
										v->key = Wunknown;
										v->indent = q->indent;
										v->parent = p;
										if (v->d = (char *) malloc (1 + strlen (r->d))) {
											v->len = strlen (r->d);
											strcpy (v->d,r->d);
											}
										v->original = 0;
										new = v;
										}
									}
								else break;
						r = r->next;
						}

				/*------------------------------------------------------*\
				 | Now r is either a line with plain text that doesn't	|
				 | begin with a greater-than sign or r is NULL, meaning	|
				 | it's the last of the children of p.					|
				\*------------------------------------------------------*/

				q = r;
				}

			/*----------------------------------------------------------*\
			 | New now replaces the entire list headed by p->child.		|
			\*----------------------------------------------------------*/

			if (new) {
				delete_children (p);
				p->child = new;
				changed = 1;
				}

			}
		}
	if (wrap_buffer) free (wrap_buffer);
	return (changed);
	}

int wrap_subtree (struct item *p, int width) {
	struct item *q;
	int result = 0;

	if (!p) return (0);
	if (p->key == Wunknown) return (0);

	/*------------------------------------------------------------------*\
	 | If the element p is a scalar, and holds text, then the first		|
	 | non-blank child of p will be of type Wunknown.  Wrap p to width.	|
	 | If the first non-blank child of p is not Wunknown, then call		|
	 | wrap_subtree on each of the children of p.						|
	\*------------------------------------------------------------------*/

	if (q = p->child) {
		while (q && q->key == Wblank) q=q->next;
		if (q->key == Wunknown)
			return (wrap_text (p,width));
		else
			for (q=p->child; q; q=q->next)
				result += wrap_subtree (q,width);
		}

	return (result);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
