# Program to create a point shapefile from a given tab-delimited
# text table, used in conjunction with Frank Warmerdam's shapelib
# utilities dbfcreate, dbfadd, shpcreate, and shpadd
#
# Usage: txt2shp text_file [XField YField]
#
# where text_file is a table, tab-delimited with field names
# in its first row.  XField is the name of the field whose
# values are X coordinates; YField is the name of the field
# whose values are Y coordinates.
#
# By default, XField is "longitude" and YField "latitude"
#
# If XField and YField are not given or those names don't match
# any of the fields in the table, the shp and shx files are not
# created, only the dbf is created.
#
# The output of this script is a single dbfcreate command,
# a single shpcreate command, and a shpadd and dbfadd command
# for each row of the table.  This can be run as a shell script
# or batch file.
#
# Field names are truncated to 10 characters if they are longer.
# Punctuation and blanks in field names are replaced by underscores.
# Field names are written in entirely upper case letters
#
# Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)

if {$argc > 0} {
	set input_file [lindex $argv 0]

	set x_field longitude
	set y_field latitude
	set x_index -1
	set y_index -1

	if {$argc > 2} {
		set x_field [lindex $argv 1]
		set y_field [lindex $argv 2]

		if {[string length $x_field] > 10} {
			set x_field [string range $x_field 0 9]
			}
		if {[string length $y_field] > 10} {
			set y_field [string range $y_field 0 9]
			}
		}

	puts "input_file: $input_file"

	# Map characters that cannot appear in DBF field names to underscore
	set map {
		\~   _
		\`   _
		\!   _
		\@   _
		\#   _
		\$   _
		\%   _
		\&   _
		\*   _
		\(   _
		\)   _
		\-   _
		\+   _
		\=   _
		\{   _
		\}   _
		\[   _
		\]   _
		\|   _
		\\   _
		\:   _
		\;   _
		\"   _
		\'   _
		\<   _
		\>   _
		\,   _
		\.   _
		\?   _
		\/   _
		" "  _
		}

	# Read headers

	set in [open $input_file r]
	set line [gets $in]
	set headers [split $line \t]
	set label_list {}
	set n 0
	foreach item $headers {
		set t [string trim $item "\""]
		set t [string map $map $t]
		if {[string length $t] > 10} {
			set t [string range $t 0 9]
			}
		set t [string toupper $t]

		if {[lsearch -exact $label_list $t] != -1} {
			puts "Field name $item, when rendered as $t would duplicate another field"
			exit
			}

		set label($n) $t
		lappend label_list $t

		if {[string compare -nocase $x_field $label($n)] == 0} {
			set x_index $n
			}
		if {[string compare -nocase $y_field $label($n)] == 0} {
			set y_index $n
			}

		set width($n) 0
		set integer($n) 1
		set double($n) 1
		set prec($n) 0
		set count($n) 0
		incr n
		}

	puts "$n fields"

	set do_shp 0
	if {$x_index != -1 && $y_index != -1} {
		puts "X coordinates in $label($x_index) (field $x_index)"
		puts "Y coordinates in $label($y_index) (field $y_index)"
		set do_shp 1
		}

	# Read data to figure out widths and types of the fields

	set nr 0
	while {[eof $in] == 0} {
		set line [gets $in]
		if {[string length $line] > 0} {
			set stuff [split $line \t]
			set i 0
			foreach item $stuff {
				set value [string trim $item "\""]
				set k [string length $value]
				if {$k > $width($i)} {
					set width($i) $k
					}
				if {$k > 0} {
					incr count($i)
					if {[string is integer $value] == 0} {
						set integer($i) 0
						}
					if {[string is double $value] && [string first . $value] != -1} {
						set point [string first . $value]
						set precision [expr $k - $point]
						if {$precision > $prec($i)} {
							set prec($i) $precision
							}
						} \
					else {
						set double($i) 0
						}
					}
				incr i
				}
			incr nr
			}
		}
	close $in
	puts "$nr records"

	for {set i 0} {$i < $n} {incr i} {
		puts -nonewline "$label($i)"
		if {$integer($i)} {
			puts -nonewline " integer ($width($i))"
			} \
		else {
			if {$double($i)} {
				puts -nonewline " double ($width($i).$prec($i))"
				} \
			else {
				puts -nonewline " string ($width($i))"
				}
			}
		puts "\t$count($i) values"
		}

	# Create a shell script to create the shapefile

	set out [open "[file rootname $input_file].sh" w]

	if {$do_shp} {
		set shpfile "[file rootname $input_file].shp"
		puts $out "shpcreate $shpfile point"
		}

	set output_file "[file rootname $input_file].dbf"
	set cmd "dbfcreate $output_file"

	for {set i 0} {$i < $n} {incr i} {
		if {$count($i) > 0} {
			if {$integer($i)} {
				append cmd " -n $label($i) $width($i) $prec($i)"
				} \
			else {
				if {$double($i)} {
					append cmd " -n $label($i) $width($i) $prec($i)"
					} \
				else {
					append cmd " -s $label($i) $width($i)"
					}
				}
			}
		}
	puts $out "$cmd"

	# re-read the input file, creating commands to store the values

	set x ""
	set y ""

	set in [open $input_file r]
	# ignore header line this time
	set line [gets $in]
	while {[eof $in] == 0} {
		set line [gets $in]
		if {[string length $line] > 0} {
			set cmd "dbfadd $output_file"
			set stuff [split $line \t]
			set i 0
			foreach item $stuff {
				if {$count($i) > 0} {
					set k [string length $item]

					if {$k > 0} {
						append cmd " $item"
						} \
					else {
						append cmd " \"\""
						}

					if {$do_shp} {
						if {$i == $x_index} {
							set x [string trim $item "\""]
							}
						if {$i == $y_index} {
							set y [string trim $item "\""]
							}
						}
					}
				incr i
				}
			while {$i < $n} {
				append cmd " \"\""
				incr i
				}

			if {$do_shp} {
				puts $out "shpadd $shpfile $x $y"
				set x ""
				set y ""
				}
			puts $out "$cmd"
			}
		}
	close $in
	close $out
	} \
else {
	puts "Program to create a point shapefile from a given tab-delimited "
	puts "text table, used in conjunction with Frank Warmerdam's shapelib"
	puts "utilities dbfcreate, dbfadd, shpcreate, and shpadd"
	puts ""
	puts "Usage: txt2shp text_file \[XField YField\]"
	puts ""
	puts "where text_file is a table, tab-delimited with field names "
	puts "in its first row.  XField is the name of the field whose"
	puts "values are X coordinates; YField is the name of the field"
	puts "whose values are Y coordinates."
	puts ""
	puts "By default, XField is \"longitude\" and YField \"latitude\""
	puts ""
	puts "If XField and YField are not given or those names don't match"
	puts "any of the fields in the table, the shp and shx files are not"
	puts "created, only the dbf is created."
	puts ""
	puts "The output of this script is a single dbfcreate command,"
	puts "a single shpcreate command, and a shpadd and dbfadd command"
	puts "for each row of the table.  This can be run as a shell script"
	puts "or batch file."
	puts ""
	puts "Field names are truncated to 10 characters if they are longer."
	puts "Punctuation and blanks in field names are replaced by underscores."
	puts "Field names are written in entirely upper case letters"
	puts ""
	puts "Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)"
	}
