# Script to read a metadata record (text, XML, or SGML) and produce
# from it a web page containing the revision history that the
# metadata record describes.
#
# Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)

# Revision history for this script:
#
# Ver  Date      Who  What
# ------------------------------------------------------------------------
# 1.2  20060302  PNS  Added CSS stylesheets
# 1.1  20030723  PNS  Added summary attribute to main table for better
#                     accessibility.
#
# 1.0  20030321  PNS  Initial version shows date and description and
#                     ignores contact and sources used and produced.
#                     Note that banner and colors for page background
#                     and table background are hard-coded here, and
#                     URLs in the Process_Description texts are not
#                     converted into hypertext links.
# ------------------------------------------------------------------------

set version 1.2

package require mq

set header "<!--#include virtual=\"header\"-->"
set footer "<!--#include virtual=\"footer\"-->"
set stylesheets "  <link rel=\"stylesheet\" type=\"text/css\" href=\"http://www.usgs.gov/styles/common.css\"/>\n\
  <link rel=\"stylesheet\" type=\"text/css\" href=\"http://geology.usgs.gov/tools/metadata/local.css\"/>"
set bgcolor "#ffeedd"

if {$argc > 0} {
	set input_file [lindex $argv 0]
	} \
else {
	puts "Name the metadata file on the command line"
	exit
	}

proc show_date {ymd} {
	set month_name [list Error Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec]
	set y [string range $ymd 0 3]
	if {[string length $ymd] > 4} {
		if {[string index $ymd 4] == 0} {
			set m [string index $ymd 5]
			} \
		else {
			set m [string range $ymd 4 5]
			}
		set mon [lindex $month_name $m]
		if {[string length $ymd] > 6} {
			if {[string index $ymd 6] == 0} {
				set d [string index $ymd 7]
				} \
			else {
				set d [string range $ymd 6 7]
				}
			return "$d-$mon-$y"
			}
		return "$mon-$y"
		}
	return "$y"
	}

if {[metadata m -parse $input_file]} {
	if {[set p [$m find_first Identification_Information]]} {
		if {[set p [$m find_in $p Citation]]} {
			if {[set p [$m find_in $p Citation_Information]]} {
				if {[set p [$m find_in $p Title]]} {
					set title [$m value_of -nonewline $p]
					}
				}
			}
		}
	if {[set p [$m find_first Data_Quality_Information]]} {
		if {[set p [$m find_in $p Lineage]]} {
			if {[set q [$m find_in $p Process_Step]]} {
				set n 0
				while {$q} {
					if {[set r [$m find_in $q Process_Description]]} {
						set desc($n) [$m value_of $r]
						if {[set r [$m find_in $q Process_Date]]} {
							set date($n) [$m value_of -nonewline $r]
							if {$date($n) == ""} {
								set $date($n) "(no date given)"
								}
							} \
						else {
							set date($n) "(no date given)"
							}
						}
					incr n
					set q [$m find_next $q]
					}

				set nm1 [expr $n - 1]

				puts "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">"
				puts "<html>"
				puts "  <title>Revision history for [string map {> &gt; < &lt; & &amp;} $title]</title>"
				puts "$stylesheets"
				puts "</head>"
				puts "<body>"

				puts "$header"

				puts "<h1>[string map {> &gt; < &lt; & &amp;} $title]</h1>"
				puts "<h2>Revision history</h2>"

				puts "<table border=\"0\" cellpadding=\"4\" cellspacing=\"2\" bgcolor=\"$bgcolor\" summary=\"shows date and description of each revision, most recent revision first\">"

				puts "  <tr>"
				puts "    <th scope=\"column\">Date</th>"
				puts "    <th scope=\"column\" align=\"left\">Description of revision</th>"
				puts "  </tr>"

				for {set k $nm1} {$k >= 0} {incr k -1} {
					puts "  <tr valign=\"top\">"
					puts "    <th scope=\"row\" title=\"$date($k)\">[show_date $date($k)]</th>"
					puts "    <td>"

					set lines [split $desc($k) "\n"]
					set line_count [llength $lines]
					set last_line [expr $line_count - 1]
					set preformatting 0
					for {set j 0} {$j < [llength $lines]} {incr j} {
						set line [lindex $lines $j]

						# At a blank line, print <p></p>, but if the next line begins with >, don't
						# Similarly, if the previous line began with >, don't

						if {$j < $last_line} {
							set next [lindex $lines [expr $j + 1]]
							} \
						else {
							set next ""
							}
						if {[string length $line] == 0} {
							if {$preformatting} {
								puts "</pre>"
								set preformatting 0
								} \
							else {
								if {[string index $next 0] != ">"} {
									puts "<br/>"
									}
								}
							} \
						else {
							if {[string index $line 0] == ">"} {
								if {!$preformatting} {
									puts "<pre>"
									set preformatting 1
									}
								set line [string range $line 1 end]
								} \
							else {
								if {$preformatting} {
									puts "</pre>"
									set preformatting 0
									}
								}
							set line [string map {> &gt; < &lt; & &amp;} $line]
							puts "$line"
							}
						}
					if {$preformatting} {
						puts "</pre>"
						}
					puts "    </td>"

					puts "  </tr>"
					}
				puts "</table>"
				puts "<br/>"
				puts "<hr>"
				puts "Generated by <a href=\"http://geo-nsdi.er.usgs.gov/metadata/whatsnew.tcl\">whatsnew.tcl</a> $version on [clock format [clock seconds] -format "%d-%b-%Y %H:%M"]<br/>"
				puts "$footer"
				puts "</body>"
				puts "</html>"
				}
			}
		}
	$m forget
	}

# end
