/*----------------------------------------------------------------------*\
 | Code to read and write XML versions of formal metadata.				|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "keyword.h"
#include "item.h"
#include "config.h"
#include "stricmp.h"
#include "actions.h"

extern char *read_text (char *input_file);
extern int is_extension (int key);

extern void set_character_encoding (char *s);
extern char *get_character_encoding (void);

extern void unescape_string (char *s);

extern FILE *out;

extern char *iso8859_1[];
extern char *iso8859_1_numeric[];
extern char *dos_codeset[];
extern char *mac_codeset[];

/*----------------------------------------------------------------------*/

extern char *ps8_tag_of (enum fgdc_keyword key);
extern enum fgdc_keyword key_of_ps8_tag (char *ps8_tag);

static char *(*xml_tag_of)(enum fgdc_keyword key);
static enum fgdc_keyword (*key_of_xml_tag) (char *xml_tag);

static char space = 0x20;

enum xml_encoding {
	UTF8,
	ISO8859_1
	};

static char *name_of_encoding (enum xml_encoding e) {
	char *name = "unknown";
	if (e == UTF8) name = "UTF-8";
	if (e == ISO8859_1) name = "ISO-8859-1";
	return (name);
	}

/*----------------------------------------------------------------------*\
 | element_order, local to this module, indicates whether the elements	|
 | should be written out in the order given by the CSDGM (default) or	|
 | as they appear in the input.  This is an enum rather than a flag so	|
 | that it might be easier to allow users to specify alternative order	|
 | in the future (for example, order ESRI might specify the order of	|
 | elements in the ESRI DTD).											|
\*----------------------------------------------------------------------*/

enum element_order {
	STANDARD,
	ASIS
	};
static enum element_order element_order = STANDARD;

/*----------------------------------------------------------------------*\
 | When reading plain text, default encoding is ISO-8859-1, but when	|
 | reading XML, default is UTF-8.										|
\*----------------------------------------------------------------------*/

static enum xml_encoding encoding = ISO8859_1;

/*----------------------------------------------------------------------*\
 | Kludgy handling of XML attributes.
\*----------------------------------------------------------------------*/

static char *copy_of (char *string) {
	char *s = NULL;
	if (string)
		if (s = (char *) malloc (1 + strlen (string)))
			strcpy (s,string);
	return (s);
	}

struct xml_attr {
	struct xml_attr *next;
	char *name;
	char *value;
	};

static struct xml_attr *parse_attr (char *string) {
	struct xml_attr *a,*attr = NULL;
	char *b, *e, *s;
	char *name, *value;

	if (! string) return (NULL);
	if (!*string) return (NULL);

	s = string + strlen (string) - 1;
	if (*s == '?') *s = 0;

	b = string;
	while (*b) {
		name = value = NULL;
		while (*b && isspace(*b)) b++;
		name = b;
		e = b;
		while (*e && !isspace(*e) && *e != '=') e++;
		while (isspace(*e)) *e++ = 0;
		if (*e == '=') {
			*e++ = 0;

			b = e;
			while (*b && isspace(*b)) b++;
			e = b+1;
			if (*b == '\"') {
				while (*e && *e != '\"') e++;
				b++;
				if (*e == '\"') *e++ = 0;
				}
			else {
				while (*e && !isspace(*e)) e++;
				if (*e) *e++ = 0;
				}
			value = b;
			}
		if (name)
			if (!attr)
				if (attr = (struct xml_attr *) malloc (sizeof (struct xml_attr))) {
					attr->next = NULL;
					attr->name = copy_of (name);
					attr->value = copy_of (value);
					}
				else
					fprintf (stderr,"Error: could not allocate memory for struct xml_attr\n");
			else {
				for (a=attr; a->next; a=a->next);
				if (a->next = (struct xml_attr *) malloc (sizeof (struct xml_attr))) {
					a = a->next;
					a->next = NULL;
					a->name = copy_of (name);
					a->value = copy_of (value);
					}
				else
					fprintf (stderr,"Error: could not allocate memory for struct xml_attr\n");
				}
		b = e;
		while (*b && isspace(*b)) b++;
		}
	return (attr);
	}

struct xml_attr_list {
	struct xml_attr_list *next;
	struct item *node;
	struct xml_attr *attr;
	};

static struct xml_attr_list *attr_list = NULL;

static struct xml_attr *xml_attributes_of (struct item *node) {
	struct xml_attr_list *list = attr_list;
	while (list)
		if (list->node == node)
			return (list->attr);
		else
			list = list->next;
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Read the string pointed to by the argument as a single character		|
 | encoded in UTF-8.  Return the unicode value of the character, and	|
 | point the argument (an address passed by reference) to the second	|
 | character in the string, which is not the second byte if the first	|
 | character is not ASCII.												|
\*----------------------------------------------------------------------*/

static int unicode_of_utf8 (char **s) {
	int j,n = 0;
	char *t = *s;

	/* ASCII */

	if ((*t & 0x80) == 0) {
		n = *t++;
		*s = t;
		return (n);
		}

	/* UTF-8 */

	for (j=0; j < 8; j++)
		if ((*t & (0x80 >> j)) == 0) break;
	if (j >= 7) return (0);

	/*------------------------------------------------------------------*\
	 | j  byte 0   byte 1   byte 2   byte 3   byte 4   byte 5
	 | 2  110xxxxx 10xxxxxx
	 | 3  1110xxxx 10xxxxxx 10xxxxxx
	 | 4  11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
	 | 5  111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
	 | 6  1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
	\*------------------------------------------------------------------*/

	n = (0xFF >> j--) & *t++;
	while (j-- > 0) {
		if ((*t & 0xC0) != 0x80) {
			return (0);
			}
		else n = (n << 6) | (0x3F & *t++);
		}
	*s = t;
	return (n);
	}

/*----------------------------------------------------------------------*\
 | Function to verify the UTF-8 encoding of the given string.			|
 | Returns nonzero if the string can be interpreted as valid UTF-8.		|
 | Returns zero if there's an inconsistency.  For ASCII returns 1.		|
 |																		|
 | The return value is the length of the UTF-8 character at the given	|
 | address; to get to the next character, add the return value to the	|
 | input address.														|
\*----------------------------------------------------------------------*/

static int check_utf8 (unsigned char *s) {
	int i,n;

	/* ASCII is valid UTF-8 */

	if ((*s & 0x80) == 0) return (1);

	/* Find out how many additional bytes make up this UTF-8 character */

	if ((0xE0 & *s) == 0xC0) n = 1;
	else
		if ((0xF0 & *s) == 0xE0) n = 2;
		else
			if ((0xF8 & *s) == 0xF0) n = 3;
			else
				if ((0xFC & *s) == 0xF8) n = 4;
				else
					if ((0xFE & *s) == 0xFC) n = 5;
					else return (0);

	/*------------------------------------------------------------------*\
	 | Check each additional byte; must have high bit set, next highest	|
	 | clear.															|
	\*------------------------------------------------------------------*/

	for (i=0; i < n; i++)
 		if ((s[1+i] & 0xC0) != 0x80) return (0);

	return (1 + n);
	}

/*----------------------------------------------------------------------*\
 | Ridiculously primitive XML parser.  This reads XML into the tree		|
 | structure used by mp.  The input XML must be perfect or it won't		|
 | work.  Specifically, no end-tags may be omitted, the nesting of		|
 | elements must be correct, and attributes in the element tags will be	|
 | ignored.  There may be character entities in the text.				|
\*----------------------------------------------------------------------*/

enum tag_type_t {
	TAG_DECLARATION,
	TAG_START,
	TAG_END,
	TAG_SINGULAR,
	TAG_COMMENT
	};

struct item *decode_xml (char *buffer) {
	int i,n;
	char *output = NULL;
	int output_size = 0;
	char *d,*s,*t,*u;
	char *buffer_end,*b,*e,*next_b;
	enum fgdc_keyword key;
	struct item *root = NULL;
	struct item *current = NULL;
	struct item *p;
	int line_number;
	struct xml_attr *a,*aa;
	struct xml_attr_list *al;

	if (!buffer) return (NULL);

	xml_tag_of = ps8_tag_of;
	key_of_xml_tag = key_of_ps8_tag;

	line_number = 1;

	if (output = (char *) malloc (8192)) {
		output_size = 8192;
		}
	else {
		fprintf (stderr,"Error: could not allocate parse text buffer\n");
		exit (1);
		}

	d = output;
	*d = 0;
	b = buffer;
	buffer_end = buffer + strlen (buffer);

	/*------------------------------------------------------------------*\
	 | Walk through the file byte by byte.  Recognize and handle tags.	|
	\*------------------------------------------------------------------*/

	while (b < buffer_end) {

		/*--------------------------------------------------------------*\
		 | When you find a tag, determine whether it is a start tag, an	|
		 | end tag, or a declaration.									|
		\*--------------------------------------------------------------*/

		if (*b == '<') {
			enum tag_type_t tag_type = TAG_START;
			b++;
			if (*b == '/') {
				b++;
				tag_type = TAG_END;
				}
			else
				if (*b == '!' || *b == '?') {
					b++;
					tag_type = TAG_DECLARATION;
					if (memcmp (b,"--",2) == 0) tag_type = TAG_COMMENT;
					}

			/*------------------------------------------------------*\
			 | Find the end of the tag.  If you get to the end of a	|
			 | line, replace the end-of-line characters with spaces	|
			 | and proceed.											|
			\*------------------------------------------------------*/

			for (e=b; e < buffer_end; e++) {
				if (*e == '\r') {
					*e = ' ';
					if (*(e+1) == '\n') {
						e++;
						*e = ' ';
						}
					line_number++;
					}
				else
					if (*e == '\n') {
						*e = ' ';
						line_number++;
						}
				if (tag_type == TAG_COMMENT) {
					if (memcmp (e,"-->",3) == 0) {
						e += 2;

						/*------------------------------------------*\
						 | If only whitespace in output buffer,		|
						 | we consider it empty.					|
						\*------------------------------------------*/

						for (u=output; *u && isspace(*u); u++);
						if (*u == 0) *output = 0;

						break;
						}
					}
				else
					if (*e == '>') break;
				}

			/*------------------------------------------------------*\
			 | If you got to the end of the file and didn't find a	|
			 | closing angle bracket, print message and return NULL	|
			\*------------------------------------------------------*/

			if (!*e) {
				fprintf (stderr,"Error parsing XML (line %d): tag not closed\n",line_number);
				return (NULL);
				}

			/*------------------------------------------------------*\
			 | At this point you have the contents of a tag between	|
			 | b and e.												|
			\*------------------------------------------------------*/

			*e = 0;

			if (tag_type == TAG_DECLARATION)
				if (*(e-1) == '!' || *(e-1) == '?') *(e-1) = 0;

			if (tag_type == TAG_START)
				if (*(e-1) == '/') {
					*(e-1) = 0;
					tag_type = TAG_SINGULAR;
					}

			/*------------------------------------------------------*\
			 | If there are attributes, parse and store them.		|
			\*------------------------------------------------------*/

			if (tag_type != TAG_COMMENT) {
				a = NULL;
				for (u=b; *u; u++)
					if (isspace(*u)) {
						*u++ = 0;
						if (tag_type != TAG_END) a = parse_attr (u);
						break;
						}
				}

			/*------------------------------------------------------*\
			 | If this is the XML declaration, find out if the file	|
			 | uses a specific encoding such as UTF-8 or ISO-8859-1	|
			\*------------------------------------------------------*/

			if (tag_type == TAG_DECLARATION)
				if (stricmp (b,"xml") == 0)
					if (a)
						for (aa=a; aa; aa=aa->next)
							if (stricmp (aa->name,"encoding") == 0) {
								if (stricmp (aa->value,"UTF-8") == 0)
									encoding = UTF8;
								if (stricmp (aa->value,"ISO-8859-1") == 0)
									encoding = ISO8859_1;
								break;
								}

			/*------------------------------------------------------*\
			 | Figure out what the tag means						|
			\*------------------------------------------------------*/

			if (tag_type == TAG_START || tag_type == TAG_END || tag_type == TAG_SINGULAR) {

				/*----------------------------------------------*\
				 | When you find a start tag, create an element	|
				 | of this type and make it the current node.	|
				\*----------------------------------------------*/

				/*----------------------------------------------*\
				 | Create a dummy extension table entry so that	|
				 | we can parse unknown extensions rather than	|
				 | just throw them away.						|
				\*----------------------------------------------*/

				if ((key = key_of_xml_tag (b)) == Wunknown) {
					key = add_extension (b,current->key);
					fprintf (out,"Warning (line %d): Unknown extension element found: %s\n",line_number,b);
					}

				if ((key = key_of_xml_tag (b)) != Wunknown) {
					if (tag_type == TAG_START || tag_type == TAG_SINGULAR) {

						if (d > output) {
							char *w;
							for (w=output; isspace (*w); w++);
							if (*w)
								fprintf (stderr,"Error (line %3d): extraneous text in XML ignored: \"%s\"\n",
									line_number,output);
							d = output;
							*d = 0;
							}
						if (!current) {
							root = allocate_item (1);
							current = root;
							current->key = key;
							current->line_number = line_number;
							}
						else {
							p = add_child (current);
							current = p;
							current->key = key;
							current->line_number = line_number;
							}

						/*----------------------------------------------*\
						 | If this element has attributes, keep track	|
						 | of them.										|
						\*----------------------------------------------*/

						if (a)
							if (!attr_list)
								if (attr_list = (struct xml_attr_list *) malloc (sizeof (struct xml_attr_list))) {
									attr_list->next = NULL;
									attr_list->node = current;
									attr_list->attr = a;
									}
								else
									fprintf (stderr,"Error: could not allocate memory for xml_attr_list\n");
							else {
								for (al=attr_list; al->next; al=al->next);
								if (al->next = (struct xml_attr_list *) malloc (sizeof (struct xml_attr_list))) {
									al = al->next;
									al->next = NULL;
									al->node = current;
									al->attr = a;
									}
								else
									fprintf (stderr,"Error: could not allocate memory for xml_attr_list\n");
								}
						}

					if (tag_type == TAG_END || tag_type == TAG_SINGULAR)
						if (key == current->key) {
							char *w;

							/*------------------------------------------*\
							 | If only whitespace in output buffer,		|
							 | we consider it empty.					|
							\*------------------------------------------*/

							for (w=output; isspace(*w); w++);
							if (*w == 0) *output = 0;

							if (strlen (output) == 0) {
								if (current->parent) current = current->parent;
								}
							else
								if (p = add_child (current)) {
									p->key = Wunknown;
									p->line_number = line_number;

									/*----------------------------------*\
									 | If encoding is UTF-8, check the	|
									 | data to make sure the characters	|
									 | are valid UTF-8.					|
									\*----------------------------------*/

									if (encoding == UTF8) {
										s = output;
										t = s + strlen (s);
										while (s < t) {
											int j = check_utf8 ((unsigned char *)s);
											if (j > 0) s += j;
											else {
												fprintf (stderr,"Error (line %3d): invalid character encoding in UTF-8 string: ",line_number);
												for (j=0; s[j] && j < 5; j++)
													fprintf (stderr," 0x%02X",(unsigned char) s[j]);
												fprintf (stderr,".  Text value will be truncated at this point.\n");

												/*----------------------*\
												 | Truncate text value	|
												 | and try to proceed.	|
												\*----------------------*/

												*s = 0;
												break;
												}
											}
										}

									/*----------------------------------*\
									 | Store the data in p->d			|
									\*----------------------------------*/

									if (p->d = (char *) malloc (1 + strlen(output))) {
										for (s = output; *s && isspace(*s); s++);
										strcpy (p->d,s);
										d = output;
										*d = 0;
										if (current->parent) current = current->parent;
										}
									else {
										fprintf (stderr,"Error (line %3d): could not allocate space for text of %s\n",
											line_number,text_of(current->key));
										exit (1);
										}
									}
								else {
									fprintf (stderr,"Error (line %3d): could not allocate scalar child of %s\n",
										line_number,text_of(current->key));
									exit (1);
									}
							}
						else {
							fprintf (stderr,"Error (line %3d): expected end-tag for %s, got %s\n",
								line_number,text_of(current->key),text_of(key));
							exit (1);
							}
					}
				else {

					/*--------------------------------------------------*\
					 | If you don't recognize the tag, output a warning	|
					 | and clear the buffer containing its contents.	|
					\*--------------------------------------------------*/

					fprintf (out,"Warning (line %3d): unrecognized %s%s>\n",
						line_number,
						((tag_type == TAG_START) ? "start tag <" : "end tag </"),
						b
						);
					d = output;
					*d = 0;
					}
				}
			else {
				/*------------------------------------------------------*\
				 | If the tag is neither a start tag, an end tag, nor a	|
				 | singular tag, ignore it.								|
				\*------------------------------------------------------*/
				}
			b = e+1;
			}
		else

			/*--------------------------------------------------*\
			 | Found a character in the content of a tag.		|
			 | Do entity substitution if you can find one.		|
			 | Copy the byte if you cannot.						|
			\*--------------------------------------------------*/

			if (*b == '&') {
				for (i=0; i < 256; i++)
					if (t = iso8859_1[i])
						if (memcmp (b+1,t,strlen(t)) == 0)
							if (*(b+1+strlen(t)) == ';') {
								*d++ = i;
								*d = 0;
								b += 2 + strlen (t);
								break;
								}
				if (i >= 256) {
					*d++ = *b++;
					*d = 0;
					}
				}
			else
				if (*b == '\r' || *b == '\n') {

					/*--------------------------------------------------*\
					 | At the end of a line, if there's anything in the	|
					 | output buffer, create a Wunknown to hold it.  If	|
					 | the line is blank and you're in the middle of a	|
					 | text value, create a Wblank element to hold it.	|
					\*--------------------------------------------------*/

					if (*b == '\r') *b++ = 0;
					if (*b == '\n') *b++ = 0;

					if (strlen (output) > 0)
						if (p = add_child (current)) {
							p->key = Wunknown;
							p->line_number = line_number;

							/*----------------------------------*\
							 | If encoding is UTF-8, check the	|
							 | data to make sure the characters	|
							 | are valid UTF-8.					|
							\*----------------------------------*/

							if (encoding == UTF8) {
								s = output;
								t = s + strlen (s);
								while (s < t) {
									int j = check_utf8 ((unsigned char *)s);
									if (j > 0) s += j;
									else {
										fprintf (stderr,"Error (line %3d): invalid character encoding in UTF-8 string: ",line_number);
										for (j=0; s[j] && j < 5; j++)
											fprintf (stderr," 0x%02X",(unsigned char) s[j]);
										fprintf (stderr,".  Text value will be truncated at this point.\n");

										/*----------------------*\
										 | Truncate text value	|
										 | and try to proceed.	|
										\*----------------------*/

										*s = 0;
										break;
										}
									}
								}


							n = strlen (output);
							if (p->d = (char *) malloc (1 + n)) {
								for (s=output; *s && isspace(*s); s++);
								strcpy (p->d,s);
								p->len = 1 + n;
								p->original = 0;
								d = output;
								*d = 0;
								}
							else {
								fprintf (stderr,"Error (line %3d): could not allocate space for text of %s\n",
									line_number,text_of(current->key));
								exit (1);
								}
							}
						else
							if (current) {
								fprintf (stderr,"Error (line %3d): could not allocate scalar child of %s\n",
									line_number,text_of(current->key));
								exit (1);
								}
							else {
								/*--------------------------------------*\
								 | We come here if the first line of XML|
								 | is a comment.  I think it's okay to	|
								 | do nothing at this point.			|
								\*--------------------------------------*/
								}
					else
						if (current)
							if (current->child)
								if (current->child->key == Wunknown)
									if (p = add_child (current)) {
										p->key = Wblank;
										p->line_number = line_number;
										}
					line_number++;
					}
				else {

					/*--------------------------------------------------*\
					 | This character is not the end of a line, and is	|
					 | not an entity.  Just append it to the output		|
					 | buffer.											|
					\*--------------------------------------------------*/

					*d++ = *b++;
					*d = 0;
					}

		/*--------------------------------------------------------------*\
		 | If you're about to write past the end of the output buffer,	|
		 | don't.  Reallocate it for more space.						|
		\*--------------------------------------------------------------*/

		if (d - output >= output_size - 1) {
			n = d - output;
			if (output = (char *) realloc (output,4096+output_size)) {
				d = output + n;
				output_size += 4096;
				}
			else {
				fprintf (stderr,"Error (line %3d): could not reallocate parse text buffer\n",
					line_number);
				exit (1);
				}
			}
		}
	if (output) free (output);
	set_character_encoding (name_of_encoding (encoding));
	return (root);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

#ifndef USE_OLD_SGML_PARSER

struct item *parse_sgml (char *input_file) {
	char *s;
	struct item *p = NULL;

	encoding = ISO8859_1; /* When reading SGML, default encoding is ISO-8859-1 */

	if (s = read_text (input_file)) {
		p = decode_xml (s);
		free (s);
		}
	return (p);
	}

#endif

struct item *parse_xml (char *input_file) {
	char *s;
	struct item *p = NULL;

	encoding = UTF8; /* When reading XML, default encoding is UTF-8 */

	if (s = read_text (input_file)) {
		p = decode_xml (s);
		free (s);
		}
	return (p);
	}

struct item *parse_xml_string (char *string) {
	encoding = UTF8; /* When parsing XML, default encoding is UTF-8 */

	return (decode_xml (string));
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static char **entity = iso8859_1;
static int output_extensions = 1;
static char *what_to_print_for_Wblank = "\n";
static char *one_indent = "";
static int indent_len = 0;
static char *empty = "";

static FILE *fp = NULL;

static void write_xml_text (unsigned char *s) {
	char *t;
	fflush (fp);
	while (*s) {
		if (*s == '<') fputs ("&lt;",fp);
		else
			if (*s == '>') fputs ("&gt;",fp);
			else
				if (*s == '&') fputs ("&amp;",fp);
				else
					if (*s == '\'') fputs ("&apos;",fp);
					else
						if (*s == '"') fputs ("&quot;",fp);
						else
							fputc (*s,fp);

#ifdef OLD_CODE
		if (t = entity[*s]) {
			fputc ('&',fp);
			while (*t) fputc (*t++,fp);
			fputc (';',fp);
			}
		else
			fputc (*s,fp);
#endif /* OLD_CODE */

		s++;
		}
	fflush (fp);
	}

static void write_xml_item (struct item *p, int n) {
	int i;
	enum fgdc_keyword *e,*ee;

	if (!p) return;
	if (!output_extensions && is_extension (p->key)) return;
	if (p->key == Wblank) {
		if (p->next && p->next->key == Wunknown)
			fprintf (fp,what_to_print_for_Wblank);
		}
	else
		if (p->key == Wunknown) {
			if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);
			write_xml_text ((unsigned char *) p->d);
			fprintf (fp,"\n");
			}
		else {
			struct item *q;
			struct xml_attr_list *al;
			struct xml_attr *a;

			if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);

			/*----------------------------------------------------------*\
			 | Write the start tag.  This used to be simple, but is now	|
			 | more complicated because attributes might be attached to	|
			 | this element.
			\*----------------------------------------------------------*/

			fprintf (fp,"<%s",xml_tag_of(p->key));
			if (a = xml_attributes_of (p))
				while (a) {
					fprintf (fp," %s=\"%s\"",a->name,a->value);
					a = a->next;
					}
			fprintf (fp,">");

			/*----------------------------------------------------------*\
			 | Write the element's content
			\*----------------------------------------------------------*/

			if (p->child)
				if (p->child->key == Wunknown && p->child->next == NULL)
					write_xml_text ((unsigned char *)p->child->d);
				else {
					fprintf (fp,"\n");

					switch (element_order) {
						case STANDARD:

							/*------------------------------------------*\
							 | Write the children of p, ordering the	|
							 | elements as specified in the CSDGM. This	|
							 | is supposed to be faithfully reflected	|
							 | in actions.c.							|
							\*------------------------------------------*/

							if (e = element_list_of (p->key)) {
								for (ee = e; *ee != Wnull; ee++)
									for (q=p->child; q; q=q->next)
										if (q->key == *ee) write_xml_item (q,n+1);

								/*--------------------------------------*\
								 | Here we write out in xml any child	|
								 | elements that aren't supposed to be	|
								 | according to actions.c.				|
								 | This presents us with a dilemma.  If	|
								 | we do output them, we make invalid	|
								 | (but well-formed) xml.  If we don't	|
								 | output them, we lose whatever info	|
								 | they and all of their children may	|
								 | contain. I opt for keeping them.		|
								\*--------------------------------------*/

								for (q=p->child; q; q=q->next) {
									for (ee = e; *ee != Wnull; ee++)
										if (q->key == *ee) break;
									if (*ee == Wnull) write_xml_item (q,n+1);
									}
								}

							else
								/*--------------------------------------*\
								 | Normally this shouldn't happen, but	|
								 | there's no reason not to cover this	|
								 | situation.  Write the children as	|
								 | given in the parse tree.				|
								\*--------------------------------------*/

								for (q=p->child; q; q=q->next) write_xml_item (q,n+1);
							break;

						case ASIS:

							/*------------------------------------------*\
							 | Write the children of p, ordering the	|
							 | elements as given in the parse tree.		|
							\*------------------------------------------*/

							for (q=p->child; q; q=q->next) write_xml_item (q,n+1);
							break;
						}


					if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);
					}

			/*----------------------------------------------------------*\
			 | Write the end tag for the element						|
			\*----------------------------------------------------------*/

			fprintf (fp,"</%s>\n",xml_tag_of(p->key));
			}
	}

#define DOCTYPE_DECLARATION	"<!DOCTYPE metadata SYSTEM \"http://www.fgdc.gov/metadata/fgdc-std-001-1998.dtd\">"

void write_xml (char *output_file, struct item *first) {
	int i = 0;
	struct item *p;
	char *s;
	void *opt;
	char *default_doctype = DOCTYPE_DECLARATION;
	char *doctype = NULL;
	char *style_type = NULL;
	char *style_href = NULL;

	if (first == NULL) return;

	if (fp = fopen (output_file,"w")) {

		xml_tag_of = ps8_tag_of;
		key_of_xml_tag = key_of_ps8_tag;

		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"xml")) {
				void *v;
				if (v = find_option (opt,"indent")) {
					char *s = text_of_option (v);
					if (!s) s = empty;
					if (*s) unescape_string (s);
					one_indent = s;
					indent_len = strlen (one_indent);
					}
				if (v = find_option (opt,"blanks")) {
					char *s = text_of_option (v);
					if (!s) s = empty;
					if (*s) unescape_string (s);
					what_to_print_for_Wblank = s;
					}
				if (v = find_option (opt,"doctype"))
					if (s = text_of_option (v))
						doctype = s;
					else
						doctype = default_doctype;
				if (v = find_option (opt,"skip_extensions"))
					output_extensions = 0;
				if (v = find_option (opt,"stylesheet")) {
					void *vv;
					if (vv = find_option (v,"type"))
						if (s = text_of_option (vv))
							if (*s) style_type = s;
					if (vv = find_option (v,"href"))
						if (s = text_of_option (vv))
							if (*s) style_href = s;
					}
				if (v = find_option (opt,"order"))
					if (s = text_of_option (v))
						if (stricmp (s,"standard") != 0) element_order = ASIS;
				}

		if (stricmp (get_character_encoding(),"UTF-8") == 0) encoding = UTF8;
		else encoding = ISO8859_1;

		fprintf (fp,"<?xml version=\"1.0\" encoding=\"%s\"?>\n",name_of_encoding (encoding));
		if (doctype) fprintf (fp,"%s\n",doctype);
		if (style_href) {
			if (!style_type) style_type = "text/xsl";
			fprintf (fp,"<?xml-stylesheet type=\"%s\" href=\"%s\"?>\n",style_type,style_href);
			}

		for (p=first; p; p=p->next) write_xml_item (p,i);
		fclose (fp);
		}
	else {
		fprintf (stderr,"Error: could not create output file %s\n",output_file);
		return;
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
