/*----------------------------------------------------------------------*\
 | Code to read and write XML versions of formal metadata.				|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <expat.h>

#include "keyword.h"
#include "item.h"
#include "config.h"
#include "stricmp.h"
#include "actions.h"

extern char *read_text (char *input_file);
extern int is_extension (int key);
extern int add_extension (const char *tag, enum fgdc_keyword parent);

extern void set_character_encoding (char *s);
extern char *get_character_encoding (void);
extern int unicode_of_utf8 (char **s);

extern void unescape_string (char *s);

extern FILE *out;

extern const char *ps8_tag_of (enum fgdc_keyword key);
extern enum fgdc_keyword key_of_ps8_tag (const char *ps8_tag);

static const char *(*xml_tag_of)(enum fgdc_keyword key);
static enum fgdc_keyword (*key_of_xml_tag) (const char *xml_tag);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static char *copy_of (const XML_Char *string) {
	char *s = NULL;
	if (string)
		if (s = (char *) malloc (1 + strlen (string)))
			strcpy (s,string);
	return (s);
	}

struct xml_attr {
	struct xml_attr *next;
	char *name;
	char *value;
	};

struct xml_attr_list {
	struct xml_attr_list *next;
	struct item *node;
	struct xml_attr *attr;
	};

static struct xml_attr_list *attr_list = NULL;

static void save_attr (struct item *node, struct xml_attr *attr) {
	struct xml_attr_list *al;
	if (!attr_list)
		if (attr_list = (struct xml_attr_list *) malloc (sizeof (struct xml_attr_list))) {
			attr_list->next = NULL;
			attr_list->node = node;
			attr_list->attr = attr;
			}
		else
			fprintf (stderr,"Error: could not allocate memory for xml_attr_list\n");
	else {
		for (al=attr_list; al->next; al=al->next);
		if (al->next = (struct xml_attr_list *) malloc (sizeof (struct xml_attr_list))) {
			al = al->next;
			al->next = NULL;
			al->node = node;
			al->attr = attr;
			}
		else
			fprintf (stderr,"Error: could not allocate memory for xml_attr_list\n");
		}
	}

static struct xml_attr *xml_attributes_of (struct item *node) {
	struct xml_attr_list *list = attr_list;
	while (list)
		if (list->node == node)
			return (list->attr);
		else
			list = list->next;
	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | New parser implementation based on expat
\*----------------------------------------------------------------------*/

int allow_unknown_extensions = 0;

static XML_Parser parser;
static struct item *root = NULL;
static struct item *current = NULL;

static XML_Char *output = NULL;
static XML_Char *output_ptr = NULL;
static int output_size = 0;

void element_start (void *userData, const XML_Char *name, const XML_Char **atts) {
	int i;
	int line_number = XML_GetCurrentLineNumber (parser);
	enum fgdc_keyword key;
	struct item *p;

	if (!out) out = stderr;

	/* Flag as an error any text in the buffer other than whitespace */

	if (output)
		if (output_ptr > output) {
			char *s = output;
			while (*s && isspace (*s)) s++;
			if (*s) {
				fprintf (out,"Error (line %d): Misplaced text in XML before element %s, will be discarded\n",line_number,name);
				output_ptr = output;
				*output_ptr = 0;
				}
			}

	key = key_of_xml_tag (name);
	if (key == Wunknown)
		if (allow_unknown_extensions) {
			key = add_extension (name,current->key);
			fprintf (out,"Warning (line %d): Unknown extension element found: %s\n",line_number,name);
			}
		else {
			fprintf (out,"Warning (line %d): Unknown extension element ignored: %s\n",line_number,name);
			return;
			}

	if (!current) {
		root = allocate_item (1);
		current = root;
		current->key = key;
		current->line_number = line_number;
		}
	else {
		p = add_child (current);
		current = p;
		current->key = key;
		current->line_number = line_number;
		}

	if (atts && *atts) {
		struct xml_attr *a,*attr = NULL;
		int i;

		if (attr = (struct xml_attr *) malloc (sizeof (struct xml_attr))) {
			attr->next = NULL;
			attr->name = copy_of(atts[0]);
			attr->value = copy_of(atts[1]);
			}
		else
			fprintf (stderr,"Error: could not allocate memory for struct xml_attr\n");
		a = attr;

		for (i=2; atts[i]; i += 2) {
			if (a->next = (struct xml_attr *) malloc (sizeof (struct xml_attr))) {
				a = a->next;
				a->next = NULL;
				a->name = copy_of(atts[i]);
				a->value = copy_of(atts[i+1]);
				}
			else
				fprintf (stderr,"Error: could not allocate memory for struct xml_attr\n");
			}

		save_attr (current,attr);
		}
	}

void element_end (void *userData, const XML_Char *name) {
	enum fgdc_keyword key;

	if (!out) out = stderr;

	key = key_of_xml_tag (name);
	if (key == Wunknown)
		if (allow_unknown_extensions == 0) {
			if (output_ptr > output) {
				char *s = output;
				while (*s && (isspace(*s) || *s == '\r' || *s == '\n')) s++;
				if (*s) fprintf (out,"Warning (line %d): Discarding text found in unknown extension %s\n",XML_GetCurrentLineNumber(parser),name);
				}
			output_ptr = output;
			if (output) *output = 0;
			return;
			}

	if (output_ptr > output) {
		char *s,*t;
		int i,n;
		struct item *p;
		while (output_ptr > output) {
			t = output_ptr - 1;
			if (*t == '\n' || *t == '\r' || isspace(*t)) {
				*t-- = 0;
				output_ptr--;
				}
			else
				break;
			}
		for (s = output; *s && isspace(*s); s++);
		while (*s == '\r' || *s == '\n') s++;
		n = output_ptr - s;
		if (n > 0) {
			char *b,*e;
			b = s;
			while (*b) {
				e = b;
				while (*e && *e != '\r' && *e != '\n') e++;
				if (*e == '\r') *e++ = ' ';

				if (e == b) {
					if (p = add_child (current)) {
						p->key = Wblank;
						p->line_number = current->line_number;
						p->d = NULL;
						p->len = 0;
						p->original = 0;
						}
					else {
						fprintf (stderr,"Error: could not allocate space to store character data node\n");
						exit (1);
						}
					}
				else
					if (t = malloc ((1+e-b) * sizeof(XML_Char))) {
						memcpy (t,b,(e-b) * sizeof(XML_Char));
						t[e-b] = 0;
						if (p = add_child (current)) {
							p->key = Wunknown;
							p->line_number = current->line_number;
							p->d = t;
							p->len = 1 + n;
							p->original = 0;
							}
						else {
							fprintf (stderr,"Error: could not allocate space to store character data node\n");
							exit (1);
							}
					}
				else {
					fprintf (stderr,"Error: could not allocate space to store character data\n");
					exit (1);
					}

				b = e;
				if (*b) b++;
				}
			}
		}
	output_ptr = output;
	if (output) *output = 0;
	if (current->parent) current = current->parent;
	}

void data_handler (void *userData, const XML_Char *s, int len) {
	int n;

	if (!output) {
		n = len * sizeof (XML_Char);
		n = 8192 * (1 + n / 8192);
		if (output = (XML_Char *) malloc (n)) {
			output_size = n;
			*output = 0;
			output_ptr = output;
			}
		else {
			fprintf (stderr,"Error: could not allocate parse text buffer\n");
			exit (1);
			}
		}

	/* How much space do we need? */

	n = 1 + (output_ptr - output) + len * sizeof(XML_Char);

	/* If we don't have enough, make more */

	if (n > output_size) {
		int m = output_ptr - output;
		n = 8192 * (1 + n / 8192);
		if (output = (XML_Char *) realloc (output,n)) {
			output_size = n;
			output_ptr = output + m;
			}
		else {
			fprintf (stderr,"Error: could not reallocate parse text buffer\n");
			exit (1);
			}
		}
	memcpy (output_ptr,s,len*sizeof(XML_Char));
	output_ptr += len;
	*output_ptr = 0;
	}

void comment_handler (void *userData, const XML_Char *s) {
/*\
	fprintf (stderr,"Comment: \"");
	while (*s) fputc (*s++,stderr);
	fprintf (stderr,"\"\n");
\*/
	}

void default_handler (void *userData, const XML_Char *s, int len) {
/*\
	fprintf (stderr,"Default_Handler called for \"");
	while (len--) fputc (*s++,stderr);
	fprintf (stderr,"\"\n");
\*/
	}

extern char *entity_list;

struct item *decode_xml (char *buffer) {
	enum XML_Status status;
	xml_tag_of = ps8_tag_of;
	key_of_xml_tag = key_of_ps8_tag;

	if (parser = XML_ParserCreate (NULL)) {
		char *s;
		XML_SetElementHandler (parser,element_start,element_end);
		XML_SetCharacterDataHandler (parser,data_handler);
		XML_SetDefaultHandler (parser,default_handler);
		XML_SetCommentHandler (parser,comment_handler);

		/*--------------------------------------------------------------*\
		 | If the buffer starts with <metadata>, then you might need a	|
		 | DOCTYPE declaration that defines character entities.			|
		 | This will be the case when parsing SGML generated by mp.		|
		\*--------------------------------------------------------------*/

		if (memcmp (buffer,"<metadata>",10) == 0)
			status = XML_Parse (parser,entity_list,strlen(entity_list),0);

		status = XML_Parse (parser,buffer,strlen(buffer),1);

		if (status != XML_STATUS_OK) {
			fprintf (stderr,"%s\n",XML_ErrorString (XML_GetErrorCode(parser)));
			fprintf (stderr,"at line   %d\n",XML_GetCurrentLineNumber(parser));
			fprintf (stderr,"   column %d\n",XML_GetCurrentColumnNumber(parser));

			exit (1);
			}
		XML_ParserFree (parser);
		}

	/* Expat always converts character data to UTF-8 */
	set_character_encoding ("UTF-8");

	return (root);
	}

/*----------------------------------------------------------------------*/

extern char *iso8859_1[];
extern char *iso8859_1_numeric[];
extern char *dos_codeset[];
extern char *mac_codeset[];

/*----------------------------------------------------------------------*/

static char space = 0x20;

enum xml_encoding {
	UTF8,
	ISO8859_1
	};

/*----------------------------------------------------------------------*\
 | When reading plain text, default encoding is ISO-8859-1, but when	|
 | reading XML, default is UTF-8.										|
\*----------------------------------------------------------------------*/

static enum xml_encoding encoding = ISO8859_1;

static char *name_of_encoding (enum xml_encoding e) {
	char *name = "unknown";
	if (e == UTF8) name = "UTF-8";
	if (e == ISO8859_1) name = "ISO-8859-1";
	return (name);
	}

/*----------------------------------------------------------------------*\
 | element_order, local to this module, indicates whether the elements	|
 | should be written out in the order given by the CSDGM (default) or	|
 | as they appear in the input.  This is an enum rather than a flag so	|
 | that it might be easier to allow users to specify alternative order	|
 | in the future (for example, order ESRI might specify the order of	|
 | elements in the ESRI DTD).											|
\*----------------------------------------------------------------------*/

enum element_order {
	STANDARD,
	ASIS
	};
static enum element_order element_order = STANDARD;

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

#ifndef USE_OLD_SGML_PARSER

struct item *parse_sgml (char *input_file) {
	char *s;
	struct item *p = NULL;

	encoding = ISO8859_1; /* When reading SGML, default encoding is ISO-8859-1 */

	if (s = read_text (input_file)) {
		char *t;

		/*--------------------------------------------------------------*\
		 | Expat is choking on the DOCTYPE declaration of SGML files,	|
		 | so skip past it if you see one.								|
		\*--------------------------------------------------------------*/

		t = strstr (s,"<metadata>");
		if (!t) t = s;

		p = decode_xml (t);
		free (s);
		}
	return (p);
	}

#endif

struct item *parse_xml (char *input_file) {
	char *s;
	struct item *p = NULL;

	encoding = UTF8; /* When reading XML, default encoding is UTF-8 */

	/*------------------------------------------------------------------*\
	 | Since this may be called after the document is closed (by Tkme),	|
	 | we need to reinitialize the root and current node pointers.		|
	\*------------------------------------------------------------------*/

	root = NULL;
	current = NULL;

	if (s = read_text (input_file)) {
		p = decode_xml (s);
		free (s);
		}
	return (p);
	}

struct item *parse_xml_string (char *string) {
	encoding = UTF8; /* When parsing XML, default encoding is UTF-8 */

	return (decode_xml (string));
	}

/*----------------------------------------------------------------------*\
 | Keep track of those elements the user has asked be omitted from the	|
 | output file.  Their keys are stored in a static array of size 1024,	|
 | which is consulted whenever an element is to be output.				|
\*----------------------------------------------------------------------*/

static enum fgdc_keyword omit_key[1024];
static int omit_key_count = -1;

static int omit (enum fgdc_keyword key) {
	int i,result = 0;
	if (omit_key_count > 0)
		for (i=0; i < omit_key_count; i++)
			if (key == omit_key[i]) {
				result = 1;
				break;
				}
	return (result);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/

static char **entity = iso8859_1;
static int output_extensions = 1;
static char *what_to_print_for_Wblank = "\n";
static char *one_indent = "";
static int indent_len = 0;
static char *empty = "";

static FILE *fp = NULL;

static void write_xml_text (unsigned char *s) {
	char *t;
	fflush (fp);
	while (*s) {
		if (*s == '<') fputs ("&lt;",fp);
		else
			if (*s == '>') fputs ("&gt;",fp);
			else
				if (*s == '&') fputs ("&amp;",fp);
				else
					if (*s == '\'') fputs ("&apos;",fp);
					else
						if (*s == '"') fputs ("&quot;",fp);
						else
							fputc (*s,fp);

		s++;
		}
	fflush (fp);
	}

static void write_xml_item (struct item *p, int n) {
	int i;
	enum fgdc_keyword *e,*ee;

	if (!p) return;
	if (omit (p->key)) return;
	if (!output_extensions && is_extension (p->key)) return;
	if (p->key == Wblank) {
		if (p->next && p->next->key == Wunknown)
			fprintf (fp,what_to_print_for_Wblank);
		}
	else
		if (p->key == Wunknown) {
			if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);
			write_xml_text ((unsigned char *) p->d);
			fprintf (fp,"\n");
			}
		else {
			struct item *q;
			struct xml_attr_list *al;
			struct xml_attr *a;

			if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);

			/*----------------------------------------------------------*\
			 | Write the start tag.  This used to be simple, but is now	|
			 | more complicated because attributes might be attached to	|
			 | this element.
			\*----------------------------------------------------------*/

			fprintf (fp,"<%s",xml_tag_of(p->key));
			if (a = xml_attributes_of (p))
				while (a) {
					fprintf (fp," %s=\"%s\"",a->name,a->value);
					a = a->next;
					}
			fprintf (fp,">");

			/*----------------------------------------------------------*\
			 | Write the element's content
			\*----------------------------------------------------------*/

			if (p->child)
				if (p->child->key == Wunknown && p->child->next == NULL)
					write_xml_text ((unsigned char *)p->child->d);
				else {
					fprintf (fp,"\n");

					switch (element_order) {
						case STANDARD:

							/*------------------------------------------*\
							 | Write the children of p, ordering the	|
							 | elements as specified in the CSDGM. This	|
							 | is supposed to be faithfully reflected	|
							 | in actions.c.							|
							\*------------------------------------------*/

							if (e = element_list_of (p->key)) {
								for (ee = e; *ee != Wnull; ee++)
									for (q=p->child; q; q=q->next)
										if (q->key == *ee) write_xml_item (q,n+1);

								/*--------------------------------------*\
								 | Here we write out in xml any child	|
								 | elements that aren't supposed to be	|
								 | according to actions.c.				|
								 | This presents us with a dilemma.  If	|
								 | we do output them, we make invalid	|
								 | (but well-formed) xml.  If we don't	|
								 | output them, we lose whatever info	|
								 | they and all of their children may	|
								 | contain. I opt for keeping them.		|
								\*--------------------------------------*/

								for (q=p->child; q; q=q->next) {
									for (ee = e; *ee != Wnull; ee++)
										if (q->key == *ee) break;
									if (*ee == Wnull) write_xml_item (q,n+1);
									}
								}

							else
								/*--------------------------------------*\
								 | Normally this shouldn't happen, but	|
								 | there's no reason not to cover this	|
								 | situation.  Write the children as	|
								 | given in the parse tree.				|
								\*--------------------------------------*/

								for (q=p->child; q; q=q->next) write_xml_item (q,n+1);
							break;

						case ASIS:

							/*------------------------------------------*\
							 | Write the children of p, ordering the	|
							 | elements as given in the parse tree.		|
							\*------------------------------------------*/

							for (q=p->child; q; q=q->next) write_xml_item (q,n+1);
							break;
						}


					if (indent_len) for (i=0; i < n; i++) fprintf (fp,one_indent);
					}

			/*----------------------------------------------------------*\
			 | Write the end tag for the element						|
			\*----------------------------------------------------------*/

			fprintf (fp,"</%s>\n",xml_tag_of(p->key));
			}
	}

#define DOCTYPE_DECLARATION	"<!DOCTYPE metadata SYSTEM \"http://www.fgdc.gov/metadata/fgdc-std-001-1998.dtd\">"

void write_xml (char *output_file, struct item *first) {
	int i = 0;
	struct item *p;
	char *s;
	void *opt;
	char *default_doctype = DOCTYPE_DECLARATION;
	char *doctype = NULL;
	char *style_type = NULL;
	char *style_href = NULL;
	char *schema = NULL;

	if (first == NULL) return;

	if (fp = fopen (output_file,"w")) {

		xml_tag_of = ps8_tag_of;
		key_of_xml_tag = key_of_ps8_tag;

		if (opt = find_option (NULL,"output"))
			if (opt = find_option (opt,"xml")) {
				void *v;
				if (v = find_option (opt,"indent")) {
					char *s = text_of_option (v);
					if (!s) s = empty;
					if (*s) unescape_string (s);
					one_indent = s;
					indent_len = strlen (one_indent);
					}
				if (v = find_option (opt,"blanks")) {
					char *s = text_of_option (v);
					if (!s) s = empty;
					if (*s) unescape_string (s);
					what_to_print_for_Wblank = s;
					}
				if (v = find_option (opt,"doctype"))
					if (s = text_of_option (v))
						doctype = s;
					else
						doctype = default_doctype;
				if (v = find_option (opt,"skip_extensions"))
					output_extensions = 0;
				if (v = find_option (opt,"stylesheet")) {
					void *vv;
					if (vv = find_option (v,"type"))
						if (s = text_of_option (vv))
							if (*s) style_type = s;
					if (vv = find_option (v,"href"))
						if (s = text_of_option (vv))
							if (*s) style_href = s;
					}
				if (v = find_option (opt,"order"))
					if (s = text_of_option (v))
						if (stricmp (s,"standard") != 0) element_order = ASIS;
				if (v = find_option (opt,"schema"))
					if (s = text_of_option(v))
						schema = s;
				}

		/*------------------------------------------------------------------*\
		 | For up to 1024 keys, the user can specify elements that will not	|
		 | be included in the output.										|
		\*------------------------------------------------------------------*/

		if (omit_key_count == -1)
			if (opt = find_option (NULL,"output"))
				if (opt = find_option (opt,"xml"))
					if (opt = find_option (opt,"omit"))
						do {
							enum fgdc_keyword key;
							s = text_of_option (opt);
							if (*s) unescape_string (s);
							if ((key = key_of (s)) != Wunknown) {
								if (omit_key_count == -1) omit_key_count = 0;
								if (omit_key_count < 1024)
									omit_key[omit_key_count++] = key;
								}
							} while (opt = find_next_option (opt,"omit"));

		if (stricmp (get_character_encoding(),"UTF-8") == 0) encoding = UTF8;
		else encoding = ISO8859_1;

		/*--------------------------------------------------------------*\
		 | If a schema document was specified in the config file, then	|
		 | attach to the root element of the tree two attributes, one	|
		 | defining the xsi namespace, the other using it to specify	|
		 | the schema document itself.									|
		\*--------------------------------------------------------------*/

		if (schema && !xml_attributes_of (root)) {
			struct xml_attr *a,*attr = NULL;

			if (attr = (struct xml_attr *) malloc (sizeof (struct xml_attr))) {
				attr->next = NULL;
				attr->name = "xmlns:xsi";
				attr->value = "http://www.w3.org/2001/XMLSchema-instance";
				}
			else
				fprintf (stderr,"Error: could not allocate memory for struct xml_attr\n");
			a = attr;

			if (a->next = (struct xml_attr *) malloc (sizeof (struct xml_attr))) {
				a = a->next;
				a->next = NULL;
				a->name = "xsi:noNamespaceSchemaLocation";
				a->value = schema;
				}
			else
				fprintf (stderr,"Error: could not allocate memory for struct xml_attr\n");

			save_attr (root,attr);
			}

		fprintf (fp,"<?xml version=\"1.0\" encoding=\"%s\"?>\n",name_of_encoding (encoding));
		if (doctype) fprintf (fp,"%s\n",doctype);
		if (style_href) {
			if (!style_type) style_type = "text/xsl";
			fprintf (fp,"<?xml-stylesheet type=\"%s\" href=\"%s\"?>\n",style_type,style_href);
			}

		for (p=first; p; p=p->next) write_xml_item (p,i);
		fclose (fp);
		}
	else {
		fprintf (stderr,"Error: could not create output file %s\n",output_file);
		return;
		}
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
