/*----------------------------------------------------------------------*\
 | Xt-based Metadata Editor												|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston, VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Viewport.h>

#include "keyword.h"
#include "item.h"
#include "config.h"
#include "actions.h"
#include "revision.h"
#include "stricmp.h"

extern void read_local (char *local_file);
extern void read_tips (char *tip_file);
extern struct item *parse_text (char *file_name, char **buffer);
extern struct item *parse_sgml (char *file_name);
extern struct item *parse_xml (char *file_name);
extern struct item *parse_string (char *string);
extern struct item *parse_xml_string (char *string);
extern int check_syntax (struct item *p);
extern void write_text (char *file_name, struct item *p);
extern void write_sgml (char *file_name, struct item *p);
extern void write_xml (char *file_name, struct item *p);
extern char *encode_text (struct item *p);
extern struct item *item_of (int line_number);

extern void read_help (char *help_file);
extern char *advice_about (enum fgdc_keyword key);

#define min(a,b)	(((a) < (b)) ? (a) : (b))
#define max(a,b)	(((a) > (b)) ? (a) : (b))

/*----------------------------------------------------------------------*\
 | Public data															|
\*----------------------------------------------------------------------*/

struct item *root = NULL;
struct item *current = NULL;
char *one_indent = "  ";
FILE *out = NULL;

/*----------------------------------------------------------------------*\
 | Static data															|
\*----------------------------------------------------------------------*/

static Display *display = NULL;
static char *input_file = NULL;
static char name_of_file[FILENAME_MAX] = "";
static int node_count = 0;
static int changed = 0;
static char print_command [FILENAME_MAX] = "lp -c";

extern char *about_text[];

static int line_height = 0;
static int top_line = 0;
static int bottom_line = 0;

static Widget
		topLevel,
			frame,
				button_box,
					file_menu_button,
						file_menu,
							save_button,
							save_as_button,
							print_button,
							quit_button,
					edit_menu_button,
						edit_menu,
							cut_button,
							copy_button,
							paste_button,
							clear_button,
							dup_button,
							swap_button,
							prune_button,
					view_menu_button,
						view_menu,
							hide_button,
							show_button,
							wrap_value_button,
					add_menu_button,
					help_menu_button,
						help_menu,
							about_button,
							help_button,
							all_button,
							tips_button,
				element_viewport,
					element_list,
				value_text,
				tip_label,
			pshell,
				info_box,
					info_button_box,
						dismiss_button,
						wrap_toggle,
					info_text,
			oshell,
				output_box,
					output_button_box,
						output_view_menu_button,
						output_view_menu,
							output_dismiss_button,
							output_update_button,
							output_wrap_toggle,
							output_search_button,
						output_jumpto,
						output_line_number_goto,
						output_line_number,
					output_text,
			qshell,
				save_as_dialog,
					save_as_dialog_cancel,
					save_as_dialog_ok,
			rshell,
				confirm_dialog,
					confirm_dialog_cancel,
					confirm_dialog_ok,
			dummy;

/*----------------------------------------------------------------------*\
 | Application resources												|
\*----------------------------------------------------------------------*/

typedef struct {
	String configFile;
	String helpFile;
	String keyboardFocusPolicy;
	String language;
	} AppData, *AppDataPtr;

#define XtNconfigFile			"configFile"
#define XtCConfigFile			"ConfigFile"

#define XtNhelpFile				"helpFile"
#define XtCHelpFile				"HelpFile"

#define XtNkeyboardFocusPolicy	"keyboardFocusPolicy"
#define XtCKeyboardFocusPolicy	"KeyboardFocusPolicy"

#define XtNlanguage				"language"
#define XtCLanguage				"Language"

static XtResource resources[] = {
	{XtNconfigFile,		XtCConfigFile,		XtRString,	sizeof(String),		XtOffset(AppDataPtr,configFile),		XtRImmediate,	(XtPointer) ".mprc"},
	{XtNhelpFile,		XtCHelpFile,		XtRString,	sizeof(String),		XtOffset(AppDataPtr,helpFile),			XtRImmediate,	(XtPointer) "xtme-en.hlp"},
	{XtNkeyboardFocusPolicy,XtCKeyboardFocusPolicy,XtRString,sizeof(String),XtOffset(AppDataPtr,keyboardFocusPolicy),XtRImmediate,	(XtPointer) "Explicit"},
	{XtNlanguage,		XtCLanguage,		XtRString,	sizeof(String),		XtOffset(AppDataPtr,language),			XtRImmediate,	(XtPointer) "en"}
	};

static String fallback_resources[] = {

"*frame*font: 8x13",
"*frame.width: 640",
"*frame.height: 480",

"*button_box.background: Navy",
"*button_box.foreground: White",

"*file_menu_button.background: Navy",
"*file_menu_button.foreground: White",
"*file_menu*background: Blue",
"*file_menu*foreground: White",
"*file_menu*translations: #override \\n\
<EnterWindow>: highlight() tip()\\n\
<LeaveWindow>: unhighlight() tip()\\n\
<BtnMotion>: highlight() tip()\\n\
<BtnUp>: MenuPopdown() notify() unhighlight() tip()",

"*edit_menu_button.background: Navy",
"*edit_menu_button.foreground: White",
"*edit_menu*background: Blue",
"*edit_menu*foreground: White",
"*edit_menu*translations: #override \\n\
<EnterWindow>: highlight() tip()\\n\
<LeaveWindow>: unhighlight() tip()\\n\
<BtnMotion>: highlight() tip()\\n\
<BtnUp>: MenuPopdown() notify() unhighlight() tip()",

"*view_menu_button.background: Navy",
"*view_menu_button.foreground: White",
"*view_menu*background: Blue",
"*view_menu*foreground: White",
"*view_menu*translations: #override \\n\
<EnterWindow>: highlight() tip()\\n\
<LeaveWindow>: unhighlight() tip()\\n\
<BtnMotion>: highlight() tip()\\n\
<BtnUp>: MenuPopdown() notify() unhighlight() tip()",

"*add_menu_button.background: Navy",
"*add_menu_button.foreground: White",
"*add_menu*background: Blue",
"*add_menu*foreground: White",

"*help_menu_button.background: Navy",
"*help_menu_button.foreground: White",
"*help_menu*background: Blue",
"*help_menu*foreground: White",
"*help_menu*translations: #override \\n\
<EnterWindow>: highlight() tip()\\n\
<LeaveWindow>: unhighlight() tip()\\n\
<BtnMotion>: highlight() tip()\\n\
<BtnUp>: MenuPopdown() notify() unhighlight() tip()",

"*element_viewport*background: LightSteelBlue",
"*element_viewport*foreground: Black",
"*element_viewport*min: 150",
"*element_list*translations: #override \\n\
<Btn2Down>: paste()\\n\
Meta<Key>s: save()\\n\
Meta<Key>q: quit()\\n\
Meta<Key>x: save() quit()\\n\
Meta<Key>d: duplicate()\\n\
Meta<Key>k: cut()\\n\
Meta<Key>p: prune()\\n\
Meta<Key>v: paste()\\n\
Meta<Key>grave: swap()\\n\
<Key>Up: list-up()\\n\
<Key>Down: list-down()\\n\
<Key>KP_8: list-up()\\n\
<Key>KP_2: list-down()\\n\
<Key>Home: list-home()\\n\
<Key>KP_7: list-home()\\n\
<Key>Left: hide()\\n\
<Key>Right: show()\\n\
<Key>KP_4: hide()\\n\
<Key>KP_6: show()\\n\
<Key>Prior: list-pageup()\\n\
<Key>Next: list-pagedown()\\n\
<Key>KP_9: list-pageup()\\n\
<Key>KP_3: list-pagedown()\\n\
<Key>KP_Subtract: collapse()\\n\
<Key>KP_Add: expand()",

"*value_text*background: LightSteelBlue",
"*value_text*foreground: Black",
"*value_text*min: 150",
"*value_text.translations: #override \\n\
!<Key>F10: check-topology()\\n\
Meta<Key>q: quit()\\n\
Meta<Key>p: prune()\\n\
Meta<Key>s: save()\\n\
Meta<Key>t: today()\\n\
Meta<Key>x: save() quit()\\n\
Meta<Key>grave: swap()\\n\
Meta<Key>Delete: select-all() delete-selection()\\n\
!<Key>Delete: delete-next-character()\\n\
!<Key>KP_1: end-of-line()\\n\
!<Key>KP_2: next-line()\\n\
!<Key>KP_3: next-page()\\n\
!<Key>KP_4: backward-character()\\n\
!<Key>KP_6: forward-character()\\n\
!<Key>KP_7: beginning-of-line()\\n\
!<Key>KP_8: previous-line()\\n\
!<Key>KP_9: previous-page()\\n\
<Key>Home: beginning-of-line()\\n\
<Key>End: end-of-line()\\n\
<Key>Prior: previous-page()\\n\
<Key>Next: next-page()\\n\
Ctrl<Key>KP_1: end-of-file()\\n\
Ctrl<Key>KP_2: next-line()\\n\
Ctrl<Key>KP_3: scroll-one-line-up()\\n\
Ctrl<Key>KP_4: backward-word()\\n\
Ctrl<Key>KP_6: forward-word()\\n\
Ctrl<Key>KP_7: beginning-of-file()\\n\
Ctrl<Key>KP_8: previous-line()\\n\
Ctrl<Key>KP_9: scroll-one-line-down()\\n\
Ctrl<Key>Prior: scroll-one-line-down()\\n\
Ctrl<Key>Next: scroll-one-line-up()\\n\
Ctrl<Key>Home: beginning-of-file()\\n\
Ctrl<Key>End: end-of-file()\\n\
Ctrl<Key>b: form-paragraph()\\n\
Meta<Key>a: add-all()\\n\
Meta<Key>d: duplicate()\\n\
Meta<Key>h: help()\\n\
Meta<Key>k: cut()\\n\
Meta<Key>v: paste()\\n\
Meta<Key>KP_8: list-up()\\n\
Meta<Key>KP_2: list-down()\\n\
Meta<Key>KP_7: list-home()\\n\
Meta<Key>KP_4: hide()\\n\
Meta<Key>KP_6: show()\\n\
Meta<Key>KP_9: list-pageup()\\n\
Meta<Key>KP_3: list-pagedown()\\n\
Meta<Key>KP_Subtract: collapse()\\n\
Meta<Key>KP_Add: expand()",

"*tip_label.font: 7x13",
"*tip_label.background: Navy",
"*tip_label.foreground: White",

"*info_box.width: 640",
"*info_box.height: 480",

"*info_button_box.background: Navy",
"*info_button_box.foreground: White",
"*info_button_box*font: 9x15",
"*dismiss_button*background: LightSteelBlue",
"*dismiss_button*foreground: Black",
"*wrap_toggle*background: LightSteelBlue",
"*wrap_toggle*foreground: Black",

"*info_text*background: LightSteelBlue",
"*info_text*foreground: Black",
"*info_text*font: 7x13",
"*info_text.translations: #override \\n\
<Key>Escape: dismiss()\\n\
Meta<Key>s: save()\\n\
Meta<Key>q: quit()\\n\
Meta<Key>x: save() quit()\\n\
<Key>KP_1: end-of-line()\\n\
<Key>KP_2: next-line()\\n\
<Key>KP_3: next-page()\\n\
<Key>KP_4: backward-character()\\n\
<Key>KP_6: forward-character()\\n\
<Key>KP_7: beginning-of-line()\\n\
<Key>KP_8: previous-line()\\n\
<Key>KP_9: previous-page()\\n\
<Key>Prior: previous-page()\\n\
<Key>Next: next-page()\\n\
Ctrl<Key>KP_1: end-of-file()\\n\
Ctrl<Key>KP_2: next-line()\\n\
Ctrl<Key>KP_3: scroll-one-line-up()\\n\
Ctrl<Key>KP_4: backward-word()\\n\
Ctrl<Key>KP_6: forward-word()\\n\
Ctrl<Key>KP_7: beginning-of-file()\\n\
Ctrl<Key>KP_8: previous-line()\\n\
Ctrl<Key>KP_9: scroll-one-line-down()\\n\
Ctrl<Key>Prior: scroll-one-line-down()\\n\
Ctrl<Key>Next: scroll-one-line-up()\\n\
Ctrl<Key>Home: beginning-of-file()\\n\
Ctrl<Key>End: end-of-file()",

"*output_box.width: 640",
"*output_box.height: 480",

"*output_button_box.background: Navy",
"*output_button_box.foreground: White",
"*output_button_box*font: 8x13",
"*output_view_menu_button.background: Navy",
"*output_view_menu_button.foreground: White",
"*output_view_menu.background: Blue",
"*output_view_menu.foreground: White",
"*output_dismiss_button*background: Navy",
"*output_dismiss_button*foreground: White",
"*output_update_button*background: Navy",
"*output_update_button*foreground: White",
"*output_wrap_toggle*background: Navy",
"*output_wrap_toggle*foreground: White",
"*output_search_button*background: Navy",
"*output_search_button*foreground: White",
"*output_jumpto*background: Navy",
"*output_jumpto*foreground: White",
"*output_line_number_goto.background: Navy",
"*output_line_number_goto.foreground: White",
"*output_line_number*background: LightSteelBlue",
"*output_line_number*foreground: Black",

"*output_text*background: LightSteelBlue",
"*output_text*foreground: Black",
"*output_text*font: 7x13",
"*output_text.translations: #override \\n\
<Btn1Down>: select-start() line-number()\\n\
<Key>Escape: dismiss-output()\\n\
Meta<Key>s: save()\\n\
Meta<Key>q: quit()\\n\
Meta<Key>x: save() quit()\\n\
<Key>KP_1: end-of-line()\\n\
<Key>KP_2: next-line() line-number()\\n\
<Key>KP_3: next-page() line-number()\\n\
<Key>KP_4: backward-character() line-number()\\n\
<Key>KP_6: forward-character() line-number()\\n\
<Key>KP_7: beginning-of-line()\\n\
<Key>KP_8: previous-line() line-number()\\n\
<Key>KP_9: previous-page() line-number()\\n\
<Key>Prior: previous-page() line-number()\\n\
<Key>Next: next-page() line-number()\\n\
<Key>Left: backward-character() line-number()\\n\
<Key>Right: forward-character() line-number()\\n\
<Key>Up: previous-line() line-number()\\n\
<Key>Down: next-line() line-number()\\n\
Ctrl<Key>KP_1: end-of-file() line-number()\\n\
Ctrl<Key>KP_2: next-line() line-number()\\n\
Ctrl<Key>KP_3: scroll-one-line-up()\\n\
Ctrl<Key>KP_4: backward-word() line-number()\\n\
Ctrl<Key>KP_6: forward-word() line-number()\\n\
Ctrl<Key>KP_7: beginning-of-file() line-number()\\n\
Ctrl<Key>KP_8: previous-line() line-number()\\n\
Ctrl<Key>KP_9: scroll-one-line-down()\\n\
Ctrl<Key>Prior: scroll-one-line-down()\\n\
Ctrl<Key>Next: scroll-one-line-up()\\n\
Ctrl<Key>Home: beginning-of-file() line-number()\\n\
Ctrl<Key>End: end-of-file() line-number()",

"*save_as_dialog*font: 9x15bold",
"*save_as_dialog.background: Navy",
"*save_as_dialog.foreground: White",
"*save_as_dialog.label.background: Navy",
"*save_as_dialog.label.foreground: White",
"*save_as_dialog.value*background: LightSteelBlue",
"*save_as_dialog.value*foreground: Black",
"*save_as_dialog_ok.background: Green",
"*save_as_dialog_ok.foreground: Black",
"*save_as_dialog_cancel.background: Red",
"*save_as_dialog_cancel.foreground: Black",

"*confirm_dialog*font: 9x15bold",
"*confirm_dialog.background: Navy",
"*confirm_dialog.foreground: White",
"*confirm_dialog.label.background: Navy",
"*confirm_dialog.label.foreground: White",
"*confirm_dialog_ok.background: Green",
"*confirm_dialog_ok.foreground: Black",
"*confirm_dialog_cancel.background: Red",
"*confirm_dialog_cancel.foreground: Black",

	NULL
	};

/*----------------------------------------------------------------------*\
 | Handle character encoding
\*----------------------------------------------------------------------*/

static char character_encoding[256] = "ISO-8859-1";

void set_character_encoding (char *new_value) {
	if (strlen (new_value) < 256)
		strcpy (character_encoding,new_value);
	else {
		memcpy (character_encoding,new_value,255);
		character_encoding[255] = 0;
		}
	}

char *get_character_encoding (void) {
	return (character_encoding);
	}

/*----------------------------------------------------------------------*\
 | Here the original item structure created by the text parser is fixed	|
 | so that the indent member holds the level of nesting, not the number	|
 | of whitespaces that preceded the element name in the input file.		|
\*----------------------------------------------------------------------*/

static void adjust_indent (struct item *p, int n) {
	struct item *q;

	if (p) {
		if (p->key != Wblank) p->indent = n;
		for (q=p->child; q; q=q->next) adjust_indent (q,n+1);
		}
	}

/*----------------------------------------------------------------------*\
 | Build a linear array of pointers to items that will be displayed.	|
\*----------------------------------------------------------------------*/

static struct item **line = NULL;
static int line_limit = 0;
static int line_count = 0;

static void build_list (struct item *p) {
	struct item *q;

	if (p) {
		if (p->key != Wunknown && p->key != Wblank && p->indent >= 0) {

			/*----------------------------------------------------------*\
			 | The line list is built as a dynamically-allocated array	|
			\*----------------------------------------------------------*/

			if (line_limit == 0) {
				line = (struct item **) calloc (1024,sizeof(struct item *));
				line_limit = 1024;
				line_count = 0;
				}

			line [line_count++] = p;

			if (line_count >= line_limit) {
				line = (struct item **) realloc (line,(line_limit + 1024)*sizeof(struct item *));
				memset (line+line_limit,0,1024*sizeof(struct item *));
				line_limit += 1024;
				}

			for (q=p->child; q; q=q->next) build_list (q);
			}
		}
	}

/*----------------------------------------------------------------------*/

static String *element = NULL;
static int element_count = 0;

static void build_element_list (void) {
	struct item *q;

	if (element_count == 0) {
		if (element = (String *) calloc (line_count,sizeof(String *))) {
			int i,n;
			char *s;

			for (i=0; i < line_count; i++) {
				n = line[i]->indent * 2;
				s = text_of(line[i]->key);
				element[i] = (String) malloc (1 + n + strlen(s));
				memset (element[i],' ',n);
				strcpy (element[i]+n,s);
				if (q = line[i]->child)
					if (q->key != Wunknown &&
						q->key != Wblank &&
						q->indent < 0) *element[i] = '+';
				}
			}
		element_count = line_count;
		}
	}

/*----------------------------------------------------------------------*/

static void rebuild_lists (void) {
	int i;
	if (element_count && element) {
		for (i=0; i < element_count; i++) {
			if (element[i]) {
				free (element[i]);
				element[i] = NULL;
				}
			}
		}
	if (element) {
		free (element);
		element = NULL;
		}
	element_count = 0;
	line_count = 0;
	build_list (root);
	build_element_list ();
	}

/*----------------------------------------------------------------------*\
 | Code, actions, and callbacks for the list widget.					|
\*----------------------------------------------------------------------*/

static void set_text (struct item *p) {
	XawTextEditType edit_type = XawtextRead;
	Boolean caret = False;
	char *buffer = NULL;
	char *string;

	/*------------------------------------------------------------------*\
	 | If the selected element is compound, just display the element	|
	 | name in the text widget, and don't allow it to be edited.		|
	 |																	|
	 | If the selected element is a scalar, build a buffer containing	|
	 | the text in its children, and pass this, editable, to the text	|
	 | widget.  If there are no children, pass an empty string for		|
	 | editing to the text widget.										|
	\*------------------------------------------------------------------*/

	if (element_list_of (p->key)) string = p->d;
	else {
		edit_type = XawtextEdit;
		caret = True;
		if (p->child) {
			struct item *q;
			int n = 0;

			for (q=p->child; q; q=q->next) n += 1 + strlen (q->d);

			if (buffer = (char *) malloc (1 + n)) {
				char *s = buffer;
				for (q=p->child; q; q=q->next) {
					strcpy (s,q->d);
					s += strlen (q->d);
					*s++ = '\n';
					*s = 0;
					}
				string = buffer;
				}
			else string = "Error: could not allocate sufficient space";
			}
		else
			string = "";
		}

	XtVaSetValues (value_text,
		XtNstring,string,
		XtNtype,XawAsciiString,
		XtNeditType,edit_type,
		XtNdisplayCaret,caret,
		NULL);
	if (buffer) free (buffer);
	}

static void get_text (struct item *p) {
	XawTextEditType edit_type;
	Widget source;
	char *string;

	XtVaGetValues (value_text,XtNeditType,&edit_type,NULL);
	if (edit_type == XawtextEdit) {
		XtVaGetValues (value_text,XtNtextSource,&source,NULL);
		if (XawAsciiSourceChanged (source)) {
			XtVaGetValues (value_text,XtNstring,&string,NULL);
			if (element_list_of (p->key))
				fprintf (stderr,"Error (internal): %s should not be editable\n",text_of(p->key));
			else {
				char *b,*e,*s;
				struct item *q;

				/*------------------------------------------------------*\
				 | To replace the value of an element, you must delete	|
				 | its existing children (i.e. the items containing the	|
				 | current text), then create new children.				|
				\*------------------------------------------------------*/

				delete_children (p);

				b = string;
				while (*b) {
					e = b;
					while (*e && *e != '\n') e++;

					s = (char *) malloc (1 + e - b);
					memcpy (s,b,e - b);
					s [e-b] = 0;

					q = add_child (p);
					q->key = Wunknown;
					q->d = s;
					q->original = 0;
					q->len = e - b + 1;
					b = e;
					if (*b) b++;
					}
				changed = 1;
				}
			}
		}
	}

static void make_add_menu (enum fgdc_keyword key);

static void list_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *q = (XawListReturnStruct *) call_data;
	struct item *p;

	get_text (current);
	p = line[q->list_index];
	set_text (p);
	make_add_menu (p->key);
	current = p;

	if (XtIsRealized (info_text)) {
		char *string = advice_about (current->key);
		XtVaSetValues (info_text,
			XtNstring,string,
			NULL);
		}
	}

/*----------------------------------------------------------------------*/

static void list_home_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	if (current != root) {
		get_text (current);
		current = root;
		XawListHighlight (element_list,0);
		set_text (current);
		make_add_menu (current->key);
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		XawViewportSetCoordinates (element_viewport,0,0);
		}
	}

static void list_down_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	int i;
	XawListReturnStruct *v;

	v = XawListShowCurrent (element_list);
	i = v->list_index;
	if (i < line_count - 1) {
		get_text (current);
		i++;
		current = line[i];
		XawListHighlight (element_list,i);
		set_text (current);
		make_add_menu (current->key);
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		if (i >= bottom_line)
			XawViewportSetCoordinates (element_viewport,0,
				line_height*(top_line + i - bottom_line + 1)
			);
		}
	}

static void list_up_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	int i;
	XawListReturnStruct *v;

	v = XawListShowCurrent (element_list);
	i = v->list_index;
	if (i > 0) {
		get_text (current);
		i--;
		current = line[i];
		XawListHighlight (element_list,i);
		set_text (current);
		make_add_menu (current->key);
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		if (i < top_line)
			XawViewportSetCoordinates (element_viewport,0,line_height*i);
		}
	}

static void list_pagedown_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	int i;
	XawListReturnStruct *v;

	v = XawListShowCurrent (element_list);
	i = v->list_index;
	if (bottom_line < line_count) {
		get_text (current);
		i = bottom_line - 1;
		current = line[i];
		XawListHighlight (element_list,i);
		set_text (current);
		make_add_menu (current->key);
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		XawViewportSetCoordinates (element_viewport,0,line_height*i);
		}
	}

static void list_pageup_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	int i;
	XawListReturnStruct *v;

	v = XawListShowCurrent (element_list);
	i = v->list_index;
	if (top_line > 0) {
		get_text (current);
		i = top_line - (bottom_line - top_line - 1);
		if (i < 0) i = 0;
		current = line[i];
		XawListHighlight (element_list,i);
		set_text (current);
		make_add_menu (current->key);
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		XawViewportSetCoordinates (element_viewport,0,line_height*i);
		}
	}

/*----------------------------------------------------------------------*\
 | Code to handle selections											|
\*----------------------------------------------------------------------*/

static struct item *selection = NULL;
static char *converted_selection = NULL;

static Boolean convert_selection (Widget w,Atom *selection_atom,Atom *target_atom,Atom *type_return,XtPointer *value_return,unsigned long *length_return,int *format_return) {
	if (*target_atom == XA_STRING) {
		if (converted_selection = encode_text (selection)) {
			*type_return = *target_atom;
			*value_return = converted_selection;
			*length_return = strlen (converted_selection);
			*format_return = 8;
			return (True);
			}
		else
			return (False);
		}
	else
		return (False);
	}

static void lose_selection (Widget w, Atom *selection_atom) {
	if (selection) {
		delete_children (selection);
		delete_item (selection);
		selection = NULL;
		}
	if (converted_selection) {
		free (converted_selection);
		converted_selection = NULL;
		}
	}

static void done_selection (Widget w, Atom *selection_atom, Atom *target_atom) {
	if (*target_atom == XA_STRING)
		if (converted_selection) {
			free (converted_selection);
			converted_selection = NULL;
			}
	}

static void paste_subtree (struct item *p) {
	XawListReturnStruct *v;
	int i;
	struct item *q,*r,*s;
	enum fgdc_keyword *key_list;
	int found = 0;

	if (!p) return;

	/* Find the current node */

	v = XawListShowCurrent (element_list);
	i = v->list_index;
	q = line [i];

	/*------------------------------------------------------------------*\
	 | Try to make the subtree p a child of the current node.  If p is	|
	 | not a potential child of q, try to make p a child of q's parent.	|
	 | If that cannot be done, keep looking up the tree for a place to	|
	 | put p until you get to the root node.  If the subtree cannot be	|
	 | a child of the root node, don't put it anywhere.					|
	\*------------------------------------------------------------------*/

	while (!found && q) {
		if (key_list = element_list_of (q->key))
			for (i=0; key_list[i] != Wnull; i++)
				if (p->key == key_list[i]) {

					/* Attach the subtree */

					if (q->child) {

						/*----------------------------------------------*\
						 | If the current node is one of the children	|
						 | of q, insert the subtree after the current	|
						 | node, otherwise make the subtree the last	|
						 | child of q.									|
						\*----------------------------------------------*/

						for (r=q->child; r->next; r=r->next)
							if (r == current) break;

						/*----------------------------------------------*\
						 | The subtree might not be headed by a single	|
						 | node.  It could be a list, and p is just the	|
						 | head of the list.  Then you need to connect	|
						 | the node after r, if there is one, to the	|
						 | last node in the list that heads the subtree	|
						 | rather than to p, the head of the list.		|
						\*----------------------------------------------*/

						for (s=p; s->next; s=s->next);
						s->next = r->next;
						if (s->next) s->next->prev = s;

						r->next = p;
						p->prev = r;
						}
					else {
						q->child = p;
						}

					/* Assign parent of the subtree */

					p->parent = q;
					if (p->next) {
						struct item *u;
						for (u=p->next; u; u=u->next) u->parent = q;
						}

					/* Adjust indentation of the subtree */

					adjust_indent (p,1+q->indent);
					if (p->next) {
						struct item *u;
						for (u=p->next; u; u=u->next) adjust_indent (u,1+q->indent);
						}

					changed = 1;
					found = 1;
					}
		q = q->parent;
		}

	/*------------------------------------------------------------------*\
	 | What happens if the subtree doesn't fit anywhere in the branches	|
	 | that lead to the current node?  Initially I think we should just	|
	 | ignore the selection.  If we traverse the entire tree looking	|
	 | for a place to put the subtree, we might well put it somewhere	|
	 | the user doesn't want it to be.  For example, if the subtree		|
	 | describes Contact_Information, that could go in any of four		|
	 | different places among the standard elements.					|
	 |
	 | In principle, we could do a linear search of the key lists that	|
	 | get returned from element_list_of(), to determine whether the	|
	 | top node in our subtree is found in only one place.  Then we		|
	 | could find the appropriate place (or make it!), and insert the	|
	 | subtree there.  This seems to me to be an awful lot of hand-		|
	 | holding for people who should really know what they are doing,	|
	 | and a(nother) complicated bit of code that can go haywire.		|
	\*------------------------------------------------------------------*/

	if (!found)
		/* Ideally we squawk a warning to the user via a dialog. */
		fprintf (stderr,"Warning: no place to paste this information\n");
		;

	/*------------------------------------------------------------------*\
	 | Adjust the list widget to reflect the addition of the subtree.	|
	\*------------------------------------------------------------------*/

	if (found) {
		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);

		current = p;
		make_add_menu (current->key);
		set_text (current);
		for (i=0; i < line_count; i++)
			if (line[i] == current) break;
		XawListHighlight (element_list,i);
		if (i >= bottom_line)
			XawViewportSetCoordinates (element_viewport,0,
				line_height*(top_line + i - bottom_line + 1));
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		}
	}

/*----------------------------------------------------------------------*\
 | This is the requestor callback.  Decode the string returned by the	|
 | selection owner and parse it, yielding a subtree.  Store the subtree	|
 | as "selection"; it can be pasted using paste_action or paste_cb.		|
\*----------------------------------------------------------------------*/

static void get_selection_cb (Widget w, XtPointer client_data, Atom *selection_atom, Atom *type_atom, XtPointer value, unsigned long *length, int *format) {
	char *s;
	struct item *p;

	get_text (current);
	if (value) {
		if (*length > 0)
			if (*type_atom == XA_STRING)

				/*------------------------------------------------------*\
				 | parse_string requires a null-terminated, local copy	|
				 | of the selection value.								|
				\*------------------------------------------------------*/

				if (s = (char *) malloc (1 + *length)) {
					memcpy (s,value,*length);
					s[*length] = 0;
					if (*s == '<')
						p = parse_xml_string (s);
					else
						p = parse_string (s);
					if (check_syntax (p) != 0) {
						/* Error: mixed element content.  Don't proceed. */
						}
					else
						paste_subtree (p);
					}
		XtFree (value);
		}
	}

/*----------------------------------------------------------------------*\
 | Callback for the add button.											|
\*----------------------------------------------------------------------*/

static void add_element (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	enum fgdc_keyword key;
	struct item *p,*q;
	char *s = XtName (w);

	if (s) {
		get_text (current);
		v = XawListShowCurrent (element_list);
		i = v->list_index;
		p = line [i];
		key = key_of (s+4);
		if (key != Wunknown && key != Wnull) {
			if (q = add_child (p)) {
				q->key = key;
				q->d = text_of (key);
				q->original = 1;
				q->len = 1 + strlen (q->d);

				rebuild_lists();
				XawListChange (element_list,element,element_count,0,True);

				current = q;
				make_add_menu (current->key);
				set_text (current);
				for (i=0; i < line_count; i++)
					if (line[i] == current) break;
				XawListHighlight (element_list,i);
				if (i >= bottom_line)
					XawViewportSetCoordinates (element_viewport,0,
						line_height*(top_line + i - bottom_line + 1));
				if (XtIsRealized (info_text)) {
					char *string = advice_about (current->key);
					XtVaSetValues (info_text,XtNstring,string,NULL);
					}
				}
			}
		else
			if (strcmp (s,"add_all") == 0) {
				enum fgdc_keyword *key_list;
				if (p)
					if (key_list = element_list_of (p->key)) {
						enum fgdc_keyword *k;
						for (k=key_list; *k != Wnull; k++) {
							if (q = add_child (p)) {
								q->key = *k;
								q->d = text_of (*k);
								q->original = 1;
								q->len = 1 + strlen (q->d);
								}
							}
						rebuild_lists();
						XawListChange (element_list,element,element_count,0,True);
						XawListHighlight (element_list,i);
					}
				}
		}
	}

struct add_menu_list {
	struct add_menu_list *next;
	enum fgdc_keyword key;
	Widget menu;
	};

static struct add_menu_list *add_menu_list = NULL;
static Widget empty_add_menu = NULL;
static XtTranslations translations = NULL;

static struct add_menu_list *create_add_menu (enum fgdc_keyword key) {
	struct add_menu_list *a = NULL;
	enum fgdc_keyword *key_list;

	if (key_list = element_list_of (key)) {
		if (a = (struct add_menu_list *) malloc (sizeof (struct add_menu_list))) {
			Widget w;
			char name[128];
			char *s;
			enum fgdc_keyword *k;
			int n;

			a->next = NULL;
			a->key = key;
			sprintf (name,"menu_%s",text_of(key));
			a->menu = XtVaCreatePopupShell (
				name,
				simpleMenuWidgetClass,
				add_menu_button,
				NULL);

			if (!translations)
				translations = XtParseTranslationTable (
					"<EnterWindow>: highlight() tip() \n\
					<LeaveWindow>: unhighlight() tip()\n\
					<BtnMotion>: highlight() tip() \n\
					<BtnUp>: MenuPopdown() notify() unhighlight() tip()"
					);

			XtOverrideTranslations (a->menu,translations);

			n = 0;
			for (k=key_list; *k != Wnull; k++) {
				s = text_of (*k);
				sprintf (name,"add_%s",s);
				w = XtVaCreateManagedWidget (
					name,
					smeBSBObjectClass,
					a->menu,
					XtNlabel,s,
					NULL);
				XtAddCallback (w,XtNcallback,add_element,NULL);
				n++;
				}

			if (n > 1) {
				w = XtVaCreateManagedWidget (
					"add_all",
					smeBSBObjectClass,
					a->menu,
					XtNlabel,"(all of the above)",
					NULL);
				XtAddCallback (w,XtNcallback,add_element,NULL);
				}
			}
		}
	else {
		if (!empty_add_menu) {
			Widget w;
			w = XtVaCreatePopupShell (
				"empty_menu",
				simpleMenuWidgetClass,
				add_menu_button,
				NULL);
			XtVaCreateManagedWidget (
				"empty_menu_item",
				smeBSBObjectClass,
				w,
				XtNlabel,"Enter value below",
				NULL);
			empty_add_menu = w;
			}
		if (a = (struct add_menu_list *) malloc (sizeof (struct add_menu_list))) {
			a->next = NULL;
			a->key = key;
			a->menu = empty_add_menu;
			}
		}
	return (a);
	}

static void make_add_menu (enum fgdc_keyword key) {
	struct add_menu_list *a;

	if (add_menu_list) {
		for (a=add_menu_list; a->next; a=a->next)
			if (a->key == key) {
				XtVaSetValues (
					add_menu_button,
					XtNmenuName,XtName(a->menu),
					NULL);
				return;
				}
		a->next = create_add_menu (key);
		XtVaSetValues (
			add_menu_button,
			XtNmenuName,XtName(a->next->menu),
			NULL);
		}
	else
		if (a = create_add_menu (key)) {
			XtVaSetValues (
				add_menu_button,
				XtNmenuName,XtName(a->menu),
				NULL);
			add_menu_list = a;
			}
	}

static Widget get_add_menu (enum fgdc_keyword key) {
	struct add_menu_list *a;

	if (add_menu_list)
		for (a=add_menu_list; a; a=a->next)
			if (a->key == key) return (a->menu);

	return (NULL);
	}

/*----------------------------------------------------------------------*\
 | Action to populate the current node with all possible children.		|
\*----------------------------------------------------------------------*/

static void add_all_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	XawListReturnStruct *v;
	int i;
	enum fgdc_keyword *key_list,*k;
	struct item *p,*q;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	p = line [i];
	if (p)
		if (key_list = element_list_of (p->key)) {
			for (k=key_list; *k != Wnull; k++) {
				if (q = add_child (p)) {
					q->key = *k;
					q->d = text_of (*k);
					q->original = 1;
					q->len = 1 + strlen (q->d);
					}
				}
			rebuild_lists();
			XawListChange (element_list,element,element_count,0,True);
			XawListHighlight (element_list,i);
			}
	}

/*----------------------------------------------------------------------*\
 | Callback for the cut button.											|
\*----------------------------------------------------------------------*/

static void cut_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	struct item *p;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	if (i > 0) {
		p = line [i];
		detach_item (p);
		changed = 1;
		selection = p;
		XtOwnSelection (element_list,
			XA_PRIMARY,
			XtLastTimestampProcessed (display),
			convert_selection,
			lose_selection,
			done_selection);

		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);

		current = line [i-1];
		set_text (current);
		XawListHighlight (element_list,i-1);
		make_add_menu (current->key);
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		}
	}

static void cut_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	cut_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the copy button.										|
\*----------------------------------------------------------------------*/

static void copy_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	struct item *p;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	p = line [i];
	if (selection = copy_item (p)) {
		XtOwnSelection (element_list,
			XA_PRIMARY,
			XtLastTimestampProcessed (display),
			convert_selection,
			lose_selection,
			done_selection);
		}
	}

static void copy_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	copy_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the paste button.										|
\*----------------------------------------------------------------------*/

static void paste_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	if (XGetSelectionOwner(display,XA_PRIMARY) == XtWindow (element_list)) {
		if (selection) {
			get_text (current);
			paste_subtree (selection);
			selection = copy_item (selection);
			}
		}
	else {
		fprintf (stderr,"Warning: The paste menu item cannot obtain metadata\n");
		fprintf (stderr,"         from another application.  Use the paste action,\n");
		fprintf (stderr,"         bound to a keystroke or button press, instead.\n");
		fprintf (stderr,"         This is a flaw of Xtme that I have been unable\n");
		fprintf (stderr,"         to fix.  If you know a lot about Xt and are willing\n");
		fprintf (stderr,"         to help, please email pschweitzer@usgs.gov.\n");
	/*\
		XtGetSelectionValue (
			w,
			XA_PRIMARY,
			XA_STRING,
			get_selection_cb,
			NULL,
			XtLastTimestampProcessed (display));
	\*/
		}
	}

static void paste_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	if (XGetSelectionOwner(display,XA_PRIMARY) == XtWindow (element_list)) {
		if (selection) {
			get_text (current);
			paste_subtree (selection);
			selection = copy_item (selection);
			}
		}
	else {
		Time t;

		if (e->type == KeyPress || e->type == KeyRelease)
			t = ((XKeyEvent *)e)->time;
		else
			if (e->type == ButtonPress || e->type == ButtonRelease)
				t = ((XButtonEvent *)e)->time;
			else
				t = XtLastTimestampProcessed (display);

		/* Request the PRIMARY selection as a STRING */

		XtGetSelectionValue (
			w,
			XA_PRIMARY,
			XA_STRING,
			get_selection_cb,
			e,
			t);
		}
	}

/*----------------------------------------------------------------------*\
 | Callback for the clear button.										|
\*----------------------------------------------------------------------*/

static void clear_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	struct item *p;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	p = line [i];
	clear_scalars (p);
	set_text (p);
	changed = 1;
	}

static void clear_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	clear_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the duplicate button.									|
\*----------------------------------------------------------------------*/

static void dup_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	struct item *p,*q;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	p = line [i];

	if (q = insert_item_after (p)) {
		q->key = p->key;
		q->indent = p->indent;

		q->d = text_of (q->key);
		q->original = 1;
		q->len = 1 + strlen (q->d);

		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);

		get_text (current);
		current = q;
		make_add_menu (current->key);
		set_text (current);
		for (i=0; i < line_count; i++)
			if (line[i] == current) break;
		XawListHighlight (element_list,i);
		if (i >= bottom_line)
			XawViewportSetCoordinates (element_viewport,0,
				line_height*(top_line + i - bottom_line + 1));
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		}
	}

static void dup_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	dup_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the swap button.										|
\*----------------------------------------------------------------------*/

static void swap_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	struct item *p,*q;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	p = line [i];

	if (q = p->prev) {
		if (q->prev) q->prev->next = p;
		if (p->next) p->next->prev = q;
		p->prev = q->prev;
		q->next = p->next;
		p->next = q;
		q->prev = p;
		if (p->parent)
			if (p->parent->child == q) p->parent->child = p;

		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);

		for (i=0; i < line_count; i++)
			if (line[i] == current) break;
		XawListHighlight (element_list,i);
		if (i < top_line)
			XawViewportSetCoordinates (element_viewport,0,line_height*i);

		changed = 1;
		}
	}

static void swap_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	swap_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the prune button.										|
\*----------------------------------------------------------------------*/

static void prune_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	struct item *p;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	p = line [i];

	if (prune (p)) {
		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);

		for (i=0; i < line_count; i++)
			if (line[i] == current) break;
		XawListHighlight (element_list,i);
		if (i < top_line)
			XawViewportSetCoordinates (element_viewport,0,line_height*i);

		changed = 1;
		}
	}

static void prune_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	prune_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the hide button.										|
\*----------------------------------------------------------------------*/

static void hide_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	struct item *p,*q;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	p = line [i];
	if (p == root) return;
	if (p->child) {
		for (q=p->child; q; q=q->next)
			if (q->indent > 0) q->indent = -q->indent;
		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);

		current = p;
		set_text (current);
		XawListHighlight (element_list,i);
		}
	}

static void hide_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	hide_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the show button.										|
\*----------------------------------------------------------------------*/

static void show_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	int i;
	struct item *p,*q;

	get_text (current);
	v = XawListShowCurrent (element_list);
	i = v->list_index;
	p = line [i];
	if (p->child) {
		for (q=p->child; q; q=q->next)
			if (q->indent < 0) q->indent = -q->indent;
		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);

		current = p;
		set_text (current);
		XawListHighlight (element_list,i);
		}
	}

static void show_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	show_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the wrap button.										|
\*----------------------------------------------------------------------*/

static void wrap_value_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawTextWrapMode wrap;
	XtVaGetValues (
		value_text,
		XtNwrap,&wrap,
		NULL);
	if (wrap == XawtextWrapWord) {
		XtVaSetValues (
			value_text,
			XtNwrap,XawtextWrapNever,
			NULL);
		XtVaSetValues (
			wrap_value_button,
			XtNlabel,"Wrap",
			NULL);
		}
	else {
		XtVaSetValues (
			value_text,
			XtNwrap,XawtextWrapWord,
			NULL);
		XtVaSetValues (
			wrap_value_button,
			XtNlabel,"No wrap",
			NULL);
		}
	}

/*----------------------------------------------------------------------*\
 | Detail callback														|
\*----------------------------------------------------------------------*/

static int current_level = 0;

static void reveal_level (struct item *p, int target) {
	struct item *q;
	if (p) {
		int level = abs (p->indent);
		p->indent = ((level <= target) ? level : -level);
		for (q=p->child; q; q=q->next) reveal_level (q,target);
		}
	}

static void detail_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	int i = (int) client_data;

	get_text (current);
	current_level = i;
	if (i == 0) i = 128;
	reveal_level (root,i);
	rebuild_lists();
	XawListChange (element_list,element,element_count,0,True);
	if (current->indent < 0) {
		while (current->indent < 0)
			if (current->parent) current = current->parent;
			else break;
		set_text (current);
		}
	for (i=0; i < line_count; i++)
		if (line[i] == current) break;
	if (i == line_count) {
		current = root;
		i = 0;
		}
	XawListHighlight (element_list,i);
	}

static void collapse_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	int i;

	if (current_level == 0) current_level = 128;
	if (current_level == 128) current_level = 4;
	if (current_level > 1) {
		get_text (current);
		current_level--;
		reveal_level (root,current_level);
		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);
		if (current->indent < 0) {
			while (current->indent < 0)
				if (current->parent) current = current->parent;
				else break;
			set_text (current);
			}
		for (i=0; i < line_count; i++)
			if (line[i] == current) break;
		if (i == line_count) {
			current = root;
			i = 0;
			}
		XawListHighlight (element_list,i);
		}
	}

static void expand_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	int i;

	if (current_level == 0) current_level = 128;
	if (current_level < 128) {
		get_text (current);
		current_level++;
		reveal_level (root,current_level);
		rebuild_lists();
		XawListChange (element_list,element,element_count,0,True);
		if (current->indent < 0) {
			while (current->indent < 0)
				if (current->parent) current = current->parent;
				else break;
			set_text (current);
			}
		for (i=0; i < line_count; i++)
			if (line[i] == current) break;
		if (i == line_count) {
			current = root;
			i = 0;
			}
		XawListHighlight (element_list,i);
		}
	}

/*----------------------------------------------------------------------*\
 | Code to handle the information window								|
\*----------------------------------------------------------------------*/

static void show_info (char *string) {
	Position x,y;
	Dimension width,height;
	XtVaSetValues (info_text,
		XtNstring,string,
		XtNtype,XawAsciiString,
		XtNeditType,XawtextRead,
		NULL);
	XtVaGetValues (
		topLevel,
		XtNwidth,&width,
		XtNheight,&height,
		NULL);
	XtTranslateCoords (
		topLevel,
		(Position) width/2,
		(Position) height/8,
		&x,&y);
	XtVaSetValues (
		pshell,
		XtNx,x,
		XtNy,y,
		NULL);
	XtPopup (pshell,XtGrabNone);
	}

static void info_done (Widget w, XtPointer client_data, XtPointer call_data) {
	XtPopdown (pshell);
	}

static void toggle_wrap (Widget w, XtPointer client_data, XtPointer call_data) {
	XawTextWrapMode wrap;
	XtVaGetValues (
		info_text,
		XtNwrap,&wrap,
		NULL);
	if (wrap == XawtextWrapWord)
		XtVaSetValues (
			info_text,
			XtNwrap,XawtextWrapNever,
			NULL);
	else
		XtVaSetValues (
			info_text,
			XtNwrap,XawtextWrapWord,
			NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the about button.										|
\*----------------------------------------------------------------------*/

static char *about_buffer = NULL;

static void about_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	if (!about_buffer) {
		char **p;
		int n = 0;
		for (p=about_text; *p; p++) n += 1 + strlen (*p);
		if (about_buffer = (char *) malloc (1+n))
			for (p=about_text; *p; p++)
				strcat (about_buffer,*p);
		else
			about_buffer = "Internal error: could not allocate about_buffer\n";
		}
	if (XtIsRealized (info_text)) {
		XtVaSetValues (info_text,XtNstring,about_buffer,NULL);
		XtPopup (pshell,XtGrabNone);
		}
	else
		show_info (about_buffer);
	}

/*----------------------------------------------------------------------*\
 | Callback for the help button.										|
\*----------------------------------------------------------------------*/

static void help_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawListReturnStruct *v;
	struct item *p;
	char *s;

	if (v = XawListShowCurrent (element_list)) {
		p = line [v->list_index];
		if (s = advice_about (p->key))
			if (XtIsRealized (info_text)) {
				XtVaSetValues (info_text,XtNstring,s,NULL);
				XtPopup (pshell,XtGrabNone);
				}
			else
				show_info (s);
		}
	}

static void help_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	help_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the dismiss button.										|
\*----------------------------------------------------------------------*/

static void dismiss_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	info_done (w,client_data,call_data);
	}

static void dismiss_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	dismiss_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Code to handle the output window										|
\*----------------------------------------------------------------------*/

static void show_output (char *string) {
	Position x,y;
	Dimension width,height;
	XtVaSetValues (output_text,
		XtNstring,string,
		XtNtype,XawAsciiString,
		XtNeditType,XawtextRead,
		/* XtNwrap,XawtextWrapWord, */
		NULL);
	XtVaGetValues (
		topLevel,
		XtNwidth,&width,
		XtNheight,&height,
		NULL);
	XtTranslateCoords (
		topLevel,
		(Position) width/2,
		(Position) height/8,
		&x,&y);
	XtVaSetValues (
		oshell,
		XtNx,x,
		XtNy,y,
		XtNtitle,name_of_file,
		NULL);
	XtPopup (oshell,XtGrabNone);
	}

static void output_dismiss_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XtPopdown (oshell);
	}

static void output_dismiss_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	XtPopdown (oshell);
	}

static void output_toggle_wrap (Widget w, XtPointer client_data, XtPointer call_data) {
	XawTextWrapMode wrap;
	XtVaGetValues (
		output_text,
		XtNwrap,&wrap,
		NULL);
	if (wrap == XawtextWrapWord)
		XtVaSetValues (
			output_text,
			XtNwrap,XawtextWrapNever,
			NULL);
	else
		XtVaSetValues (
			output_text,
			XtNwrap,XawtextWrapWord,
			NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the Output button.										|
\*----------------------------------------------------------------------*/

static char *output_string = NULL;
static int output_string_length = 0;

static void output_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	get_text (current);
	if (output_string) {
		free (output_string);
		output_string = NULL;
		output_string_length = 0;
		}
	if (output_string = encode_text (root)) {
		if (XtIsRealized (output_text)) {
			XtVaSetValues (output_text,XtNstring,output_string,NULL);
			XtPopup (oshell,XtGrabNone);
			}
		else
			show_output (output_string);
		output_string_length = strlen (output_string);
		}
	}

static int line_number_at (long position) {
	if (output_string) {
		char *s = output_string;
		char *t = s + position;
		int n = 1;
		while (s < t) {
			if (*s == '\n') n++;
			s++;
			}
		return (n);
		}
	else return (0);
	}

static void line_number_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	char number[8];
	if (w == output_text) {
		XawTextPosition position = XawTextGetInsertionPoint (output_text);
		if (position < 0                   ) position = 0;
		if (position > output_string_length) position = output_string_length;
		sprintf (number,"%d",line_number_at (position));
		XtVaSetValues (
			output_line_number,
			XtNstring,number,
			NULL);
		}
	}

static void line_number_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	long n,position;
	char *string,*s;
	XtVaGetValues (output_line_number,XtNstring,&string,NULL);
	n = strtol (string,NULL,0);
	if (n > 0)
		if (output_string) {
			n--;
			for (s=output_string; n > 0; s++) if (*s == '\n') n--;
			position = s - output_string;
			XawTextSetInsertionPoint (output_text,position);
			}
	}

static void print_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	get_text (current);
	if (output_string) {
		free (output_string);
		output_string = NULL;
		output_string_length = 0;
		}
	if (output_string = encode_text (root)) {
		int fd;
		char print_file [FILENAME_MAX];
		strcpy (print_file,"xtmeXXXXXX");
		output_string_length = strlen (output_string);
		if (fd = mkstemp (print_file)) {
			FILE *pf;
			if (pf = fdopen (fd,"w")) {
				char cmd [FILENAME_MAX];
				fwrite (output_string,1,output_string_length,pf);
				fclose (pf);
				sprintf (cmd,"%s %s",print_command,print_file);
				system (cmd);
				}
			}
		}
	}

/*----------------------------------------------------------------------*\
 | jumpto_cb allows you to set the current element in the list widget	|
 | according to a line in the output window.							|
\*----------------------------------------------------------------------*/

static void jumpto_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	int i;
	long n;
	char *string;
	struct item *p,*q,*r;
	int do_list_update = 0;

	/*------------------------------------------------------------------*\
	 | Find out what line of the output text window you're on.			|
	\*------------------------------------------------------------------*/

	XtVaGetValues (output_line_number,XtNstring,&string,NULL);
	n = strtol (string,NULL,0);

	/* In the window, they start from 1; the array index starts at 0 */

	n--;

	/*------------------------------------------------------------------*\
	 | Find the item that is associated with this line.					|
	\*------------------------------------------------------------------*/

	if (p = item_of (n)) {

		if (!valid_node (p,root)) return;

		/*--------------------------------------------------------------*\
		 | If the line you specified in the output window is Wblank,	|
		 | move backwards until you find Wunknown or an element or NULL	|
		 | If you end up with nothing, do nothing and return.  That		|
		 | should never happen, of course, but recover well anyway.		|
		\*--------------------------------------------------------------*/

		while (p->key == Wblank)
			if (p->prev) p = p->prev;
		if (p->key == Wunknown || p->key == Wblank) p = p->parent;
		if (!p) return;

		/*--------------------------------------------------------------*\
		 | There's a good chance you're going to change the current		|
		 | element, so save the current contents of value_text.			|
		\*--------------------------------------------------------------*/

		get_text (current);

		/*--------------------------------------------------------------*\
		 | The item you're after could be hidden.  If so, unhide it,	|
		 | rebuild the lists, and update the list widget's list.		|
		\*--------------------------------------------------------------*/

		for (q=p; q; q=q->parent)
			if (q->indent < 0) {
				q->indent = -q->indent;
				do_list_update = 1;
				for (r=q->child; r; r=r->next)
					if (r->indent < 0) r->indent = -r->indent;
				}

		if (do_list_update) {
			rebuild_lists();
			XawListChange (element_list,element,element_count,0,True);
			}

		current = p;
		make_add_menu (current->key);
		set_text (current);
		for (i=0; i < line_count; i++)
			if (line[i] == current) break;
		XawListHighlight (element_list,i);

		if (i < top_line)
			XawViewportSetCoordinates (element_viewport,0,line_height*i);
		if (i >= bottom_line)
			XawViewportSetCoordinates (element_viewport,0,
				line_height*(top_line + i - bottom_line + 1));
		if (XtIsRealized (info_text)) {
			char *string = advice_about (current->key);
			XtVaSetValues (info_text,XtNstring,string,NULL);
			}
		}
	}

static void jumpto_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	jumpto_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | output_search_cb calls the search() action of output_text.			|
\*----------------------------------------------------------------------*/

static void output_search_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	String param;
	param = "forward";
	XtCallActionProc (output_text,"search",NULL,&param,1);
	}

/*----------------------------------------------------------------------*\
 | Set Metadata_Date to date when the program is run.					|
\*----------------------------------------------------------------------*/

static void update_metadata_date (struct item *r) {
	struct item *p,*q;
	time_t t;

	if (p = find_key (r,WMetadata_Date))
		if (q = p->child)
			if (q->key == Wunknown)
				if (q->d) {
					if (q->len < 9) {
						if (q->original == 0) free (q->d);
						if (q->d = (char *) malloc (9)) {
							q->len = 9;
							q->original = 0;
							*q->d = 0;
							}
						else
							return;
						}
					t = time (NULL);
					strftime (q->d,9,"%Y%m%d",localtime(&t));
					}
	}

/*----------------------------------------------------------------------*\
 | Callbacks for the save as button.									|
\*----------------------------------------------------------------------*/

static void show_save_as_dialog (void) {
	Position x,y;
	Dimension width,height;
	XtVaGetValues (
		topLevel,
		XtNwidth,&width,
		XtNheight,&height,
		NULL);
	XtTranslateCoords (
		topLevel,
		(Position) width/8,
		(Position) height/8,
		&x,&y);
	XtVaSetValues (
		qshell,
		XtNx,x,
		XtNy,y,
		NULL);
	XtPopup (qshell,XtGrabExclusive);
	}

static void save_as_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	get_text (current);
	show_save_as_dialog();
	}

static void save_as_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	get_text (current);
	show_save_as_dialog();
	}

static void save_as_done (Widget w, XtPointer client_data, XtPointer call_data) {
	int ok = (int) client_data;
	char *s,*t;

	if (ok) {
		if (s = XawDialogGetValueString (save_as_dialog))
			if (*s) {
				strcpy (name_of_file,s);
				if (changed) update_metadata_date (root);

				if (t = strrchr (name_of_file,'.'))
					if ((stricmp (t,".sgml") == 0) || (stricmp (t,".sgm") == 0))
						write_sgml (name_of_file,root);
					else
						if (stricmp (t,".xml") == 0)
							write_xml (name_of_file,root);
						else
							write_text (name_of_file,root);
				else
					write_text (name_of_file,root);

				XtVaSetValues (
					topLevel,
					XtNtitle,name_of_file,
					NULL);
				XtVaSetValues (
					save_as_dialog,
					XtNvalue,name_of_file,
					NULL);
				changed = 0;
				}
		XtPopdown (qshell);
		}
	else
		XtPopdown (qshell);
	}

static void ok_save_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	save_as_done (w,(XtPointer)1,NULL);
	}

static void cancel_save_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	save_as_done (w,(XtPointer)0,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the save button.										|
\*----------------------------------------------------------------------*/

static void save_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	get_text (current);
	if (changed) {
		update_metadata_date (root);
		if (strcmp (name_of_file,"Untitled") == 0) show_save_as_dialog();
		else {
			char *t;
			if (t = strrchr (name_of_file,'.'))
				if ((stricmp (t,".sgml") == 0) || (stricmp (t,".sgm") == 0))
					write_sgml (name_of_file,root);
				else
					if (stricmp (t,".xml") == 0)
						write_xml (name_of_file,root);
					else
						write_text (name_of_file,root);
			else
				write_text (name_of_file,root);
			changed = 0;
			}
		}
	}

static void save_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	save_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Callback for the quit button.										|
\*----------------------------------------------------------------------*/

static void show_confirm_dialog (void) {
	Position x,y;
	Dimension width,height;
	XtVaGetValues (
		topLevel,
		XtNwidth,&width,
		XtNheight,&height,
		NULL);
	XtTranslateCoords (
		topLevel,
		(Position) width/8,
		(Position) height/8,
		&x,&y);
	XtVaSetValues (
		rshell,
		XtNx,x,
		XtNy,y,
		NULL);
	XtPopup (rshell,XtGrabExclusive);
	}

static void confirm_done (Widget w, XtPointer client_data, XtPointer call_data) {
	int ok = (int) client_data;

	if (ok) {
		XtPopdown (rshell);
		exit (0);
		}
	else
		XtPopdown (rshell);
	}

static void quit_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	if (changed) {
		get_text (current);
		show_confirm_dialog();
		}
	else {
		forget_buffer (root);
		exit (0);
		}
	}

static void quit_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	quit_cb (w,NULL,NULL);
	}

/*----------------------------------------------------------------------*\
 | Viewport callback													|
\*----------------------------------------------------------------------*/

static void viewport_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	XawPannerReport *r = (XawPannerReport *) call_data;

	/*------------------------------------------------------------------*\
	if (r) {
		char status[8];
		char *s;

		s = status;
		*s++ = ((r->changed & XawPRSliderX)      ? 'x' : ' ');
		*s++ = ((r->changed & XawPRSliderY)      ? 'y' : ' ');
		*s++ = ((r->changed & XawPRSliderWidth)  ? 'w' : ' ');
		*s++ = ((r->changed & XawPRSliderHeight) ? 'h' : ' ');
		*s++ = ((r->changed & XawPRCanvasWidth)  ? 'W' : ' ');
		*s++ = ((r->changed & XawPRCanvasHeight) ? 'H' : ' ');
		*s = 0;
		printf ("Status: %s  x: %3d  y: %3d  w: %3d  h: %3d  W: %3d  H: %3d\n",
			status,
			r->slider_x,
			r->slider_y,
			r->slider_width,
			r->slider_height,
			r->canvas_width,
			r->canvas_height
			);
		}
	\*------------------------------------------------------------------*/

	if (r) {
		double yt,yb;
		yt = ((double)r->slider_y)/((double)line_height);
		yb = ((double)(r->slider_y + r->slider_height))/((double)line_height);
		top_line = (int) ceil (yt);
		bottom_line = (int) floor (yb);
		}
	}

/*----------------------------------------------------------------------*\
 | Change tip															|
\*----------------------------------------------------------------------*/

static Widget current_choice = NULL;

static struct known_t {
	Widget w;
	char *text;
	} known_tips [1024];
static int known_count = 0;

static int compare (const void *e1, const void *e2) {
	const struct known_t *t1 = (const struct known_t *) e1;
	const struct known_t *t2 = (const struct known_t *) e2;
	return ((unsigned)t1->w - (unsigned)t2->w);
	}

static void cache_tip (Widget w, char *text) {
	if (known_count < 1024) {
		known_tips [known_count].w = w;
		known_tips [known_count].text = text;
		known_count++;
		qsort (known_tips,known_count,sizeof (struct known_t),compare);
		}
	}

static char *tip_from_cache (Widget w) {
	struct known_t target;
	struct known_t *t;
	target.w = w;
	t = (struct known_t *) bsearch (&target,known_tips,known_count,sizeof (struct known_t),compare);
	if (t) return (t->text);
	else return (NULL);
	}

extern char *tip_text (char *parent_name, char *child_name);

static char *empty = "";

static void tip_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	XawListReturnStruct *v;
	int i;
	struct item *p;
	Widget menu,choice;
	char *s;

	if (XtIsManaged (tip_label) == False) return;

	if (w == file_menu || w == edit_menu || w == view_menu || w == help_menu) {
		if (choice = XawSimpleMenuGetActiveEntry (w)) {
			if (choice != current_choice) {
				s = tip_from_cache (choice);
				if (!s) {
					if (w == file_menu) {
						if (choice == save_button)
							s = "Writes current file to disk";
						if (choice == save_as_button)
							s = "Prompts user for a new name, then writes file";
						if (choice == quit_button)
							s = "Exits, prompting for confirmation if file changed.";
						}
					if (w == edit_menu) {
						if (choice == cut_button)
							s = "Detaches a subtree, remembering it for paste";
						if (choice == copy_button)
							s = "Remembers a subtree for later paste";
						if (choice == paste_button)
							s = "Places a cut or copied subtree into metadata";
						if (choice == clear_button)
							s = "Removes the text belonging to a subtree";
						if (choice == dup_button)
							s = "Makes and places a copy of the current element";
						if (choice == swap_button)
							s = "Switches the order of the current subtree with the preceding sibling";
						if (choice == prune_button)
							s = "Deletes all subtrees that do not contain text";
						}
					if (w == view_menu) {
						if (choice == hide_button)
							s = "Hides children of the current element";
						else
						if (choice == show_button)
							s = "Shows children of the current node";
						else
						if (choice == wrap_value_button)
							s = "Turns word-wrap on or off in the value widget";
						else {
							char *t = XtName (choice);
							if (memcmp (t,"detail_button_",14) == 0)
								switch (*(t+14) - '0') {
									case 0:
										s = "Shows the entire hierarchy";
										break;
									case 1:
										s = "Hides all but the major metadata sections";
										break;
									case 2:
										s = "Shows the major sections plus one level down";
										break;
									case 3:
										s = "Shows major sections plus two levels down";
										break;
									}
							}
						}
					if (w == help_menu) {
						if (choice == about_button)
							s = "Show version information and general hints";
						if (choice == help_button)
							s = "Show specific information about the current element";
						if (choice == all_button)
							s = "Show the output as it will appear";
						if (choice == tips_button)
							s = "Show or hide the tips line";
						}
					if (s) cache_tip (choice,s);
					}
				if (!s) s = empty;
				XtVaSetValues (
					tip_label,
					XtNlabel,s,
					NULL);
				current_choice = choice;
				}
			}
		else {
			XtVaSetValues (
				tip_label,
				XtNlabel,empty,
				NULL);
			current_choice = NULL;
			}
		}
	else {
		v = XawListShowCurrent (element_list);
		i = v->list_index;
		if (p = line [i])
			if (menu = get_add_menu (p->key))
				if (choice = XawSimpleMenuGetActiveEntry (menu)) {
					if (choice != current_choice) {
						current_choice = choice;
						s = tip_from_cache (choice);
						if (!s)
							if (strcmp (XtName (choice)+4,"all") == 0)
								s = "Add all elements (may add too many, but you can always cut)";
							else
								s = tip_text (XtName(menu)+5,XtName(choice)+4);
						if (!s) s = empty;
						cache_tip (choice,s);
						XtVaSetValues (
							tip_label,
							XtNlabel,s,
							NULL);
						}
					}
				else {
					XtVaSetValues (
						tip_label,
						XtNlabel,empty,
						NULL);
					current_choice = NULL;
					}
		}
	}

static void tips_cb (Widget w, XtPointer client_data, XtPointer call_data) {
	if (XtIsManaged (tip_label)) {
		XtUnmanageChild (tip_label);
		XtVaSetValues (
			tips_button,
			XtNlabel,"Show tips",
			NULL);
		XtVaSetValues (
			value_text,
			XtNshowGrip,True,
			NULL);
		XtVaSetValues (
			value_text,
			XtNshowGrip,False,
			NULL);
		}
	else {
		XtManageChild (tip_label);
		XtVaSetValues (
			tips_button,
			XtNlabel,"Hide tips",
			NULL);
		}
	}

/*----------------------------------------------------------------------*\
 | Insert today's date into the value widget.
\*----------------------------------------------------------------------*/

static void today_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	char date_string [16];
	char *params[1];
	time_t now;
	now = time (NULL);
	strftime (date_string,16,"%Y%m%d",localtime(&now));
	params[0] = date_string;

	XtCallActionProc (value_text,"insert-string",NULL,params,1);
	}

/*----------------------------------------------------------------------*\
 | For debugging, allow user to trigger a check of the tree topology.	|
\*----------------------------------------------------------------------*/

static void check_topology_action (Widget w, XEvent *e, String *args, Cardinal *n) {
	XawListReturnStruct *v;
	struct item *p;

	if (v = XawListShowCurrent (element_list)) {
		if (p = line [v->list_index])
			check_topology (p);
		else
			fprintf (stderr,"Topology error: No item at line[%d]\n",v->list_index);
		}
	else
		fprintf (stderr,"Warning: no item selected\n");
	}

/*----------------------------------------------------------------------*\
 | Global action table													|
\*----------------------------------------------------------------------*/

static XtActionsRec global_action[] = {
	{"check-topology",	check_topology_action},
	{"save",			save_action},
	{"save-as",			save_as_action},
	{"ok-save",			ok_save_action},
	{"cancel-save",		cancel_save_action},
	{"quit",			quit_action},
	{"cut",				cut_action},
	{"copy",			copy_action},
	{"paste",			paste_action},
	{"clear",			clear_action},
	{"duplicate",		dup_action},
	{"swap",			swap_action},
	{"prune",			prune_action},
	{"add-all",			add_all_action},
	{"help",			help_action},
	{"dismiss",			dismiss_action},
	{"dismiss-output",	output_dismiss_action},
	{"list-home",		list_home_action},
	{"list-up",			list_up_action},
	{"list-down",		list_down_action},
	{"list-pageup",		list_pageup_action},
	{"list-pagedown",	list_pagedown_action},
	{"collapse",		collapse_action},
	{"expand",			expand_action},
	{"hide",			hide_action},
	{"show",			show_action},
	{"tip",				tip_action},
	{"line-number",		line_number_action},
	{"jump-to",			jumpto_action},
	{"today",			today_action}
	};

/*----------------------------------------------------------------------*\
 | main procedure														|
\*----------------------------------------------------------------------*/

int main (int argc, char **argv) {
	int i;
	void *opt;
	char *ext_file,*tip_file,*s;
	XtAppContext app_context;
	AppData app_data;
	char *text_buffer = NULL;
	char *language = NULL;
	char *profile_name = NULL;

	out = stderr;

	/*------------------------------------------------------------------*\
	 | Initialize the application and interpret its resources.			|
	\*------------------------------------------------------------------*/

	topLevel = XtVaAppInitialize (
		&app_context,
		"Xtme",
		NULL,0,
		&argc,argv,
		fallback_resources,
		XtNtitle,"Xt Metadata Editor",
		NULL);

	display = XtDisplay (topLevel);

	XtGetApplicationResources (
		topLevel,
		&app_data,
		resources,
		XtNumber(resources),
		NULL,
		0);

	/*------------------------------------------------------------------*\
	 | Read any extra command-line arguments.							|
	\*------------------------------------------------------------------*/

	if (argc > 1)
		for (i=1; i < argc; i++)
			if (memcmp (argv[i],"-c",2) == 0) {
				i++;
				if (i < argc) app_data.configFile = argv[i];
				}
			else
				if (memcmp (argv[i],"-h",2) == 0) {
					i++;
					if (i < argc) app_data.helpFile = argv[i];
					}
				else
					if (memcmp (argv[i],"-l",2) == 0) {
						i++;
						if (i < argc) language = argv[i];
						}
					else
						input_file = argv[i];

	/*------------------------------------------------------------------*\
	 | Read the config file
	\*------------------------------------------------------------------*/

	read_configuration (app_data.configFile);

	/*------------------------------------------------------------------*\
	 | Use requested language.  Command-line argument overrides what is	|
	 | specified in the config file, config file overrides app default.	|
	\*------------------------------------------------------------------*/

	if (!language)
		if (opt = find_option (NULL,"input"))
			if (opt = find_option (opt,"language"))
				if (s = text_of_option (opt))
					language = s;

	if (!language) language = app_data.language;

	/*------------------------------------------------------------------*\
	 | Use requested profile.
	\*------------------------------------------------------------------*/

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"profile"))
			if (s = text_of_option (opt))
				profile_name = s;

	use_element_names (language,profile_name);		/* in keyword.c */
	if (profile_name) use_profile (profile_name);	/* in actions.c */

	/*------------------------------------------------------------------*\
	 | Read the extensions, tips, and help file
	\*------------------------------------------------------------------*/

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"extensions")) {
			if (ext_file = text_of_option (opt)) read_local (ext_file);
			while (opt = find_next_option (opt,"extensions"))
				if (ext_file = text_of_option (opt)) read_local (ext_file);
			}

	if (opt = find_option (NULL,"input"))
		if (opt = find_option (opt,"tips")) {
			if (tip_file = text_of_option (opt)) read_tips (tip_file);
			while (opt = find_next_option (opt,"tips"))
				if (tip_file = text_of_option (opt)) read_tips (tip_file);
			}

	read_help (app_data.helpFile);

	/*------------------------------------------------------------------*\
	 | Parse the input file.  Check the extension to see how.			|
	\*------------------------------------------------------------------*/

	root = NULL;
	if (input_file)
		if (s = strrchr (input_file,'.'))
			if (stricmp (s,".sgml") == 0 || stricmp (s,".sgm") == 0)
				root = parse_sgml (input_file);
			else
				if (stricmp (s,".xml") == 0)
					root = parse_xml (input_file);

	if (!root) root = parse_text (input_file, &text_buffer);

	/*------------------------------------------------------------------*/

	if (root) {
		strcpy (name_of_file,input_file);

		/*--------------------------------------------------------------*\
		 | The parser classifies all nodes by their keys.  If a valid	|
		 | element name appears in a textual value (at the beginning of	|
		 | a line), the parser classifies that node according to the	|
		 | element name.  But the node should be treated as text.  To	|
		 | correct this problem requires an understanding of the syntax	|
		 | of the standard.  Consequently it is necessary to minimally	|
		 | check the syntax to catch and recode elements that are found	|
		 | in textual values.											|
		\*--------------------------------------------------------------*/

		if (check_syntax (root) != 0) {
			fprintf (stderr,"Error: Mixed content is not permitted in elements.  Structural information would be lost if you continued.\n");
			exit (1);
			}

		/*--------------------------------------------------------------*\
		 | If the top node isn't Metadata, then you won't be able to	|
		 | add any of the immediate children of Metadata to the record.	|
		 | To solve this problem, create a Metadata node at the top if	|
		 | one is not present.											|
		 |																|
		 | More ecumenically, we should allow people to nest Metadata	|
		 | in their extensions.  So we only plant a Metadata node at	|
		 | the root if there isn't a Metadata node somewhere in the		|
		 | tree; it could be buried several levels down.				|
		\*--------------------------------------------------------------*/

		if (find_key (root,WMetadata) == NULL) {
			struct item *p,*q;

			p = allocate_item (1);
			p->key = WMetadata;
			p->d = text_of (p->key);
			p->original = 1;
			p->len = strlen (p->d);
			p->indent = 0;
			if (root->indent == p->indent)
				for (q=root; q; q=q->next) indent (q);
			p->child = root;
			root->parent = p;
			root = p;
			for (q=root->child; q; q=q->next) q->parent = root;
			}
		}
	else {
		struct item *p;

		/*--------------------------------------------------------------*\
		 | If for whatever reason there are no nodes to begin with,		|
		 | create a top node of type Metadata to hang everything on.	|
		\*--------------------------------------------------------------*/

		p = allocate_item (1);
		p->key = WMetadata;
		p->d = text_of (p->key);
		p->original = 1;
		p->len = strlen (p->d);
		p->indent = 0;
		root = p;
		if (input_file)
			strcpy (name_of_file,input_file);
		else
			strcpy (name_of_file,"Untitled");
		}

	if (text_buffer) remember_buffer (root,text_buffer);

	node_count = item_count (root);
	adjust_indent (root,0);
	build_list (root);
	build_element_list ();
	current = root;

	/*------------------------------------------------------------------*\
	\*------------------------------------------------------------------*/

	XtVaSetValues (
		topLevel,
		XtNtitle,name_of_file,
		NULL);

	frame = XtVaCreateManagedWidget (
		"frame",
		panedWidgetClass,
		topLevel,
		XtNorientation,XtorientVertical,
		NULL);

	button_box = XtVaCreateManagedWidget (
		"button_box",
		boxWidgetClass,
		frame,
		XtNorientation,XtorientHorizontal,
		XtNshowGrip,False,
		NULL);

	/*------------------------------------------------------------------*\
	 | File menu														|
	\*------------------------------------------------------------------*/

	file_menu_button = XtVaCreateManagedWidget (
		"file_menu_button",
		menuButtonWidgetClass,
		button_box,
		XtNlabel,"File",
		XtNmenuName,"file_menu",
		XtNborderWidth,0,
		NULL);

	file_menu = XtVaCreatePopupShell (
		"file_menu",
		simpleMenuWidgetClass,
		file_menu_button,
		NULL);

	save_button = XtVaCreateManagedWidget (
		"save_button",
		smeBSBObjectClass,
		file_menu,
		XtNlabel,"Save",
		NULL);
	XtAddCallback (save_button,XtNcallback,save_cb,NULL);

	save_as_button = XtVaCreateManagedWidget (
		"save_as_button",
		smeBSBObjectClass,
		file_menu,
		XtNlabel,"Save As",
		NULL);
	XtAddCallback (save_as_button,XtNcallback,save_as_cb,NULL);

	print_button = XtVaCreateManagedWidget (
		"print_button",
		smeBSBObjectClass,
		file_menu,
		XtNlabel,"Print",
		NULL);
	XtAddCallback (print_button,XtNcallback,print_cb,NULL);

	quit_button = XtVaCreateManagedWidget (
		"quit_button",
		smeBSBObjectClass,
		file_menu,
		XtNlabel,"Quit",
		NULL);
	XtAddCallback (quit_button,XtNcallback,quit_cb,NULL);

	/*------------------------------------------------------------------*\
	 | Edit menu														|
	\*------------------------------------------------------------------*/

	edit_menu_button = XtVaCreateManagedWidget (
		"edit_menu_button",
		menuButtonWidgetClass,
		button_box,
		XtNlabel,"Edit",
		XtNmenuName,"edit_menu",
		XtNborderWidth,0,
		NULL);

	edit_menu = XtVaCreatePopupShell (
		"edit_menu",
		simpleMenuWidgetClass,
		edit_menu_button,
		NULL);

	cut_button = XtVaCreateManagedWidget (
		"cut_button",
		smeBSBObjectClass,
		edit_menu,
		XtNlabel,"Cut",
		NULL);
	XtAddCallback (cut_button,XtNcallback,cut_cb,NULL);

	copy_button = XtVaCreateManagedWidget (
		"copy_button",
		smeBSBObjectClass,
		edit_menu,
		XtNlabel,"Copy",
		NULL);
	XtAddCallback (copy_button,XtNcallback,copy_cb,NULL);

	paste_button = XtVaCreateManagedWidget (
		"paste_button",
		smeBSBObjectClass,
		edit_menu,
		XtNlabel,"Paste",
		NULL);
	XtAddCallback (paste_button,XtNcallback,paste_cb,NULL);

	clear_button = XtVaCreateManagedWidget (
		"clear_button",
		smeBSBObjectClass,
		edit_menu,
		XtNlabel,"Clear",
		NULL);
	XtAddCallback (clear_button,XtNcallback,clear_cb,NULL);

	dup_button = XtVaCreateManagedWidget (
		"dup_button",
		smeBSBObjectClass,
		edit_menu,
		XtNlabel,"Duplicate",
		NULL);
	XtAddCallback (dup_button,XtNcallback,dup_cb,NULL);

	swap_button = XtVaCreateManagedWidget (
		"swap_button",
		smeBSBObjectClass,
		edit_menu,
		XtNlabel,"Swap",
		NULL);
	XtAddCallback (swap_button,XtNcallback,swap_cb,NULL);

	prune_button = XtVaCreateManagedWidget (
		"prune_button",
		smeBSBObjectClass,
		edit_menu,
		XtNlabel,"Prune",
		NULL);
	XtAddCallback (prune_button,XtNcallback,prune_cb,NULL);

	/*------------------------------------------------------------------*\
	 | View menu														|
	\*------------------------------------------------------------------*/

	view_menu_button = XtVaCreateManagedWidget (
		"view_menu_button",
		menuButtonWidgetClass,
		button_box,
		XtNlabel,"View",
		XtNmenuName,"view_menu",
		XtNborderWidth,0,
		NULL);

	view_menu = XtVaCreatePopupShell (
		"view_menu",
		simpleMenuWidgetClass,
		view_menu_button,
		NULL);

	for (i=0; i < 4; i++) {
		Widget w;
		char name [16], label[12];
		sprintf (name,"detail_button_%d",i);
		if (i == 0)
			strcpy (label,"All levels");
		else {
			strcpy (label,"  level");
			label[0] = '0' + i;
			if (i > 1) {
				label[7] = 's';
				label[8] = 0;
				}
			}
		w = XtVaCreateManagedWidget (
			name,
			smeBSBObjectClass,
			view_menu,
			XtNlabel,label,
			NULL);
		XtAddCallback (w,XtNcallback,detail_cb,(XtPointer)i);
		}

	dummy = XtVaCreateManagedWidget (
		"view_separator",
		smeLineObjectClass,
		view_menu,
		NULL);

	hide_button = XtVaCreateManagedWidget (
		"hide_button",
		smeBSBObjectClass,
		view_menu,
		XtNlabel,"Hide",
		NULL);
	XtAddCallback (hide_button,XtNcallback,hide_cb,NULL);

	show_button = XtVaCreateManagedWidget (
		"show_button",
		smeBSBObjectClass,
		view_menu,
		XtNlabel,"Show",
		NULL);
	XtAddCallback (show_button,XtNcallback,show_cb,NULL);

	wrap_value_button = XtVaCreateManagedWidget (
		"wrap_value_button",
		smeBSBObjectClass,
		view_menu,
		XtNlabel,"Wrap",
		NULL);
	XtAddCallback (wrap_value_button,XtNcallback,wrap_value_cb,NULL);

	/*------------------------------------------------------------------*\
	 | Add menu															|
	\*------------------------------------------------------------------*/

	add_menu_button = XtVaCreateManagedWidget (
		"add_menu_button",
		menuButtonWidgetClass,
		button_box,
		XtNborderWidth,0,
		XtNlabel,"Add",
		NULL);
	make_add_menu (WMetadata);

	/*------------------------------------------------------------------*\
	 | Help menu														|
	\*------------------------------------------------------------------*/

	help_menu_button = XtVaCreateManagedWidget (
		"help_menu_button",
		menuButtonWidgetClass,
		button_box,
		XtNborderWidth,0,
		XtNlabel,"Help",
		XtNmenuName,"help_menu",
		NULL);

	help_menu = XtVaCreatePopupShell (
		"help_menu",
		simpleMenuWidgetClass,
		help_menu_button,
		NULL);

	about_button = XtVaCreateManagedWidget (
		"about_button",
		smeBSBObjectClass,
		help_menu,
		XtNlabel,"Version",
		NULL);
	XtAddCallback (about_button,XtNcallback,about_cb,NULL);

	help_button = XtVaCreateManagedWidget (
		"help_button",
		smeBSBObjectClass,
		help_menu,
		XtNlabel,"Element",
		NULL);
	XtAddCallback (help_button,XtNcallback,help_cb,NULL);

	all_button = XtVaCreateManagedWidget (
		"all_button",
		smeBSBObjectClass,
		help_menu,
		XtNlabel,"Output",
		NULL);
	XtAddCallback (all_button,XtNcallback,output_cb,NULL);

	tips_button = XtVaCreateManagedWidget (
		"tips_button",
		smeBSBObjectClass,
		help_menu,
		XtNlabel,"Hide tips",
		NULL);
	XtAddCallback (tips_button,XtNcallback,tips_cb,NULL);

	/*------------------------------------------------------------------*\
	 | Element viewport depicting metadata structure					|
	\*------------------------------------------------------------------*/

	element_viewport = XtVaCreateManagedWidget (
		"element_viewport",
		viewportWidgetClass,
		frame,
		XtNallowVert,TRUE,
		XtNuseRight,TRUE,
		NULL);
	XtAddCallback (element_viewport,XtNreportCallback,viewport_cb,NULL);

	element_list = XtVaCreateManagedWidget (
		"element_list",
		listWidgetClass,
		element_viewport,
		XtNdefaultColumns,1,
		XtNforceColumns,TRUE,
		NULL);
	XtAddCallback (element_list,XtNcallback,list_cb,NULL);
	XawListChange (element_list,element,element_count,0,True);
	XawListHighlight (element_list,0);

	if (element_list) {
		int font_height;
		Dimension row_spacing;
		XFontStruct *font;
		XtVaGetValues (element_list,XtNfont,&font,NULL);
		font_height = font->max_bounds.ascent + font->max_bounds.descent;
		XtVaGetValues (element_list,XtNrowSpacing,&row_spacing,NULL);
		line_height = font_height + row_spacing;
		}

	value_text = XtVaCreateManagedWidget (
		"value_text",
		asciiTextWidgetClass,
		frame,
		XtNscrollVertical,XawtextScrollWhenNeeded,
		XtNuseRight,TRUE,
		XtNshowGrip,False,
		NULL);
	set_text (current);

	if (stricmp (app_data.keyboardFocusPolicy,"explicit") == 0)
		XtSetKeyboardFocus (frame,value_text);

	tip_label = XtVaCreateManagedWidget (
		"tip_label",
		labelWidgetClass,
		frame,
		XtNlabel,"Xtme: X Toolkit Metadata Editor by Peter Schweitzer (USGS)",
		XtNmin,20,
		XtNmax,20,
		XtNjustify,XtJustifyLeft,
		NULL);

	/*------------------------------------------------------------------*\
	\*------------------------------------------------------------------*/

	pshell = XtVaCreatePopupShell (
		"pshell",
		transientShellWidgetClass,
		topLevel,
		XtNtitle,"Xt Metadata Editor Help",
		NULL);

	info_box = XtVaCreateManagedWidget (
		"info_box",
		panedWidgetClass,
		pshell,
		XtNorientation,XtorientVertical,
		NULL);

	info_button_box = XtVaCreateManagedWidget (
		"info_button_box",
		boxWidgetClass,
		info_box,
		XtNborderWidth,0,
		XtNshowGrip,False,
		NULL);

	dismiss_button = XtVaCreateManagedWidget (
		"dismiss_button",
		commandWidgetClass,
		info_button_box,
		XtNlabel,"Dismiss",
		NULL);
	XtAddCallback (dismiss_button,XtNcallback,dismiss_cb,NULL);

	wrap_toggle = XtVaCreateManagedWidget (
		"wrap_toggle",
		toggleWidgetClass,
		info_button_box,
		XtNlabel,"Wrap",
		XtNstate,True,
		NULL);
	XtAddCallback (wrap_toggle,XtNcallback,toggle_wrap,NULL);

	info_text = XtVaCreateManagedWidget (
		"info_text",
		asciiTextWidgetClass,
		info_box,
		XtNscrollVertical,XawtextScrollWhenNeeded,
		XtNuseRight,TRUE,
		NULL);

	/* XtSetKeyboardFocus (info_box,info_text); */

	/*------------------------------------------------------------------*\
	\*------------------------------------------------------------------*/

	oshell = XtVaCreatePopupShell (
		"oshell",
		transientShellWidgetClass,
		topLevel,
		XtNtitle,"Xtme output",
		NULL);

	output_box = XtVaCreateManagedWidget (
		"output_box",
		panedWidgetClass,
		oshell,
		XtNorientation,XtorientVertical,
		NULL);

	output_button_box = XtVaCreateManagedWidget (
		"output_button_box",
		boxWidgetClass,
		output_box,
		XtNborderWidth,0,
		XtNshowGrip,False,
		NULL);

	output_view_menu_button = XtVaCreateManagedWidget (
		"output_view_menu_button",
		menuButtonWidgetClass,
		output_button_box,
		XtNborderWidth,0,
		XtNlabel,"View",
		XtNmenuName,"output_view_menu",
		NULL);

	output_view_menu = XtVaCreatePopupShell (
		"output_view_menu",
		simpleMenuWidgetClass,
		output_view_menu_button,
		NULL);

	output_update_button = XtVaCreateManagedWidget (
		"output_update_button",
		smeBSBObjectClass,
		output_view_menu,
		XtNlabel,"Update",
		NULL);
	XtAddCallback (output_update_button,XtNcallback,output_cb,NULL);

	output_search_button = XtVaCreateManagedWidget (
		"output_search_button",
		smeBSBObjectClass,
		output_view_menu,
		XtNlabel,"Search",
		NULL);
	XtAddCallback (output_search_button,XtNcallback,output_search_cb,NULL);

	output_wrap_toggle = XtVaCreateManagedWidget (
		"output_wrap_toggle",
		smeBSBObjectClass,
		output_view_menu,
		XtNlabel,"Wrap",
		NULL);
	XtAddCallback (output_wrap_toggle,XtNcallback,output_toggle_wrap,NULL);

	output_dismiss_button = XtVaCreateManagedWidget (
		"output_dismiss_button",
		smeBSBObjectClass,
		output_view_menu,
		XtNlabel,"Dismiss",
		NULL);
	XtAddCallback (output_dismiss_button,XtNcallback,output_dismiss_cb,NULL);

	output_jumpto = XtVaCreateManagedWidget (
		"output_jumpto",
		commandWidgetClass,
		output_button_box,
		XtNlabel,"Edit",
		NULL);
	XtAddCallback (output_jumpto,XtNcallback,jumpto_cb,NULL);

	output_line_number_goto = XtVaCreateManagedWidget (
		"output_line_number_goto",
		commandWidgetClass,
		output_button_box,
		XtNlabel,"Line:",
		NULL);
	XtAddCallback (output_line_number_goto,XtNcallback,line_number_cb,NULL);

	output_line_number = XtVaCreateManagedWidget (
		"output_line_number",
		asciiTextWidgetClass,
		output_button_box,
		XtNtype,XawAsciiString,
		XtNeditType,XawtextEdit,
		XtNstring,"1",
		NULL);

	XtSetKeyboardFocus (output_button_box,output_line_number);

	output_text = XtVaCreateManagedWidget (
		"output_text",
		asciiTextWidgetClass,
		output_box,
		XtNscrollVertical,XawtextScrollWhenNeeded,
		XtNuseRight,TRUE,
		NULL);

	/*------------------------------------------------------------------*\
	\*------------------------------------------------------------------*/

	qshell = XtVaCreatePopupShell (
		"qshell",
		transientShellWidgetClass,
		topLevel,
		XtNtitle,"Save As ...",
		NULL);

	save_as_dialog = XtVaCreateManagedWidget (
		"save_as_dialog",
		dialogWidgetClass,
		qshell,
		XtNlabel,"Enter name for metadata file:",
		XtNvalue,name_of_file,
		NULL);

	save_as_dialog_cancel = XtVaCreateManagedWidget (
		"save_as_dialog_cancel",
		commandWidgetClass,
		save_as_dialog,
		XtNlabel,"Cancel",
		NULL);
	XtAddCallback (save_as_dialog_cancel,XtNcallback,save_as_done,(XtPointer)0);

	save_as_dialog_ok = XtVaCreateManagedWidget (
		"save_as_dialog_ok",
		commandWidgetClass,
		save_as_dialog,
		XtNlabel,"OK",
		NULL);
	XtAddCallback (save_as_dialog_ok,XtNcallback,save_as_done,(XtPointer)1);

	/*------------------------------------------------------------------*\
	\*------------------------------------------------------------------*/

	rshell = XtVaCreatePopupShell (
		"rshell",
		transientShellWidgetClass,
		topLevel,
		XtNtitle,"Confirm Quit",
		NULL);

	confirm_dialog = XtVaCreateManagedWidget (
		"confirm_dialog",
		dialogWidgetClass,
		rshell,
		XtNlabel,"File has changed.\nQuit without saving?",
		/* XtNvalue,name_of_file, */
		NULL);

	confirm_dialog_cancel = XtVaCreateManagedWidget (
		"confirm_dialog_cancel",
		commandWidgetClass,
		confirm_dialog,
		XtNlabel,"Cancel",
		NULL);
	XtAddCallback (confirm_dialog_cancel,XtNcallback,confirm_done,(XtPointer)0);

	confirm_dialog_ok = XtVaCreateManagedWidget (
		"confirm_dialog_ok",
		commandWidgetClass,
		confirm_dialog,
		XtNlabel,"Quit",
		NULL);
	XtAddCallback (confirm_dialog_ok,XtNcallback,confirm_done,(XtPointer)1);

	/*------------------------------------------------------------------*\
	\*------------------------------------------------------------------*/

	XtAppAddActions (app_context,global_action,XtNumber(global_action));

	XtRealizeWidget (topLevel);
	XtAppMainLoop (app_context);
	return (0);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
